#!/bin/sh

# compare_rep.sh -- Henrique Moreira

# Compares two sets of files (in two dirs - D1 and D2):
#	* the original repository;
#	* the repository on specified directory.
#
# The values are hard-coded, but could be changed to args.

[ "$BASE_REP" = "" ] && BASE_REP=~paula/tmp-gprog/gprog

# In the case of xpfweb the directory (D2) gprog is created by a
# makefile, and then also gprog/gsrc is just a soft-link to lib,
# where lib contains itself the gobj directory: our main object of
# our comparison.

[ "$isVERBOSE" = "" ] && isVERBOSE=1

#OPT_DIFF=-y


n_args ()
{
 local N=$#
 [ $N -ge 255 ] && return 255
 return $N
}

check_dir_exists ()
{
 local D
 for D in $* ; do
    [ "$D" = "" ] && continue
    [ -d $D ] && continue
    echo "Directory does not exist: $D"
    exit 2
 done
 return 0
}

compare_files ()
{
 local F
 local D2
 local F2
 local B1
 local N
 local N_DIFFS=0
 local N_TOTAL=0
 local LST_DIFF=""

 D2=$1
 [ "$1" = "" ] && return 0
 shift

 # Note this function only compares files that exist on D2
 for F in $* ; do
    [ "$F" = "" ] && continue
    [ ! $F ] && return 2
    B1=`basename $F`
    F2="`find $D2 -type f -name $B1`"
    n_args $F2
    N=$?
    if [ $N != 1 ]; then
	[ $N = 0 ] && echo "No D2 for file: $B1"
	[ $N != 0 ] && echo "Too many D2: $B1"
	continue
    fi
    diff_file $F $F2
    if [ $? = 0 ]; then
	[ $isVERBOSE != 0 ] && echo "Checked: $B1 : $F"
	N_TOTAL=$(expr $N_TOTAL + 1)
    else
	N_DIFFS=$(expr $N_DIFFS + 1)
	LST_DIFF="$LST_DIFF $B1"
    fi
 done
 echo "::: Total files $N_TOTAL, differing: $N_DIFFS"
 [ "$LST_DIFF" ] && echo "$LST_DIFF"
 [ $N_DIFFS != 0 ] && return 1  # At least one diff
 return 0
}

diff_file ()
{
 local RES
 local N=$#
 [ "$N" != "2" ] && return 4  # Wrong usage of function
 diff $1 $2 > /dev/null
 RES=$?
 [ $RES = 0 ] && return 0
 echo "::: diff $1 $2"
 diff $OPT_DIFF $1 $2
 RES=$?
 return $RES
}

RES2=0
LST=""
if [ 0 = 0 ]; then
for S in \
	gsrc/gobj \
	gsrc/gbhtml \
	gsrc/gdhtml \
	\
	; do
	D1=$BASE_REP/gprog/$S
	D2=gprog/$S
	LST_1=`find $D1 -type f -name '*.h'`
	LST_2=`find $D1 -type f -name '*.cpp'`
	LST_3=`find $D1 -type f -name '*.dev'`
	echo "Comparing files: $D1 $D2"
	compare_files $D2 $LST_1 $LST_2 $LST_3
	if [ $? != 0 ]; then
		RES2=1
	fi
done
fi
# Now main progs:

### netinfo
D1=$BASE_REP/gprog/gNetInfo2/src
D2=netinfo/gNetInfo2/src
LST_1=`find $D1 -type f -name '*.h'`
LST_2=`find $D1 -type f -name '*.cpp'`
echo "Comparing progs: $D1 $D2"
compare_files $D2 $LST_1 $LST_2 $LST_3

### unweb:
D1=$BASE_REP/gprog/gUnweb/src
D2=unweb/gUnweb/src
LST_1=`find $D1 -type f -name '*.h'`
LST_2=`find $D1 -type f -name '*.cpp'`
LST_3=`find $D1 -type f -name '*.dev'`
echo "Comparing progs: $D1 $D2"
compare_files $D2 $LST_1 $LST_2 $LST_3

### uncoder:
D1=$BASE_REP/gprog/gUnCoder/src
D2=uncoder/gUnCoder/src
LST_1=`find $D1 -type f -name '*.h'`
LST_2=`find $D1 -type f -name '*.cpp'`
LST_3=`find $D1 -type f -name '*.dev'`
echo "Comparing progs: $D1 $D2"
compare_files $D2 $LST_1 $LST_2 $LST_3



# Exit status
STAT=0
[ $RES2 != 0 ] && STAT=1
exit $STAT

