// gBHttp.cpp -- Version 0.8

#include <string.h>  //memcpy, ...
#include "gBHttp.h"
#include "gstringext.h"
#include "garg.h"
////////////////////////////////////////////////////////////
gXHttpReplyHeader::gXHttpReplyHeader ()
    : htmlCode( -1 ),
      isTransferEncodingChunked( false )
{
}

bool gXHttpReplyHeader::IsText ()
{
 gString sKeyText;
 // If no key is found, we assume binary
 if ( GetLine( "Content-Type", sKeyText )==false ) return false;
 gParam paramText( sKeyText, "/" );
 paramText.SetCaseSense( false );
 return paramText.Match( "text" );
}

bool gXHttpReplyHeader::GetLine (char* sKind, gString& sResult)
{
 unsigned idx;
 idx = kindsL.Match( sKind );
 if ( idx==0 ) return false;
 sResult.Set( linesL.Str(idx) );
 return true;
}

bool gXHttpReplyHeader::AddLine (gString& s)
{
 char* str;
 unsigned nParam, n=N();
 gParam aParam( s, ": ", gParam::e_StopSplitOnFirst );
 nParam = aParam.N();
 if ( nParam<2 ) {
     if ( n==0 ) {
	 gParam firstLine( s, " " );
	 nParam = firstLine.N();
	 if ( nParam>=2 ) {
	     t_uint32 vRes;
	     str = firstLine.Str(2);
	     if ( gStrControl::Self().ConvertToUInt32( str, vRes )==0 ) {
		 htmlCode = (int)vRes;
		 sHtmlCode.SetEmpty();
		 for (unsigned iterStr=3; iterStr<=nParam; iterStr++) {
		     str = firstLine.Str( iterStr );
		     if ( str!=nil && str[0]!=0 ) {
			 if ( iterStr>3 ) sHtmlCode.Add( " " );
			 sHtmlCode.Add( str );
		     }
		 }
	     }
	     else {
		 htmlCode = -1;
	     }
	     return thisAddLine( firstLine.Str(1), str );
       	 }
     }
     kindsL.Add( "\0" );
     linesL.Add( "\0" );
     Add( s );
     return false;
 }
 return thisAddLine( aParam.Str(1), aParam.Str(2) );
}

t_uint32 gXHttpReplyHeader::GetLength (bool& hasFound, bool& isLengthOk)
{
 int error;
 t_uint32 vRes=0;
 gString sValue;

 hasFound = false;
 isLengthOk = false;

 if ( GetLine( "Transfer-Encoding", sValue ) ) {
     isTransferEncodingChunked = sValue.Find( "chunked" );
 }

 if ( GetLine( "Content-Length", sValue )==false ) return 0;

 hasFound = true;
 error = gStrControl::Self().ConvertToUInt32( sValue.Str(), vRes );
 isLengthOk = error==0 && vRes>0 && vRes<MAX_HTTP_CHUNK_SIZE;
 if ( isLengthOk ) return vRes;
 // RFC2616 refers:
 // {If a message is received with both a
 //  Transfer-Encoding header field and a Content-Length header field,
 //  the latter MUST be ignored.}
 // Therefore when 'isTransferEncodingChunked'=True, we must ignore the length.
 return 0;
}

int gXHttpReplyHeader::thisAddLine (char* sKind, char* sContent)
{
 gString s1( sKind );
 gString s2( sContent );
 return thisAddLine( s1, s2 );
}

int gXHttpReplyHeader::thisAddLine (gString& sKind, gString& sContent)
{
 gString aLine;
 kindsL.Add( sKind );
 linesL.Add( sContent );
 aLine = sKind;
 aLine.Add( ": " );
 aLine.AddString( sContent );
 Add( aLine );
 return 0;
}
////////////////////////////////////////////////////////////
gXHttpGeneric::gXHttpGeneric (gTcpConnect& connection)
    : gHttpGeneric( connection ),
      ioOpErrCode( 0 ),
      contentBytesRead( 0 ),
      totalBytesRead( 0 ),
      keepAliveTimeout( 0 ),  // for chunks, Apache usually sets 15 secs
      keepAliveMax( 0 ),  // for chunks, typically 100 secs
      fRepErr( nil ),
      pBuf( nil ),
      sProtocol( 40, '\0' )
{
 sProtoVersion.Set( "1.1" );
 // Default HTTP version shoud be set to 1.1! See RFC2145 (...HTTP Version Numbers), ch.2.3
}

gXHttpGeneric::~gXHttpGeneric ()
{
 delete pBuf;
}

gBigBuffer& gXHttpGeneric::GetBuffer ()
{
 ASSERTION(pBuf!=nil,"pBuf!=nil");
 return *pBuf;
}

gUCharBuffer& gXHttpGeneric::GetBaseBuffer ()
{
 gUCharBuffer* sBuf;
 sBuf = GetBuffer().baseBuf;
 ASSERTION(sBuf!=nil,"sBuf!=nil");
 return *sBuf;
}

gString& gXHttpGeneric::GetProtocolString ()
{
 snprintf( sProtocol.Str(), 38, "HTTP/%s", sProtoVersion.Str() );
 return sProtocol;
}

gVersion& gXHttpGeneric::GetVersion ()
{
 httpVersion.SetVersion( sProtoVersion );
 return httpVersion;
}

gTcpConnect& gXHttpGeneric::Connection ()
{
 ASSERTION(pConnection!=nil,"Connection()");
 return *pConnection;
}

gTcpConnect* gXHttpGeneric::ConnectionPtr ()
{
 ASSERTION(pConnection!=nil,"ConnectionPtr()");  // optional assert
 return pConnection;
}

bool gXHttpGeneric::GetContent (eContentMethod method,
				gString& sHost,
				gString& sPath,
				gString& sReply)
{
 getHeaderL.Delete();
 return true;
}

bool gXHttpGeneric::SetProtocolVersion (char* str)
{
 if ( str==nil ) return SetProtocolVersion( "1.1" );
 sProtoVersion.Set( str );
 return sProtoVersion.Length()<40;
}

int gXHttpGeneric::AddHead (gString& s, bool doCrNl)
{
 // By default doCrNl=true, meaning a Carriage-Return and New-Line (13d,10d) will be appended to 's'
 gString sSend( s );
 if ( doCrNl ) sSend.Add( "\r\n" );
 DBGPRINT_MIN("DBG_WEB: AddHead: %s! (%s)\n",s.Str(),doCrNl?"CR&LF":"\0");
 if ( sSend.IsEmpty() ) return 0;  // Nothing to write
 getHeaderL.Add( sSend );
 return 0;
}

int gXHttpGeneric::AddHeadStr (char* str)
{
 ASSERTION(str!=nil,"str!=nil");
 gString s( str );
 return AddHead( s );
}

int gXHttpGeneric::WriteHead (bool doFlush)
{
 // Write HTTP header request only once: more efficient
 gString s;
 char* str;
 unsigned i, n;
 unsigned pos;

 for (i=1, n=getHeaderL.N(); i<=n; i++) {
     str = getHeaderL.Str( i );
     pos = lReq.FindHeadFullFirst( str );
     gString sNewHead( str );
     if ( pos>0 ) {
	 DBGPRINT("DBG: list-pos=%u, of '%s', using: '%s'\n\n",
		  pos,
		  str,
		  lReq.Str(pos));
	 sNewHead = lReq.Str( pos );
	 lReq.Delete( pos, pos );
     }
     if ( sNewHead[ sNewHead.Length() ]!='\n' )
	 sNewHead.Add( "\r\n" );
     s.AddString( sNewHead );
 }

 for (i=1, n=lReq.N(); i<=n; i++) {
     str = lReq.Str( i );
     gString sNewHead( str );
     if ( sNewHead[ sNewHead.Length() ]!='\n' )
	 sNewHead.Add( "\r\n" );
     s.AddString( sNewHead );
 }

 if ( doFlush ) {
     s.Add( "\r\n" );

     for (i=1, n=lReq.NLines(); i<=n; i++) {
	 str = lReq.lText.Str( i );
	 ASSERTION(str,"str");
	 gString sNewHead( str );
	 if ( sNewHead[ sNewHead.Length() ]!='\n' )
	     sNewHead.Add( "\r\n" );
	 s.AddString( sNewHead );
     }
 }

 DBGPRINT_HIO("\nWRITE >>>>\n%s<<<< EOF\n\n",s.Str());

 return Connection().Write( s );
}

int gXHttpGeneric::SecSleep (t_uint32 aSec)
{
 DBGPRINT("DBG: SecSleep(%u)\n",(unsigned)aSec);
 return gFileControl::Self().SecSleep( aSec );
}

int gXHttpGeneric::MiliSecSleep (t_uint32 aMiliSec)
{
 DBGPRINT("DBG: MiliSecSleep(%u)/%d\n",(unsigned)aMiliSec,keepAliveTimeout);
 gFileControl::Self().MiliSecSleep( aMiliSec );
 return (int)aMiliSec;
}

int gXHttpGeneric::thisReadHeadReply (gTcpConnect& connection, unsigned& count, gString& sResult)
{
 int error;
 t_uchar c, lastChr;
 gString sLine;

 for (count=0, lastChr=0; ; ) {
     error = connection.Read( c )==false;
     if ( error!=0 ) return 1; //Broken connection
     count++;
     if ( c=='\r' ) continue;
     if ( c=='\n' ) {
	 //printf("DBG: REPLY '%s'\n",sLine.Str());
	 if ( c==lastChr ) return 0;
	 replyHeader.AddLine( sLine );
	 sLine.SetEmpty();
     }
     else {
	 sLine.Add( c );
     }
     lastChr = c;
     sResult.Add( c );
 }
 return 0;
}

int gXHttpGeneric::thisParseServerReply (gString& sReply, gXHttpReplyHeader& serverReply)
{
 int count=0;
 unsigned pos;
 char* str;
 t_int32 value;

 // printf("kindsL: "); serverReply.kindsL.Show();
 // printf("linesL: "); serverReply.linesL.Show();
 // E.g. of a chunked response
 //	kindsL: ("HTTP/1.1" "Date" "Server" "X-Powered-By" "Keep-Alive" "Connection" "Transfer-Encoding" "Content-Type")
 //	linesL: ("OK" "Sun, 24 Sep 2006 13:45:02 GMT" "Apache/2.0.52 (Fedora)" "PHP/4.3.10" "timeout=15, max=100" "Keep-Alive" "chunked" "text/html; charset=iso-8859-1")

 pos = serverReply.kindsL.Match( "Keep-Alive" );
 if ( pos==0 ) return 0;
 gParam timers( serverReply.linesL.Str( pos ), "," );
 // E.g.: {"timeout=15", "max=100"}
 str = timers.Str( 1 );
 pos = gStrControl::Self().Find( str, '=' );
 if ( pos>0 && gStrControl::Self().ConvertToInt32( str+pos, value )==0 ) {
     keepAliveTimeout = value<32000 ? (int)value : 0;
     count++;
 }
 str = timers.Str( 2 );
 pos = gStrControl::Self().Find( str, '=' );
 if ( pos>0 && gStrControl::Self().ConvertToInt32( str+pos, value )==0 ) {
     keepAliveMax = value<32000 ? (int)value : 0;
     count++;
 }

 DBGPRINT("DBG: keepAlive={%d,%d}\n",keepAliveTimeout,keepAliveMax);
 return count;
}
////////////////////////////////////////////////////////////
gXHttpCont::gXHttpCont (gTcpConnect& connection)
    : gXHttpGeneric( connection ),
      isContText( false )
{
 // I chose this agent: please set 'strUserAgent' empty for no "User-Agent" head writing
 strUserAgent.Set( "Lynx/2.8.4rel.1 libwww-FM/2.14 SSL-MM/1.4.1 OpenSSL/0.9.6b" );
}

gXHttpCont::~gXHttpCont ()
{
}

bool gXHttpCont::GetContent (eContentMethod method,
			     gString& sHost,
			     gString& sPath,
			     gString& sReply)
{
 sHostname = sHost;
 if ( sHostname.Find('/')>0 ) {
     ioOpErrCode = 5;
     return false;
 }
 gXHttpGeneric::GetContent( method, sHost, sPath, sReply );
 SetError( thisGetHtmlContrl( method, sPath, sReply ) );
 return lastOpError==0;
}

char* gXHttpCont::GetHeadUserAgentStr ()
{
 if ( strUserAgent.IsEmpty() ) return nil;
 sHeadTempStr.Set( "User-Agent: " );
 sHeadTempStr.AddString( strUserAgent );
 return sHeadTempStr.Str();
}

char* gXHttpCont::GetHeadRefererStr ()
{
 if ( sHostname.IsEmpty() ) return nil;
 sHeadTempStr.Set( "Referer: http://" );
 sHeadTempStr.AddString( sHostname );
 sHeadTempStr.Add( "/" );
 return sHeadTempStr.Str();
}

int gXHttpCont::thisGetHtmlContrl (eContentMethod method,
				   gString& sPath,
				   gString& sReply)
{
 int error;
 int parsedArgs;
 unsigned count;
 unsigned iL, nL;
 gString sendStr;
 eContentImport contImport;
 bool isGreaterThan1Dot0;
 bool doFlush = true;
 char* str;
 short vIdx;  // Stat index

 ioOpErrCode = 0;
 ASSERTION(pConnection!=nil,"pConnection!=nil");
 if ( sPath.IsEmpty() ) return -1;

 gVersion baseVersion( 1, 0 );
 isGreaterThan1Dot0 = GetVersion() > baseVersion;

 if ( method==e_Get )
     sendStr.Add( "GET " );
 else
     sendStr.Add( "POST " );
 sendStr.AddString( sPath );
 sendStr.Add( ' ' );
 //sendStr.Add( "HTTP/1.0" )  ==> May not send e.g. 'Host: yahoo.com'
 sendStr.AddString( GetProtocolString() );
 error = AddHead( sendStr );
 if ( error!=0 ) return 1;
 ioOpErrCode = 2;
 if ( isGreaterThan1Dot0 ) {
     // The client sends "Proxy-Connection: Keep-Alive" and "Pragma: No-Cache" headers,
     // which are HTTP/1.0 headers that have been replaced by:
     // "Connection: Keep-Alive" in HTTP/1.1.
     AddHeadStr( "Connection: Keep-Alive" );
 }

 gString sHostLine( "Host: " );
 sHostLine.AddString( sHostname );
 AddHead( sHostLine );
 for (iL=1, nL=strHeaderL.N(); iL<=nL; iL++) {
     str = strHeaderL.Str( iL );
     if ( str==nil || str[0]==0 ) continue;
     AddHeadStr( str );
 }
 str = GetHeadUserAgentStr();
 if ( str!=nil ) AddHeadStr( str );
 str = GetHeadRefererStr();
 if ( str!=nil ) AddHeadStr( str );
 if ( WriteHead( doFlush )!=0 ) return 1;

 ioOpErrCode = 3;
 if ( doFlush==false ) error = Connection().Write( "\r\n" );
 if ( error!=0 ) return 1;
 ioOpErrCode = 4;

 for (vIdx=0; vIdx<nStat; vIdx++) mStat[vIdx].Start();

 error = thisReadHeadReply( *pConnection, count, sReply );
 parsedArgs = thisParseServerReply( sReply, replyHeader );
 ioOpErrCode = 8;
 totalBytesRead += (t_uint64)count;

 DBGPRINT_MIN("DBG: first reply: error=%d, bytes-read=%u, count=%u, parsed-args=%d\n",
	      error,
	      (unsigned)totalBytesRead,
	      (unsigned)count,
	      parsedArgs);

 vIdx = 1;
 mStat[vIdx].CpuTics(); vStat[vIdx].SetUInt( mStat[vIdx].GetMilisec() ); vIdx++;
 if ( error!=0 ) return 1;

 bool hasFound, isLengthOk;
 t_uint16 uSize = replyHeader.GetLength( hasFound, isLengthOk );
 isLengthOk = isLengthOk && hasFound;
 bool isChunked = replyHeader.IsChunked();

 delete pBuf;
 pBuf = nil;

 // Allocate buffer
 if ( isLengthOk )
     pBuf = new gBigBuffer( (t_uint32)uSize );
 else
     pBuf = new gBigBuffer;
 ASSERTION(pBuf!=nil,"pBuf!=nil");

 isContText = replyHeader.IsText();

 DBGPRINT_MIN("TODO: first reply: isChunked:%c, isLengthOk:%c, isContText:%c\n",
	      ISyORn(isChunked),
	      ISyORn(isLengthOk),
	      ISyORn(isContText));

 if ( isChunked ) {
     // Set correct 'isContText'
     error = ReadReplyChunked( *pConnection, *pBuf, count );
     DBGPRINT_MIN("TODO: ReadReplyChunked(%u), error=%d\n",count,error);
 }
 else {
     // If there is a key 'Content-Type: text/plain', this method returns true
     contImport = isContText ? gXHttpGeneric::e_TextPlain : gXHttpGeneric::e_Binary;

     if ( isLengthOk ) {
	 DBGPRINT_HIO("DBG: uSize=%u, bBuf.size=%u\n",
		      uSize,
		      pBuf->baseBuf->size);
	 if ( isContText )
	     error = ReadReplyText( *pConnection, contImport, *pBuf, count );
	 else
	     error = ReadReplyBin( *pConnection, *pBuf, count );
     }
     else {
	 error = ReadReplyUnsized( *pConnection, contImport, *pBuf, count );
	 DBGPRINT_HIO("DBG: ReadReplyUnsized, error=%d\n",error);
	 if ( error!=0 ) return ioOpErrCode = 32; //Unable to save buffer (for 'Unsized')
     }

     ioOpErrCode = 0;
     DBGPRINT_HIO("DBG: uSize=%u, count=%u\n",uSize,count);
     if ( uSize>0 && (unsigned)uSize!=count+1 ) {
	 ioOpErrCode = 16;
	 // Note: no error because this is a non-blocking error
     }
 }//end IF (not chunked)

 // vIdx is 2
 mStat[vIdx].CpuTics(); vStat[vIdx].SetUInt( mStat[vIdx].GetMilisec() ); vIdx++;
 // vIdx 3 is used for our own client pace

 contentBytesRead += (t_uint64)count;
 totalBytesRead += (t_uint64)count;

 DBGPRINT_MIN("DBG: gXHttpCont::thisGetHtmlContrl returning %d (lastOpError=%d, ioOpErrCode=%d)\n",
	      error,
	      lastOpError,
	      ioOpErrCode);
 return error;
}

int gXHttpCont::ReadReplyText (gTcpConnect& connection,
			      eContentImport contImport,
			      gBigBuffer& bBuf,
			      unsigned& count)
{
 static char chBuf[8], ch2Buf[2];
 t_uchar c;
 t_uchar* uBuf;
 bool isTextData;
 unsigned usedCount, size;

 uBuf = bBuf.baseBuf->uBuf;
 size = bBuf.baseBuf->size;

 if ( connection.IsOk()==false ) return 2;
 switch ( contImport ) {
 case e_TextPlain:
     isTextData = true;
     break;
 case e_Binary:  //Not for this method!
 default:
     ASSERTION_FALSE("ReadReply(1)");
 }

 for (count=1, usedCount=0; count<size; ) {
     DBGPRINT_HIO("DBG: reading-txt count=%u/%u BEF\n",count,size);
     if ( connection.Read( c )==false ) break;
     count++;
     if ( fRepErr!=nil && count%64==0 ) {
	 sprintf(chBuf,"0x%02X",c);
	 ch2Buf[0] = c;
	 fprintf(fRepErr,":%4s %u/%u\n",c<' '||c>126?chBuf:ch2Buf,count,size);
     }
     if ( c=='\r' ) continue;
     uBuf[usedCount++] = c;
 }

 bBuf.usedSize = (t_uint64)usedCount;
 DBGPRINT_HIO("DBG: reading-txt count=%u/%u, usedCount=%u FINAL\n",count,size,usedCount);
 // Pad end with chr 0x0
 uBuf[ usedCount ] = 0x0;

 if ( fRepErr!=nil ) fprintf(fRepErr,":Done %u/%u\n",count,size);

 return count>=size ? 0 : -1;
}

int gXHttpCont::ReadReplyBin (gTcpConnect& connection,
			     gBigBuffer& bBuf,
			     unsigned& count)
{
 t_uchar c;
 t_uchar* uBuf;
 unsigned size;

 uBuf = bBuf.baseBuf->uBuf;
 size = bBuf.baseBuf->size;
 ASSERTION(size>0,"size>0");
 size--;  // Buffer is one byte wider than needed.

 for (count=0; count<size; ) {
     DBGPRINT_HIO("DBG: reading count=%u/%u BEF\n",count,size);
     bBuf.usedSize = (t_uint64)count;  // Must be here!
     if ( connection.Read( c )==false ) {
	 DBGPRINT_HIO("DBG: reading count=%u/%u AFT(ERROR)\n",count,size);
	 return -1;
     }
     uBuf[ count++ ] = c;
     uBuf[ count ] = 0;
     DBGPRINT_HIO("DBG: reading count=%u/%u AFT(OK) [%02X]\t%c\n",
		  count, size,
		  c,
		  c<' ' || c>=127 ? '.' : c);
 }
 DBGPRINT_HIO("DBG: reading count=%u/%u FINAL\n",count,size);
 // Pad end with chr 0x0
 return 0;
}

int gXHttpCont::ReadReplyUnsized (gTcpConnect& connection,
				 eContentImport contImport,
				 gBigBuffer& bBuf,
				 unsigned& count)
{
 // Returns 0 on success
 t_uchar c;
 t_uchar* uBuf;
 unsigned idxCount=0;
 t_uint64 bCount=0;
 bool isReadOk;
 bool isFlushOk=true;

 bBuf.usedSize = 0;
 uBuf = bBuf.baseBuf->uBuf;

 for (count=0; isFlushOk==true; ) {
     c = 0; //TODO:cut
     isReadOk = connection.Read( c );
     //dbg...: if ( c<' ' || c=='[' || c==']' ) printf("[%02X]\t",c); else printf("%c\t",c);
     DBGPRINT_HIO("DBG: read-unsized ok?%c, count=%u, usedSize=%lu\n",
		  ISyORn( isReadOk ),
		  count,
		  (long unsigned)bBuf.usedSize);
     if ( isReadOk==false ) {
	 bBuf.usedSize = bCount;
	 isFlushOk = bBuf.Flush( idxCount );
	 return isFlushOk ? 0 : -1;
     }
     count++;
     bCount++;
     uBuf[idxCount] = c;
     idxCount++;

     // Check buffer overflow
     if ( idxCount>=bBuf.baseBuf->size ) {
	 bBuf.usedSize++;
	 isFlushOk = bBuf.Flush( idxCount );
	 idxCount = 0;
	 // Important: buffer re-created, thus uBuf re-assigned
	 // (and uBuf is for performance reasons)
	 uBuf = bBuf.baseBuf->uBuf;
     }
 }
 // Only reaches here if buffering failed
 return -1;
}

int gXHttpCont::ReadReplyChunked (gTcpConnect& connection,
				 gBigBuffer& bBuf,
				 unsigned& count)
{
 // Returns 0 on success, 1 if chunk-size is invalid, 2 if read failed, -1 on writting error.
 // 4 indicates no CR-LF sequence before chunk
 unsigned iter=0;
 int error;
 t_uint32 chunkSize=0;
 t_uint32 timeBefore=vStat[1].GetUInt();
 t_uint32 timeNew, timeChunk, waited=0;
 bool isFlushOk;
 bool isHugeBuffer;
 gBigBuffer* ptrBuf;

 bBuf.NewFile();
 //printf("DBG: todo, dbg\n");

 for (count=0, error=0, isFlushOk=true;
      error==0 && isFlushOk==true && (chunkSize = thisReadChunkSize( connection, error ))>0;
      count+=chunkSize, iter++) {
     ASSERTION(chunkSize<MAX_HTTP_HUGE_CHUNK_SIZE,"ReadReplyChunked(0)");
     isHugeBuffer = chunkSize>=MAX_HTTP_CHUNK_SIZE;
     DBGPRINT_MIN("DBG: ReadReplyChunked()[%u]=%d chunkSize=%ld%s\n",iter,error,(long)chunkSize,isHugeBuffer?" (HUGE-CHUNK)":"\0");
     if ( error!=0 ) break;
     ptrBuf = new gBigBuffer( isHugeBuffer ? 10 : chunkSize );
     ASSERTION(ptrBuf!=nil,"ReadReplyChunked(1)");
     if ( isHugeBuffer ) {
	 delete[] ptrBuf->baseBuf->uBuf;
	 ptrBuf->baseBuf->uBuf = new t_uchar[ chunkSize+1 ];
	 ASSERTION(ptrBuf->baseBuf->uBuf!=nil,"Not enough memory to read big chunks");
     }

     // Pace of server respected (acc.RFC2616, ch.8.2.4)
#ifdef gDOS_SPEC
     timeBefore /= 1000;  // (CHECK THIS)
#else
     timeBefore /= 100;  // 1/10th seconds before...
#endif //gDOS_SPEC

     timeChunk = chunkSize>8192;
     timeChunk += 10*(chunkSize>MAX_HTTP_CHUNK_SIZE);

#ifdef gDOS_SPEC
     timeNew = timeBefore + 2*(chunkSize>3000) + (chunkSize>1024)/*+1 would be an ugly try*/;
     waited = timeNew * 1000;
     DBGPRINT_MIN("DBG: [gXHttpCont] (DOS) Pace: %u s\n",(unsigned)timeNew);
     SecSleep( timeNew );
#else
     timeNew = timeBefore + timeChunk;
     waited = timeNew*10+chunkSize/8;
     DBGPRINT_MIN("DBG: [gXHttpCont] (non-DOS) Pace: %u ms\n",(unsigned)waited);
     MiliSecSleep( waited );
#endif

     if ( fRepErr )
	 fprintf(fRepErr,"Pace (ms): %u = (time=%u/%u [chunk=%ld, timeChunk=%u])\n",
		 waited,
		 (unsigned)timeNew,
		 (unsigned)timeBefore,
		 (long)chunkSize,
		 (unsigned)timeChunk);

     // Just increment the vStat[3] (pace)
     vStat[3].SetUInt( vStat[3].GetUInt() + waited );  // Miliseconds made (pace) on our own HTTP client
     DBGPRINT("DBG: (iter=%u) stat=%u, waited=%lu (total_wait:%u)\n",
	      iter,
	      vStat[1].GetUInt(),
	      (unsigned long)waited,
	      vStat[3].GetUInt());
     // ==>>old?	mStat[3/*Hard-coded:pace*/].SetUInt( timeNew * 1000 );
     // Read from socket...
     error = connection.Read( ptrBuf->baseBuf->uBuf, chunkSize ) ? 0 : 2;

     DBGPRINT_MIN("DBG: timeNew=%u = (%u + %u [chunk=%ld]),error=%d\n",
	    (unsigned)timeNew,
	    (unsigned)timeBefore,
	    (unsigned)timeChunk,
	    (long)chunkSize,
	    error);

     {
	int dosCount=0;
	for ( ; error!=0 && dosCount<3/* 200 ??*/; dosCount++) {
	    DBGPRINT("DBG: READ(iter=%u,dosCount=%d):err=%d, size=%u\n",iter,dosCount,error,(unsigned)chunkSize);
	    error = connection.Read( ptrBuf->baseBuf->uBuf, chunkSize ) ? 0 : 2;
	    vStat[3].SetUInt( vStat[3].GetUInt() + 1 );  // Just loose some time, and increment statistical pace
	}
#ifdef DEBUG
	if ( dosCount>0 ) printf("DBG: READ(iter=%u,dosCount=%d):err=%d (FINAL)\n",iter,dosCount,error);
#endif //DEBUG
     }

     if ( error==0 ) {
	 // ...and write to file
	 isFlushOk = bBuf.WriteBuf( bBuf.GetFile().fHandle, *(ptrBuf->baseBuf), chunkSize );
	 // Read CR-LF
	 memset( ptrBuf->baseBuf->uBuf, 0x0, 10 );
	 error = connection.Read( ptrBuf->baseBuf->uBuf, 2 )==false;
	 if ( error==0 ) error = ptrBuf->baseBuf->uBuf[1]=='\n' ? 0 : 4;
     }
     delete ptrBuf;
     ptrBuf = nil;
 }//end FOR

 if ( isFlushOk==false ) return -1;
 if ( error==-1 ) return 1;
 if ( error>=2 ) return error;
 return count==0;
}

t_uint32 gXHttpCont::thisReadChunkSize (gTcpConnect& connection, int& error)
{
 gString sChunk;
 t_uint32 numHex;

 error = 1;
 if ( connection.ReadLine( sChunk )==false ) return 0;
 error = gStrControl::Self().ConvertHexToUInt32( sChunk.Str(), e_DigConvAny, numHex );
 DBGPRINT("DBG: [gXHttpCont] Read chunk %s%lu\n",
	  error<=-1 ? "(failed) " : "\0",
	  (unsigned long)numHex);
 // error is -1 when the conversion fails
 if ( error!=0 ) return 0;
 return numHex;
}
////////////////////////////////////////////////////////////

