#include "gHtmlAttr.h"
#include "gstringext.h"
////////////////////////////////////////////////////////////
gHAttrList::gHAttrList ()
    : nAttr( 0 ),
      originalLst( nil ),
      pAttrAssign( nil )
{
}

gHAttrList::~gHAttrList ()
{
 Reset();
}

char* gHAttrList::Str ()
{
 sAttr.SetEmpty();
 for (short idx=1; idx<=nAttr; idx++) {
     gParam* pAssign = pAttrAssign[idx];
     ASSERTION(pAssign!=nil,"pAssign!=nil");
     unsigned n = pAssign->N();
     if ( n==0 ) continue;
     if ( idx>1 ) sAttr.Add( ' ' );
     sAttr.Add( pAssign->Str(1) );
     if ( n>1 ) {
	 sAttr.Add( '=' );
	 sAttr.Add( pAssign->Str(2) );
     }
 }
 return sAttr.Str();
}

unsigned gHAttrList::N ()
{
 ASSERTION(nAttr>=0,"nAttr>=0");
 return (unsigned)nAttr;
}

gParam* gHAttrList::GetAttrParam (short iAttr)
{
 ASSERTION(iAttr>=1 && iAttr<=nAttr,"iAttr>=1");
 gParam* pAssign = pAttrAssign[iAttr];
 ASSERTION(pAssign!=nil,"pAssign!=nil");
 return pAssign;
}

char* gHAttrList::GetAttr (short iAttr)
{
 unsigned iter, n;
 ASSERTION(iAttr>=1,"iAttr>=1");
 if ( iAttr>nAttr ) return nil;
 sOAttr.SetEmpty();
 gParam* pAssign = pAttrAssign[iAttr];
 for (iter=1, n=pAssign->N(); iter<=n; iter++) {
     sOAttr.Add( pAssign->Str(iter) );
     if ( iter<n ) sOAttr.Add( ' ' );
 }
 return sOAttr.Str();
}

char* gHAttrList::GetAttrValue (short iAttr)
{
 unsigned len;

 // Same as GetAttr, except:
 //   If the attribute contains quotes, then this result is merely the attribute.
 // Examples:
 //   <A HREF="abc" Attr2=Val2>
 //   There are two attributes: HREF="abc" and Attr2=Val2, i.e. nAttr=2.
 //   The result from GetAttr and GetAttrValue is the same.
 //   Now the statement <!DOCTYPE HTML PUBLIC -//W3C//DTD HTML 4.01 Transitional//EN>
 //   contains 6 attributes. But typically this is encoded with quotes, like:
 //   <!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
 //   with 3 attributes, GetAttrValue(3) returns exactly the unquoted string:
 //	-//W3C//DTD HTML 4.01 Transitional//EN
 //
 if ( GetAttr(iAttr)==nil ) return nil;
 len = sOAttr.Length();
 if ( sOAttr[1]=='"' && sOAttr[len]=='"' ) {
     gString sTemp;
     sTemp.CopyFromTo( sOAttr, 2, len-1 );
     sOAttr = sTemp;
 }
 return sOAttr.Str();
}

char* gHAttrList::GetAttrLValue (short iAttr)
{
 gParam* pAssign = GetAttrParam( iAttr );
 return pAssign->Str( 1 );
}

char* gHAttrList::GetAttrRValue (short iAttr)
{
 gParam* pAssign = GetAttrParam( iAttr );
 return pAssign->Str( 2 );
}

int gHAttrList::ValidateValue (char* s)
{
 // Return 0 if ok
 ASSERTION(s!=nil,"s!=nil");
 gString aStr( s );
 return thisValidateValue( aStr );
}

void gHAttrList::Reset ()
{
 gControl::Reset();
 delete originalLst;
 for (short idx=1; pAttrAssign!=nil && idx<=nAttr; idx++) delete pAttrAssign[idx];
 delete[] pAttrAssign;
 nAttr = 0;
}

void gHAttrList::Set (char* s)
{
 lastOpError = 0;
 if ( s==nil ) {
     Reset();
     return;
 }
 DBGPRINT_MIN("DBG:.\nDBG: ATTR is: %s\n",s);
 ASSERTION(originalLst==nil,"Set(1)");
 originalLst = new gParam( s, " ", gParam::e_NormalQuoted );
 ASSERTION(originalLst!=nil,"Set(2)");
 nAttr = originalLst->N();
 ASSERTION(nAttr>=1,"Set(3): nAttr>=1");
 //printf("DBG: LST:");originalLst->Show(true);printf("!!\n");
 pAttrAssign = new gParam*[ nAttr+1 ];
 pAttrAssign[ 0 ] = nil;  // Not used index0
 for (short idx=1; idx<=nAttr; idx++) {
     gParam* pAssign;
     char* str = originalLst->Str(idx);
     char* str2 = nil;
     pAssign = new gParam( str, "=", gParam::e_StopSplitOnFirst );
     ASSERTION(pAssign!=nil,"Set(4): pAssign!=nil");
     pAttrAssign[idx] = pAssign;
     unsigned n = pAssign->N();
     ASSERTION(n==1 || n==2,"Set(5): n==1 || n==2");
     if ( n==2 ) {
	 lastOpError += ValidateValue( str2 = pAssign->Str(2) )!=0;
     }
     else {
	 lastOpError += ValidateValue( str )!=0;
     }
     DBGPRINT_MIN("DBG: ATTR: [%s] n=%u (IDX=%u/%u) AttrVal:[%s|%s], nerrors=%d\n",str,pAssign->N(),idx,nAttr,pAssign->Str(1),str2,lastOpError);
 }
 // lastOpError contains the number of errors found in attr-list
 DBGPRINT_MIN("DBG: ATTR [[end]]\n");
}

void gHAttrList::CopyAttr (gHAttrList& copy)
{
 if ( copy.IsEmpty() ) return;
 DBGPRINT_MIN("DBG:.\nDBG: ATTR [[copy]]");
 Set( copy.Str() );
}

char* gHAttrList::Find (char* strAttr, bool doUnquote, gList& otherL)
{
 short iAttr;
 gParam* pAttr;
 gString s;
 unsigned len;

 for (iAttr=1; iAttr<=nAttr; iAttr++) {
     pAttr = pAttrAssign[ iAttr ];
     ASSERTION(pAttr!=nil,"pAttr!=nil");
     if ( gStrControl::Self().Match( strAttr, pAttr->Str(1), true ) ) {
	 s.Set( pAttr->Str(2) );
	 len = s.Length();
	 sThisVal = s;
	 if ( len<=2 ) return sThisVal.Str();
	 if ( doUnquote ) {
	     if ( (s[1]=='"' && s[len]=='"') ||
		  s[1]=='\'' && s[len]=='\'' ) {
		 sThisVal.CopyFromTo( s, 2, len-1 );
	     }
	 }
	 return sThisVal.Str();
     }
 }
 return nil;
}

int gHAttrList::thisValidateValue (gString& s)
{
 unsigned len=s.Length();
 if ( s[1]=='=' || s[len]=='=' ) return 32;
 // No further check is made, for now...
 return 0;
}
////////////////////////////////////////////////////////////

