#include "gHtmlHATypes.h"
#include "gHtmlCtrl.h"
#include "gstringext.h"
////////////////////////////////////////////////////////////
gHAType::gHAType (t_int16 aTypeId, char* strValue)
    : typeId( aTypeId ),
      sOrig( strValue ),
      sUnquoted( thisUnquotedStr( strValue ) )
{
 //Unquote( sOrig, sUnquoted );
}

gHAType::~gHAType ()
{
}

bool gHAType::IsOk ()
{
 return typeId>e_HAT_None && typeId<=e_HATp_EnumChoice;
}

t_int16 gHAType::GetTypeId ()
{
 return typeId;
}

eHAttrType gHAType::GetType ()
{
 ASSERTION(typeId<=e_HATp_EnumChoice,"GetType(1)");
 ASSERTION(typeId>e_HAT_None,"GetType(2)");
 return (eHAttrType)typeId;
}

void gHAType::Reset ()
{
 gList::Reset();
}

int gHAType::Unquote (gString& sIn, gString& sResult)
{
 // Return non-0 if any error exists
 unsigned len = sIn.Length();

 if ( len<2 ) return 0;
 if ( sIn[1]=='"' || sIn[1]=='\'' ) {
     if ( sIn[len]=='"' || sIn[len]=='\'') {
	 if ( sIn[1]!=sIn[len] ) return -1;
	 sResult.CopyFromTo( sIn, 2, len-1 );
     }
 }
 else {
     sResult = sIn;
 }
 return 0;
}

char* gHAType::thisUnquotedStr (char* str)
{
 gString s( str );
 Unquote( s, sAuxOut );
 return sAuxOut.Str();
}
////////////////////////////////////////////////////////////
gHATypeGString::gHATypeGString (t_int16 aTypeId, char* strValue)
    : gHAType( aTypeId, strValue )
{
}

gHATypeGString::~gHATypeGString ()
{
}
////////////////////////////////////////////////////////////
gHATypeStringSingle::gHATypeStringSingle (t_int16 aTypeId, char* strValue)
    : gHATypeGString( aTypeId, strValue )
{
}

gHATypeStringSingle::~gHATypeStringSingle ()
{
}

bool gHATypeStringSingle::IsOk ()
{
 if ( sUnquoted.Find( ' ' ) ) return false;
 return gHATypeGString::IsOk();
}
////////////////////////////////////////////////////////////
gHATypeURI::gHATypeURI (char* strValue)
    : gHATypeStringSingle( e_HATp_URI, strValue )
{
}

gHATypeURI::~gHATypeURI ()
{
}
////////////////////////////////////////////////////////////
gHATypeSimple::gHATypeSimple (t_int16 aTypeId, char* strValue, t_int16 val)
    : gHAType( aTypeId, strValue ),
      v( val )
{
}

gHATypeSimple::~gHATypeSimple ()
{
}

bool gHATypeSimple::IsOk ()
{
 return gHAType::IsOk() && v>=1;
}

t_int32 gHATypeSimple::GetInt32 ()
{
 DBGPRINT_WEBA2("{DBG: GetInt32as16bit()=%u}",v);
 return (t_int32)v;
}

gHATypeSimple32::gHATypeSimple32 (t_int16 aTypeId, char* strValue, t_int32 val)
    : gHAType( aTypeId, strValue ),
      vS32( val )
{
 if ( vS32==HATYPE_S32_INVALID_VAL ) {
     gStrControl::Self().ConvertToInt32( sUnquoted.Str(), vS32 );
 }
}

gHATypeSimple32::~gHATypeSimple32 ()
{
}

bool gHATypeSimple32::IsOk ()
{
 return vS32!=HATYPE_S32_INVALID_VAL;
}

t_int32 gHATypeSimple32::GetInt32 ()
{
 DBGPRINT_WEBA2("{DBG: GetInt32()=%ldLD}",(long)vS32);
 return vS32;
}

gHATypeSimpleU32::gHATypeSimpleU32 (t_int16 aTypeId, char* strValue, t_uint32 val, bool isHex)
    : gHAType( aTypeId, strValue ),
      vU32( val )
{
 char* str = sUnquoted.Str();
 if ( isHex ) {
     if ( ConvertHexStrToUInt32( str, vU32 )!=0 )
	 vU32 = HATYPE_U32_INVALID_VAL;
 }
 else {
     if ( gStrControl::Self().ConvertToUInt32( str, vU32 )!=0 )
	 vU32 = HATYPE_U32_INVALID_VAL;
 }
}

gHATypeSimpleU32::~gHATypeSimpleU32 ()
{
}

bool gHATypeSimpleU32::IsOk ()
{
 return vU32!=HATYPE_U32_INVALID_VAL;
}

t_int32 gHATypeSimpleU32::GetInt32 ()
{
 DBGPRINT_WEBA2("{DBG: GetInt32()=%luLU}",(unsigned long)vU32);
 if ( vU32>=(t_uint32)MAX_DUINT32 ) return -1;
 return (t_int32)vU32;
}

int gHATypeSimpleU32::ConvertHexStrToUInt32 (char* strValue, t_uint32& val)
{
 // Return 0 if o.k.
 if ( strValue==nil || strValue[0]==0 ) return -1;  // Not necessary: just for consistence
 return gStrControl::Self().ConvertHexToUInt32( strValue, e_DigConvAny, val );
}
////////////////////////////////////////////////////////////
gHATypeColor::gHATypeColor (char* strValue)
    : gHATypeSimpleU32( e_HATp_Color, strValue, 0, true )
{
 DBGPRINT_MIN("DBG0::: color### [%s]:%ld\n",strValue,(long)vU32);
 if ( vU32==HATYPE_U32_INVALID_VAL ) {
     gStrControl::Self().ConvertHexToUInt32( sUnquoted.Str()+1, e_DigConvAny, vU32 );
 }
 DBGPRINT_MIN("DBG1::: color### [%s]:%ld\n",strValue,(long)vU32);
}

gHATypeColor::~gHATypeColor ()
{
}

gHATypeLength::gHATypeLength (char* strValue)
    : gHATypeSimple32( e_HATp_Length, strValue, HATYPE_S32_INVALID_VAL ),
      percent( -1 )
{
 unsigned len;
 if ( vS32==HATYPE_S32_INVALID_VAL ) {
     sAuxOut = sUnquoted;
     len = sAuxOut.Length();
     sAuxOut.Trim();
     if ( sAuxOut[len]=='%' ) {
	 sAuxOut[len] = '\0';
	 if ( gStrControl::Self().ConvertToInt32( sAuxOut.Str(), vS32 )==0 )
	     percent = (int)vS32;
     }
 }
}

gHATypeLength::~gHATypeLength ()
{
}
////////////////////////////////////////////////////////////
gHATypeGenLst::gHATypeGenLst (t_int16 aTypeId, char* strValue, short codeAttrSingle)
    : gHAType( aTypeId, strValue ),
      hAttrSingle( codeAttrSingle )
{
}

gHATypeGenLst::~gHATypeGenLst ()
{
}
////////////////////////////////////////////////////////////
gHATypeLstVal::gHATypeLstVal (t_int16 aTypeId, char* strValue, short codeAttrSingle)
    : gHATypeGenLst( aTypeId, strValue, codeAttrSingle )
{
}

gHATypeLstVal::~gHATypeLstVal ()
{
}
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
gHATypeLstStr::gHATypeLstStr (t_int16 aTypeId, char* strValue, short codeAttrSingle)
    : gHATypeGenLst( aTypeId, strValue, codeAttrSingle )
{
}

gHATypeLstStr::~gHATypeLstStr ()
{
}
////////////////////////////////////////////////////////////
gHATypeCharsets::gHATypeCharsets (char* strValue, short codeAttrSingle)
    : gHATypeLstStr( codeAttrSingle==e_HAT_Single ? e_HATp_Charset : codeAttrSingle, strValue, codeAttrSingle )
{
}

gHATypeCharsets::~gHATypeCharsets ()
{
}

gHATypeContentTypes::gHATypeContentTypes (char* strValue, short codeAttrSingle)
    : gHATypeLstStr( codeAttrSingle==e_HAT_Single ? e_HATp_ContentType : codeAttrSingle, strValue, codeAttrSingle )
{
}

gHATypeContentTypes::~gHATypeContentTypes ()
{
}

gHATypeStyleSheet::gHATypeStyleSheet (char* strValue)
    : gHATypeLstStr( e_HATp_StyleSheet, strValue, e_HAT_Single )
{
}

gHATypeStyleSheet::~gHATypeStyleSheet ()
{
}
////////////////////////////////////////////////////////////
gHATypeMultiLengths::gHATypeMultiLengths (char* strValue, short codeAttrSingle)
    : gHATypeLstVal( codeAttrSingle==e_HAT_Single ? e_HATp_MultiLength : codeAttrSingle, strValue, codeAttrSingle )
{
}

gHATypeMultiLengths::~gHATypeMultiLengths ()
{
}
////////////////////////////////////////////////////////////
gHATypeCDATA::gHATypeCDATA (char* strValue)
    : gHAType( e_HAT_CDATA, strValue )
{
}

gHATypeCDATA::~gHATypeCDATA ()
{
}

gHATypeID::gHATypeID (char* strValue)
    : gHAType( e_HAT_ID, strValue )
{
}

gHATypeID::~gHATypeID ()
{
}

gHATypeIDREF::gHATypeIDREF (char* strValue)
    : gHAType( e_HAT_IDREF, strValue )
{
}

gHATypeIDREF::~gHATypeIDREF ()
{
}

gHATypeIDREFS::gHATypeIDREFS (char* strValue)
    : gHAType( e_HAT_IDREFS, strValue )
{
}

gHATypeIDREFS::~gHATypeIDREFS ()
{
}

gHATypeNAME::gHATypeNAME (char* strValue)
    : gHAType( e_HAT_NAME, strValue )
{
}

gHATypeNAME::~gHATypeNAME ()
{
}

gHATypeNUMBER::gHATypeNUMBER (char* strValue)
    : gHATypeSimple32( e_HAT_NUMBER, strValue, HATYPE_S32_INVALID_VAL )
{
}

gHATypeNUMBER::~gHATypeNUMBER ()
{
}
////////////////////////////////////////////////////////////
gHATypePixels::gHATypePixels (char* strValue)
    : gHATypeSimple32( e_HATp_Pixels, strValue, HATYPE_S32_INVALID_VAL )
{
}

gHATypePixels::~gHATypePixels ()
{
}
////////////////////////////////////////////////////////////
gHATypeText::gHATypeText (char* strValue)
    : gHATypeGString( e_HATp_Text, strValue )
{
 DBGPRINT_MIN("DBG: gHATypeText(%s)\n",strValue);
}

gHATypeText::~gHATypeText ()
{
}
////////////////////////////////////////////////////////////
gHATypeFrameTarget::gHATypeFrameTarget (char* strValue)
    : gHATypeStringSingle( e_HATp_FrameTarget, strValue ),
      isBlank( sUnquoted.Match( "_blank", true ) )
{
}

gHATypeFrameTarget::~gHATypeFrameTarget ()
{
}
////////////////////////////////////////////////////////////
gHATypeEnum::gHATypeEnum (t_int16 aTypeId, char* strValue, t_int16 val)
    : gHATypeSimple( aTypeId, strValue, val )
{
 if ( v>=1 ) sUnquoted.DownString();
}

gHATypeEnum::gHATypeEnum (t_int16 aTypeId, char* strValue, gSmartList& lEnumStr)
    : gHATypeSimple( aTypeId, strValue, -1 )
{
 unsigned pos = lEnumStr.MatchWhere( sUnquoted.Str() );
 v = (t_int16)pos;
 // Replace if o.k. by downcase (e.g. 'Center'(strValue)==>Center=>(sUnquoted)=>>center)
 if ( v>=1 ) sUnquoted.DownString();
 DBGPRINT_WEBA2("DBG: gHATypeEnum(%s|%s)=%d\n",strValue,GetStr(),v);
}

gHATypeEnum::~gHATypeEnum ()
{
}
////////////////////////////////////////////////////////////

