#ifndef gHTML_HATYPES_X_H
#define gHTML_HATYPES_X_H

#include "glistext.h"

#define e_HAT_Single 0

#define HATYPE_S32_INVALID_VAL MAX_DLINT32
#define HATYPE_U32_INVALID_VAL MAX_DUINT32
////////////////////////////////////////////////////////////
enum eHAttrType {
    e_HAT_None,
    e_HAT_CDATA,
    e_HAT_ID,
    e_HAT_IDREF,
    e_HAT_IDREFS,
    e_HAT_NAME,
    e_HAT_NUMBER,
    e_HATp_CAlign,
    e_HATp_Character,
    e_HATp_Charset,
    e_HATp_Charsets,
    e_HATp_Color,
    e_HATp_ContentType,
    e_HATp_ContentTypes,
    e_HATp_Coords,
    e_HATp_DateTime,
    e_HATp_FrameTarget,
    e_HATp_HtmlVersion,
    e_HATp_IAlign,
    e_HATp_InputType,
    e_HATp_LAlign,
    e_HATp_LanguageCode,
    e_HATp_Length,
    e_HATp_LinkTypes,
    e_HATp_LIStyle,
    e_HATp_MediaDesc,
    e_HATp_MultiLength,
    e_HATp_MultiLengths,
    e_HATp_OLStyle,
    e_HATp_Pixels,
    e_HATp_Scope,
    e_HATp_Script,
    e_HATp_Shape,
    e_HATp_StyleSheet,
    e_HATp_TAlign,
    e_HATp_Text,
    e_HATp_TFrame,
    e_HATp_TRules,
    e_HATp_ULStyle,
    e_HATp_URI,
    e_HATp_EnumChoice = 254 // e.g. "left|center|right"
};
////////////////////////////////////////////////////////////
struct sHAttrType {
    t_int16 typeFamily;
    char* strType;
    eHAttrType hAttrType;
    short hAttrSingle;  // Refer to basic-type, when is a set of several
};
////////////////////////////////////////////////////////////
class gHAType : public gList {
public:
    virtual ~gHAType () ;

    // Get methods
    virtual bool IsOk () ;

    virtual t_int16 GetTypeId () ;
    virtual eHAttrType GetType () ;

    virtual char* GetStr () {
	return GetAttrStr();
    }

    virtual char* GetAttrStr () {
	return sUnquoted.Str();
    }

    virtual t_int32 GetInt32 () {
	return 0;
    }

    virtual bool IsValue () {
	return false;
    }

    virtual bool IsValueAbsolute () {
	return false;
    }

    // Set methods
    virtual void Reset () ;

    int Unquote (gString& sIn, gString& sResult) ;

protected:
    gHAType (t_int16 aTypeId, char* strValue) ;
    t_int16 typeId;
    gString sAuxOut;  // Must be declared before any string
    gString sOrig;
    gString sUnquoted;

    char* thisUnquotedStr (char* str) ;

private:
    // Operators,empty
    gHAType (gHAType& ) ; //empty
    gHAType& operator= (gHAType& ) ; //empty
};
////////////////////////////////////////////////////////////
class gHATypeGString : public gHAType {
public:
    virtual ~gHATypeGString () ;

protected:
    gHATypeGString (t_int16 aTypeId, char* strValue) ;

private:
    // Operators,empty
    gHATypeGString (gHATypeGString& ) ; //empty
    gHATypeGString& operator= (gHATypeGString& ) ; //empty
};
////////////////////////////////////////////////////////////
class gHATypeStringSingle : public gHATypeGString {
public:
    virtual ~gHATypeStringSingle () ;

    // Get methods
    virtual bool IsOk () ;

protected:
    gHATypeStringSingle (t_int16 aTypeId, char* strValue) ;

private:
    // Operators,empty
    gHATypeStringSingle (gHATypeStringSingle& ) ; //empty
    gHATypeStringSingle& operator= (gHATypeStringSingle& ) ; //empty
};
////////////////////////////////////////////////////////////
class gHATypeURI : public gHATypeStringSingle {
public:
    gHATypeURI (char* strValue) ;
    virtual ~gHATypeURI () ;

private:
    // Operators,empty
    gHATypeURI (gHATypeURI& ) ; //empty
    gHATypeURI& operator= (gHATypeURI& ) ; //empty
};
////////////////////////////////////////////////////////////
class gHATypeSimple : public gHAType {
public:
    virtual ~gHATypeSimple () ;

    // Public data-members
    t_int16 v;

    // Get methods
    virtual bool IsOk () ;

    virtual t_int32 GetInt32 () ;

    virtual bool IsValue () {
	return true;
    }

protected:
    gHATypeSimple (t_int16 aTypeId, char* strValue, t_int16 val) ;

private:
    // Operators,empty
    gHATypeSimple (gHATypeSimple& ) ; //empty
    gHATypeSimple& operator= (gHATypeSimple& ) ; //empty
};
////////////////////////////////////////////////////////////
class gHATypeSimple32 : public gHAType {
public:
    virtual ~gHATypeSimple32 () ;

    // Public data-members
    t_int32 vS32;

    // Get methods
    virtual bool IsOk () ;

    virtual t_int32 GetInt32 () ;

    virtual bool IsValue () {
	return true;
    }

protected:
    gHATypeSimple32 (t_int16 aTypeId, char* strValue, t_int32 val) ;

private:
    // Operators,empty
    gHATypeSimple32 (gHATypeSimple32& ) ; //empty
    gHATypeSimple32& operator= (gHATypeSimple32& ) ; //empty
};

class gHATypeSimpleU32 : public gHAType {
public:
    virtual ~gHATypeSimpleU32 () ;

    // Public data-members
    t_uint32 vU32;

    // Get methods
    virtual bool IsOk () ;

    virtual t_int32 GetInt32 () ;

    virtual bool IsValue () {
	return true;
    }

    int ConvertHexStrToUInt32 (char* strValue, t_uint32& val) ;

protected:
    gHATypeSimpleU32 (t_int16 aTypeId, char* strValue, t_uint32 val, bool isHex) ;

private:
    // Operators,empty
    gHATypeSimpleU32 (gHATypeSimpleU32& ) ; //empty
    gHATypeSimpleU32& operator= (gHATypeSimpleU32& ) ; //empty
};
////////////////////////////////////////////////////////////
class gHATypeColor: public gHATypeSimpleU32 {
public:
    gHATypeColor (char* strValue) ;
    virtual ~gHATypeColor () ;

private:
    // Operators,empty
    gHATypeColor (gHATypeColor& ) ; //empty
    gHATypeColor& operator= (gHATypeColor& ) ; //empty
};

class gHATypeLength : public gHATypeSimple32 {
public:
    gHATypeLength (char* strValue) ;
    virtual ~gHATypeLength () ;

    // Public data-members
    int percent;  // -1 stands for absolute value

    virtual bool IsValueAbsolute () {
	return percent==-1;
    }

private:
    // Operators,empty
    gHATypeLength (gHATypeLength& ) ; //empty
    gHATypeLength& operator= (gHATypeLength& ) ; //empty
};
////////////////////////////////////////////////////////////
class gHATypeGenLst : public gHAType {
public:
    virtual ~gHATypeGenLst () ;

    // Public data-members
    short hAttrSingle;

protected:
    gHATypeGenLst (t_int16 aTypeId, char* strValue, short codeAttrSingle) ;

private:
    // Operators,empty
    gHATypeGenLst (gHATypeGenLst& ) ; //empty
    gHATypeGenLst& operator= (gHATypeGenLst& ) ; //empty
};
////////////////////////////////////////////////////////////
class gHATypeLstStr : public gHATypeGenLst {
public:
    virtual ~gHATypeLstStr () ;

protected:
    gHATypeLstStr (t_int16 aTypeId, char* strValue, short codeAttrSingle) ;

private:
    // Operators,empty
    gHATypeLstStr (gHATypeLstStr& ) ; //empty
    gHATypeLstStr& operator= (gHATypeLstStr& ) ; //empty
};
////////////////////////////////////////////////////////////
class gHATypeLstVal : public gHATypeGenLst {
public:
    virtual ~gHATypeLstVal () ;

    virtual bool IsValue () {
	return hAttrSingle==e_HAT_Single;
    }

protected:
    gHATypeLstVal (t_int16 aTypeId, char* strValue, short codeAttrSingle) ;

private:
    // Operators,empty
    gHATypeLstVal (gHATypeLstVal& ) ; //empty
    gHATypeLstVal& operator= (gHATypeLstVal& ) ; //empty
};
////////////////////////////////////////////////////////////
class gHATypeCharsets : public gHATypeLstStr {
public:
    gHATypeCharsets (char* strValue, short codeAttrSingle) ;
    virtual ~gHATypeCharsets () ;

private:
    // Operators,empty
    gHATypeCharsets (gHATypeCharsets& ) ; //empty
    gHATypeCharsets& operator= (gHATypeCharsets& ) ; //empty
};

class gHATypeContentTypes : public gHATypeLstStr {
public:
    gHATypeContentTypes (char* strValue, short codeAttrSingle) ;
    virtual ~gHATypeContentTypes () ;

private:
    // Operators,empty
    gHATypeContentTypes (gHATypeContentTypes& ) ; //empty
    gHATypeContentTypes& operator= (gHATypeContentTypes& ) ; //empty
};

class gHATypeStyleSheet : public gHATypeLstStr {
public:
    gHATypeStyleSheet (char* strValue) ;
    virtual ~gHATypeStyleSheet () ;

private:
    // Operators,empty
    gHATypeStyleSheet (gHATypeStyleSheet& ) ; //empty
    gHATypeStyleSheet& operator= (gHATypeStyleSheet& ) ; //empty
};
////////////////////////////////////////////////////////////
class gHATypeMultiLengths : public gHATypeLstVal {
public:
    gHATypeMultiLengths (char* strValue, short codeAttrSingle) ;
    virtual ~gHATypeMultiLengths () ;

private:
    // Operators,empty
    gHATypeMultiLengths (gHATypeMultiLengths& ) ; //empty
    gHATypeMultiLengths& operator= (gHATypeMultiLengths& ) ; //empty
};
////////////////////////////////////////////////////////////
class gHATypeCDATA : public gHAType {
public:
    gHATypeCDATA (char* strValue) ;
    virtual ~gHATypeCDATA () ;

private:
    // Operators,empty
    gHATypeCDATA (gHATypeCDATA& ) ; //empty
    gHATypeCDATA& operator= (gHATypeCDATA& ) ; //empty
};

class gHATypeID : public gHAType {
public:
    gHATypeID (char* strValue) ;
    virtual ~gHATypeID () ;

private:
    // Operators,empty
    gHATypeID (gHATypeID& ) ; //empty
    gHATypeID& operator= (gHATypeID& ) ; //empty
};

class gHATypeIDREF : public gHAType {
public:
    gHATypeIDREF (char* strValue) ;
    virtual ~gHATypeIDREF () ;

private:
    // Operators,empty
    gHATypeIDREF (gHATypeIDREF& ) ; //empty
    gHATypeIDREF& operator= (gHATypeIDREF& ) ; //empty
};

class gHATypeIDREFS : public gHAType {
public:
    gHATypeIDREFS (char* strValue) ;
    virtual ~gHATypeIDREFS () ;

private:
    // Operators,empty
    gHATypeIDREFS (gHATypeIDREFS& ) ; //empty
    gHATypeIDREFS& operator= (gHATypeIDREFS& ) ; //empty
};

class gHATypeNAME : public gHAType {
public:
    gHATypeNAME (char* strValue) ;
    virtual ~gHATypeNAME () ;

private:
    // Operators,empty
    gHATypeNAME (gHATypeNAME& ) ; //empty
    gHATypeNAME& operator= (gHATypeNAME& ) ; //empty
};

class gHATypeNUMBER : public gHATypeSimple32 {
public:
    gHATypeNUMBER (char* strValue) ;
    virtual ~gHATypeNUMBER () ;

private:
    // Operators,empty
    gHATypeNUMBER (gHATypeNUMBER& ) ; //empty
    gHATypeNUMBER& operator= (gHATypeNUMBER& ) ; //empty
};
////////////////////////////////////////////////////////////
class gHATypePixels : public gHATypeSimple32 {
public:
    gHATypePixels (char* strValue) ;
    virtual ~gHATypePixels () ;

private:
    // Operators,empty
    gHATypePixels (gHATypePixels& ) ; //empty
    gHATypePixels& operator= (gHATypePixels& ) ; //empty
};
////////////////////////////////////////////////////////////
class gHATypeText : public gHATypeGString {
public:
    gHATypeText (char* strValue) ;
    virtual ~gHATypeText () ;

private:
    // Operators,empty
    gHATypeText (gHATypeText& ) ; //empty
    gHATypeText& operator= (gHATypeText& ) ; //empty
};

class gHATypeFrameTarget : public gHATypeStringSingle {
public:
    gHATypeFrameTarget (char* strValue) ;
    virtual ~gHATypeFrameTarget () ;

    bool isBlank;

private:
    // Operators,empty
    gHATypeFrameTarget (gHATypeFrameTarget& ) ; //empty
    gHATypeFrameTarget& operator= (gHATypeFrameTarget& ) ; //empty
};
////////////////////////////////////////////////////////////
class gHATypeEnum : public gHATypeSimple {
public:
    gHATypeEnum (t_int16 aTypeId, char* strValue, t_int16 val=-1) ;
    gHATypeEnum (t_int16 aTypeId, char* strValue, gSmartList& lEnumStr) ;
    virtual ~gHATypeEnum () ;

private:
    // Operators,empty
    gHATypeEnum (gHATypeEnum& ) ; //empty
    gHATypeEnum& operator= (gHATypeEnum& ) ; //empty
};
////////////////////////////////////////////////////////////
#endif //gHTML_HATYPES_X_H

