#include "gHtmlHAttr.h"
#include "gHtmlHATypes.h"
#include "gHtmlCtrl.h"
////////////////////////////////////////////////////////////
gHAttrValue::gHAttrValue (char* strL, gHAType* pValue)
    : gString( strL ),
      valuePtr( pValue )
{
}

gHAttrValue::~gHAttrValue ()
{
 delete valuePtr;
}

gHAType& gHAttrValue::GetValue ()
{
 ASSERTION(valuePtr!=nil,"valuePtr!=nil");
 return *valuePtr;
}

gHAType* gHAttrValue::GetValuePtr ()
{
 return valuePtr;
}
////////////////////////////////////////////////////////////
gHAttrs::gHAttrs ()
{
}

gHAttrs::~gHAttrs ()
{
 Reset();
}

bool gHAttrs::IsOk ()
{
 // Returns true if all attributes contained are o.k.
 unsigned i, n;
 for (i=1, n=N(); i<=n; i++) {
     gStorage* pObjx = GetObjectPtr( i );
     ASSERTION(pObjx!=nil,"pObjx!=nil");
     if ( pObjx->IsOk()==false ) return false;
 }
 // Note for the code above:
 // Since usually BuildFromLRLists is done with addOnError=false, no bad attributes exist.
 return gList::IsOk();
}

gHAttrValue& gHAttrs::GetAttrValue (unsigned idx)
{
 gHAttrValue* aLR = GetAttrValuePtr( idx );
 ASSERTION(aLR!=nil,"aLR!=nil");
 return *aLR;
}

gHAttrValue* gHAttrs::GetAttrValuePtr (unsigned idx)
{
 gHAttrValue* aLR = (gHAttrValue*)GetObjectPtr( idx );
 return aLR;
}

gHAType& gHAttrs::GetAttr (unsigned idx)
{
 gHAType* pValue = GetAttrPtr( idx );
 ASSERTION(pValue!=nil,"pValue!=nil");
 return *pValue;
}

gHAType* gHAttrs::GetAttrPtr (unsigned idx)
{
 if ( idx<1 || idx>N() ) return nil;
 // Note: indexes are also validated by gList::GetObjectPtr.
 gStorage* pObx = GetObjectPtr( idx );
 if ( pObx==nil ) return nil;
 gHAType* pValue = ((gHAttrValue*)pObx)->GetValuePtr();
 ASSERTION(pValue!=nil,"pValue!=nil");
 return pValue;
}

sHAttrType* gHAttrs::TypeName2HAttrType (char* strType)
{
 bool isBasicType;  // Those at lHAttrTypes with typeFamily=1
 t_int16 idxType, typeFamily, nTypes;
 t_uchar lastChr;
 gHtmlParser* pParser = gHtmlCtrl::Self().ParserPtr();
 sHAttrType* pAttrType;

 if ( strType==nil || strType[0]==0 ) return nil;

 isBasicType = strType[0]!='%' && strType[0]!='(';
 strType += (isBasicType==false);  // Skip '%'or'('

 gString s( strType );
 if ( (lastChr = s[ s.Length() ])==';' || lastChr==')' ) s[ s.Length() ] = '\0';

 for (idxType=1, nTypes=pParser->GetNAttrType(); idxType<=nTypes++; idxType++) {
     pAttrType = pParser->GetAttrType( idxType );
     DBGPRINT_MIN("TypeName2HAttrType(%s,%s): %d/%d\n",strType,s.Str(),idxType,nTypes);
     ASSERTION(pAttrType!=nil,"pAttrType!=nil");
     typeFamily = pAttrType->typeFamily;
     if ( isBasicType ) {
	 if ( typeFamily!=1 ) return nil;
	 if ( s.Find( pAttrType->strType )==1 ) return pAttrType;
     }
     else {
	 if ( s.Match( pAttrType->strType ) ) return pAttrType;
     }
 }

 return nil;
}

int gHAttrs::TypeEnumToSmartL (sHAttrType* pAttrType, gSmartList& lEnumResult)
{
 ASSERTION(pAttrType!=nil,"pAttrType!=nil");
 lEnumResult.Delete();
 lEnumResult.SetSmartChr( '|' );
 lEnumResult.SetCaseSense( false );  // Non-case-sensitive!
 // E.g. "left|center|right" copyied into the list (3 elements)
 return lEnumResult.AddFromStr( pAttrType->strType );
}

void gHAttrs::Reset ()
{
 gList::Reset();
 Delete();
}

int gHAttrs::BuildFromLRLists (char* strTag, gList& lLVal, gList& lRVal, bool addOnError, unsigned iLine)
{
 unsigned i, n;

 gHtmlParser* pParser = gHtmlCtrl::Self().ParserPtr();
 sAttrRefer* lstAttrRef = pParser->GetAttrRef();
 sAttrDef* dAttr;
 sHAttrType* pAttrType;
 t_int16 idx, uniqIdx;
 char* strL;
 char* strR;
 FILE* fErr = gHtmlCtrl::Self().hLog.fRepErr;

 ASSERTION(lLVal.N()==lRVal.N(),"lLVal.N()==lRVal.N()");

 for (i=1, n=lLVal.N(); i<=n; i++) {
     strL = lLVal.Str( i );
     strR = lRVal.Str( i );
     if ( strL==nil || strL[0]==0 ) continue;
     idx = lstAttrRef->FindAttr( strL, strTag, uniqIdx );
     ASSERTION(idx>=0,"idx>=0");
     ASSERTION(uniqIdx>=0,"uniqIdx>=0");
     dAttr = pParser->GetAttrDef( idx );
     pAttrType = TypeName2HAttrType( dAttr->strType );
     if ( pAttrType==nil ) {
	 HTML_LOG( stderr, LOG_ERROR, "Internal-error: Unable to cross-check attribute: '%s' for tag: %s\n",dAttr->strType,strTag);
	 continue;
     }

     DBGPRINT_WEBA1("DBG: iLine:%u strL:%s=%s (TypeId:%s=%d)!\n",iLine,strL,strR,pAttrType->strType,pAttrType->hAttrType);

     gHAType* newValue = NewType( strR, pAttrType );

     // DEBUG only: TODO: cut
     if ( newValue==nil ) {
	 DBGPRINT_WEBA1("DBG: <<TODO:attr:%s>>\n",strL);
	 continue;
     }
     ASSERTION(newValue!=nil,"newValue!=nil");

     //if ( pAttrType->hAttrType==e_HATp_EnumChoice ) printf("DBG: enum(%s):%d:OK?%c\n",newValue->GetStr(),((gHATypeSimple*)newValue)->v,ISyORn(newValue->IsOk()));

     if ( newValue->IsOk()==false ) {
	 HTML_LOG( fErr, LOG_NOTICE, "Line %u: Invalid value (%s) for attribute: %s, of type %s\n", iLine, strR, strL, dAttr->strType );
	 if ( addOnError==false ) {
	     delete newValue;
	     newValue = nil;
	 }
     }

     if ( newValue!=nil ) {
	 gHAttrValue* newLR = new gHAttrValue( strL, newValue );
	 ASSERTION(newLR!=nil,"newLR!=nil");
	 AppendObject( newLR );
     }
 }

 return 0;
}

gHAType* gHAttrs::NewType (char* strValue, sHAttrType* pAttrType)
{
 eHAttrType hAttrType;
 gSmartList lEnumResult;
 short hAttrSingle;

 ASSERTION(strValue!=nil,"strValue!=nil");
 if ( pAttrType==nil ) return nil;

 hAttrType = pAttrType->hAttrType;
 hAttrSingle = pAttrType->hAttrSingle;

 switch ( pAttrType->typeFamily ) {
 case 1:  // Basic attribute-types
     switch ( hAttrType ) {
     case e_HAT_CDATA:
	 return new gHATypeCDATA( strValue );
     case e_HAT_ID:
	 return new gHATypeID( strValue );
     case e_HAT_IDREF:
	 return new gHATypeIDREF( strValue );
     case e_HAT_IDREFS:
	 return new gHATypeIDREFS( strValue );
     case e_HAT_NAME:
	 return new gHATypeNAME( strValue );
     case e_HAT_NUMBER:
	 return new gHATypeNUMBER( strValue );
	 break;
     default:
	 ASSERTION_FALSE("NewType(Family1)");
     }

 case 0:  // Compound attribute-types
     switch ( hAttrType ) {
     case e_HATp_CAlign:
	 break;
     case e_HATp_Character:
	 break;
     case e_HATp_Charset:
     case e_HATp_Charsets:
	 return new gHATypeCharsets( strValue, hAttrSingle );
     case e_HATp_Color:
	 return new gHATypeColor( strValue );
     case e_HATp_ContentType:
     case e_HATp_ContentTypes:
	 return new gHATypeContentTypes( strValue, hAttrSingle );
     case e_HATp_Coords:
     case e_HATp_DateTime:
	 break;
     case e_HATp_FrameTarget:
	 return new gHATypeFrameTarget( strValue );
     case e_HATp_HtmlVersion:
	 break;
     case e_HATp_IAlign:
	 break;
     case e_HATp_InputType:
	 break;
     case e_HATp_LAlign:
	 break;
     case e_HATp_LanguageCode:
	 break;
     case e_HATp_Length:
	 return new gHATypeLength( strValue );
     case e_HATp_LinkTypes:
	 break;
     case e_HATp_LIStyle:
	 break;
     case e_HATp_MediaDesc:
	 break;
     case e_HATp_MultiLength:
     case e_HATp_MultiLengths:
	 return new gHATypeMultiLengths( strValue, hAttrSingle );
     case e_HATp_OLStyle:
	 break;
     case e_HATp_Pixels:
	 return new gHATypePixels( strValue );
     case e_HATp_Scope:
	 break;
     case e_HATp_Script:
	 break;
     case e_HATp_Shape:
	 break;
     case e_HATp_StyleSheet:
	 return new gHATypeStyleSheet( strValue );
     case e_HATp_TAlign:
	 break;
     case e_HATp_Text:
	 return new gHATypeText( strValue );
     case e_HATp_TFrame:
	 break;
     case e_HATp_TRules:
	 break;
     case e_HATp_ULStyle:
	 break;
     case e_HATp_URI:
	 return new gHATypeURI( strValue );
     default:
	 ASSERTION_FALSE("NewType(Family0)");
     }//end inner-CASE
     break;

 case 10:  // Enum attribute-types
     TypeEnumToSmartL( pAttrType, lEnumResult );
     return new gHATypeEnum( hAttrType, strValue, lEnumResult );
 default:
     break;
 };//end outer-CASE

 return nil;
}

void gHAttrs::Show (bool doShowAll)
{
 unsigned i, n=N();
 char* str;
 const char* strNL=doShowAll?"\n":"\0";
 gHAType* pValue;

 DBGPRINT_WEBA1("DBG: gHAttrs::Show%s (N=%u)\n",doShowAll?":ALL":"\0",n);

 for (i=1; i<=n; i++) {
     pValue = GetAttrPtr( i );
     str = pValue->GetStr();
     DBGPRINT_MIN("typeId%d",pValue->GetTypeId());
     if ( pValue->IsValue() )
	 printf("[%s|%ld]",str,(long)pValue->GetInt32());
     else
	 printf("[%s]",str);
     printf("%s",i<n?" ":strNL);
 }
}
////////////////////////////////////////////////////////////

