// gHtmlOpt, part of gUnweb

#include <string.h>
#include "gHtmlOpt.h"
#include "gHtmlCtrl.h"
#include "gstringext.h"
////////////////////////////////////////////////////////////
char* sHOutFormat::GetBeginLineStr (t_uint16 state, t_uint16 rLine)
{
 bool isFirst = rLine<=1;
 if ( state>=e_HS_Last ) return nil;
 if ( isFirst==true || state==e_HS_HeadAfter ) state = e_HS_Last; //empty string
 return sBeginLine[state].Str();
}

void sHOutFormat::Prepare ()
{
 t_uint16 state;

 // sBeginLine[e_HS_Last] is always empty
 for (state=e_HS_Start; state<e_HS_Last; state++) {
     sBeginLine[state].SetEmpty();
     switch ( state ) {
     case e_HS_Start:
     case e_HS_StartAfter:
     case e_HS_Html:
     case e_HS_End:
	 ;
	 break;
     default:
	 if ( oIndentLevel1.IsOn() ) {
	     SetIndentString( cIndentLevel, 4, sBeginLine[state] );
	 }
	 break;
     }//END case
 }
}

int sHOutFormat::SetIndentString (char cIndent, t_uint16 nBlanks, gString& sResult)
{
 t_uint16 uIdx;

 if ( cIndent=='\t' ) {
     sResult.Add( '\t' );
     return 0;
 }
 for (uIdx=0; uIdx<nBlanks; uIdx++)
     sResult.Add( cIndent );

 return 0;
}

void sHOutFormat::CopyHOutFormat (sHOutFormat& copy)
{
 cIndentLevel = copy.cIndentLevel;
 oIndentLevel1.SetOn( copy.oIndentLevel1.IsOn() );
 Prepare();
}
////////////////////////////////////////////////////////////
int sAttrsTagNorm::CopyDefaultAttrNorms (t_int16 idTag0, t_int16 idTag1, bool doDefault)
{
 t_int16 idxNorm;
 t_int16 idTag;
 sAttrNorm* pNorm;
 gSmartList* pList;

 Reset();
 if ( idTag0>idTag1 ) return -1;
 lstNorms = new gSmartList[ idTag1+1 ];
 ASSERTION(lstNorms!=nil,"lstNorms!=nil");

 pNorm = gHtmlCtrl::Self().ParserPtr()->GetDefaultAttrNorm( 0 );
 pList = pNorm->pAttrSeq;
 if ( pList!=nil ) {
     // There are norms for all-tags, here stated:
     allTagNorm.CopyList( *pList );
 }

 for (idxNorm=1;
      (pNorm = gHtmlCtrl::Self().ParserPtr()->GetDefaultAttrNorm( idxNorm ))!=nil;
      idxNorm++) {
     idTag = pNorm->idTag;
     if ( idTag<idTag0 || idTag>idTag1 ) continue;
     pList = pNorm->pAttrSeq;
     if ( pList==nil ) continue;
     lstNorms[ idTag ].CopyList( *pList );
 }

 return 0;
}
////////////////////////////////////////////////////////////
gHInOpt::gHInOpt ()
    : vLineLTE( -1 ),
      vLineGTE( -1 ),
      vLineSupLTE( -1 ),
      vLineSupGTE( -1 ),
      hasLineAddRestriction( false )
{
}

gHInOpt::~gHInOpt ()
{
}

bool gHInOpt::HasToSkip (int iLine)
{
 // Check if LTE or GTE rules apply to 'iLine'
 if ( (vLineSupLTE>-1 && iLine<=vLineSupLTE) ||
     (vLineSupGTE>-1 && iLine>=vLineSupGTE) ) return true;

 // Now check if we want to skip individual lines
 if ( lLineSuppress.FindInt( iLine )>0 ) return true;

 return thisHasToShow( iLine )==false;
}

bool gHInOpt::ConvertToInt (char* str, int& v)
{
 t_uint32 vRes;
 lastOpError = gStrControl::Self().ConvertToUInt32( str, vRes )!=0;
 if ( lastOpError!=0 ) return false;
 v = (int)vRes;
 return true;
}

bool gHInOpt::SetFromString (char* str)
{
 char c;
 char* s = str;
 char* sAux;
 unsigned n;
 int v, vEnd;
 bool doExclude;

 DBGPRINT_MIN("gHInOpt::SetFromString: %s\n",str);

 if ( s==nil ) return true;

 // E.g.:
 //  @12     Suppress line 12
 //  @12 @14 Suppress line 12 and 14
 //  @12-14  Suppress line 12 to 14
 //  @12-    Suppress line 12 till the end
 //  12-15   Shows line 12 till 15, including 12 and 15.
 for ( ; (c = *s)!=0; ) {
     if ( c==' ' ) {
	 s++;
	 continue;
     }
     doExclude = c=='@';
     if ( doExclude ) s++;
     c = *s;
     if ( c==0 ) return false;
     gParam aParam( s, " ", gParam::e_StopSplitOnFirst );
     sAux = aParam.Str( 1 );
     gParam aParamSup( sAux, "-" );
     n = aParamSup.N();
     sAux = aParamSup.Str( 1 );
     if ( n<1 ) return false;
     if ( ConvertToInt( sAux, v )==false ) return false;
     if ( n==1 ) {
	 gString sOne( s );
	 if ( sOne.FindBack('-')>0 ) {
	     // e.g.: @12-
	     if ( thisAddLineSupGTE( v, doExclude )!=0 ) return false;
	 }
	 else {
	     if ( thisAddLineSup( v, doExclude )!=0 ) return false;
	 }
     }
     else {
	 if ( n!=2 ) return false;
	 if ( ConvertToInt( aParamSup.Str(2), vEnd )==false ) return false;
	 if ( v>vEnd ) return false;
	 for ( ; v<=vEnd; v++) {
	     thisAddLineSup( v, doExclude );
	 }
     }
     for ( ; (c = *s)!=0 && c!=' '; ) s++;
 }//end FOR (str...)

 DBGPRINT_MIN("gHInOpt::SetFromString: %s (END)\n",str);

 return true;
}

int gHInOpt::thisAddLineSup (int v, bool doExclude)
{
 if ( v<1 ) return -1;
 if ( doExclude ) {
     lLineSuppress.Append( v );
 }
 else {
     lLineAdd.Append( v );
     hasLineAddRestriction = true;
 }
 return 0;
}

int gHInOpt::thisAddLineSupGTE (int v, bool doExclude)
{
 // e.g.: @12- (Greater Than or Equal... than 12)
 if ( doExclude ) {
     // returns error (-1) if already specified GTE
     if ( vLineSupGTE>-1 ) return -1;
     vLineSupGTE = v;
 }
 else {
     if ( vLineGTE>-1 ) return -1;
     vLineGTE = v;
     hasLineAddRestriction = true;
 }
 return 0;
}

bool gHInOpt::thisHasToShow (int iLine)
{
 // Returns true if the line is to be shown
 if ( hasLineAddRestriction==false ) return true;

 if ( vLineGTE>-1 && iLine>=vLineGTE ) return true;
 return lLineAdd.FindInt( iLine )>0;
}
////////////////////////////////////////////////////////////
void sTidyArray::Copy (sTidyArray& copy)
{
 unsigned i;
 Release();
 n = copy.n;
 doTidyAll = copy.doTidyAll;
 if ( copy.lst==nil ) return;
 lst = new sTidyElem[ n+1 ];
 ASSERTION(lst!=nil,"lst!=nil");
 for (i=1; i<=n; i++) {
     lst[i].Copy( copy.lst[i] );
 }
}
////////////////////////////////////////////////////////////
gHOutOpt::gHOutOpt (FILE* afRepErr)
    : fRepErr( afRepErr ),
      myConfig( nil )
{
}

gHOutOpt::~gHOutOpt ()
{
 delete myConfig;
}

gConfig& gHOutOpt::Config ()
{
 ASSERTION(myConfig!=nil,"myConfig!=nil");
 return *myConfig;
}

bool gHOutOpt::HasToSkip (char* sFullTagStr)
{
 unsigned pos;
 if ( sFullTagStr==nil ) return false;
 gString s( sFullTagStr );
 s.UpString();
 if ( lSkipFSt.Match( s.Str() )>0 ) return true;
 pos = s.Find( " " );
 if ( pos>0 ) s[ pos ] = '\0';
 return lSkipStt.Match( s.Str() )>0;
}

void gHOutOpt::CopyHOutOpt (gHOutOpt& copy)
{
 fRepErr = copy.fRepErr;
 // myConfig must not be copied
 aTidy.Copy( copy.aTidy );
 lSkipTag.CopyList( copy.lSkipTag );
 lSkipFSt.CopyList( copy.lSkipFSt );
 lSkipStt.CopyList( copy.lSkipStt );
}

int gHOutOpt::CopyDefaultAttrs ()
{
 return CopyDefaultAttrNorms( 0, gHtmlCtrl::Self().ParserPtr()->GetTagMaxId(), true );
}

int gHOutOpt::CopyDefaultAttrNorms (t_int16 idTag0, t_int16 idTag1, bool doDefault)
{
 return lAttrsNorm.CopyDefaultAttrNorms( idTag0, idTag1, doDefault );
}

int gHOutOpt::ConfigFromFile (char* fName)
{
 ASSERTION(fName!=nil,"fName!=nil");
 delete myConfig;
 myConfig = new gConfig( gConfig::e_ConfSectionStrict, fName, true );
 ASSERTION(myConfig!=nil,"myConfig!=nil");
 lastOpError = myConfig->lastOpError;
 if ( lastOpError!=0 ) return lastOpError;
 return thisConfig( *myConfig );
}

bool gHOutOpt::thisHasTidy (char* strTag, unsigned& idx)
{
 unsigned i;
 gString s( strTag );
 sHtmlElement* pElem;

 if ( aTidy.doTidyAll ) return true;

 s.UpString();
 pElem = gHtmlCtrl::Self().ParserPtr()->FindTag( s.Str() );
 DBGPRINT_MIN("DBG: thisHasTidy-possibly(%s)?%c\n",strTag,ISyORn(pElem!=nil));
 for (i=1; pElem!=nil && i<=aTidy.n; i++) {
     if ( pElem==aTidy.lst[i].pElem ) {
	 idx = i;
	 DBGPRINT("DBG: thisHasTidy(%s|%s,attr:%s):%u/%u\n",strTag,pElem->elemName,aTidy.lst[i].sAttr.Str(),i,aTidy.n);
	 return true;
     }
 }
 return false;
}

bool gHOutOpt::thisHasTidyAttr (char* strTag, gString& sAttr, unsigned& idx)
{
 unsigned i;
 gString s( strTag );
 sHtmlElement* pElem;

 if ( aTidy.doTidyAll ) return true;

 s.UpString();
 pElem = gHtmlCtrl::Self().ParserPtr()->FindTag( s.Str() );
 for (i=1; pElem!=nil && i<=aTidy.n; i++) {
     if ( pElem==aTidy.lst[i].pElem ) {
	 if ( aTidy.lst[i].AnyAttr() || aTidy.lst[i].sAttr.Match( sAttr ) ) {
	     idx = i;
	     DBGPRINT("DBG: thisHasTidyAttr(%s|%s,attr:%s):%u/%u\n",strTag,pElem->elemName,sAttr.Str(),i,aTidy.n);
	     return true;
	 }
     }
 }
 return false;
}

int gHOutOpt::thisConfig (gConfig& cfg)
{
 unsigned i, n, nTotal=cfg.nConf;
 unsigned k, len;
 int tLine;
 char* str;
 char* strWord1;
 char* strWord2;
 gList* pSkip;
 sHtmlElement* pElem;

 for (i=1; i<=nTotal; i++) {
     tLine = cfg.nIdxSectL.GetInt( i );
     str = cfg.sectL.Str( i );
     gString sCmdTrim( str );
     sCmdTrim.Trim();
     str = sCmdTrim.Str();
     gParam aParam( str, " " );
     strWord1 = aParam.Str( 1 );
     strWord2 = aParam.Str( 2 );
     if ( strcmp(str,"SKIP")!=0 && strcmp(strWord1,"TIDY")!=0 ) {
	 cfg.iLineError = tLine;
	 return 16;
     }
     ASSERTION(strWord2!=nil,"strWord2!=nil");
     if ( strWord2[0]!=0 ) {
	 if ( strcmp( strWord2, "ATTR" )!=0 ) {
	     cfg.iLineError = tLine;
	     return 16;
	 }
     }
 }//end FOR (first check)

 pSkip = cfg.GetSection( "SKIP" );
 if ( pSkip!=nil ) {
     tLine = cfg.nIdxSectL.GetInt( cfg.GetSectionFromName( "SKIP" ) );

     // Check if all upper-case skipped tags are valid elems
     for (i=1, n=pSkip->N(); i<=n; i++) {
	 tLine++;
	 str = pSkip->Str( i );
	 gString sTemp( str );
	 bool isAllUppercase = true;
	 for (k=1, len=sTemp.Length(); k<=len && isAllUppercase==true; k++)
	     if ( sTemp[k]<'A' || sTemp[k]>'Z' ) isAllUppercase = false;
	 ;
	 if ( len<=0 ) continue;
	 DBGPRINT_WEB2("DBG: opt(SKIP): '%s'\n",str);
	 sTemp.UpString();
	 pElem = gHtmlCtrl::Self().ParserPtr()->FindTag( str );
	 cfg.iLineError = tLine;  // Assume an error for line tLine
	 if ( isAllUppercase ) {
	     if ( pElem==nil ) {
		 fprintf(fRepErr,"SKIP: Invalid tag: %s\n",str);
		 return 32;
	     }
	     // Check if not already skipped
	     switch ( pElem->ctrl ) {
	     case -1:
		 fprintf(fRepErr,"SKIP: Duplicated tag: %s\n",str);
		 return 34;
	     case 1://tidy
		 fprintf(fRepErr,"SKIP: Tidy applied on tag: %s\n",str);
		 return 34;
	     case 0:
	     default:  // 0 is default...
		 pElem->ctrl = -1;  // Do skip tag
		 break;
	     }
	 }
	 else {
	     if ( pElem!=nil ) {
		 return 33;
	     }
	 }
	 cfg.iLineError = 0;
	 // Ugly (but quick) replacement of skipped tag, by upper-case
	 strcpy( str, sTemp.Str() );
	 lSkipTag.Add( str );
	 gString sSkipStart;
	 sSkipStart.Add( "<" );
	 gString sSkipEnd = sSkipStart;
	 sSkipStart.Add( str );
	 sSkipEnd.Add( "/" );
	 sSkipEnd.Add( str );
	 lSkipStt.Add( sSkipStart );
	 sSkipStart.Add( ">" );
	 sSkipEnd.Add( ">" );
	 lSkipFSt.Add( sSkipStart );
	 lSkipFSt.Add( sSkipEnd );
     }//end FOR i
 }//end token:SKIP

 str = "TIDY ATTR";
 pSkip = cfg.GetSection( str );
 if ( pSkip!=nil ) {
     tLine = cfg.nIdxSectL.GetInt( cfg.GetSectionFromName( str ) );

     n = pSkip->N();
     aTidy.Release();
     aTidy.lst = new sTidyElem[ n+1 ];  // highly-maximized array (1..n-indexed)
     ASSERTION(aTidy.lst!=nil,"aTidy.lst!=nil");

     for (i=1; i<=n; i++) {
	 tLine++;
	 str = pSkip->Str( i );
	 if ( str==nil || str[0]==0 ) continue;
	 DBGPRINT("DBG: TIDY: str=%s!\n",str);
	 gString sUp( str );
	 sUp.UpString();
	 str = sUp.Str();
	 gParam aParam( str, ".", gParam::e_StopSplitOnFirst );
	 strWord1 = aParam.Str( 1 );
	 strWord2 = aParam.Str( 2 );
	 ASSERTION(strWord2!=nil,"strWord2!=nil");
	 // strWord1 must be a valid tag
	 // Example:
	 //   IMG.SRC ==>Param<==> IMG , SRC  as Word1 , Word2
	 //   <IMG SRC=abc> is tidyed as <IMG SRC="abc">  (quoted)
	 pElem = gHtmlCtrl::Self().ParserPtr()->FindTag( strWord1 );
	 if ( pElem==nil ) {
	     fprintf(fRepErr,"TIDY: Invalid tag%s%s\n",strWord1[0]==0 ? "." : ": ",strWord1);
	     return 36;
	 }
	 if ( pElem->ctrl<0 ) {
	     fprintf(fRepErr,"TIDY: Tag was marked to be skipped: %s\n",strWord1);
	     return 38;
	 }
	 pElem->ctrl = 1; //tidy
	 unsigned iter = ++aTidy.n;
	 DBGPRINT("DBG: aTidy(%s|%s,n=%u (n=%u,nSections=%u)\n",strWord1,strWord2,iter,n,nTotal);
	 aTidy.lst[iter].pElem = pElem;
	 aTidy.lst[iter].sAttr.Set( strWord2 );
	 aTidy.lst[iter].sAttr.UpString();
     }//end FOR i
 }//end token:TIDY ATTR

 return 0;
}

void gHOutOpt::Show (bool doShowAll)
{
 unsigned i;
 gList* pSection;

 if ( myConfig==nil ) return;
 printf("nConf=%d\n",myConfig->nConf); myConfig->sectL.Show(true);
 for (i=1; i<=myConfig->nConf; i++) {
     printf("DBG: i=%u: ",i); myConfig->conpL[i].Show(true);
 }
 if ( doShowAll ) {
     printf("nIdxSectL: "); myConfig->nIdxSectL.Show( true );
 }
 pSection = myConfig->GetSection( "SKIP" );
 if ( pSection!=nil ) {
     printf("SKIP: ");
     pSection->Show( false );
     printf("!\n");
 }
 printf("lSkipFSt: "); lSkipFSt.Show( true );
}
////////////////////////////////////////////////////////////
gHtmlOpt::gHtmlOpt ()
    : outOpt( stderr )
{
}

gHtmlOpt::~gHtmlOpt ()
{
}

char* gHtmlOpt::GetBaseHRef ()
{
 ;
 // "<BASE HREF=sBaseHREF>", but just returning attributes
 sStrAux.Set( "HREF=" );
 sStrAux.Add( '"' );
 sStrAux.AddString( sBaseHREF );
 sStrAux.Add( '"' );
 return sStrAux.Str();
}

void gHtmlOpt::CopyOptions (gHtmlOpt& copy)
{
 short idx;
 sBaseHREF = copy.sBaseHREF;
 oErrStateAllSuppress.SetOn( copy.oErrStateAllSuppress.IsOn() );
 for (idx=0; idx<10; idx++) oErrStateSuppress[idx].SetOn( copy.oErrStateSuppress[idx].IsOn() );
 lIdTagOptEnd.CopyList( copy.lIdTagOptEnd );
 outFmt.CopyHOutFormat( copy.outFmt );
 outOpt.CopyHOutOpt( copy.outOpt );
}
////////////////////////////////////////////////////////////

