#ifndef gHTML_OPT_X_H
#define gHTML_OPT_X_H

#include "glistext.h"
#include "gstring.h"
#include "gconfig.h"

// Forward declarations
struct sHtmlElement;
////////////////////////////////////////////////////////////
enum eHState {
    e_HS_Start = 0,
    e_HS_StartAfter = 1,
    e_HS_Html = 2,
    e_HS_Head = 3,
    e_HS_HeadAfter = 4,
    e_HS_Body = 5,
    e_HS_BodyAfter = 6,
    e_HS_End = 7,
    e_HS_Last,  // Not used
    e_HS_Sp_FakeBody
};
////////////////////////////////////////////////////////////
struct sHOutFormat {
    sHOutFormat ()
	: cIndentLevel( '\t' ),
	  oIndentLevel1( 'T' ) {
	Prepare();
    }
    ~sHOutFormat () {
    }

    char cIndentLevel;
    gSwitch oIndentLevel1;

    gString sBeginLine[e_HS_Last+1];

    // Get methods
    char* GetBeginLineStr (t_uint16 state, t_uint16 rLine) ;

    // Set methods
    void Prepare () ;
    int SetIndentString (char cIndent, t_uint16 nBlanks, gString& sResult) ;

    void CopyHOutFormat (sHOutFormat& copy) ;
};
////////////////////////////////////////////////////////////
struct sAttrNorm {
    t_int16 idTag;         // -2 for the first element (all tags norms)
    char* strAttrSeq;      // comma separated list of attributes, uppercase: e.g. "SRC, ALT" (for IMG tags)
    gSmartList* pAttrSeq;  // depends on strAttrSeq
};

struct sAttrsTagNorm {
    sAttrsTagNorm ()
	: lstNorms( nil ) {
    }
    ~sAttrsTagNorm () {
	delete[] lstNorms;
    }

    gSmartList allTagNorm;
    gSmartList* lstNorms;  // [0:idxTag=0, gHtmlParser::nElems:last-tag]

    // Set methods
    void Reset () {
	delete[] lstNorms;
	lstNorms = nil;
    }

    int CopyDefaultAttrNorms (t_int16 idTag0, t_int16 idTag1, bool doDefault) ;
};
////////////////////////////////////////////////////////////
class gHInOpt : public gControl {
public:
    gHInOpt () ;
    virtual ~gHInOpt () ;

    // Public data-members
    gListInt lLineAdd;
    int vLineLTE, vLineGTE;
    gListInt lLineSuppress;  // suppress
    int vLineSupLTE, vLineSupGTE; // suppress

    // Get methods
    virtual bool HasToSkip (int iLine) ;

    // Special methods
    virtual bool ConvertToInt (char* str, int& v) ;
    virtual bool SetFromString (char* str) ;

protected:
    bool hasLineAddRestriction;

    int thisAddLineSup (int v, bool doExclude) ;
    int thisAddLineSupGTE (int v, bool doExclude) ;
    bool thisHasToShow (int iLine) ;

private:
    // Operators,empty
    gHInOpt (gHInOpt& ) ; //empty
    gHInOpt& operator= (gHInOpt& ) ; //empty
};
////////////////////////////////////////////////////////////
struct sTidyElem {
    sTidyElem ()
	: pElem( nil ) {
    }
    ~sTidyElem () {
    }
    sHtmlElement* pElem;  // never allocated
    gString sAttr;
    bool AnyAttr () {
	return sAttr.IsEmpty();
    }
    void Copy (sTidyElem& copy) {
	pElem = copy.pElem;
	sAttr = copy.sAttr;
    }
};

struct sTidyArray {
    sTidyArray ()
	: n( 0 ),
	  lst( nil ),
	  doTidyAll( false ) {
    }
    ~sTidyArray () {
	Release();
    }
    unsigned n;
    sTidyElem* lst;  // 1..n tidy elements
    bool doTidyAll;  // all tags&attrs

    // Methods
    void Release () {
	n = 0;
	delete[] lst;
	lst = nil;
    }
    void Copy (sTidyArray& copy) ;
};
////////////////////////////////////////////////////////////
class gHOutOpt : public gControl {
public:
    gHOutOpt (FILE* afRepErr) ;
    virtual ~gHOutOpt () ;

    // Public data-members
    FILE* fRepErr;
    sTidyArray aTidy;
    sAttrsTagNorm lAttrsNorm;  // List of configured attributes norms

    // Get methods
    virtual gConfig& Config () ;

    virtual bool HasToSkip (char* sFullTagStr) ;

    virtual bool IsTagSkipped (char* sTagStr) {
	return lSkipTag.Match( sTagStr )>0;
    }

    bool HasTidy (char* strTag) {
	unsigned idx;
	return thisHasTidy( strTag, idx );
    }
    bool HasTidy (char* strTag, unsigned& idx) {
	return thisHasTidy( strTag, idx );
    }
    bool HasTidyAttr (char* strTag, gString& sAttr) {
	unsigned idx;
	return thisHasTidyAttr( strTag, sAttr, idx );
    }
    bool HasTidyAttr (char* strTag, gString& sAttr, unsigned& idx) {
	return thisHasTidyAttr( strTag, sAttr, idx );
    }

    // Set methods
    void CopyHOutOpt (gHOutOpt& copy) ;
    virtual int CopyDefaultAttrs () ;
    virtual int CopyDefaultAttrNorms (t_int16 idTag0, t_int16 idTag1, bool doDefault) ;

    // Special methods
    virtual int ConfigFromFile (char* fName) ;

    // Show methods
    virtual void Show (bool doShowAll=true) ;

protected:
    gConfig* myConfig;
    gList lSkipTag;  // List of tags skipped (only tag, e.g. B)
    gList lSkipFSt;  // List of full string skipped tags (e.g. <B> and </B>)
    gList lSkipStt;  // List of full string skipped tag begining (e.g. <B>)

    bool thisHasTidy (char* strTag, unsigned& idx) ;
    bool thisHasTidyAttr (char* strTag, gString& sAttr, unsigned& idx) ;

    int thisConfig (gConfig& cfg) ;

private:
    // Operators,empty
    gHOutOpt (gHOutOpt& ) ; //empty
    gHOutOpt& operator= (gHOutOpt& ) ; //empty
};
////////////////////////////////////////////////////////////
class gHtmlOpt : public gControl {
public:
    gHtmlOpt () ;
    virtual ~gHtmlOpt () ;

    // Public data-members
    gString sBaseHREF;
    gSwitch oErrStateAllSuppress;
    gSwitch oErrStateSuppress[10];  // nr of states (<10)
    gSwitch oErrInvTagWrite;  // Write out invalid tags
    gList lIdTagOptEnd;  // set of id's of tags

    sHOutFormat outFmt;
    gHInOpt inOpt;
    gHOutOpt outOpt;

    // Get methods
    virtual bool DoBaseHRef () {
	return sBaseHREF.IsEmpty()==false;
    }
    virtual char* GetBaseHRef () ;

    // Set methods
    virtual bool AddIdTagOptEnd (t_int16 id) {
	if ( id<0 ) return false;
	lIdTagOptEnd.Add( (int)id );
	return true;
    }

    void CopyOptions (gHtmlOpt& copy) ;

protected:
    gString sStrAux;

private:
    // Operators,empty
    gHtmlOpt (gHtmlOpt& ) ; //empty
    gHtmlOpt& operator= (gHtmlOpt& ) ; //empty
};
////////////////////////////////////////////////////////////
#endif //gHTML_OPT_X_H

