// gHtmlParser, main part for libgdhtml

#include <string.h>
#include "gHtmlParser.h"
#include "gHtmlCtrl.h"
#include "gHtmlSeq.h"
#include "gstringext.h"

////////////////////////////////////////////////////////////
// Static members
const char* gHParsed::tblStateStr[e_HS_Last]={
	"START",
	"START_",
	"HTML",
	"HEAD",
	"HEAD_",
	"BODY",
	"BODY_",
	"HTML_"};

t_int16 gHtmlParser::nElems=-1;
sHtmlElement gHtmlParser::lElems[]={
	{0, "A",	"anchor",	'@',	'@',	'@',	'@',	'@'},  // 0
	{0, "ABBR",	"abbreviated form",	'@',	'@',	'@',	'@',	e_HtmlElementPhrase},
	{0, "ACRONYM",	"acronym",	'@',	'@',	'@',	'@',	e_HtmlElementPhrase},
	{0, "ADDRESS",	"information on author",	'@',	'@',	'@',	'@',	'@'},
	{0, "APPLET",	"Java applet",	'@',	'@',	'D',	'L',	'@'},
	{0, "AREA",	"client-side image map",	'@',	'F',	'@',	'@',	'@'},
	{0, "B",	"bold text style",	'@',	'@',	'@',	'@',	'@'},
	{0, "BASE",	"document base URI",	'@',	'F',	'@',	'@',	'@'},  // 7
	{0, "BASEFONT",	"base font size",	'@',	'F',	'D',	'L',	'@'},
	{0, "BDO",	"I18N BiDi over-ride",	'@',	'@',	'@',	'@',	'@'},
	{0, "BIG",	"large text style",	'@',	'@',	'@',	'@',	'@'},
	{0, "BLOCKQUOTE",	"long quotation",	'@',	'@',	'@',	'@',	'@'},
	{0, "BODY",	"document body",	'O',	'O',	'@',	'@',	'@'},  // 12
	{0, "BR",	"forced line break",	'@',	'F',	'@',	'@',	'@'},
	{0, "BUTTON",	"push button",	'@',	'@',	'@',	'@',	'@'},
	{0, "CAPTION",	"table caption",	'@',	'@',	'@',	'@',	'@'},
	{0, "CENTER",	"center line (deprecated)",	'@',	'@',	'D',	'L',	'@'},
	{0, "CITE",	"citation",	'@',	'@',	'@',	'@',	e_HtmlElementPhrase},
	{0, "CODE",	"computer code fragment",	'@',	'@',	'@',	'@',	e_HtmlElementPhrase},
	{0, "COL",	"table column",	'@',	'F',	'@',	'@',	'@'},
	{0, "COLGROUP",	"table column group",	'@',	'O',	'@',	'@',	'@'},
	{0, "DD",	"definition description",	'@',	'O',	'@',	'@',	'@'},
	{0, "DEL",	"deleted text",	'@',	'@',	'@',	'@',	'@'},
	{0, "DFN",	"instance definition",	'@',	'@',	'@',	'@',	e_HtmlElementPhrase},
	{0, "DIR",	"directory list",	'@',	'@',	'D',	'L',	'@'},
	{0, "DIV",	"generic language/style",	'@',	'@',	'@',	'@',	'@'},
	{0, "DL",	"definition list",	'@',	'@',	'@',	'@',	'@'},
	{0, "DT",	"definition term",	'@',	'O',	'@',	'@',	'@'},
	{0, "EM",	"emphasis",	'@',	'@',	'@',	'@',	e_HtmlElementPhrase},
	{0, "FIELDSET",	"form control group",	'@',	'@',	'@',	'@',	'@'},
	{0, "FONT",	"local change to font",	'@',	'@',	'D',	'L',	'@'},
	{0, "FORM",	"interactive form",	'@',	'@',	'@',	'@',	'@'},
	{0, "FRAME",	"subwindow",	'@',	'F',	'@',	'F',	'@'},
	{0, "FRAMESET",	"window subdivision",	'@',	'@',	'@',	'F',	'@'},
	{0, "H1",	"heading",	'@',	'@',	'@',	'@',	'@'},
	{0, "H2",	"heading",	'@',	'@',	'@',	'@',	'@'},
	{0, "H3",	"heading",	'@',	'@',	'@',	'@',	'@'},
	{0, "H4",	"heading",	'@',	'@',	'@',	'@',	'@'},
	{0, "H5",	"heading",	'@',	'@',	'@',	'@',	'@'},
	{0, "H6",	"heading",	'@',	'@',	'@',	'@',	'@'},
	{0, "HEAD",	"document head",	'O',	'O',	'@',	'@',	'@'},  // 40
	{0, "HR",	"horizontal rule",	'@',	'F',	'@',	'@',	'@'},
	{0, "HTML",	"document root element",	'O',	'O',	'@',	'@',	'@'},  // 42
	{0, "I",	"italic text style",	'@',	'@',	'@',	'@',	'@'},
	{0, "IFRAME",	"inline subwindow",	'@',	'@',	'@',	'L',	'@'},
	{0, "IMG",	"Embedded image",	'@',	'F',	'@',	'@',	'@'},
	{0, "INPUT",	"form control",	'@',	'F',	'@',	'@',	'@'},
	{0, "INS",	"inserted text",	'@',	'@',	'@',	'@',	'@'},
	{0, "ISINDEX",	"single line prompt",	'@',	'F',	'D',	'L',	'@'},
	{0, "KBD",	"text to be entered by user",	'@',	'@',	'@',	'@',	e_HtmlElementPhrase},
	{0, "LABEL",	"form field label text",	'@',	'@',	'@',	'@',	'@'},
	{0, "LEGEND",	"fieldset legend",	'@',	'@',	'@',	'@',	'@'},
	{0, "LI",	"list item",	'@',	'O',	'@',	'@',	'@'},
	{0, "LINK",	"media-independent link",	'@',	'F',	'@',	'@',	'@'},
	{0, "MAP",	"client-side image map",	'@',	'@',	'@',	'@',	'@'},
	{0, "MENU",	"menu list",	'@',	'@',	'D',	'L',	'@'},
	{0, "META",	"generic meta info",	'@',	'F',	'@',	'@',	'@'},
	{0, "NOFRAMES",	"alternate content",	'@',	'@',	'@',	'F',	'@'},
	{0, "NOSCRIPT",	"alternate content",	'@',	'@',	'@',	'@',	'@'},
	{0, "OBJECT",	"generic embedded object",	'@',	'@',	'@',	'@',	'@'},
	{0, "OL",	"ordered list",	'@',	'@',	'@',	'@',	'@'},
	{0, "OPTGROUP",	"option group",	'@',	'@',	'@',	'@',	'@'},
	{0, "OPTION",	"selectable choice",	'@',	'O',	'@',	'@',	'@'},
	{0, "P",	"paragraph",	'@',	'O',	'@',	'@',	'@'},
	{0, "PARAM",	"named property value",	'@',	'F',	'@',	'@',	'@'},
	{0, "PRE",	"preformatted text",	'@',	'@',	'@',	'@',	'@'},
	{0, "Q",	"short inline quotation",	'@',	'@',	'@',	'@',	'@'},
	{0, "S",	"strike-through text style",	'@',	'@',	'D',	'L',	'@'},
	{0, "SAMP",	"sample program output",	'@',	'@',	'@',	'@',	e_HtmlElementPhrase},
	{0, "SCRIPT",	"script statements",	'@',	'@',	'@',	'@',	'@'},
	{0, "SELECT",	"option selector",	'@',	'@',	'@',	'@',	'@'},
	{0, "SMALL",	"small text style",	'@',	'@',	'@',	'@',	'@'},
	{0, "SPAN",	"generic language/style",	'@',	'@',	'@',	'@',	'@'},
	{0, "STRIKE",	"strike-through text",	'@',	'@',	'D',	'L',	'@'},
	{0, "STRONG",	"strong emphasis",	'@',	'@',	'@',	'@',	e_HtmlElementPhrase},
	{0, "STYLE",	"style info",	'@',	'@',	'@',	'@',	'@'},
	{0, "SUB",	"subscript",	'@',	'@',	'@',	'@',	'@'},
	{0, "SUP",	"superscript",	'@',	'@',	'@',	'@',	'@'},
	{0, "TABLE",	"table",	'@',	'@',	'@',	'@',	'@'},
	{0, "TBODY",	"table body",	'O',	'O',	'@',	'@',	'@'},
	{0, "TD",	"table data cell",	'@',	'O',	'@',	'@',	'@'},
	{0, "TEXTAREA",	"multi-line text field",	'@',	'@',	'@',	'@',	'@'},
	{0, "TFOOT",	"table footer",	'@',	'O',	'@',	'@',	'@'},
	{0, "TH",	"table header cell",	'@',	'O',	'@',	'@',	'@'},
	{0, "THEAD",	"table header",	'@',	'O',	'@',	'@',	'@'},
	{0, "TITLE",	"document title",	'@',	'@',	'@',	'@',	'@'},
	{0, "TR",	"table row",	'@',	'O',	'@',	'@',	'@'},
	{0, "TT",	"teletype or monospaced",	'@',	'@',	'@',	'@',	'@'},
	{0, "U",	"underlined text style",	'@',	'@',	'D',	'L',	'@'},
	{0, "UL",	"unordered list",	'@',	'@',	'@',	'@',	'@'},
	{0, "VAR",	"instance of a variable",	'@',	'@',	'@',	'@',	e_HtmlElementPhrase},
	{0, NULL,	NULL,	'@',	'@',	'@',	'@',	'\0'}};


sAttrDef gHtmlParser::lAttrs[]={
	{0, "abbr",	"TD, TH",	"%Text;",	"#I",	'n',	"",	"abbreviation for header cell"},
	{0, "accept-charset",	"FORM",	"%Charsets;",	"#I",	'n',	"",	"list of supported charsets"},
	{0, "accept",	"FORM, INPUT",	"%ContentTypes;",	"#I",	'n',	"",	"list of MIME types for file upload"},
	{0, "accesskey",	"A, AREA, BUTTON, INPUT, LABEL, LEGEND, TEXTAREA",	"%Character;",	"#I",	'n',	"",	"accessibility key character"},
	{0, "action",	"FORM",	"%URI;",	"#R",	'n',	"",	"server-side form handler"},
	{0, "align",	"CAPTION",	"%CAlign;",	"#I",	'D',	"L",	"relative to table"},
	{0, "align",	"APPLET, IFRAME, IMG, INPUT, OBJECT",	"%IAlign;",	"#I",	'D',	"L",	"vertical or horizontal alignment"},
	{0, "align",	"LEGEND",	"%LAlign;",	"#I",	'D',	"L",	"relative to fieldset"},
	{0, "align",	"TABLE",	"%TAlign;",	"#I",	'D',	"L",	"table position relative to window"},
	{0, "align",	"HR",	"(left|center|right)",	"#I",	'D',	"L",	""},
	{0, "align",	"DIV, H1, H2, H3, H4, H5, H6, P",	"(left|center|right|justify)",	"#I",	'D',	"L",	"align, text alignment"},
	{0, "align",	"COL, COLGROUP, TBODY, TD, TFOOT, TH, THEAD, TR",	"(left|center|right|justify|char)",	"#I",	'n',	"",	""},
	{0, "alink",	"BODY",	"%Color;",	"#I",	'D',	"L",	"color of selected links"},
	{0, "alt",	"APPLET",	"%Text;",	"#I",	'D',	"L",	"short description"},
	{0, "alt",	"AREA, IMG",	"%Text;",	"#R",	'n',	"",	"short description"},
	{0, "alt",	"INPUT",	"CDATA",	"#I",	'n',	"",	"short description"},
	{0, "archive",	"APPLET",	"CDATA",	"#I",	'D',	"L",	"comma-separated archive list"},
	{0, "archive",	"OBJECT",	"CDATA",	"#I",	'n',	"",	"space-separated list of URIs"},
	{0, "axis",	"TD, TH",	"CDATA",	"#I",	'n',	"",	"comma-separated list of related headers"},
	{0, "background",	"BODY",	"%URI;",	"#I",	'D',	"L",	"texture tile for document background"},
	{0, "bgcolor",	"TABLE",	"%Color;",	"#I",	'D',	"L",	"background color for cells"},
	{0, "bgcolor",	"TR",	"%Color;",	"#I",	'D',	"L",	"background color for row"},
	{0, "bgcolor",	"TD, TH",	"%Color;",	"#I",	'D',	"L",	"cell background color"},
	{0, "bgcolor",	"BODY",	"%Color;",	"#I",	'D',	"L",	"document background color"},
	{0, "border",	"TABLE",	"%Pixels;",	"#I",	'n',	"",	"controls frame width around table"},
	{0, "border",	"IMG, OBJECT",	"%Pixels;",	"#I",	'D',	"L",	"link border width"},
	{0, "cellpadding",	"TABLE",	"%Length;",	"#I",	'n',	"",	"spacing within cells"},
	{0, "cellspacing",	"TABLE",	"%Length;",	"#I",	'n',	"",	"spacing between cells"},
	{0, "char",	"COL, COLGROUP, TBODY, TD, TFOOT, TH, THEAD, TR",	"%Character;",	"#I",	'n',	"",	"alignment char, e.g. char :"},
	{0, "charoff",	"COL, COLGROUP, TBODY, TD, TFOOT, TH, THEAD, TR",	"%Length;",	"#I",	'n',	"",	"offset for alignment char"},
	{0, "charset",	"A, LINK, SCRIPT",	"%Charset;",	"#I",	'n',	"",	"char encoding of linked resource"},
	{0, "checked",	"INPUT",	"(checked)",	"#I",	'n',	"",	"for radio buttons and check boxes"},
	{0, "cite",	"BLOCKQUOTE, Q",	"%URI;",	"#I",	'n',	"",	"URI for source document or msg"},
	{0, "cite",	"DEL, INS",	"%URI;",	"#I",	'n',	"",	"info on reason for change"},
	{0, "class",	"@BASE, BASEFONT, HEAD, HTML, META, PARAM, SCRIPT, STYLE, TITLE",	"CDATA",	"#I",	'n',	"",	"space-separated list of classes"},
	{0, "classid",	"OBJECT",	"%URI;",	"#I",	'n',	"",	"identifies an implementation"},
	{0, "clear",	"BR",	"(left|all|right|none)",	"none",	'D',	"L",	"control of text flow"},
	{0, "code",	"APPLET",	"CDATA",	"#I",	'D',	"L",	"applet class file"},
	{0, "codebase",	"OBJECT",	"%URI;",	"#I",	'n',	"",	"base URI for classid, data, archive"},
	{0, "codebase",	"APPLET",	"%URI;",	"#I",	'D',	"L",	"optional base URI for applet"},
	{0, "codetype",	"OBJECT",	"%ContentType;",	"#I",	'n',	"",	"content type for code"},
	{0, "color",	"BASEFONT, FONT",	"%Color;",	"#I",	'D',	"L",	"text color"},
	{0, "cols",	"FRAMESET",	"%MultiLengths;",	"#I",	'n',	"F",	"list of lengths, default: 100% (1 col)"},
	{0, "cols",	"TEXTAREA",	"NUMBER",	"#R",	'n',	"",	""},
	{0, "colspan",	"TD, TH",	"NUMBER",	"1",	'n',	"",	"number of cols spanned by cell"},
	{0, "compact",	"DIR, DL, MENU, OL, UL",	"(compact)",	"#I",	'D',	"L",	"reduced interitem spacing"},
	{0, "content",	"META",	"CDATA",	"#R",	'n',	"",	"associated information"},
	{0, "coords",	"AREA",	"%Coords;",	"#I",	'n',	"",	"comma-separated list of lengths"},
	{0, "coords",	"A",	"%Coords;",	"#I",	'n',	"",	"for use with client-side image maps"},
	{0, "data",	"OBJECT",	"%URI;",	"#I",	'n',	"",	"reference to object$s data"},
	{0, "datetime",	"DEL, INS",	"%Datetime;",	"#I",	'n',	"",	"date and time of change"},
	{0, "declare",	"OBJECT",	"(declare)",	"#I",	'n',	"",	"declare but don$t instantiate flag"},
	{0, "defer",	"SCRIPT",	"(defer)",	"#I",	'n',	"",	"UA may defer execution of script"},
	{0, "dir",	"@APPLET, BASE, BASEFONT, BDO, BR, FRAME, FRAMESET, IFRAME, PARAM, SCRIPT",	"(ltr|rtl)",	"#I",	'n',	"",	"direction for weak/neutral text"},
	{0, "dir",	"BDO",	"(ltr|rtl)",	"#R",	'n',	"",	"directionality"},
	{0, "disabled",	"BUTTON, INPUT, OPTGROUP, OPTION, SELECT, TEXTAREA",	"(disabled)",	"#I",	'n',	"",	"unavailable in this context"},
	{0, "enctype",	"FORM",	"%ContentType;",	"application/x-www- form-urlencoded",	'n',	"",	""},
	{0, "face",	"BASEFONT, FONT",	"CDATA",	"#I",	'D',	"L",	"comma-separated list of font names"},
	{0, "for",	"LABEL",	"IDREF",	"#I",	'n',	"",	"matches field ID value"},
	{0, "frame",	"TABLE",	"%TFrame;",	"#I",	'n',	"",	"which parts of frame to render"},
	{0, "frameborder",	"FRAME, IFRAME",	"(1|0)",	"1",	'n',	"F",	"request frame borders?"},
	{0, "headers",	"TD, TH",	"IDREFS",	"#I",	'n',	"",	"list of id$s for header cells"},
	{0, "height",	"IFRAME",	"%Length;",	"#I",	'f',	"L",	"frame height"},  // f in "Depr." means deprecated frame-set
	{0, "height",	"TD, TH",	"%Length;",	"#I",	'D',	"L",	"height for cell"},
	{0, "height",	"IMG, OBJECT",	"%Length;",	"#I",	'n',	"",	"override height"},
	{0, "height",	"APPLET",	"%Length;",	"#R",	'D',	"L",	"initial height"},
	{0, "href",	"A, AREA, LINK",	"%URI;",	"#I",	'n',	"",	"URI for linked resource"},
	{0, "href",	"BASE",	"%URI;",	"#I",	'n',	"",	"URI that acts as base URI"},
	{0, "hreflang",	"A, LINK",	"%LanguageCode;",	"#I",	'n',	"",	"language code"},
	{0, "hspace",	"APPLET, IMG, OBJECT",	"%Pixels;",	"#I",	'D',	"L",	"horizontal gutter"},
	{0, "http-equiv",	"META",	"NAME",	"#I",	'n',	"",	"HTTP response header name"},
	{0, "id",	"@BASE, HEAD, HTML, META, SCRIPT, STYLE, TITLE",	"ID",	"#I",	'n',	"",	"document-wide unique id"},
	{0, "ismap",	"IMG, INPUT",	"(ismap)",	"#I",	'n',	"",	"use server-side image map"},
	{0, "label",	"OPTION",	"%Text;",	"#I",	'n',	"",	"for use in hierarchical menus"},
	{0, "label",	"OPTGROUP",	"%Text;",	"#R",	'n',	"",	"for use in hierarchical menus"},
	{0, "lang",	"@APPLET, BASE, BASEFONT, BR, FRAME, FRAMESET, IFRAME, PARAM, SCRIPT",	"%LanguageCode;",	"#I",	'n',	"",	"language code"},
	{0, "language",	"SCRIPT",	"CDATA",	"#I",	'D',	"L",	"predefined script language name"},
	{0, "link",	"BODY",	"%Color;",	"#I",	'D',	"L",	"color of links"},
	{0, "longdesc",	"IMG",	"%URI;",	"#I",	'n',	"",	"link to long description (complements alt)"},
	{0, "longdesc",	"FRAME, IFRAME",	"%URI;",	"#I",	'n',	"F",	"link to long description (complements title)"},
	{0, "marginheight",	"FRAME, IFRAME",	"%Pixels;",	"#I",	'n',	"F",	"margin height in pixels"},
	{0, "marginwidth",	"FRAME, IFRAME",	"%Pixels;",	"#I",	'n',	"F",	"margin widths in pixels"},
	{0, "maxlength",	"INPUT",	"NUMBER",	"#I",	'n',	"",	"max chars for text fields"},
	{0, "media",	"STYLE",	"%MediaDesc;",	"#I",	'n',	"",	"designed for use with these media"},
	{0, "media",	"LINK",	"%MediaDesc;",	"#I",	'n',	"",	"for rendering on these media"},
	{0, "method",	"FORM",	"(GET|POST)",	"GET",	'n',	"",	"HTTP method used to submit the form"},
	{0, "multiple",	"SELECT",	"(multiple)",	"#I",	'n',	"",	"default is single selection"},
	{0, "name",	"BUTTON, TEXTAREA",	"CDATA",	"#I",	'n',	"",	""},
	{0, "name",	"APPLET",	"CDATA",	"#I",	'D',	"L",	"allows applets to find each other"},
	{0, "name",	"SELECT",	"CDATA",	"#I",	'n',	"",	"field name"},
	{0, "name",	"FORM",	"CDATA",	"#I",	'n',	"",	"name of form for scripting"},
	{0, "name",	"FRAME, IFRAME",	"CDATA",	"#I",	'n',	"F",	"name of frame for targetting"},
	{0, "name",	"IMG",	"CDATA",	"#I",	'n',	"",	"name of image for scripting"},
	{0, "name",	"A",	"CDATA",	"#I",	'n',	"",	"named link end"},
	{0, "name",	"INPUT, OBJECT",	"CDATA",	"#I",	'n',	"",	"submit as part of form"},
	{0, "name",	"MAP",	"CDATA",	"#R",	'n',	"",	"for reference by usemap"},
	{0, "name",	"PARAM",	"CDATA",	"#R",	'n',	"",	"property name"},
	{0, "name",	"META",	"NAME",	"#I",	'n',	"",	"metainformation name"},
	{0, "nohref",	"AREA",	"(nohref)",	"#I",	'n',	"",	"this region has no action"},
	{0, "noresize",	"FRAME",	"(noresize)",	"#I",	'n',	"F",	"allow users to resize frames?"},
	{0, "noshade",	"HR",	"(noshade)",	"#I",	'D',	"L",	""},
	{0, "nowrap",	"TD, TH",	"(nowrap)",	"#I",	'D',	"L",	"suppress word wrap"},
	{0, "object",	"APPLET",	"CDATA",	"#I",	'D',	"L",	"serialized applet file"},
	{0, "onblur",	"A, AREA, BUTTON, INPUT, LABEL, SELECT, TEXTAREA",	"%Script;",	"#I",	'n',	"",	"the element lost the focus"},
	{0, "onchange",	"INPUT, SELECT, TEXTAREA",	"%Script;",	"#I",	'n',	"",	"the element value was changed"},
	{0, "onclick",	"@APPLET, BASE, BASEFONT, BDO, BR, FONT, FRAME, FRAMESET, HEAD, HTML, IFRAME, ISINDEX, META, PARAM, SCRIPT, STYLE, TITLE",	"%Script;",	"#I",	'n',	"",	"a pointer button was clicked"},
	{0, "ondblclick",	"@APPLET, BASE, BASEFONT, BDO, BR, FONT, FRAME, FRAMESET, HEAD, HTML, IFRAME, ISINDEX, META, PARAM, SCRIPT, STYLE, TITLE",	"%Script;",	"#I",	'n',	"",	"a pointer button was double clicked"},
	{0, "onfocus",	"A, AREA, BUTTON, INPUT, LABEL, SELECT, TEXTAREA",	"%Script;",	"#I",	'n',	"",	"the element got the focus"},
	{0, "onkeydown",	"@APPLET, BASE, BASEFONT, BDO, BR, FONT, FRAME, FRAMESET, HEAD, HTML, IFRAME, ISINDEX, META, PARAM, SCRIPT, STYLE, TITLE",	"%Script;",	"#I",	'n',	"",	"a key was pressed down"},
	{0, "onkeypress",	"@APPLET, BASE, BASEFONT, BDO, BR, FONT, FRAME, FRAMESET, HEAD, HTML, IFRAME, ISINDEX, META, PARAM, SCRIPT, STYLE, TITLE",	"%Script;",	"#I",	'n',	"",	"a key was pressed and released"},
	{0, "onkeyup",	"@APPLET, BASE, BASEFONT, BDO, BR, FONT, FRAME, FRAMESET, HEAD, HTML, IFRAME, ISINDEX, META, PARAM, SCRIPT, STYLE, TITLE",	"%Script;",	"#I",	'n',	"",	"a key was released"},
	{0, "onload",	"FRAMESET",	"%Script;",	"#I",	'n',	"F",	"all the frames have been loaded"},
	{0, "onload",	"BODY",	"%Script;",	"#I",	'n',	"",	"the document has been loaded"},
	{0, "onmousedown",	"@APPLET, BASE, BASEFONT, BDO, BR, FONT, FRAME, FRAMESET, HEAD, HTML, IFRAME, ISINDEX, META, PARAM, SCRIPT, STYLE, TITLE",	"%Script;",	"#I",	'n',	"",	"a pointer button was pressed down"},
	{0, "onmousemove",	"@APPLET, BASE, BASEFONT, BDO, BR, FONT, FRAME, FRAMESET, HEAD, HTML, IFRAME, ISINDEX, META, PARAM, SCRIPT, STYLE, TITLE",	"%Script;",	"#I",	'n',	"",	"a pointer was moved within"},
	{0, "onmouseout",	"@APPLET, BASE, BASEFONT, BDO, BR, FONT, FRAME, FRAMESET, HEAD, HTML, IFRAME, ISINDEX, META, PARAM, SCRIPT, STYLE, TITLE",	"%Script;",	"#I",	'n',	"",	"a pointer was moved away"},
	{0, "onmouseover",	"@APPLET, BASE, BASEFONT, BDO, BR, FONT, FRAME, FRAMESET, HEAD, HTML, IFRAME, ISINDEX, META, PARAM, SCRIPT, STYLE, TITLE",	"%Script;",	"#I",	'n',	"",	"a pointer was moved onto"},
	{0, "onmouseup",	"@APPLET, BASE, BASEFONT, BDO, BR, FONT, FRAME, FRAMESET, HEAD, HTML, IFRAME, ISINDEX, META, PARAM, SCRIPT, STYLE, TITLE",	"%Script;",	"#I",	'n',	"",	"a pointer button was released"},
	{0, "onreset",	"FORM",	"%Script;",	"#I",	'n',	"",	"the form was reset"},
	{0, "onselect",	"INPUT, TEXTAREA",	"%Script;",	"#I",	'n',	"",	"some text was selected"},
	{0, "onsubmit",	"FORM",	"%Script;",	"#I",	'n',	"",	"the form was submitted"},
	{0, "onunload",	"FRAMESET",	"%Script;",	"#I",	'n',	"F",	"all the frames have been removed"},
	{0, "onunload",	"BODY",	"%Script;",	"#I",	'n',	"",	"the document has been removed"},
	{0, "profile",	"HEAD",	"%URI;",	"#I",	'n',	"",	"named dictionary of meta info"},
	{0, "prompt",	"ISINDEX",	"%Text;",	"#I",	'D',	"L",	"prompt message"},
	{0, "readonly",	"TEXTAREA",	"(readonly)",	"#I",	'n',	"",	""},
	{0, "readonly",	"INPUT",	"(readonly)",	"#I",	'n',	"",	"for text and passwd"},
	{0, "rel",	"A, LINK",	"%LinkTypes;",	"#I",	'n',	"",	"forward link types"},
	{0, "rev",	"A, LINK",	"%LinkTypes;",	"#I",	'n',	"",	"reverse link types"},
	{0, "rows",	"FRAMESET",	"%MultiLengths;",	"#I",	'n',	"F",	"list of lengths, default: 100% (1 row)"},
	{0, "rows",	"TEXTAREA",	"NUMBER",	"#R",	'n',	"",	""},
	{0, "rowspan",	"TD, TH",	"NUMBER",	"1",	'n',	"",	"number of rows spanned by cell"},
	{0, "rules",	"TABLE",	"%TRules;",	"#I",	'n',	"",	"rulings between rows and cols"},
	{0, "scheme",	"META",	"CDATA",	"#I",	'n',	"",	"select form of content"},
	{0, "scope",	"TD, TH",	"%Scope;",	"#I",	'n',	"",	"scope covered by header cells"},
	{0, "scrolling",	"FRAME, IFRAME",	"(yes|no|auto)",	"auto",	'n',	"F",	"scrollbar or none"},
	{0, "selected",	"OPTION",	"(selected)",	"#I",	'n',	"",	""},
	{0, "shape",	"AREA",	"%Shape;",	"rect",	'n',	"",	"controls interpretation of coords"},
	{0, "shape",	"A",	"%Shape;",	"rect",	'n',	"",	"for use with client-side image maps"},
	{0, "size",	"HR",	"%Pixels;",	"#I",	'D',	"L",	""},
	{0, "size",	"FONT",	"CDATA",	"#I",	'D',	"L",	"[+|-]nn e.g. size=+1, size=4"},
	{0, "size",	"INPUT",	"CDATA",	"#I",	'n',	"",	"specific to each type of field"},
	{0, "size",	"BASEFONT",	"CDATA",	"#R",	'D',	"L",	"base font size for FONT elements"},
	{0, "size",	"SELECT",	"NUMBER",	"#I",	'n',	"",	"rows visible"},
	{0, "span",	"COL",	"NUMBER",	"1",	'n',	"",	"COL attributes affect N columns"},
	{0, "span",	"COLGROUP",	"NUMBER",	"1",	'n',	"",	"default number of columns in group"},
	{0, "src",	"SCRIPT",	"%URI;",	"#I",	'n',	"",	"URI for an external script"},
	{0, "src",	"INPUT",	"%URI;",	"#I",	'n',	"",	"for fields with images"},
	{0, "src",	"FRAME, IFRAME",	"%URI;",	"#I",	'n',	"F",	"source of frame content"},
	{0, "src",	"IMG",	"%URI;",	"#R",	'n',	"",	"URI of image to embed"},
	{0, "standby",	"OBJECT",	"%Text;",	"#I",	'n',	"",	"message to show while loading"},
	{0, "start",	"OL",	"NUMBER",	"#I",	'D',	"L",	"starting sequence number"},
	{0, "style",	"@BASE, BASEFONT, HEAD, HTML, META, PARAM, SCRIPT, STYLE, TITLE",	"%StyleSheet;",	"#I",	'n',	"",	"associated style info"},
	{0, "summary",	"TABLE",	"%Text;",	"#I",	'n',	"",	"purpose/structure for speech output"},
	{0, "tabindex",	"A, AREA, BUTTON, INPUT, OBJECT, SELECT, TEXTAREA",	"NUMBER",	"#I",	'n',	"",	"position in tabbing order"},
	{0, "target",	"A, AREA, BASE, FORM, LINK",	"%FrameTarget;",	"#I",	'f',	"L",	"render in this frame"},
	{0, "text",	"BODY",	"%Color;",	"#I",	'D',	"L",	"document text color"},
	{0, "title",	"@BASE, BASEFONT, HEAD, HTML, META, PARAM, SCRIPT, TITLE",	"%Text;",	"#I",	'n',	"",	"advisory title"},
	{0, "type",	"A, LINK",	"%ContentType;",	"#I",	'n',	"",	"advisory content type"},
	{0, "type",	"OBJECT",	"%ContentType;",	"#I",	'n',	"",	"content type for data"},
	{0, "type",	"PARAM",	"%ContentType;",	"#I",	'n',	"",	"content type for value when valuetype=ref"},
	{0, "type",	"SCRIPT",	"%ContentType;",	"#R",	'n',	"",	"content type of script language"},
	{0, "type",	"STYLE",	"%ContentType;",	"#R",	'n',	"",	"content type of style language"},
	{0, "type",	"INPUT",	"%InputType;",	"TEXT",	'n',	"",	"what kind of widget is needed"},
	{0, "type",	"LI",	"%LIStyle;",	"#I",	'D',	"L",	"list item style"},
	{0, "type",	"OL",	"%OLStyle;",	"#I",	'D',	"L",	"numbering style"},
	{0, "type",	"UL",	"%ULStyle;",	"#I",	'D',	"L",	"bullet style"},
	{0, "type",	"BUTTON",	"(button|submit|reset)",	"submit",	'n',	"",	"for use as form button"},
	{0, "usemap",	"IMG, INPUT, OBJECT",	"%URI;",	"#I",	'n',	"",	"use client-side image map"},
	{0, "valign",	"COL, COLGROUP, TBODY, TD, TFOOT, TH, THEAD, TR",	"(top|middle|bottom|baseline)",	"#I",	'n',	"",	"vertical alignment in cells"},
	{0, "value",	"INPUT",	"CDATA",	"#I",	'n',	"",	"Specify for radio buttons and checkboxes"},
	{0, "value",	"OPTION",	"CDATA",	"#I",	'n',	"",	"defaults to element content"},
	{0, "value",	"PARAM",	"CDATA",	"#I",	'n',	"",	"property value"},
	{0, "value",	"BUTTON",	"CDATA",	"#I",	'n',	"",	"sent to server when submitted"},
	{0, "value",	"LI",	"NUMBER",	"#I",	'D',	"L",	"reset sequence number"},
	{0, "valuetype",	"PARAM",	"(DATA|REF|OBJECT)",	"DATA",	'n',	"",	"How to interpret value"},
	{0, "version",	"HTML",	"CDATA",	"%HTML.Version;",	'D',	"L",	"Constant"},
	{0, "vlink",	"BODY",	"%Color;",	"#I",	'D',	"L",	"color of visited links"},
	{0, "vspace",	"APPLET, IMG, OBJECT",	"%Pixels;",	"#I",	'D',	"L",	"vertical gutter"},
	{0, "width",	"HR",	"%Length;",	"#I",	'D',	"L",	""},
	{0, "width",	"IFRAME",	"%Length;",	"#I",	'f',	"L",	"frame width"},
	{0, "width",	"IMG, OBJECT",	"%Length;",	"#I",	'n',	"",	"override width"},
	{0, "width",	"TABLE",	"%Length;",	"#I",	'n',	"",	"table width"},
	{0, "width",	"TD, TH",	"%Length;",	"#I",	'D',	"L",	"width for cell"},
	{0, "width",	"APPLET",	"%Length;",	"#R",	'D',	"L",	"initial width"},
	{0, "width",	"COL",	"%MultiLength;",	"#I",	'n',	"",	"column width specification"},
	{0, "width",	"COLGROUP",	"%MultiLength;",	"#I",	'n',	"",	"default width for enclosed COLs"},
	{0, "width",	"PRE",	"NUMBER",	"#I",	'D',	"L",	"predefined characters width"},
	{0, NULL, NULL, NULL, NULL, '\0', NULL, NULL}};
sAttrRefer* gHtmlParser::lAttrRef=nil;

t_int16 gHtmlParser::nDefAttrNorms=-1;
sAttrNorm gHtmlParser::lDefAttrNorms[]={
	{ -2, "\0", NULL },  // all tags norm-behaviour
	{ XH_IDTAG_ANCHOR, "HREF", NULL },
	{ XH_IDTAG_IMG, "SRC, ALT", NULL },
	{ -1, NULL, NULL }};

// The following table was based on HTML-4.01:
//		http://www.w3.org/TR/html401/index/attributes.html
//
sHAttrType gHtmlParser::lHAttrTypes[]={
	{ -2, "%HTML-Attribute-types", e_HAT_None, e_HAT_Single }, // typeFamily here indicates nr of types
	{ 1, "CDATA", e_HAT_CDATA, e_HAT_Single },
	{ 1, "ID", e_HAT_ID, e_HAT_Single },
	{ 1, "IDREF", e_HAT_IDREF, e_HAT_Single },
	{ 1, "IDREFS", e_HAT_IDREFS, e_HAT_IDREF },
	{ 1, "NAME", e_HAT_NAME, e_HAT_Single },
	{ 1, "NUMBER", e_HAT_NUMBER, e_HAT_Single },
	{ 0, "CAlign", e_HATp_CAlign, e_HAT_Single },  // "%CAlign;" referred; below, "%Character;", etc.
	{ 0, "Character", e_HATp_Character, e_HAT_Single },
	{ 0, "Charset", e_HATp_Charset, e_HAT_Single },
	{ 0, "Charsets", e_HATp_Charsets, e_HATp_Charset },
	{ 0, "Color", e_HATp_Color, e_HAT_Single },
	{ 0, "ContentType", e_HATp_ContentType, e_HAT_Single },
	{ 0, "ContentTypes", e_HATp_ContentTypes, e_HATp_ContentType },
	{ 0, "Coords", e_HATp_Coords, e_HAT_Single },
	{ 0, "Datetime", e_HATp_DateTime, e_HAT_Single },
	{ 0, "FrameTarget", e_HATp_FrameTarget, e_HAT_Single },
	{ 0, "HTML.Version", e_HATp_HtmlVersion, e_HAT_Single },
	{ 0, "IAlign", e_HATp_IAlign, e_HAT_Single },
	{ 0, "InputType", e_HATp_InputType, e_HAT_Single },
	{ 0, "LAlign", e_HATp_LAlign, e_HAT_Single },
	{ 0, "LanguageCode", e_HATp_LanguageCode, e_HAT_Single },
	{ 0, "Length", e_HATp_Length, e_HAT_Single },
	{ 0, "LinkTypes", e_HATp_LinkTypes, e_HAT_Single },
	{ 0, "LIStyle", e_HATp_LIStyle, e_HAT_Single },
	{ 0, "MediaDesc", e_HATp_MediaDesc, e_HAT_Single },
	{ 0, "MultiLength", e_HATp_MultiLength, e_HAT_Single },
	{ 0, "MultiLengths", e_HATp_MultiLengths, e_HATp_MultiLength },
	{ 0, "OLStyle", e_HATp_OLStyle, e_HAT_Single },
	{ 0, "Pixels", e_HATp_Pixels, e_HAT_Single },
	{ 0, "Scope", e_HATp_Scope, e_HAT_Single },
	{ 0, "Script", e_HATp_Script, e_HAT_Single },
	{ 0, "Shape", e_HATp_Shape, e_HAT_Single },
	{ 0, "StyleSheet", e_HATp_StyleSheet, e_HAT_Single },
	{ 0, "TAlign", e_HATp_TAlign, e_HAT_Single },
	{ 0, "Text", e_HATp_Text, e_HAT_Single },
	{ 0, "TFrame", e_HATp_TFrame, e_HAT_Single },
	{ 0, "TRules", e_HATp_TRules, e_HAT_Single },
	{ 0, "ULStyle", e_HATp_ULStyle, e_HAT_Single },
	{ 0, "URI", e_HATp_URI, e_HAT_Single },
	{ 10, "1|0", e_HATp_EnumChoice, -1 },	// Here start the choices
	{ 10, "button|submit|reset", e_HATp_EnumChoice, -1 },
	{ 10, "checked", e_HATp_EnumChoice, -1 },
	{ 10, "compact", e_HATp_EnumChoice, -1 },
	{ 10, "DATA|REF|OBJECT", e_HATp_EnumChoice, -1 },
	{ 10, "declare", e_HATp_EnumChoice, -1 },
	{ 10, "defer", e_HATp_EnumChoice, -1 },
	{ 10, "disabled", e_HATp_EnumChoice, -1 },
	{ 10, "GET|POST", e_HATp_EnumChoice, -1 },
	{ 10, "ismap", e_HATp_EnumChoice, -1 },
	{ 10, "left|all|right|none", e_HATp_EnumChoice, -1 },
	{ 10, "left|center|right|justify|char", e_HATp_EnumChoice, -1 },
	{ 10, "left|center|right|justify", e_HATp_EnumChoice, -1 },
	{ 10, "left|center|right", e_HATp_EnumChoice, -1 },
	{ 10, "ltr|rtl", e_HATp_EnumChoice, -1 },
	{ 10, "multiple", e_HATp_EnumChoice, -1 },
	{ 10, "nohref", e_HATp_EnumChoice, -1 },
	{ 10, "noresize", e_HATp_EnumChoice, -1 },
	{ 10, "noshade", e_HATp_EnumChoice, -1 },
	{ 10, "nowrap", e_HATp_EnumChoice, -1 },
	{ 10, "readonly", e_HATp_EnumChoice, -1 },
	{ 10, "selected", e_HATp_EnumChoice, -1 },
	{ 10, "top|middle|bottom|baseline", e_HATp_EnumChoice, -1 },
	{ 10, "yes|no|auto", e_HATp_EnumChoice, -1 },
	{ -2, "4.01", e_HAT_None, -1 },
	{ -2, "v0.0", e_HAT_None, -1 }};

// 7591 is prime, see http://www.math.utah.edu/~alfeld/math/primelist.html
gHashTriple* gHtmlParser::hElems=nil;// SIZE( 7591 )
////////////////////////////////////////////////////////////
sAttrRefer::sAttrRefer (t_int16 nAttrs)
    : nlAttrs( nAttrs ),
      sAttrs( nil ),
      lstRelated( nil ),
      idxUniqMax( -1 ),
      sLUniqs( nil ),
      idxLUniqs( nil )
{
 sAttrs = new gString[ nlAttrs ];
 ASSERTION(sAttrs!=nil,"sAttrs!=nil");
 lstRelated = new gSmartList[ nlAttrs ];
 ASSERTION(lstRelated!=nil,"lstRelated!=nil");
 lstIsAllButEtc = new gSwitch[ nlAttrs ];
 ASSERTION(lstIsAllButEtc!=nil,"lstIsAllButEtc!=nil");
 sLUniqs = new gString[ nlAttrs ];
 ASSERTION(sLUniqs!=nil,"sLUniqs!=nil");
 idxLUniqs = new t_int16[ nlAttrs ];
 ASSERTION(idxLUniqs!=nil,"idxLUniqs!=nil");
 for (t_int16 idx=0; idx<nlAttrs; idx++) {
     idxLUniqs[ idx ] = -1;
 }
}

sAttrRefer::~sAttrRefer ()
{
 delete[] sAttrs;
 delete[] lstRelated;
 delete[] lstIsAllButEtc;
 delete[] sLUniqs;
 delete[] idxLUniqs;
}

t_int16 sAttrRefer::FindAttr (char* attrName, t_int16& uniqIdx)
{
 // Returns the first absolute index (for lAttrRef), or -1 if not found.
 for (uniqIdx=0; uniqIdx<=idxUniqMax; uniqIdx++) {
     if ( sLUniqs[ uniqIdx ].Match( attrName ) ) {
	 return idxLUniqs[ uniqIdx ];
     }
 }
 uniqIdx = -1;
 return -1;
}

t_int16 sAttrRefer::FindAttr (char* attrName, char* strTag, t_int16& uniqIdx)
{
 bool isAllBut;
 t_int16 idxAbs, firstIdx = FindAttr( attrName, uniqIdx );

 sOutHelper.SetEmpty();

 if ( firstIdx<0 ) return -1;
 ASSERTION(uniqIdx>=0,"uniqIdx>=0");

 // Within attributes firstIdx to the last attribute with the same name,
 // compare 'strTag' with respective attribute 'related tags'
 for (idxAbs=firstIdx; idxAbs<nlAttrs; idxAbs++) {
     isAllBut = lstIsAllButEtc[ idxAbs ].IsOn();
     if ( isAllBut==false ) {
	 // Usual case...
	 if ( lstRelated[ idxAbs ].Match( strTag )>0 )
	     return idxAbs;  // Note uniqIdx surely matches!
     }
     else {
	 if ( lstRelated[ idxAbs ].Match( strTag )==0 )
	     return idxAbs;
	 sOutHelper.Set( lstRelated[ idxAbs ].Str(1) );
	 // TODO: replicate for all other finds
     }

     // We can afford to make the following evaluation here
     // (since the first match occurred)
     if ( sAttrs[ idxAbs ].Match( attrName )==false ) break;
 }
 // No attribute for the tag: 'strTag'
 return -1;
}
////////////////////////////////////////////////////////////
gHtmlCouple::gHtmlCouple (unsigned lineNr, char* sText)
    : iLine( lineNr ),
      idTag( XH_NOTAG ),
      idEndTag( -1 ),
      pHStr( nil ),
      pElem( nil ),
      synError( 0 ),
      coupleId( -1 ),
      theDocType( 0 ),
      oCouple( nil )
{
 if ( sText!=nil ) AddText( sText );
}

gHtmlCouple::gHtmlCouple (unsigned lineNr, char* strTag, char* sAttrLst, bool doAddSkippedTags)
    : iLine( lineNr ),
      idTag( -1 ),
      idEndTag( -1 ),
      pHStr( nil ),
      pElem( nil ),
      synError( 0 ),
      coupleId( -1 ),
      theDocType( 0 ),
      oCouple( nil )
{
 DBGPRINT_MIN("DBG: Couple:%s, sAttrLst:%s\n",strTag,sAttrLst);
 AddTag( strTag );
 if ( synError==0 || (synError==XH_SKIP_TAG && doAddSkippedTags==true) ) {
     if ( sAttrLst!=nil ) {
	 Add( sAttrLst );
	 // Update attrL with the list of attributes found in string 'sAttrLst'
	 attrL.Set( sAttrLst );
     }
 }
}

gHtmlCouple::~gHtmlCouple ()
{
 delete pHStr;
 delete oCouple;
}

bool gHtmlCouple::IsOk ()
{
 ASSERTION(IsText() || (IsText()==false && ((idTag>=0 && pElem!=nil) || (idTag==-1 && synError==-1 && pElem==nil))),"gHtmlCouple::IsOk");
 return synError==0;
}

char* gHtmlCouple::GetStr ()
{
 char* str;
 sWholeTag.SetEmpty();
 if ( IsText() ) {
     str = Str( 1 );
     if ( pHStr==nil ) return str;
     char* strX = pHStr->Str();
     if ( strX==nil ) return nil;  // Never happens
     if ( strX[0]==0 ) return str;  // May happen if there is an invalid symbol within the string (e.g. amp: &)
     return strX;
 }
 sWholeTag.Add( '<' );
 if ( IsTagEnd() ) sWholeTag.Add( '/' );
 sWholeTag.AddString( sTag );
 if ( N()==2 ) {
     sWholeTag.Add( ' ' );
     // attrL.Str provides the full concatenated string of the attribute assignments
     sWholeTag.Add( attrL.Str() );
 }
 sWholeTag.Add( '>' );
 return sWholeTag.Str();
}

char* gHtmlCouple::GetStrForTree ()
{
 char* str;

 sKeepStr = "TODO:::TODO:::";
 sWholeTag.SetEmpty();
 if ( IsText() ) {
     str = Str( 1 );
     if ( pHStr==nil ) return str;
     char* strX = pHStr->Str();
     if ( strX==nil ) return nil;  // Never happens
     if ( strX[0]==0 ) return str;  // May happen if there is an invalid symbol within the string (e.g. amp: &)
     return strX;
 }
 sWholeTag.Add( '<' );
 if ( IsTagEnd() ) sWholeTag.Add( '/' );
 sWholeTag.AddString( sTag );
 if ( N()==2 ) {
     sWholeTag.Add( ' ' );
     // attrL.Str provides the full concatenated string of the attribute assignments
     sWholeTag.Add( attrL.Str() );
 }
 sWholeTag.Add( '>' );
 sprintf(sKeepStr.Str(),"[%u]",N());
 sWholeTag.AddString( sKeepStr );
 return sWholeTag.Str();
}

gString& gHtmlCouple::TagString (bool forceEnd)
{
 sTagStr.SetEmpty();
 if ( IsText() ) return sTagStr;
 sTagStr.Add( '<' );
 if ( forceEnd || IsTagEnd() ) sTagStr.Add( '/' );
 sTagStr.AddString( sTag );
 sTagStr.Add( '>' );
 return sTagStr;
}

char* gHtmlCouple::GetHRef ()
{
 if ( IsText()==true ) return nil;
 if ( IsAnchor()==false || IsTagEnd()==true ) return nil;
 return attrL.Find( "HREF", true );
}

unsigned gHtmlCouple::Add (char* s)
{
 gString sTrim( s );
 sTrim.Trim();
 return gList::Add( sTrim );
}

unsigned gHtmlCouple::Add (gString& copy)
{
 return Add( copy.Str() );
}

unsigned gHtmlCouple::AddText (char* s)
{
 // Return 1 iff the text is valid
 Add( s );
 ASSERTION(N()==1,"N()==1");
 s = Str( 1 );  // Trimmed text string
 pHStr = new gHtmlString( s );
 ASSERTION(pHStr!=nil,"pHStr!=nil");
 return (unsigned)pHStr->IsOk();
}

unsigned gHtmlCouple::AddTag (char* strTag)
{
 t_uchar uChr;
 // Return 0 if tag is empty, or basically invalid; 1 otherwise
 ASSERTION(strTag!=nil,"strTag!=nil");
 gString s( strTag );
 s.Trim();
 if ( s.IsEmpty() ) return 0;
 gString sUp( s );
 uChr = s[1];
 if ( uChr=='!' || uChr=='?' ) {
     synError = XH_SKIP_TAG;
     theDocType = (t_int16)sUp.Match("!DOCTYPE");
 }
 // synError<-1 not yet implemented (at least here not used)
 sUp.UpString();
 gList::Add( sUp ); // Add(char*) is not called, it is not virtual, ok
 sTag = sUp;
 return 1;
}
////////////////////////////////////////////////////////////
gHtmlContent::gHtmlContent ()
    : nLines( 0 ),
      theHtmlOpt( nil )
{
}

gHtmlContent::~gHtmlContent ()
{
}

gHtmlCouple* gHtmlContent::GetCouple (unsigned idx)
{
 ASSERTION(idx>0,"idx>0");
 if ( IsValidIndex(idx)==false ) return nil;
 gStorage* pObjx = GetObjectPtr( idx );
 ASSERTION(pObjx!=nil,"pObjx!=nil");
 return (gHtmlCouple*)pObjx;
}

char* gHtmlContent::Str (unsigned idx)
{
 gHtmlCouple* pObjC = GetCouple( idx );
 if ( pObjC==nil ) return nil;
 // => simplified method: now GetStr made by HtmlCouple!
 return pObjC->GetStr();
}

unsigned gHtmlContent::Add (char* s)
{
 nLines++;
 if ( s==nil || s[0]==0 ) return 0;
 thisAddHmtlLine( nLines, s );
 return nLines;
}

int gHtmlContent::UpCaseAttributes (char* strTag, gString& sRes)
{
 // Return 0 if ok; note it will eventually manipulates sRes (if tidyed)
 unsigned i, n;
 unsigned idx;
 t_uchar uChr;
 short quoteCount=0;
 bool tidyApplied=false;

 ASSERTION(strTag!=nil,"strTag!=nil");

 // Check if tidy was requested
 if ( GetHtmlOpt().outOpt.HasTidy( strTag, idx ) ) {
     // Apply tidy
     gString s;
     // Tidy blanks first
     for (i=1, n=sRes.Length(); i<=n; i++) {
	 uChr = sRes[i];
	 if ( uChr=='"' ) quoteCount = quoteCount==0;
	 if ( uChr=='=' && quoteCount==0 ) {
	     s.TrimRight();
	 }
	 if ( uChr==' ' ) {
	     if ( sRes[i-1]=='=' ) continue;
	 }
	 s.Add( uChr );
     }
     sRes = s;
     DBGPRINT("DBG: Apply tidy for [%s|%s] (idx=%u)\n",strTag,sRes.Str(),idx);
     s.SetEmpty();
     gParam aParam( sRes, " ", gParam::e_NormalQuoted );
     // Now split each param, and add them
     for (i=1, n=aParam.N(); i<=n; i++) {
	 gString sAssign;
	 gString sTemp( aParam.Str(i) );
	 sTemp.Trim();
	 gParam paramVal( sTemp, "=", gParam::e_StopSplitOnFirst );
	 unsigned nVal = paramVal.N();
	 ASSERTION(nVal==1 || nVal==2,"nVal...");
	 sAssign.Set( paramVal.Str(1) );
	 if ( nVal>=2 ) {
	     sAssign.UpString();  // Only the assign-lval (e.g. SRC of a <IMG SRC=...>)
	     // Check if attribute-value is to be tidyed
	     bool doTidyAttr = GetHtmlOpt().outOpt.HasTidyAttr( strTag, sAssign );
	     gString sVal( paramVal.Str(2) );
	     sVal.Trim();  // academic
	     if ( doTidyAttr ) {
		 if ( sVal[1]!='"' && sVal[1]!='\'' ) {
		     gString sTempVal( "\"" );
		     sTempVal.AddString( sVal );
		     sTempVal.Add( "\"" );
		     sVal = sTempVal;
		 }
		 else {
		     switch ( sVal[1] ) {
		     case '"':
			 if ( sVal[sVal.Length()]=='\'' ) {
			     sVal[sVal.Length()] = '"';
			 }
			 break;
		     case '\'':
			 // Tidy 'abc' to "ABC" if double-quote not found
			 if ( sVal.Find( '"' )==0 && sVal[sVal.Length()]=='\'' ) {
			     sVal[1] = '"';
			     sVal[sVal.Length()] = '"';
			 }
		     default:
			 break; //nothing
		     }
		 }
	     }//end IF doTidy...
	     sAssign.Add( "=" );
	     sAssign.AddString( sVal );
	 }
	 DBGPRINT("DBG: param: [%s]\n",sAssign.Str());
	 if ( i>1 ) s.Add( " " );
	 s.AddString( sAssign );
     }
     tidyApplied = sRes.Match( s )==false;

     DBGPRINT_WEB3("DBG: Applied tidy(%c): [%s|%s]\n",ISyORn(tidyApplied),sRes.Str(),s.Str());

     // Replace the result
     sRes = s;
 }

 for (i=1, n=sRes.Length(), quoteCount=0; i<=n; i++) {
     uChr = sRes[i];
     if ( uChr=='"' ) quoteCount = quoteCount==0;
     if ( quoteCount ) continue;
     if ( uChr>='a' && uChr<='z' ) sRes[i] = uChr-32;
 }

 return quoteCount!=0;
}

gHtmlOpt& gHtmlContent::GetHtmlOpt ()
{
 ASSERTION(theHtmlOpt!=nil,"theHtmlOpt!=nil");
 return *theHtmlOpt;
}

bool gHtmlContent::SetHtmlOpt (gHtmlOpt* pHtmlOpt)
{
 return (theHtmlOpt = pHtmlOpt)!=nil;
}

int gHtmlContent::TagError (unsigned lineNr, int error, char* sLine, char* sShortMsg)
{
 FILE* fRepErr = stderr;

 if ( error==0 ) return 0;
 HTML_LOG(fRepErr,LOG_ERROR,"Line %u: Invalid syntax (int-ref %d): '%s'. %s\n",
	  lineNr,
	  error,
	  sLine,
	  sShortMsg);
 return error;
}

void gHtmlContent::Show (bool doShowAll)
{
 unsigned i, n=N();

 for (i=1; i<=n; i++) {
     if ( doShowAll ) printf("C%u/%u: ",i,n);
     printf("%s\n",Str(i));
 }
}

int gHtmlContent::thisAddHmtlLine (unsigned lineNr, char* s)
{
 gString sLine( s ), sRes;
 unsigned pos, posEnd, posTagEnd, len=sLine.Length();

 ASSERTION(len>0,"len>0");
 sLine.Trim();  // Not absolutely necessary
 pos = sLine.Find('<');
 posEnd = sLine.Find('>');
 if ( pos>0 ) {
     if ( posEnd>0 && posEnd+1>pos ) {
	 sRes.CopyFromTo( sLine, pos+1, posEnd-1 );
	 sRes.Trim();
	 posTagEnd = sRes[1]=='/';
	 return thisAddHtmlTag( lineNr, sRes.Str()+posTagEnd, posTagEnd>0 );
     }
 }
 return thisAddHtmlText( lineNr, sLine.Str() );
}

int gHtmlContent::thisAddHtmlText (unsigned lineNr, char* sText)
{
 // Returns always 0
 gHtmlCouple* pCouple;

 pCouple = new gHtmlCouple( lineNr, sText );
 thisAddCouple( pCouple, *this );
 return 0;
}

int gHtmlContent::thisAddHtmlTag (unsigned lineNr, char* strTag, bool isEndTag)
{
 // Return 0 if all ok
 int error=0;
 unsigned pos;
 gHtmlCouple* pCouple;

 ASSERTION(strTag!=nil,"strTag!=nil");

 // sTag can be just, e.g. "BODY"
 // or eventually "BODY LINK=xyz"
 pos = gStrControl::Self().Find( strTag, " " );
 if ( pos==0 ) {
     pCouple = new gHtmlCouple( lineNr, strTag, nil );
 }
 else {
     gString sMainTag( strTag );
     sMainTag[ pos ] = 0;
     DBGPRINT("DBG: thisAddHtmlTag: %s [%s|%s]\n",strTag,sMainTag.Str(),strTag);
     strTag += pos;
     gString sAttr( strTag );
     sAttr.Trim();
     UpCaseAttributes( sMainTag.Str(), sAttr );
     pCouple = new gHtmlCouple( lineNr, sMainTag.Str(), sAttr.Str() );
     error = pCouple->attrL.lastOpError;
     TagError( lineNr, error, strTag, "Invalid attribute" );
 }
 pCouple->idEndTag = isEndTag ? XH_ENDTAG : -1;  // Instead of being limitted to pos==0.
 thisAddCouple( pCouple, *this );
 return error;
}

int gHtmlContent::thisAddCouple (gHtmlCouple* pCouple, gList& oL)
{
 ASSERTION(pCouple!=nil,"pCouple!=nil");
 oL.AppendObject( pCouple );
 return 0;
}
////////////////////////////////////////////////////////////
gHtmlCouple* gHParsed::CurrentCouple ()
{
 gHtmlCouple* pCouple;
 ASSERTION(Depth()>0,"Depth()>0");
 pCouple = (gHtmlCouple*)GetCurrent();
 ASSERTION(pCouple!=nil,"pCouple!=nil");
 return pCouple;
}

char* gHParsed::Str (unsigned idx)
{
 return GetCouple( idx )->GetStr();
}

char* gHParsed::StrMust (unsigned idx)
{
 gHtmlCouple* pCouple;
 pCouple = (gHtmlCouple*)kMust.GetObjectPtr( idx );
 ASSERTION(pCouple!=nil,"pCouple!=nil");
 return pCouple->GetStr();
}

gHtmlCouple* gHParsed::GetCouple (unsigned idx)
{
 ASSERTION(IsValidIndex(idx),"IsValidIndex(idx)");
 gStorage* pObx = GetObjectPtr( idx );
 ASSERTION(pObx!=nil,"pObx!=nil");
 return (gHtmlCouple*)pObx;
}

gHtmlCouple* gHParsed::FindCouple (t_int16 idTag)
{
 unsigned idx, n=N();
 gHtmlCouple* pCouple;

 if ( idTag<0 ) return nil;
 for (idx=1; idx<=n; idx++) {
     pCouple = GetCouple( idx );
     if ( pCouple->idTag==idTag ) return pCouple;
 }
 return nil;
}

void gHParsed::PushTagOptEnd (gHtmlCouple& couple)
{
 thisPushCouple( couple, true );
}

int gHParsed::PushTag (gHtmlCouple& couple, bool doCheckOnly)
{
 t_int16 id = couple.idTag;
 int error=0;

 if ( doCheckOnly==false ) thisPushCouple( couple, false );

 ///printf("DBG: ###PUSH: "); for (unsigned dbgIdx=1; dbgIdx<=kMust.N(); dbgIdx++) printf("[%s]",StrMust(dbgIdx)); printf("DBG: ###(%s).\n",couple.sTag.Str());

 switch ( id ) {
 case XH_IDTAG_BASE:
     // Cowardly not assuming an error,
     // because here you base use, e.g.: '<BASE TARGET="_blank">'
     // and also '<BASE HREF...>'; therefore we choose not to check.
     // COMMENTED, see above: if ( hasBaseHRef ) return 101;  // More than one BASE_HREF
     hasBaseHRef = true;
     if ( state!=e_HS_Head ) return 102;  // Badly placed BASE_HREF
     break;
 case XH_IDTAG_BODY:
     error = state==e_HS_HeadAfter ? 0 : -103; // BODY not after HEAD
     state = e_HS_Body;
     break;
 case XH_IDTAG_HEAD:
     error = state==e_HS_Html || state==e_HS_StartAfter ? 0 : -104; // HEAD not after HTML
     state = e_HS_Head;
     break;
 case XH_IDTAG_HTML:
     error = state==e_HS_Start ? 0 : -105; // <HTML> not at begining
     state = e_HS_Html;
     break;
 default:
     if ( state==e_HS_Start ) {
	 state = e_HS_StartAfter;
	 return -106;  // <HTML> optionally expected
     }
     if ( state==e_HS_End ) return 107;  // </HTML> already parsed
     if ( state==e_HS_HeadAfter ) return -108;  // Invalid tag after </HEAD>
     if ( state==e_HS_BodyAfter ) return -109;  // Invalid tag after </BODY>
     return 0;
 }

 return error;
}

int gHParsed::PopTag (bool hasOptEnd)
{
 t_int16 id;
 int error=0;
 gHtmlCouple* pCouple = CurrentCouple();

 ///printf("DBG: ###POP_: "); for (unsigned dbgIdx=1; dbgIdx<=kMust.N(); dbgIdx++) printf("[%s]",StrMust(dbgIdx)); printf("DBG: ### line %u:(%s).\n",((gHtmlCouple*)kMust.EndPtr()->me)->iLine,((gHtmlCouple*)kMust.EndPtr()->me)->sTag.Str());

 id = pCouple->idTag;

 Pop();
 if ( hasOptEnd ) return 0;

 kMust.Pop();

 switch ( id  ) {
 case XH_IDTAG_BODY:
     error = state==e_HS_Body ? 0 : 121;
     state = e_HS_BodyAfter;
     break;
 case XH_IDTAG_HEAD:
     error = state==e_HS_Head ? 0 : 122;
     state = e_HS_HeadAfter;
     break;
 case XH_IDTAG_HTML:
     error = state==e_HS_BodyAfter ? 0 : 123;
     state = e_HS_End;
     break;
 default:
     break;
 }

 return 0;
}

void gHParsed::TrashLast ()
{
 DBGPRINT("DBG: trash: %s (N=%u kMustN=%u)\n",CurrentCouple()->sTag.Str(),N(),kMust.N());
 Pop();
}

int gHParsed::thisPush (gHtmlCouple& couple, gStack& aStack)
{
 gHtmlCouple* newCouple;
 newCouple = new gHtmlCouple( couple.iLine, couple.sTag.Str(), nil );
 ASSERTION(newCouple!=nil,"newCouple!=nil");
 newCouple->CopyTag( couple );
 aStack.Push( newCouple );
 return 0;
}

int gHParsed::thisPushCouple (gHtmlCouple& couple, bool hasOptEnd)
{
 thisPush( couple, *this );
 if ( hasOptEnd==false ) thisPush( couple, kMust );
 //printf("DBG: PUSH: %s (hasOptEnd?%d)\n",couple.sTag.Str(),hasOptEnd);
 return 0;
}
////////////////////////////////////////////////////////////
gHtmlCouple* gHList::GetCouple (unsigned idx)
{
 if ( idx>0 ) return gHtmlContent::GetCouple( idx );
 ASSERTION(coupleFakeBody!=nil,"coupleFakeBody!=nil");
 return coupleFakeBody;
}

void gHList::AppendCouple (gHtmlCouple& couple)
{
 gHtmlCouple* newCouple;
 newCouple = new gHtmlCouple( couple.iLine, couple.sTag.Str(), nil );
 ASSERTION(newCouple!=nil,"newCouple!=nil");
 // If there are attributes, copy them also
 if ( couple.attrL.IsEmpty()==false ) {
     newCouple->Add( couple.attrL.Str() );
 }
 newCouple->CopyTag( couple );
 AppendObject( newCouple );
}
////////////////////////////////////////////////////////////
gHtmlParser::gHtmlParser (gUnweb* ptrUnweb)
    : docType( 0 ),
      nErrorsSyntax( 0 ),
      nErrorsOther( 0 ),
      nWarnings( 0 ),
      lastWarnOpCode( 0 ),
      pUnweb( ptrUnweb )
{
 lastOpError = thisInitTbl( nElems );
 ASSERTION(lastOpError==0,"lastOpError==0");
}

gHtmlParser::~gHtmlParser ()
{
}

sHtmlElement* gHtmlParser::GetTagElement (t_int16 idxTag)
{
 if ( idxTag<0 || idxTag>nElems ) return nil;
 return &lElems[ idxTag ];
}

sAttrDef* gHtmlParser::GetAttrDef (t_int16 idxAttr)
{
 ASSERTION(lAttrRef!=nil,"lAttrRef!=nil");
 ASSERTION(idxAttr>=0 && idxAttr<lAttrRef->nlAttrs,"GetAttrDef(3)");
 return &lAttrs[ idxAttr ];
}

sAttrRefer* gHtmlParser::GetAttrRef ()
{
 ASSERTION(lAttrRef!=nil,"lAttrRef!=nil");
 return lAttrRef;
}

sAttrNorm* gHtmlParser::GetDefaultAttrNorm (t_int16 idxNorm)
{
 if ( idxNorm<0 || idxNorm>nDefAttrNorms ) return nil;
 return &lDefAttrNorms[ idxNorm ];
}

sHAttrType* gHtmlParser::GetAttrType (t_int16 idxType)
{
 t_int16 n = GetNAttrType();
 if ( idxType<1 || idxType>n ) return nil;
 return &lHAttrTypes[ idxType ];
}

void gHtmlParser::ReleaseHash ()
{
 t_int16 idx;
 // Delete hash for tags
 delete hElems; hElems = nil;
 nElems = -1;
 // Delete "Related Elements" lists
 delete lAttrRef; lAttrRef = nil;

 for (idx=0; idx<=nDefAttrNorms; idx++) {
     delete lDefAttrNorms[ idx ].pAttrSeq;
     lDefAttrNorms[ idx ].pAttrSeq = nil;
 }
}

bool gHtmlParser::SetOptions (gHtmlOpt& copy)
{
 htmlOpt.CopyOptions( copy );

 t_uint16 i, n=htmlOpt.lIdTagOptEnd.N();
 t_int16 id, nMax = nElems;
 for (i=1; i<=n; i++) {
     id = htmlOpt.lIdTagOptEnd.GetInt( i );
     if ( id<0 || id>=nMax ) return false;
     char cOpt = lElems[id].optEnd;
     DBGPRINT("DBG: CFG_OPTEND: id=%d: %s (%c)\n",id,lElems[id].elemName,cOpt);
     if ( cOpt!='@' && cOpt!=XH_TBL_TAG_OPTEND ) return false;
     lElems[id].optEnd = XH_TBL_TAG_OPTEND_CFG;
 }

 htmlInput.SetHtmlOpt( &htmlOpt );

 return true;
}

int gHtmlParser::Parse (FILE* fRepErr)
{
 int error = thisParse( fRepErr );

 return error;
}

int gHtmlParser::SetError (int opError)
{
  lastOpError = opError;
  gControl::SetError( 0 );  // sStrError init'zed
  if ( opError==0 ) return 0;
  return lastOpError = opError;
}

int gHtmlParser::SetWarn (int opError)
{
 lastWarnOpCode = opError;
 if ( opError==0 ) return 0;

 nWarnings++;
 return opError;
}

int gHtmlParser::ShowTree (FILE* fRepErr)
{
 // Return 0 on success; -1 if there is nothing to show
 short p, nFilled;
 unsigned n;
 gHSeq hSeq( fRepErr );
 eHState aState, showState=e_HS_Body;

 hSeq.SetBaseHRef( myBaseHRef );

#ifdef DEBUG
 for (p=0; p<(short)e_HS_Last; p++) {
     n = lParts[p].N();
     if ( n==0 ) continue;
     printf("ShowPart(%d)--->START<---\n",p);
     hSeq.ShowPart( lParts[p], (eHState)p, htmlOpt );
     printf("ShowPart(%d)--->END<---\n",p);
 }
#endif //DEBUG(_...)

 for (p=0, nFilled=0, aState=e_HS_Body; p<(short)e_HS_Last; p++) {
     n = lParts[p].N();
     if ( n==0 ) continue;
     if ( nFilled==0 ) {
	 aState = (eHState)p;
     }
     nFilled++;
     DBGPRINT("DBG: nFilled=%d, aState=%d, showState=%d\n",
	      (int)nFilled,
	      aState,
	      showState);
 }

 if ( nFilled<=0 ) return -1;

 if ( nFilled==1 ) {
     // Simulating we have a BODY statement at begin of doc
     showState = e_HS_Sp_FakeBody;
 }
 else {
     // Usually this is the normal case, LTD formatted docs contain in this order:
     //   HTML, TITLE, HEAD, BODY
     // and we want to show the doc body.
     aState = showState;
 }

 hSeq.Build( lParts[aState], showState );
 hSeq.hTree.Show( true );

 return 0;
}


#ifdef DEBUG
int gHtmlParser::Show_dbg (bool doShowAll)
{
 int ex;
 t_int16 id;
 for (ex=1; ex<=(int)htmlInput.N(); ex++) {
     gStorage* pObx = htmlInput.GetObjectPtr(ex);
     gHtmlCouple* pObC = (gHtmlCouple*)pObx;
     thisFindTag( pObC->sTag.Str(), id ) ;
     printf("DBG: line %u: %d/%d [%s]",
	    pObC->iLine,
	    ex, (short)htmlInput.N(),
	    htmlInput.Str(ex));
     if ( pObC->IsText() )
	 printf("\n");
     else
	 printf(" TAG:%s {id:%d}\n",pObC->sTag.Str(),id);
     /// OLD DEBUG:
     ///printf("DBG:[%s], @@@%d (OriginalInputLine=%u, synErr=%d %s)\n",pObC->Str(1),pObC->idTag,pObC->iLine,pObC->synError,pObC->IsTagEnd()?"IsTagEnd":(pObC->IsText()?"TXT":"Tag"));
     ///if ( pObC->pHStr!=nil ) printf("DBG:TEXT: [%s]\n\n",pObC->pHStr->Str());
 }
 return 0;
}
#endif //DEBUG

int gHtmlParser::thisFillFromUnweb (gUnweb& unweb, gHtmlContent& hInput)
{
 unsigned i, nPost = unweb.coordSerial.N();

 for (i=1; i<=nPost; i++) {
     hInput.Add( unweb.Str( i ) );
 }
 return 0;
}

sHtmlElement* gHtmlParser::thisFindTag (char* strTag, t_int16& idxTag)
{
 // Returns the element, or null if not found.
 sHtmlElement* pElem;
 gHashElemTriple* pHashElem;

 idxTag = -1;
 if ( strTag==nil || strTag[0]==0 ) return nil;

 //for (idxTag=0; idxTag<nElems; idxTag++) {
 //    pElem = &lElems[idxTag];
 //    if ( gStrControl::Self().Match(strTag,pElem->elemName) )
 //	 return pElem;
 //}

 gKey hKey( strTag );
 unsigned idx;
 int x;
 pHashElem = hElems->Find( hKey, idx, x );
 if ( pHashElem==nil ) return nil;
 // Particularity of this hash-function: for HTML tags, usually one per hash-row!
 // x is the row from the hash-table: usually 1.
 idxTag = (t_int16)pHashElem->iVal;
 ASSERTION(idxTag<nElems,"idxTag<nElems");
 pElem = &lElems[idxTag];

 /* --- Just show the entries on hash...
 if ( x==1 ) return pElem;
 printf("DBG::: thisFindTag: '%s', x=%d (iVal=%d)\n",strTag,x,pHashElem->iVal);
 for (idxTag=0; idxTag<=90; idxTag++) {
     gKey xKey(lElems[idxTag].elemName);
     pHashElem = hElems->Find(xKey,idx,x);
     printf("DBG::: idxTag=%d, {idx=%d:x=%d}, NAME=%s\n",idxTag,idx,x,lElems[idxTag].elemName);
 }
 */

 return pElem;
}

int gHtmlParser::thisInitTbl (t_int16& size)
{
 t_int16 idx, iN;
 sHtmlElement* pElem;
 sAttrDef* pAttrDef;
 char* elemName;
 char* attrName, *lastName;
 char* str;

 if ( size>0 ) return 0;

 if ( hElems==nil ) {
     hElems = new gHashTriple( 7591 );
     ASSERTION(hElems!=nil,"hElems!=nil");
 }
 if ( hElems->IsEmpty()==false ) return 0;

 for (idx=0, size=0;
      (pElem = &(lElems[idx]))!=nil &&
	  (elemName = pElem->elemName)!=nil;
      idx++) {
     size++;
     gKey hKey( elemName );
     hElems->AddTriple( hKey, (int)idx );
     switch ( pElem->family ) {
     case '@':
	 pElem->family = 0;
     case (char)e_HtmlElementPhrase:
	 break;
     default:
	 return -1;
     }
     // Check optStart is either 'O' or '@': 1 resp. 0
     switch ( pElem->optStart ) {
     case 'O':
	 pElem->optStart = 1;
	 break;
     case '@':
	 pElem->optStart = 0;
	 break;
     default:
	 ASSERTION_FALSE("optStart(1)");
	 break;
     }
     // isDeprecated becomes either 0 or non-zero
     char isDeprecated = pElem->isDeprecated;
     if ( isDeprecated=='@' ) isDeprecated = pElem->isDeprecated = 0;
     // If is deprecated, then for sure kindDTD is 'L'oose,
     // but not the otherway round (e.g. IFRAME is Loose, but not deprecated)
     if ( pElem->kindDTD=='@' ) pElem->kindDTD = 0;
     if ( isDeprecated!=0 && pElem->kindDTD!='L' ) return -1;
 }//end FOR

 for (idx=0; (attrName = lAttrs[idx].attrName)!=NULL; ) idx++;

 ASSERTION(lAttrRef==nil,"lAttrRef==nil");  // Un-initialized
 lAttrRef = new sAttrRefer( iN = idx );
 ASSERTION(lAttrRef!=nil,"lAttrRef!=nil");

 for (idx=0, lastName="@"; idx<iN; idx++) {
     pAttrDef = &lAttrs[ idx ];
     attrName = pAttrDef->attrName;
     gString s( attrName );
     s.UpString();
     lAttrRef->sAttrs[ idx ] = s;
     if ( strcmp( attrName, lastName ) ) {
	 // New attribute unique name
	 lAttrRef->idxUniqMax++;
	 lAttrRef->sLUniqs[ lAttrRef->idxUniqMax ] = s;
	 lAttrRef->idxLUniqs[ lAttrRef->idxUniqMax ] = idx;
     }
     elemName = pAttrDef->strRelatedLst;
     bool isAllBut = elemName[0]=='@';  // Indicates all tags but these...
     elemName += isAllBut;
     lAttrRef->lstIsAllButEtc[ idx ].SetOn( isAllBut );
     lAttrRef->lstRelated[ idx ].AddFromStr( elemName );

     // Keep last attribute name
     lastName = attrName;
 }//end FOR

 sAttrNorm* pNormAttr;

 ASSERTION(nDefAttrNorms==-1,"nDefAttrNorms==-1");
 for (idx=0, nDefAttrNorms=0;
      (pNormAttr = &lDefAttrNorms[idx])!=nil && pNormAttr->idTag!=-1;
      idx++) {
     str = pNormAttr->strAttrSeq;
     if ( str!=nil && str[0]!=0 ) {
	 pNormAttr->pAttrSeq = new gSmartList;
	 ASSERTION(pNormAttr->pAttrSeq!=nil,"pNormAttr->pAttrSeq!=nil");
	 pNormAttr->pAttrSeq->AddFromStr( str );
     }
     nDefAttrNorms++;
 }

 // Now refresh lHAttrTypes (list of attribute types)
 for (idx=1;
      lHAttrTypes[idx].typeFamily>=0;
     ) idx++;

 lHAttrTypes[ 0 ].typeFamily = idx;

 return 0;
}

int gHtmlParser::thisParse (FILE* fRepErr)
{
 int error;
 unsigned i, n;
 unsigned currLine = 0;
 t_int16 id;
 gHtmlCouple* pObjC;
 sHtmlElement* pElem;
 char* str;
 char* msgError;
 bool isSkippedTag;
 bool isInvalidTag;

 if ( pUnweb==nil ) return -1;
 if ( thisFillFromUnweb( *pUnweb, htmlInput )!=0 ) return -1;

 // Pre-parsing, setting id's
 for (i=1, n=htmlInput.N(); i<=n; i++) {
     pObjC = htmlInput.GetCouple( i );
     ASSERTION(pObjC!=nil,"pObjC!=nil");
     currLine = pObjC->iLine;
     if ( pObjC->IsText() ) {
	 // If pHStr->IsOk() is false, perhaps issue an error?
	 DBGPRINT_MIN("DBG: [%s]: pHStr:%c OK-pHStr:%c\n",pObjC->Str(1),ISyORn(pObjC->pHStr!=nil),ISyORn(pObjC->pHStr==nil?0:pObjC->pHStr->IsOk()));
	 continue;
     }
     str = pObjC->sTag.Str();
     /////printf("DBG: STRTAG: %s (%d) synError=%d\n",str,pObjC->idTag,pObjC->synError);
     if ( pObjC->theDocType!=0 ) {
	 // E.g.: HTML PUBLIC -//W3C//DTD HTML 4.01 Transitional//EN http://www.w3.org/TR/html4/loose.dtd
	 // (Given '...4.01...' string without quotes by attrL.GetAttrValue(3))
	 DBGPRINT("DBG: DocType: %s [%s]\n",pObjC->attrL.Str(),pObjC->attrL.GetAttrValue(3));
	 if ( docType==0 ) {
	     docType = pObjC->theDocType;  // Always 1
	 }
	 else {
	     SetWarn( 20 );
	     HTML_LOG(fRepErr,LOG_WARNING,"Line %u: %s\n",currLine,"Duplicated DOCTYPE declaration");
	     SetWarn( 0 );
	 }
     }
     pElem = thisFindTag( str, pObjC->idTag );
     pObjC->pElem = pElem;
     if ( pElem==nil ) {
	 // Check if tag is configured to be skipped
	 if ( htmlOpt.outOpt.IsTagSkipped( str ) )
	     pObjC->synError = XH_SKIP_TAG_FORCE;
     }
     else {
	 if ( pElem->ctrl==-1 )
	     pObjC->synError = XH_SKIP_TAG_FORCE;
     }
 }

 gHParsed kParsed;
 //eHState lastState = kParsed.state;

 // Parsing 'htmlInput'
 for (i=1; i<=n; i++) {
     pObjC = htmlInput.GetCouple( i );
     id = pObjC->idTag;
     str = htmlInput.Str( i );
     ASSERTION(str!=nil,"str!=nil");
     currLine = pObjC->iLine;
     if ( pObjC->IsText() ) {
	 // The state will pass to StartAfter if at least one string is present.
	 // The state change in this class is an exception.
	 if ( kParsed.state==e_HS_Start ) kParsed.state = e_HS_StartAfter;
	 if ( pObjC->pHStr!=nil && pObjC->pHStr->Str()!=nil && pObjC->pHStr->Str()[0]==0 ) {
	     HTML_LOG(fRepErr,LOG_NOTICE,"Line %u: invalid string, but used: '%s'\n",currLine,str);
	 }
	 thisAddedParsedLine( str, *pObjC, kParsed.state, true );
     }
     else {
	 error = thisParseLine( *pObjC, currLine, str, kParsed );
	 msgError = GetErrorStr();
	 isInvalidTag = lastOpError<0;
	 isSkippedTag = pObjC->IsSkippedTag();
	 if ( isSkippedTag ) {
	     ; // Skip tag
	     if ( pObjC->synError!=XH_SKIP_TAG_FORCE )
		 thisAddedParsedLine( str, *pObjC, kParsed.state, true );
	     continue;
	 }
	 if ( isInvalidTag ) {
	     nErrorsSyntax++;
	     HTML_LOG(fRepErr,LOG_ERROR,"Line %u: Invalid tag: %s\n",currLine,str);
	     if ( htmlOpt.oErrInvTagWrite.IsOn()==false ) continue;
	 }

	 char* strBaseHRef;

	 if ( id==XH_IDTAG_HEAD && kParsed.state==e_HS_HeadAfter && htmlOpt.DoBaseHRef()==true ) {
	     // Build and add new Html-Couple based on given option...
	     strBaseHRef = htmlOpt.GetBaseHRef();
	     myBaseHRef.Set( strBaseHRef );
	     gHtmlCouple newObjC( currLine, "BASE", strBaseHRef );
	     thisAddedParsedLine( newObjC.GetStr(), newObjC, kParsed.state, true );
	 }
	 thisAddedParsedLine( str, *pObjC, kParsed.state, lastOpError>=0 );
	 if ( lastOpError==0 ) {
	     if ( lastWarnOpCode==0 ) {
		 if ( id==XH_IDTAG_BASE ) {
		     strBaseHRef = pObjC->attrL.Find("HREF");
		     if ( strBaseHRef!=nil ) myBaseHRef.Set( strBaseHRef );
		 }
	     }
	     else {
		 HTML_LOG(fRepErr,LOG_WARNING,"Line %u: %s\n",currLine,msgError);
		 SetWarn( 0 );
	     }
	     continue;
	 }
	 if ( isInvalidTag==false ) {
	     nErrorsOther++;
	     HTML_LOG(fRepErr,LOG_ERROR,"Line %u: bad sequence (int-reg %d): %s\n%s%s%s",
		      currLine,
		      lastOpError,
		      str,
		      msgError[0]==0 ? "\0" : "ERROR: ", msgError, msgError[0]==0 ? "\0" : "\n");
	 }
     }//end ELSE IsText
 }//end FOR: main iter

 error = nErrorsSyntax!=0 || nErrorsOther!=0;
 if ( error!=0 ) return 1;

 n = kParsed.kMust.N();
 for (i=1; i<=n; i++) {
     pObjC = (gHtmlCouple*)kParsed.kMust.GetObjectPtr( i );
     ASSERTION(pObjC!=nil,"pObjC"); //Academic assert
     nErrorsOther++;
     HTML_LOG(fRepErr,LOG_ERROR,"Line %u: unclosed tag (int-ref %d): %s, opened at line %u\n",
	      currLine,
	      10,
	      pObjC->TagString().Str(),
	      pObjC->iLine);
     DBGPRINT("DBG: ATTRS: %s (%s)\n",pObjC->attrL.Str(),pObjC->attrL.Find("HREF"));
 }

 return 0;
}

int gHtmlParser::thisParseLine (gHtmlCouple& inCouple, unsigned lineNr, char* s, gHParsed& kParsed)
{
 // Return <=-8 on syntax error, -1/-2 on skipped tag
 // or non-zero for semantic error.
 //
 int error;
 t_int16 id;
 sHtmlElement* pInElem;
 gHtmlCouple* pCouple;
 bool inRefEnd, inCannotEnd, inOptEnd;

 ASSERTION(s!=nil,"s!=nil");
 SetError( 0 );
 id = inCouple.idTag;
 if ( inCouple.IsSkippedTag() ) return inCouple.synError;
 if ( id<0 ) return SetError( -8 );
 pInElem = inCouple.pElem;

 inRefEnd = inCouple.IsTagEnd();
 inCannotEnd = pInElem->CannotEndTag();
 inOptEnd = pInElem->MayEndTag();

 DBGPRINT_MIN("DBG: thisParseLine:%u [%s] inCouple=%s\n",lineNr,s,inCouple.sTag.Str());

 if ( inRefEnd ) {
     // Input was like '</B>'
     if ( inCannotEnd ) return SetError( 12 );  // E.g. '</BR>'
     // was it e.g. <A ...><B>XYZ</B></A> ?
     if ( kParsed.Depth()<=0 ) return SetError( 13 ); // E.g. </B> without <B>
     pCouple = kParsed.CurrentCouple();
     if ( id==pCouple->idTag ) {
	 error = kParsed.PopTag( inOptEnd );
	 ASSERTION(error==0,"kParsed.PopTag(1)");
	 return 0;
     }
     if ( inOptEnd ) {
	 // One opt-end-tag, and another different opt-end-tag? Check: ...
	 if ( id!=pCouple->idTag ) {
	     int countTrash=0;
	     int iOptIdx, iOptN=(int)kParsed.N();
	     for (iOptIdx=iOptN; iOptIdx>0 && countTrash==0; iOptIdx--) {
		 gHtmlCouple* pObjC = kParsed.GetCouple( (unsigned)iOptIdx );
		 ///printf("DBG: ###CURm: "); for (unsigned dbgIdx=1; dbgIdx<=kParsed.kMust.N(); dbgIdx++) printf("[%s]",kParsed.StrMust(dbgIdx)); printf("DBG: ### line %u:(%s).\n",((gHtmlCouple*)kParsed.kMust.EndPtr()->me)->iLine,((gHtmlCouple*)kParsed.kMust.EndPtr()->me)->sTag.Str());
		 ///printf("DBG: ###CURo: "); for (unsigned dbgIdx=1; dbgIdx<=kParsed.N(); dbgIdx++) printf("[%s]",((gHtmlCouple*)kParsed.GetObjectPtr(dbgIdx))->GetStr()); printf("DBG: ### line %u:(%s).\n",((gHtmlCouple*)kParsed.EndPtr()->me)->iLine,((gHtmlCouple*)kParsed.EndPtr()->me)->sTag.Str());
		 ///printf("DBG: sweep(%d|%d): %s (N=%u kMustN=%u)\n",id,pObjC->idTag,kParsed.CurrentCouple()->sTag.Str(),kParsed.N(),kParsed.kMust.N());
		 if ( id==pObjC->idTag ) {
		     kParsed.Delete( iOptIdx, iOptIdx );
		     countTrash++;
		 }
	     }// end FOR iOptIdx
	 }
     }
     else {
	 // Here something like '</A>' must pop any pending optional-ended tags
	 while ( kParsed.N()>kParsed.kMust.N() ) {
	     pCouple = kParsed.CurrentCouple();
	     if ( id==pCouple->idTag ) break;
	     kParsed.TrashLast();
	 }
     }
     // Update pointer
     pCouple = kParsed.CurrentCouple();
     // Now check consistency of current 'id', and the stack (kParsed)
     if ( id!=pCouple->idTag ) {
	 //printf("DBG: WOW optend?%d (id={%d:%s},current={%d:%s}): ",inOptEnd,id,s,pCouple->idTag,pCouple->sTag.Str()); kParsed.Show(true); printf("STK_MUST IS: "); kParsed.kMust.Show(true);
	 SetError( 14 ); // E.g. </B> ending <A>
	 snprintf(sStrError,200,
		  "Expected %s [line %u] but got %s",
		  pCouple->TagString(true).Str(),
		  pCouple->iLine,
		  s);
	 return 14;
     }
     // All ok, e.g. <B> ends here with </B>
     error = kParsed.PopTag( inOptEnd );
     ASSERTION(error==0,"kParsed.PopTag(2)");
 }
 else {
     // Three cases:
     // 1. plain '<BR>' (no end)
     // 2. optional-ended tag (e.g. <DD>)
     // 3. must-be-ended tag (e.g. <B>)
     if ( inOptEnd==true && inCannotEnd==false ) {
	 // 2nd case
	 kParsed.PushTagOptEnd( inCouple );
	 return 0;
     }
     // 1st/3rd case
     // Note: 1st case: inCannotEnd=true
     //
     // Error when issuing <A>, ..., <A> (without ending the first anchor)
     //
     if ( id==XH_IDTAG_ANCHOR ) {
	 gHtmlCouple* pCoupleAnchor = kParsed.FindCouple( XH_IDTAG_ANCHOR );
	 if ( pCoupleAnchor!=nil ) {
	     error = 129;
	     SetError( error );
	     snprintf(sStrError,200,
		      "Doubled anchor on line %u: already started on line %u (int-seq:%d)",
		      inCouple.iLine,
		      pCoupleAnchor->iLine,
		      error);
	     return error;
	 }
     }
     // Now pushing
     error = kParsed.PushTag( inCouple, inCannotEnd );

     if ( error==0 ) return 0;
     if ( error<0 ) {
	 if ( htmlOpt.oErrStateAllSuppress.IsOn() )
	     return 0;
	 switch ( error ) {
	 case -103:
	     if ( htmlOpt.oErrStateSuppress[e_HS_HeadAfter].IsOn() )
		 return 0;
	     break;
	 case -104:
	     if ( htmlOpt.oErrStateSuppress[e_HS_Html].IsOn() )
		 return 0;
	     break;
	 case -105:
	     if ( htmlOpt.oErrStateSuppress[e_HS_Start].IsOn() )
		 return 0;
	     break;
	 case -106:
	     if ( htmlOpt.oErrStateSuppress[e_HS_Body].IsOn() )
		 return 0;
	 default:
	     break;
	 }
	 SetWarn( 21 );
	 snprintf(sStrError,200,
		  "Not in LTD sequence: %s (int-seq:%d)",
		  s,
		  -error);
     }// end IF error<0
     else {
	 // ...then error>0
	 return SetError( error );
     }
 }
 return 0;
}

int gHtmlParser::thisAddedParsedLine (char* s, gHtmlCouple& inCouple, eHState state, bool doAccept)
{
 ASSERTION(s!=nil,"s!=nil");
 if ( s[0]==0 ) return 0;
 if ( fOutAll.f!=nil ) fprintf(fOutAll.f,"%s%s",s,fOutAll.NewLine());
 lOut.Add( s );
 lParts[state].AppendCouple( inCouple );
 return 0;
}
////////////////////////////////////////////////////////////

