#ifndef gHTML_PARSER_X_H
#define gHTML_PARSER_X_H

#include "gweb.h"
#include "ghash.h"
#include "glistext.h"
#include "gstack.h"

#include "gHtmlOpt.h"
#include "gHtmlAttr.h"
#include "gHtmlHATypes.h"
#include "gdNetStrings.h"

#define XH_NOTAG -9
#define XH_ENDTAG -11
#define XH_TBL_TAG_OPTEND 'O'
#define XH_TBL_TAG_OPTEND_CFG 'o'
#define XH_SKIP_TAG -1
#define XH_SKIP_TAG_FORCE -2

#define XH_SKIP_ATTR -1

// Definitions below must be adjusted if 'lElems' changes
#define XH_IDTAG_ANCHOR 0
#define XH_IDTAG_BASE 7
#define XH_IDTAG_BODY 12
#define XH_IDTAG_FONTx 30
#define XH_IDTAG_H1 34
#define XH_IDTAG_HL 39
#define XH_IDTAG_HEAD 40
#define XH_IDTAG_HTML 42
#define XH_IDTAG_IMG 45
#define XH_IDTAG_TABLEx 78
////////////////////////////////////////////////////////////
enum eHtmlElementFamily {
    e_HtmlElementPhrase = 1
};

struct sHtmlElement {
    t_int16 ctrl;   // zero(default): in use; 1: tidy; -1: skipped
    char* elemName;
    char* elemInfo;
    char optStart;  // letter-O: may not be present, or may not end (e.g. <BODY>) ::0|1
    char optEnd;    // letter-O: may not have end tag; F: must be uniquely starting (e.g. <BR>)
    char isDeprecated;  // D: deprecated (not allowed in transitional)
    char kindDTD;       // L: Loose DTD; F: belongs to frameset (i.e. frame elements, like <FRAME>)
    char family;

    bool CannotEndTag () {
	return optEnd=='F';
    }
    bool MayEndTag () {
	return (optEnd==XH_TBL_TAG_OPTEND || optEnd==XH_TBL_TAG_OPTEND_CFG) && optStart==0;
    }
    bool IsDeprecated () {
	return isDeprecated!=0;
    }
    short DeprecatedLevel () {
	return (short)(isDeprecated-'A');
    }
    bool IsFrameset () {
	return kindDTD=='F';
    }
    bool IsLoose () {
	return kindDTD=='L';
    }
};
////////////////////////////////////////////////////////////
struct sAttrDef {
    t_int16 ctrl;   // zero(default): in use; 1: tidy; -1: skipped
    char* attrName;
    char* strRelatedLst;
    char* strType;
    char* strImplReq;  // #IMPLIED, #REQUIRED, DATA, "%str", or "1"
    char chrDeprecated;
    char* strKindDTD;
    char* attrInfo;

    bool IsDeprecated () {
	return chrDeprecated=='D';
    }
};

struct sAttrRefer {
    sAttrRefer (t_int16 nAttrs) ;
    ~sAttrRefer () ;

    t_int16 nlAttrs;
    gString* sAttrs;         // [0:first attribute name, all-uppercase; (nlAttrs-1):last]
    gSmartList* lstRelated;  // [0:first attribute 'related elements'; (nlAttrs-1):last]
    gSwitch* lstIsAllButEtc; // see above, when related are 'all but' these elements
    t_int16 idxUniqMax;  // Maximum number of index for unique names (of attrs)
    gString* sLUniqs;    // List of unique names (majorated nlAttrs)
    t_int16* idxLUniqs;  // Indexes (see above)
    gString sOutHelper;  // Output helper string

    t_int16 FindAttr (char* attrName) {
	// Returns the first absolute index (for lAttrRef)
	t_int16 uniqIdx;
	return FindAttr( attrName, uniqIdx );
    }

    t_int16 FindAttr (char* attrName, t_int16& uniqIdx) ;

    t_int16 FindAttr (char* attrName, char* strTag) {
	t_int16 uniqIdx;
	return FindAttr( attrName, strTag, uniqIdx );
    }

    t_int16 FindAttr (char* attrName, char* strTag, t_int16& uniqIdx) ;
};
////////////////////////////////////////////////////////////
class gHtmlCouple : public gList {
public:
    gHtmlCouple (unsigned lineNr, char* sText) ;
    gHtmlCouple (unsigned lineNr, char* strTag, char* sAttrLst, bool doAddSkippedTags=true) ;
    virtual ~gHtmlCouple () ;

    // Public data-members
    unsigned iLine;
    t_int16 idTag;  // XH_NOTAG=-9 for text
    t_int16 idEndTag;
    gString sTag;   // empty for text
    gHtmlString* pHStr;  // nil for non-text
    sHtmlElement* pElem; // nil for text (never allocated)
    gHAttrList attrL;    // attribute list
    int synError;        // -1: skipped tag, 1 basically invalid tag
    t_int16 coupleId;    // only used for coupling
    t_int16 theDocType;
    // Special members
    gHtmlCouple* oCouple;
    gTagCoord tcoord;

    // Get methods
    virtual bool IsOk () ;
    virtual bool IsText () {
	return idTag==XH_NOTAG;
    }
    virtual bool IsTagEnd () {
	return idEndTag==XH_ENDTAG;
    }
    virtual bool IsAnchor () {
	return idTag==0;
    }
    virtual bool IsSkippedTag () {
	return synError==XH_SKIP_TAG || synError==XH_SKIP_TAG_FORCE;
    }

    virtual char* GetStr () ;
    virtual char* GetStrForTree () ;
    virtual gString& TagString (bool forceEnd=false) ;

    virtual char* GetHRef () ;

    // Set methods
    unsigned Add (char* s) ;
    unsigned Add (gString& copy) ;
    virtual unsigned AddText (char* s) ;
    virtual unsigned AddTag (char* strTag) ;

    virtual void CopyTag (gHtmlCouple& copy) {
	Reset();
	sTag = copy.sTag;
	idTag = copy.idTag;
	idEndTag = copy.idEndTag;
	pElem = copy.pElem;
	attrL.CopyAttr( copy.attrL );
	synError = copy.synError;
	coupleId = copy.coupleId;
	theDocType = copy.theDocType;
	if ( copy.IsText() ) AddText( copy.Str(1) );
    }

protected:
    gString sTagStr;
    gString sWholeTag;
    gString sKeepStr;

private:
    // Operators,empty
    gHtmlCouple (gHtmlCouple& ) ; //empty
    gHtmlCouple& operator= (gHtmlCouple& ) ; //empty
};
////////////////////////////////////////////////////////////
class gHtmlContent : public gList {
public:
    gHtmlContent () ;
    virtual ~gHtmlContent () ;

    // Get methods
    virtual gHtmlCouple* GetCouple (unsigned idx) ;
    virtual char* Str (unsigned idx) ;

    // Set methods
    unsigned Add (char* s) ;
    virtual int UpCaseAttributes (char* strTag, gString& sRes) ;

    // Specific methods
    virtual gHtmlOpt& GetHtmlOpt () ;
    virtual bool SetHtmlOpt (gHtmlOpt* pHtmlOpt) ;
    virtual int TagError (unsigned lineNr, int error, char* sLine, char* sShortMsg) ;

    // Show methods
    virtual void Show (bool doShowAll=true) ;

protected:
    unsigned nLines;
    gHtmlOpt* theHtmlOpt;

    int thisAddHmtlLine (unsigned lineNr, char* s) ;
    int thisAddHtmlText (unsigned lineNr, char* sText) ;
    int thisAddHtmlTag (unsigned lineNr, char* strTag, bool isEndTag) ;
    int thisAddCouple (gHtmlCouple* pCouple, gList& oL) ;

private:
    // Operators,empty
    gHtmlContent (gHtmlContent& ) ; //empty
    gHtmlContent& operator= (gHtmlContent& ) ; //empty
};
////////////////////////////////////////////////////////////
class gHParsed : public gStack {
public:
    // Stack-like class for all pending-closed HTML tags

    gHParsed ()
	: state( e_HS_Start ),
	  hasBaseHRef( false ) {
    }
    virtual ~gHParsed () {
    }

    static const char* tblStateStr[e_HS_Last];

    // Public data-members
    eHState state;
    bool hasBaseHRef;
    gStack kMust;  // Must-end tags stack

    // Get methods
    virtual int Depth () {
	return (int)N();
    }

    virtual char* Str (unsigned idx) ;
    virtual char* StrMust (unsigned idx) ;

    virtual gHtmlCouple* CurrentCouple () ;
    virtual gHtmlCouple* GetCouple (unsigned idx) ;

    virtual gHtmlCouple* FindCouple (t_int16 idTag) ;

    // Set methods
    virtual void PushTagOptEnd (gHtmlCouple& couple) ;
    virtual int PushTag (gHtmlCouple& couple, bool doCheckOnly=false) ;
    virtual int PopTag (bool hasOptEnd) ;
    virtual void TrashLast () ;

protected:
    int thisPush (gHtmlCouple& couple, gStack& aStack) ;
    int thisPushCouple (gHtmlCouple& couple, bool hasOptEnd) ;

private:
    // Operators,empty
    gHParsed (gHParsed& ) ; //empty
    gHParsed& operator= (gHParsed &) ; //empty
};
////////////////////////////////////////////////////////////
class gHList : public gHtmlContent {
public:
    // gHList class suits for parts of HTML, as a list
    gHList ()
	: coupleFakeBody( nil ) {
    }
    virtual ~gHList () {
	delete coupleFakeBody;
    }

    // Public data-members
    gHtmlCouple* coupleFakeBody;

    // Get methods
    virtual gHtmlCouple* GetCouple (unsigned idx) ;

    // Set methods
    virtual void AppendCouple (gHtmlCouple& couple) ;

private:
    // Operators,empty
    gHList (gHList& ) ; //empty
    gHList& operator= (gHList& ) ; //empty
};
////////////////////////////////////////////////////////////
class gHtmlParser : public gControl {
public:
    gHtmlParser (gUnweb* ptrUnweb) ;
    virtual ~gHtmlParser () ;

    // Public data-members
    gFileOut fOutAll;
    t_int16 docType;
    int nErrorsSyntax;
    int nErrorsOther;
    int nWarnings;
    int lastWarnOpCode;
    gHtmlOpt htmlOpt;
    gList lOut;
    gHList lParts[e_HS_Last];

    // Get methods
    virtual t_int16 GetTagMaxId () {
	return nElems;
    }

    virtual sHtmlElement* GetTagElement (t_int16 idxTag) ;

    virtual sHtmlElement* FindTag (char* strTag) {
	t_int16 idxTag;
	return thisFindTag( strTag, idxTag );
    }

    virtual int FindTagIdx (char* strTag, t_int16& idxTag) {
	int val;
	thisFindTag( strTag, idxTag );
	val = (int)idxTag;
	return val;
    }

    virtual sAttrDef* GetAttrDef (t_int16 idxAttr) ;

    virtual sAttrRefer* GetAttrRef () ;

    virtual sAttrNorm* GetDefaultAttrNorm (t_int16 idxNorm) ;

    virtual t_int16 GetNAttrType () {
	return lHAttrTypes[ 0 ].typeFamily;
    }

    virtual sHAttrType* GetAttrType (t_int16 idxType) ;

    // Set methods
    virtual void ReleaseHash () ;
    virtual bool SetOptions (gHtmlOpt& copy) ;

    // Specific methods
    virtual int Parse (FILE* fRepErr) ;
    virtual int SetError (int opError) ;
    virtual int SetWarn (int opError) ;

    // Show & debug methods
    int ShowTree (FILE* fRepErr) ;
    int Show_dbg (bool doShowAll=true) ;

protected:
    gString myBaseHRef;
    gUnweb* pUnweb;
    gHtmlContent htmlInput;

    // These-methods
    int thisFillFromUnweb (gUnweb& unweb, gHtmlContent& hInput) ;
    sHtmlElement* thisFindTag (char* strTag, t_int16& idxTag) ;
    int thisParse (FILE* fRepErr) ;

private:
    static t_int16 nElems;
    static sHtmlElement lElems[];
    static sAttrDef lAttrs[];
    static sAttrRefer* lAttrRef;
    static t_int16 nDefAttrNorms;
    static sAttrNorm lDefAttrNorms[];
    static sHAttrType lHAttrTypes[];

    static gHashTriple* hElems;
    // Note: hElems should not be an allocated object (but rather a pointer),
    // otherwise the gTop collection will return an error.

    int thisInitTbl (t_int16& size) ;
    int thisParseLine (gHtmlCouple& inCouple, unsigned lineNr, char* s, gHParsed& kParsed) ;
    int thisAddedParsedLine (char* s, gHtmlCouple& inCouple, eHState state, bool doAccept) ;

    // Operators,empty
    gHtmlParser (gHtmlParser& ) ; //empty
    gHtmlParser& operator= (gHtmlParser& ) ; //empty
};
////////////////////////////////////////////////////////////
#endif //gHTML_PARSER_X_H

