// gdNetNames.cpp -- Version 1.0

#include "gdNetNames.h"
// HOW-TO URIs:
//
// Consider the following URI:
//	http://pt.freescosoft.net/images/background.jpg
//	GetDomainStr()="pt.freescosoft.net", same as Str()
//	GetPathStr()="/images/background.jpg"
//	GetRelativePathStr()="images/background.jpg" (same as GetPathStr except leading slashes)
//	GetDomainAndPathStr()="pt.freescosoft.net/images/"
//	GetPathNameStr()="background.jpg"
// .
////////////////////////////////////////////////////////////
gdURI::gdURI (char* strURI)
{
 if ( strURI!=nil )  SetString( strURI );
}

gdURI::gdURI (gURI::eScheme aScheme, char* strURI)
    : gURI( aScheme )
{
 ASSERTION(strURI!=nil,"strURI!=nil");
 SetString( strURI );
}

gdURI::~gdURI ()
{
}

bool gdURI::IsOk ()
{
 bool isOk;

 // This IsOk is far less relaxed than gURI
 isOk = IsValidDomain();

 //printf("DBG: [[[DOMAIN:%s]]]:valid?%d,gURI::IsOk()?%d, SCHEME_INVALID?%d\n",GetDomainStr(),isOk,gURI::IsOk(),GetScheme()==e_invalid);
 ;
 return isOk && IsOkScheme() && gURI::IsOk();  // True if domain and scheme are o.k.
}

bool gdURI::IsValidDomain ()
{
 bool isOk = gURI::IsValidDomain();
 gString sDomain( GetDomainStr() );

 if ( isOk==false ) return false;
 isOk = sDomain.FindExcept(
     "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-." )==0;

 return isOk;
}

char* gdURI::GetPathStr ()
{
 return gURI::GetPathStr();
 // ...repeated in a dummy way: would not be redefined.
}

char* gdURI::GetRelativePathStr (eSwift aSwift)
{
 unsigned pos;
 gString sTemp( GetPathStr() );
 pos = sTemp.FindExcept( "/" );
 // If there would be well-behaved URLs, we would only compute the pad as:
 //	sPad.CopyFromTo( sTemp, 1+(sTemp[1]=='/') )
 // but we want to be relaxed with URLs, and allow multiple leading-slashes:
 sPad.CopyFromTo( sTemp, pos );
 // Note: pos=0 is interpreted as equal to pos=1, see gString...
 ASSERTION(aSwift==e_URI_NormalSwift,"TODO other 'swift's");
 return sPad.Str();
}

char* gdURI::GetDomainAndPathStr (eSwift aSwift)
{
 unsigned pos;
 gString sTemp( GetDomainStr() );
 gString sRelative( GetRelativePathStr() );

 sPad.SetEmpty();
 pos = sRelative.FindBack( "/" );

 if ( pos>0 ) {
     sPad.CopyFromTo( sRelative, 1, pos-1 );
 }

 if ( sTemp.IsEmpty()==false ) {
     gString sDomain( sTemp );
     sDomain.Add( "/" );
     sPad.Insert( sDomain, 1 );
 }

 return sPad.Str();
}

char* gdURI::GetPathNameStr (eSwift aSwift)
{
 unsigned pos;
 gString sPathFind( GetPathStr() );
 gString sComp( Str() );

 sPad.SetEmpty();

 pos = sComp.FindBack( sPathFind );
 sPathFind.CopyFromTo( sComp, pos+1 );
 if ( sPathFind.IsEmpty() ) return sPad.Str();

 sPad = sPathFind;
 pos = sPad.FindBack( "/" );
 if ( pos==0 ) {
     return sPad.Str();
 }

 sPad.CopyFromTo( sPathFind, pos+1 );

 return sPad.Str();
}

bool gdURI::Update ()
{
 if ( IsOk() ) return false;  // No update
 // URI is fairly o.k., but e_invalid means no prefix 'http://' was given.
 // We try to recover situations like: 'file://abc/def'
 return true;
}
////////////////////////////////////////////////////////////

