#include "gdSnarfCont.h"  // Snarf-contents
#include "gHtmlCtrl.h"
#include "gstringext.h"
////////////////////////////////////////////////////////////
gdHCouple::gdHCouple (unsigned lineNr)
    : gHtmlCouple( lineNr, nil )
{
}

gdHCouple::gdHCouple (unsigned lineNr, char* sText)
    : gHtmlCouple( lineNr, sText )
{
}

gdHCouple::gdHCouple (unsigned lineNr, char* strTag, char* sAttrLst)
    : gHtmlCouple( lineNr, strTag==nil ? nil : strTag+(strTag[0]=='/'), sAttrLst )
{
 // strTag not nil, asserted by: gHtmlCouple::AddTag
 thisInitTag( strTag[0]=='/' );
}

gdHCouple::~gdHCouple ()
{
}

bool gdHCouple::IsOk ()
{
 if ( attrs.IsOk()==false ) return false;
 return gHtmlCouple::IsOk();
}

char* gdHCouple::GetStr ()
{
 // Redefinition of gHtmlCouple
 ;
 unsigned iAttr, nAttrs;
 char* strL;
 char* strR;

 sWholeTag.SetEmpty();
 if ( IsText() ) {
     if ( pHStr==nil ) return Str(1);
     return pHStr->Str();
 }
 sWholeTag.Add( '<' );
 if ( IsTagEnd() ) sWholeTag.Add( '/' );
 sWholeTag.AddString( sTag );
 ;
 for (iAttr=1, nAttrs=attrs.N(); iAttr<=nAttrs; iAttr++) {
     strL = attrs.GetAttrValue( iAttr ).Str();
     strR = attrs.GetAttr( iAttr ).GetStr();
     //printf("ATTRLIST=%u:",nAttrs);attrs.Show(true);printf(".{%s=%s}\n",attrs.GetAttrValue(1).Str(),attrs.GetAttr(1).GetStr());printf("DBG:LST-end!\n");
     sWholeTag.Add( ' ' );
     sWholeTag.Add( strL );
     if ( strR!=nil && strR[0]!=0 ) {
	 sWholeTag.Add( '=' );
	 sWholeTag.Add( '"' );
	 sWholeTag.Add( strR );
	 sWholeTag.Add( '"' );
     }
 }
 sWholeTag.Add( '>' );
 return sWholeTag.Str();
}

sHtmlElement* gdHCouple::GetTagElement (t_int16 idxTag)
{
 sHtmlElement* ptrElem;

 ptrElem = gHtmlCtrl::Self().ParserPtr()->GetTagElement( idxTag );
 // Only difference to gHtmlParser: the assertion below
 ASSERTION(ptrElem!=nil,"ptrElem!=nil");
 return ptrElem;
}

char* gdHCouple::TagStr (bool forceEnd)
{
 // Note: 'sTagStr' from gHtmlCouple is allocated: this return is valid
 return TagString( forceEnd ).Str();
}

int gdHCouple::FindTagIdx (char* strTag, t_int16& idxTag)
{
 return gHtmlCtrl::Self().ParserPtr()->FindTagIdx( strTag, idxTag );
}

char* gdHCouple::GetHRef ()
{
 if ( idTag!=XH_IDTAG_ANCHOR ) return nil;
 gHAType* pValue = attrs.GetAttrPtr( 1 );
 if ( pValue==nil ) return nil;
 if ( pValue->GetType()!=e_HATp_URI ) return nil;
 return pValue->GetStr();
}

int gdHCouple::CopyNormalizeAttr (char* strTag, t_int16 idxTag, gHAttrList& copyL)
{
 // Return 0 if no attribute has a basic error
 ;
 short i, iOk, n=copyL.NAttr();
 short k, kN;
 char* str;
 char* strL, *strR;
 gString* sLVal, *sRVal;
 gList lLVal, lRVal;
 FILE* fErr = gHtmlCtrl::Self().hLog.fRepErr;
 gHtmlParser* pParser = gHtmlCtrl::Self().ParserPtr();

 t_int16 idx, uniqIdx;
 sAttrRefer* lstAttrRef = pParser->GetAttrRef();
 sAttrDef* dAttr;

 ASSERTION(idxTag>=0,"AttrNorm(0)");

 // Copy is only allowed if current list is empty
 ASSERTION(attrs.N()==0,"AttrNorm(1)");

 sLVal = new gString[ n+1 ];
 sRVal = new gString[ n+1 ];
 ASSERTION(sLVal!=nil && sRVal!=nil,"AttrNorm(2)");

 for (i=1, iOk=0; i<=n; i++) {
     strL = copyL.GetAttrLValue( i );
     strR = copyL.GetAttrRValue( i );
     ASSERTION(strL!=nil,"strL!=nil");
     // Check if attr 'strL' (LValue of assignment, e.g. HREF=uri, strL="HREF")
     // is o.k. for the tag
     idx = lstAttrRef->FindAttr( strL, strTag, uniqIdx );
     // If the attribute exists, but not suited for that tag, then uniqIdx>=0.
     if ( uniqIdx<0 ) {
	 if ( gStrControl::Self().Match(strL,"/") )
	     HTML_LOG( fErr, LOG_NOTICE, "Line %u: attribute not DTD: '%s'\n", iLine, strL );
	 else
	     HTML_LOG( fErr, LOG_ERROR, "Line %u: invalid attribute: '%s'\n", iLine, strL );
	 continue;
     }
     else {
	 if ( idx<0 ) {
	     HTML_LOG( fErr, LOG_ERROR, "Line %u: attribute '%s' not valid for tag: %s\n", iLine, strL, strTag );
	     //fprintf(stderr,"DBG: sOutHelper=[@(All-tags-but):%s,...]\n",lstAttrRef->sOutHelper.Str());
	     continue;
	 }
     }
     ASSERTION(idx>=0,"idx>=0");
     dAttr = pParser->GetAttrDef( idx );
     // ...or if
     // is to be skipped
     if ( dAttr->ctrl==XH_SKIP_ATTR ) {
	 HTML_LOG( fErr, LOG_NOTICE, "Line %u: attribute '%s' skipped (tag: %s)\n", iLine, strL, strTag );
	 continue;
     }
     iOk++;
     sLVal[ iOk ].Set( strL );
     sRVal[ iOk ].Set( strR );
 }

 //printf("DBG: LOG_ERROR:\n");log.lLog[LOG_ERROR].Show(true);printf("DBG: LOG_ERROR-END.\n");
 //printf("DBG: LOG_NOTICE:\n");log.lLog[LOG_NOTICE].Show(true);printf("DBG: LOG_NOTICE-END.\n");

 // Copy attributes from the order specified in gHOutOpt::lAttrsNorm (struct sAttrsTagNorm)
 gSmartList* lstNorms = pParser->htmlOpt.outOpt.lAttrsNorm.lstNorms;
 gSmartList* pList = nil;

 ASSERTION(lstNorms!=nil,"AttrNorm(4):temporary");
 if ( lstNorms!=nil ) {
     pList = &lstNorms[ idxTag ];
 }

 if ( pList!=nil ) {
     for (k=1, kN=pList->N(); k<=kN; k++) {
	 str = pList->Str( k );
	 for (i=1; i<=iOk; i++) {
	     if ( sLVal[ i ].Match( str ) ) {
		 strR = sRVal[ i ].Str();
		 lLVal.Add( str );
		 lRVal.Add( strR );
		 sLVal[ i ].SetEmpty();
	     }
	 }
     }
 }

 // We have copied by order, the norm'ed attrs, now copy remaining
 for (i=1; i<=iOk; i++) {
     if ( sLVal[ i ].IsEmpty() ) continue;
     lLVal.Add( sLVal[ i ] );
     lRVal.Add( sRVal[ i ] );
 }

 delete[] sLVal;
 delete[] sRVal;

 gControl logValues;
 attrs.BuildFromLRLists( strTag, lLVal, lRVal, false, iLine );

 return 0;
}

void gdHCouple::Show (bool doShowAll)
{
 DBGPRINT_WEBA1("DBG: gdHCouple::Show%s (N=%u)\n",doShowAll?":ALL":"\0",N());
 if ( doShowAll ) {
     printf("gdHCouple:START:\n");
 }
 gHtmlCouple::Show( doShowAll );
 if ( doShowAll ) {
     attrs.Show( true );
     printf("gdHCouple:END.\n");
 }
}

int gdHCouple::thisInitTag (bool isEndTag)
{
 char* strTag = sTag.Str();
 idEndTag = isEndTag ? XH_ENDTAG : 0;
 if ( FindTagIdx( strTag, idTag )<0 ) return -1;
 pElem = GetTagElement( idTag );
 // From here, pElem points correctly
 ASSERTION(pElem!=nil,"pElem!=nil");
 return 0;
}
////////////////////////////////////////////////////////////
gdSnarfCont::gdSnarfCont (gHtmlOpt* pHtmlOpt)
{
 if ( pHtmlOpt==nil ) pHtmlOpt = new gHtmlOpt;
 ASSERTION(pHtmlOpt!=nil,"pHtmlOpt!=nil");
 pHtmlOpt->outOpt.aTidy.doTidyAll = true;  // Tidy all tags&attrs by default
 SetHtmlOpt( pHtmlOpt );
}

gdSnarfCont::~gdSnarfCont ()
{
 delete theHtmlOpt;
 theHtmlOpt = nil;
}

gdHCouple* gdSnarfCont::GetOCouple (unsigned idx)
{
 gHtmlCouple* pObjC = GetCouple( idx );
 gdHCouple* pHObjC;
 if ( pObjC==nil ) return nil;
 pHObjC = (gdHCouple*)pObjC->oCouple;
 return pHObjC;
}

char* gdSnarfCont::Str (unsigned idx)
{
 // Redefinition of gHtmlContent
 gHtmlCouple* pObjC = GetCouple( idx );
 gdHCouple* pHObjC;
 if ( pObjC==nil ) return nil;
 pHObjC = (gdHCouple*)pObjC->oCouple;
 if ( pHObjC==nil ) return pObjC->GetStr();
 return pHObjC->GetStr();
}

unsigned gdSnarfCont::Add (char* s)
{
 // Returns 0 on error, or the actual number of lines.
 // This method allows to:
 //   add a complete tag line (e.g. "<A HREF=abc>")
 ;
 // It is very similar to its parent, but also sets the
 // couple idTag and 'pElem' correctly.
 ;
 unsigned uResult;
 gHtmlCouple* pCouple;
 gdHCouple* pNewCouple;
 char* strTag;
 t_int16 idxTag=XH_NOTAG;

 uResult = gHtmlContent::Add( s );
 if ( uResult==0 ) return 0;

 pCouple = (gHtmlCouple*)GetLastObjectPtr();
 ASSERTION(pCouple!=nil,"pCouple!=nil");

 if ( pCouple->IsText() ) return uResult;

 strTag = pCouple->sTag.Str();

 // If pCouple contains two entries, the second entry contains all attributes
 // pNewCouple = new gdHCouple( uResult, strTag, pCouple->N()<=1 ? nil : pCouple->Str(2) );

 pNewCouple = new gdHCouple( uResult, strTag, nil );  // No attributes (nil)
 ASSERTION(pNewCouple!=nil,"pNewCouple!=nil");
 if ( pNewCouple->FindTagIdx( strTag, idxTag )<0 ) {
     pCouple->synError = -1;
     delete pNewCouple;
     return 0;
 }

 ASSERTION(idxTag>=0,"idxTag>=0");

 pCouple->idTag = idxTag;
 pCouple->pElem = pNewCouple->GetTagElement( idxTag );

 ASSERTION(pCouple->oCouple==nil,"pCouple->oCouple==nil");
 if ( pCouple->IsTagEnd() ) {
     // Check if there are no attributes.
     // If no check was made, then these attrs would be silently discarded...
     delete pNewCouple;
     if ( pCouple->attrL.N()>0 ) {
	 HTML_LOG( nil, LOG_WARNING, "Line %u: tag-end (%s) with %u attribute(s)\n",
		   uResult, strTag, pCouple->attrL.N() );
     }
 }
 else {
     // Syntax o.k. (and it's a starting tag)
     pNewCouple->CopyNormalizeAttr( strTag, idxTag, pCouple->attrL );
     pCouple->oCouple = pNewCouple;
 }

 return uResult;
}

int gdSnarfCont::AddFromStr (char* s)
{
 // Returns 0 on success
 return Add( s )==0;
}

int gdSnarfCont::AddFromString (gString& aS)
{
 return AddFromStr( aS.Str() );
}

int gdSnarfCont::AddFromList (gList& aL)
{
 unsigned i, n;

 for (i=1, n=aL.N(); i<=n; i++) {
     gString s( aL.Str( i ) );
     AddFromString( s );
 }
 return 0;
}
////////////////////////////////////////////////////////////

