#include "gdSnarfHttp.h"
////////////////////////////////////////////////////////////
gdHttpURI::gdHttpURI (gURI& aURI, gString& sInProxy, t_gPort myPort)
    : uPort( myPort ),
      isHttpOk( false )
{
 int parseOp = thisProcessHttpURI( aURI, sInProxy, uPort );
 isHttpOk = parseOp==0;
}

gdHttpURI::gdHttpURI (gURI& aURI)
    : uPort( 80 ),
      isHttpOk( false )
{
 gString sEmpty;
 int parseOp = thisProcessHttpURI( aURI, sEmpty, uPort );
 isHttpOk = parseOp==0;
}

gdHttpURI::~gdHttpURI ()
{
}

bool gdHttpURI::IsOk ()
{
 return isHttpOk;
}

int gdHttpURI::thisProcessHttpURI (gURI& aURI, gString& sInProxy, t_gPort& myPort)
{
 char* str = aURI.GetDomainStr();
 char* inHostName = aURI.IsValidDomain() ? str : nil;
 char* connectHostName = inHostName;
 gString sUriPath( aURI.GetPathStr() );
 t_gPort portURI=0;
 bool isPxyOk = true;

  // If the path is empty (e.g. URI=google.com, instead of google.com/)
 // then assume '/' (root)
 if ( sUriPath.IsEmpty() ) sUriPath.Set( "/" );

 DBGPRINT("DBG: GetDomainStr()=%s, inHostName=%s, connectHostName=%s\n",str,inHostName,connectHostName);

 if ( sInProxy.IsEmpty() ) {
     sPath = sUriPath;
 }
 else {
     inHostName = ParseProxyStr( sInProxy, isPxyOk, proxy );
     sPath.Set( "http://" );
     sPath.Add( aURI.GetDomainStr() );
     if ( myPort!=80 ) portURI = myPort;
     myPort = proxy.port;
     if ( portURI!=0 ) {
	 // e.g.: http://moreira.dnsalias.net:82/abc/... (when --proxy option and -p are used)
	 sPath.Add( ":" );
	 sPath.Add( (unsigned)portURI );
     }
     sPath.Add( aURI.GetPathStr() );
     if ( isPxyOk==false ) return 1;
 }

 if ( inHostName==nil || inHostName[0]==0 || connectHostName==nil || connectHostName[0]==0 ) return 1;

 sConnectHost.Set( inHostName );
 sContentHost.Set( connectHostName );

 DBGPRINT("DBG: sConnectHost=%s, sContentHost=%s, sPath='%s', port=%u\n",inHostName,connectHostName,sPath.Str(),myPort);

 return 0;
}

char* gdHttpURI::ParseProxyStr (gString& sInProxy, bool& isProxyOk, sOptProxy& resultProxy)
{
 ;
 // Returns nil if format is invalid, otherwise it returns the host string
 //
 // A valid format is:  HOST:PORT
 // .
 unsigned n;
 char* str = sInProxy.Str();
 gParam aParam( str, ":" );

 n = aParam.N();
 resultProxy.Reset();
 isProxyOk = n==2;
 if ( isProxyOk==false ) return nil;
 // Check if host is in a good format
 gString sHost( aParam.Str(1) );
 isProxyOk = sHost.FindExcept( "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.-_", true )==0;
 if ( isProxyOk==false ) return nil;
 // Check if port is in a good format (decimal positive)
 t_uint32 vRes=0;
 isProxyOk = gStrControl::Self().ConvertToUInt32( aParam.Str(2), vRes )==0;
 if ( isProxyOk==false ) return nil;
 isProxyOk = vRes>0 && vRes<=MAX_UINT16_U;
 resultProxy.sHost = sHost;
 resultProxy.port = (t_gPort)vRes;
 return resultProxy.sHost.Str();
}
////////////////////////////////////////////////////////////
gdHttpSnarf::gdHttpSnarf (gTcpConnect& connection)
    : gXHttpCont( connection ),
      usedSize( 0 )
{
}

gdHttpSnarf::~gdHttpSnarf ()
{
}

gString& gdHttpSnarf::GetReply ()
{
 // replyHeader contains the list of replies
 return sReply;
}

int gdHttpSnarf::HttpCode ()
{
 /* Not necessary to convert: use htmlCode within replyHeader --->cut
 char* str = HttpCodeStr();
 t_uint32 vRes;
 if ( gStringControl::Self().ConvertToUInt32( str, vRes )!=0 )
     return -1;
 if ( vRes > 505 ) return -1;
 return (t_int16)vRes;
 <--- cut */
 return replyHeader.htmlCode;
}

unsigned gdHttpSnarf::GetHeaderLines (char* sKind, gList& resultL)
{
 unsigned i, n, pos=0;
 gString s;

 if ( sKind==nil ) return 0;
 for (i=1, n=replyHeader.kindsL.N(); i<=n; i++) {
     gString s( replyHeader.kindsL.Str(i) );
     if ( s.Match( sKind ) ) {
	 if ( pos==0 ) pos = i;
	 resultL.Add( replyHeader.linesL.Str(i) );
     }
 }
 return pos;
}

bool gdHttpSnarf::SnarfContent (gString& sHost, gString& sPath)
{
 bool resultOk = GetContent( e_Get, sHost, sPath, sReply );
 return resultOk;
}

int gdHttpSnarf::FileInOut (gFileTemp& fT, FILE* fOut)
{
 int fHandle = fT.fHandle;
 t_uchar c;

 if ( fT.Rewind()==false ) return -1;
 for ( ; read(fHandle,&c,1)==1; ) {
     fprintf(fOut,"%c",c);
 }
 return 0;
}

int gdHttpSnarf::DoPrint (FILE* fOut)
{
 int thisError=0;
 t_uchar* uBuf = UBuffer();
 unsigned k, nod, nNodes, maxSize;

 //static const char *strBufKind[gBigBuffer::e_InFile+1]={"RAM","NodesRAM","InFile"};
 //fprintf(stderr,"BufferKind: %s (%s)\n",strBufKind[BufferKind()],IsText()?"Text":"Bin");

 switch ( BufferKind() ) {
 case gBigBuffer::e_NormalRAM:
     if ( IsText() ) {
	 fprintf(fOut,"%s",uBuf);
     }
     else {
	 for (k=0, maxSize=(unsigned)usedSize; k<maxSize; k++) {
	     fprintf(fOut,"%c",uBuf[k]);
	 }
     }
 case gBigBuffer::e_NodesRAM:
     for (nod=0, nNodes=GetBuffer().GetNumberNodes(); nod<nNodes; nod++) {
	 uBuf = GetBuffer().GetNode(nod)->uBuf;
	 for (k=0, maxSize=GetBuffer().GetNode(nod)->size; k<maxSize; k++)
	     fprintf(fOut,"%c",uBuf[k]);
     }
     break;
 case gBigBuffer::e_InFile:
     thisError = FileInOut( GetBuffer().GetFile(), fOut )!=0;
 default:
     break;
 }

 return thisError;
}
////////////////////////////////////////////////////////////

