// garray.cpp -- Version 0.0

#include "garray.h"
#include <string.h>

// Static members
int gArray::defaultFmtDigits=3;
short gArrayCollect::iCtrl=0;
gArrayCollect gArrayCollect::myself;
// .

////////////////////////////////////////////////////////////
gVector::gVector (unsigned nMax)
    : gInt( (int)nMax ),
      n( nMax )
{
 v = new int[ n+1 ];
 ASSERTION(v!=nil,"v!=nil");
 v[ 0 ] = 0;
 Zero();
}

gVector::gVector (gVector& copy)
    : gInt( copy.N() ),
      n( copy.N() )
{
 unsigned iter;
 v = new int[ n+1 ];
 ASSERTION(v!=nil,"v!=nil");
 for (iter=1; iter<=n; iter++) v[ iter ] = copy.GetNumInt( iter );
}

gVector::~gVector ()
{
 Release();
}

char* gVector::Str ()
{
 // Calling (nil,nil) is the same as ("%d","(%s)")
 return StrFormat( nil, nil );
}

char* gVector::StrFormat (char* fmtOptions, char* fmtString)
{
 static char sBuf[30];
 unsigned iter;
 unsigned len;
 char* strFmtOptions = fmtOptions==nil ? (char*)"%d" : fmtOptions;
 char* strFmtString = fmtString==nil ? (char*)"(%s)" : fmtString;

 gString sTemp;
 for (iter=1; iter<=n; iter++) {
	sprintf( sBuf, strFmtOptions, v[iter] );
	if ( iter<n ) strcat( sBuf, " " );
	sTemp.Add( sBuf );
 }
 gString sTempCopy( len = (sTemp.Length()+(unsigned)strlen(strFmtString)+20), '\0' );
 snprintf( sTempCopy.Str(), len, strFmtString, sTemp.Str() );
 sVector = sTempCopy;
 // Return data-member (otherwise, on local var, content would be lost)
 return sVector.Str();
}

void gVector::Release ()
{
 delete[] v;
 v = nil;
}

void gVector::Resize (unsigned nMax)
{
 unsigned iter;
 gVector vTemp = *this;
 Release();
 v = new int[ nMax+1 ];
 ASSERTION(v!=nil,"v!=nil");
 v[ 0 ] = 0;
 for (iter=1; iter<=nMax; iter++) {
	v[ iter ] = (iter<=n) ? vTemp[ iter ] : 0;
 }
 n = nMax;
}

void gVector::Set (int iVal)
{
 unsigned iter;
 for (iter=1; iter<=n; iter++) v[ iter ] = iVal;
}

gVector& gVector::Sort (bool isAscending)
{
 unsigned iter, k, iterBest;
 int thisBest, aWorse = isAscending ? GetMaxInt() : GetMinInt();
 gVector vTemp( n );
 ;
 vTemp.Set( aWorse );

 for (k=1; k<=n; k++) {
	iterBest = 0;
	thisBest = aWorse;
	for (iter=1; isAscending==true && iter<=n; iter++) {
		if ( v[iter] < thisBest ) {
			thisBest = v[iter];
			iterBest = iter;
		}
	}
	for (iter=1; isAscending==false && iter<=n; iter++) {
		if ( v[iter] > thisBest ) {
			thisBest = v[iter];
			iterBest = iter;
		}
	}
	if ( iterBest<=0 ) continue;
	v[ iterBest ] = aWorse;
	vTemp[ k ] = thisBest;
 }

 *this = vTemp;
 return *this;
}

gVector& gVector::operator= (gVector& copy)
{
 unsigned iter;
 Release();
 n = copy.n;
 v = new int[ n+1 ];
 ASSERTION(v!=nil,"v!=nil");
 v[0] = 0;
 for (iter=1; iter<=n; iter++) v[ iter ] = copy.GetNumInt( iter );
 return *this;
}

gVector& gVector::operator+ (gVector& copy)
{
 unsigned iter;
 if ( n!=copy.N() ) {
	GARRAY_ANY_LOG( LOG_ERROR, "operator+" );
	return *this;
 }

 gVector* newA = new gVector( n );
 gArrayCollect::Self().Garbage( newA );

 for (iter=1; iter<=n; iter++) (*newA)[ iter ] = v[ iter ] + copy.GetNumInt( iter );

 return *newA;
}

gVector& gVector::operator- (gVector& copy)
{
 unsigned iter;
 if ( n!=copy.N() ) {
	GARRAY_ANY_LOG( LOG_ERROR, "operator+" );
	return *this;
 }

 gVector* newA = new gVector( n );
 gArrayCollect::Self().Garbage( newA );

 for (iter=1; iter<=n; iter++) (*newA)[ iter ] = v[ iter ] - copy.GetNumInt( iter );

 return *newA;
}

gVector& gVector::operator* (gVector& copy)
{
 unsigned iter;
 int iVal;

 if ( n!=copy.N() ) {
	GARRAY_ANY_LOG( LOG_ERROR, "operator*" );
	return *this;
 }

 gVector* newA = new gVector( n );
 gArrayCollect::Self().Garbage( newA );

 for (iter=1; iter<=n; iter++) {
	iVal = v[ iter ] * copy.GetNumInt( iter );
	(*newA)[ iter ] = iVal;
 }

 return *newA;
}

int& gVector::operator[] (unsigned idx)
{
 ASSERTION(idx>=1,"idx>=1");
 ASSERTION(idx<=n,"idx<=n");
 return v[ idx ];
}

bool gVector::operator== (gVector& copy)
{
 unsigned iter;
 if ( n!=copy.N() ) return false;
 for (iter=1; iter<=n; iter++) {
	if ( v[ iter ] != copy.GetNumInt( iter ) ) return false;
 }
 return true;
}

bool gVector::operator!= (gVector& copy)
{
 return ((*this) == copy)==false;
}
////////////////////////////////////////////////////////////
gArray::gArray (unsigned y, unsigned x)
    : gInt( y ),
      pV( nil )
{
 thisNewDim( y, x );
}

gArray::gArray (gArray& copy)
    : gInt( copy.GetInt() ),
      pV( nil )
{
 thisCopyArray( copy );
}

gArray::~gArray ()
{
 Release();
}

char* gArray::Str ()
{
 unsigned y;

 sArray.SetEmpty();
 sArray.Add( '[' );
 for (y=1; y<=dimY && IsOk()==true; y++) {
	sArray.Add( pV[ y ].Str() );
	if ( y<dimY ) sArray.Add( ' ' );
 }
 sArray.Add( ']' );
 return sArray.Str();
}

char* gArray::StrFormat (char* fmtOptions, char* fmtString)
{
 unsigned y;
 // For printing a 2-row, 3-column array (2x3) like this
 //    [ 12 345 678|
 //    |  4   5   6]
 // use the following: fmtOptions="%3d", fmtString="[@|\n|@]\n"
 // (where '@' means each 'vector' contents).

 // The format string "[@|\n|@]\n" is the default.
 // The second format puts an array into a single line ("[|]")

 static char sBuf[30];
 gString sFmtOptions( fmtOptions==nil ? thisFmtOptions( defaultFmtDigits, 'd' ) : fmtOptions );
 char* strFmtOptions = sFmtOptions.Str();
 char* strFmtString = fmtString==nil ? (char*)"[@|\n|@]\n" : fmtString;
 unsigned fmtLen = (unsigned)strlen( strFmtString );
 unsigned iter;
 bool isSecondFmt = fmtLen>=3 && strFmtString[1]=='|';

 for (iter=0; iter<fmtLen; iter++) {
	if ( strFmtString[iter]=='F' ) strFmtString[iter] = '\n';
 }

 sArray.SetEmpty();

 // E.g. of a fmtString:
 //     01234567
 //     [@|F|@]F
 // len:12345678
 //
 for (y=1; y<=dimY; y++) {
    if ( isSecondFmt ) {
	char endBuf[8];
	char* vectorFmtString = strFmtString[1]=='|' ? (char*)"%s" : nil;
	endBuf[ 0 ] = y==dimY ? strFmtString[2] : '\0';
	endBuf[ 1 ] = 0;
	sprintf( sBuf, "%c%s%s",
		 y==1 ? strFmtString[0] : strFmtString[1],
		 (*this)[y].StrFormat( strFmtOptions, vectorFmtString ),
		 endBuf );
	sArray.Add( sBuf );
    }
    else {
	if ( y==1 ) {
		sprintf( sBuf, "%c%s%c", strFmtString[0], "%s", strFmtString[1]=='@' ? strFmtString[2] : '.' );
	}
	else {
		if ( y==dimY ) {
			if ( fmtLen>6 )
				sprintf( sBuf, "%c%s%c", strFmtString[4], "%s", strFmtString[6] );
			else
				sprintf( sBuf, "[%s]", "%s" );
		}
		else {
			sprintf( sBuf, "%c%s%c", fmtLen>2 ? strFmtString[2] : '[', "%s", strFmtString[1]=='@' ? strFmtString[2] : ']' );
		}
	}
	sArray.Add( (*this)[y].StrFormat( strFmtOptions, sBuf ) );
	if ( fmtLen>3 ) {
		sArray.Add( strFmtString[3] );
	}
    }
 }

 return sArray.Str();
}

void gArray::Release ()
{
 delete[] pV;
 pV = nil;
}

void gArray::Set (int iVal)
{
 unsigned y;
 if ( IsOk()==false ) return;
 for (y=1; y<=dimY; y++) pV[ y ].Set( iVal );
}

void gArray::Set (gVector& iVector)
{
 Resize( 1, iVector.N() );
 pV[ 1 ] = iVector;
}

void gArray::TransposeMe ()
{
 unsigned y, x, newY, newX;

 newY = dimX;
 newX = dimY;
 gArray aTemp( newY, newX );
 for (y=1; y<=newY; y++) {
    for (x=1; x<=newX; x++) {
	aTemp[ y ][ x ] = pV[ x ][ y ];
    }
 }
 *this = aTemp;
}

gArray& gArray::Transpose ()
{
 gArray* newA;
 newA = new gArray( *this );
 gArrayCollect::Self().Garbage( newA );
 newA->TransposeMe();
 return *newA;
}

gArray& gArray::AddRow (gVector& aV)
{
 unsigned y;
 gArray newB( dimY+1, dimX );
 if ( aV.N()!=dimX ) {
    GARRAY_ANY_LOG( LOG_ERROR, "AddRow" );
    return *this;
 }
 for (y=1; y<=dimY; y++) {
    newB[y] = (*this)[y];
 }
 newB[ y ] = aV;
 thisCopyArray( newB );
 return *this;
}

gArray& gArray::AddEmptyRow (unsigned iter)
{
 unsigned aIter;
 gVector aV( dimX );
 for (aIter=1; aIter<=iter; aIter++) AddRow( aV );
 return *this;
}

gArray& gArray::AddColumn (gVector& aV)
{
 unsigned y, x, width = aV.N();
 gArray newB( dimY, dimX+1 );
 if ( width!=dimY ) {
    GARRAY_ANY_LOG( LOG_ERROR, "AddColumn" );
    return *this;
 }
 for (y=1; y<=dimY; y++) {
    for (x=1; x<=dimX; x++) {
	newB[ y ][ x ] = GetNumInt( y, x );
    }
    newB[ y ][ x ] = aV[ y ];
 }
 thisCopyArray( newB );
 return *this;
}

gArray& gArray::operator= (gArray& copy)
{
 thisCopyArray( copy );
 return *this;
}

gArray& gArray::AddEmptyColumn (unsigned iter)
{
 unsigned aIter;
 gVector aV( dimY );
 for (aIter=1; aIter<=iter; aIter++) AddColumn( aV );
 return *this;
}

gArray& gArray::JoinArrayRight (gArray& copy, bool doForceDim)
{
 unsigned y, x, newX = dimX+copy.DimX(), newY = dimY;
 ;
 if ( doForceDim==false ) {
	if ( dimY!=copy.DimY() ) {
		GARRAY_ANY_LOG( LOG_ERROR, "JoinArrayRight (wrong dimY)" );
		return *this;
	}
 }
 ASSERTION(doForceDim==false,"doForceDim==false: TODO");
 ;
 gArray* newA = new gArray( newY, newX );
 gArrayCollect::Self().Garbage( newA );
 ;
 for (y=1; y<=newY; y++) {
	gVector aV( newX );
	for (x=1; x<=dimX && x<=newX; x++) {
		aV[ x ] = GetNumInt( y, x );
	}
	for (unsigned copyX=1; x<=newX; x++) {
		aV[ x ] = copy.GetNumInt( y, copyX++ );
	}
	(*newA)[ y ] = aV;
 }
 return *newA;
}

gArray& gArray::JoinArrayDown (gArray& copy, bool doForceDim)
{
 unsigned y, newX = dimX, newY = dimY + copy.DimY();
 ;
 if ( doForceDim==false ) {
	if ( dimX!=copy.DimX() ) {
		GARRAY_ANY_LOG( LOG_ERROR, "JoinArrayDown (wrong dimX)" );
		return *this;
	}
 }
 ASSERTION(doForceDim==false,"doForceDim==false: TODO");
 ;
 gArray* newA = new gArray( newY, newX );
 gArrayCollect::Self().Garbage( newA );
 ;
 for (y=1; y<=dimY; y++) {
	(*newA)[ y ] = (*this)[ y ];
 }
 for (unsigned copyY=1; y<=newY; y++) {
	(*newA)[ y ] = copy[ copyY ];
 }
 return *newA;
}

gArray& gArray::operator+ (gArray& copy)
{
 unsigned y;
 if ( DimX()!=copy.DimX() ) {
	GARRAY_ANY_LOG( LOG_ERROR, "operator+ (DimX!=copy.DimX)" );
	return *this;
 }
 if ( DimY()!=copy.DimY() ) {
	GARRAY_ANY_LOG( LOG_ERROR, "operator+ (DimY!=copy.DimY)" );
	return *this;
 }

 gArray* newA = new gArray( dimY, dimX );
 gArrayCollect::Self().Garbage( newA );

 for (y=1; y<=dimY; y++) {
	(*newA)[ y ] = pV[ y ] + copy[ y ];
 }

 return *newA;
}

gArray& gArray::operator- (gArray& copy)
{
 unsigned y;
 if ( DimX()!=copy.DimX() ) {
	GARRAY_ANY_LOG( LOG_ERROR, "operator- (DimX!=copy.DimX)" );
	return *this;
 }
 if ( DimY()!=copy.DimY() ) {
	GARRAY_ANY_LOG( LOG_ERROR, "operator- (DimY!=copy.DimY)" );
	return *this;
 }

 gArray* newA = new gArray( dimY, dimX );
 gArrayCollect::Self().Garbage( newA );

 for (y=1; y<=dimY; y++) {
	(*newA)[ y ] = pV[ y ] - copy[ y ];
 }

 return *newA;
}

gArray& gArray::operator* (gArray& copy)
{
 unsigned /*y, x,*/ newY, newX;
 if ( DimX()!=copy.DimY() ) {
	GARRAY_ANY_LOG( LOG_ERROR, "operator* (DimX!=copy.DimY)" );
	return *this;
 }
 ;
 newY = DimY();
 newX = copy.DimX();

 gArray* newA = new gArray( newY, newX );
 gArrayCollect::Self().Garbage( newA );
 ;
 // TODO
 return *newA;
}

gArray& gArray::operator| (gArray& copy)
{
 return JoinArrayRight( copy, false );
}

gVector& gArray::operator[] (unsigned idx)
{
 if ( idx<1 || idx>dimY ) {
	GARRAY_ANY_LOG( LOG_ERROR, "operator[] (wrong index)" );
	idx = 0;  // Arbitrary error...
 }
 return pV[ idx ];
}

bool gArray::operator== (gArray& copy)
{
 unsigned y;
 if ( dimY!=copy.DimY() || dimX!=copy.DimX() ) return false;
 for (y=1; y<=dimY; y++) {
	if ( (pV[ y ]==copy[ y ])==false ) return false;
 }
 return true;
}

bool gArray::operator!= (gArray& copy)
{
 return ((*this) == copy)==false;
}

int gArray::thisNewDim (unsigned y, unsigned x)
{
 dimY = y;
 dimX = x;
 Release();
 pV = new gVector[ dimY+1 ];
 ASSERTION(pV!=nil,"pV!=nil");
 for (y=1; y<=dimY; y++) pV[ y ].Resize( x );
 return 0;
}

int gArray::thisCopyArray (gArray& copy)
{
 unsigned y;
 thisNewDim( copy.DimY(), copy.DimX() );
 for (y=1; y<=dimY; y++) {
	pV[ y ] = copy[ y ];
 }
 return 0;
}

char* gArray::thisFmtOptions (int nrFmtDigits, char fmtChar)
{
 gString sTemp( 30, '\0' );
 snprintf( sTemp.Str(), 30, "%%%d%c", nrFmtDigits, fmtChar );
 sArray = sTemp;
 return sArray.Str();
}
////////////////////////////////////////////////////////////
gArrayCollect::gArrayCollect ()
{
 if ( iCtrl==0 ) {
	iCtrl = 1;
	gStorageControl::Self().StaticAlloc( "gArrayCollect:lGarbage", 1 );
 }
}

gArrayCollect::~gArrayCollect ()
{
 ReleaseAll();
}

int gArrayCollect::Garbage (gStorage* newA)
{
 return thisAddIntoList( newA, lGarbage );
}

int gArrayCollect::ReleaseAll ()
{
 lGarbage.Delete();
 return 0;
}

int gArrayCollect::thisAddIntoList (gStorage* newA, gList& aList)
{
 ASSERTION(newA!=nil,"newA!=nil");
 aList.AppendObject( newA );
 return 0;
}
////////////////////////////////////////////////////////////

