#ifndef gARRAY_X_H
#define gARRAY_X_H

#include "gstring.h"
#include "gcontrol.h"

#define GDEF_DIM 100

#define GARRAY_LOGF stderr
#define GARRAY_ANY_LOG(level,y) { \
	/* fprintf(GARRAY_LOGF,":::%s:%u:%s\n",__FILE__,__LINE__,y); */ ; \
	if ( level!=LOG_NONE ) gStorageControl::Self().Log( GARRAY_LOGF, level, "%s:%u:%s\n", __FILE__, __LINE__, y ); \
}
////////////////////////////////////////////////////////////
class gVector : public gInt {
public:
    gVector (unsigned nMax=0) ;
    gVector (gVector& copy) ;
    virtual ~gVector () ;

    // Public data-members
    gString sVector;

    // Get methods
    virtual unsigned N () {
	return n;
    }

    virtual int GetNumInt (unsigned iter) {
	return (iter>=1 && iter<=n && IsOk()==true) ? v[ iter ] : -1;
    }

    virtual bool IsOk () {
	return v!=nil && gInt::IsOk();
    }

    virtual char* Str () ;
    virtual char* StrFormat (char* fmtOptions, char* fmtString) ;

    // Set methods
    virtual void Reset () {
	Zero();
    }

    virtual void Release () ;

    virtual void Resize (unsigned nMax) ;

    void Zero () {
	Set( 0 );
    }

    void Set (int iVal) ;

    gVector& Sort (bool isAscending=true) ;

    // All operators
    virtual gVector& operator= (gVector& copy) ;
    virtual gVector& operator+ (gVector& copy) ;
    virtual gVector& operator- (gVector& copy) ;
    virtual gVector& operator* (gVector& copy) ;
    virtual int& operator[] (unsigned idx) ;
    virtual bool operator== (gVector& copy) ;
    virtual bool operator!= (gVector& copy) ;

protected:
    unsigned n;
    int* v;
};
////////////////////////////////////////////////////////////
class gArray : public gInt {
public:
    gArray (unsigned y=0, unsigned x=0) ;
    gArray (gArray& copy) ;
    virtual ~gArray () ;

    // Public data-members
    static int defaultFmtDigits;
    gString sArray;

    // Get methods
    virtual unsigned DimY () {
	return dimY;
    }

    virtual unsigned DimX () {
	return dimX;
    }

    virtual int GetNumInt (unsigned y, unsigned x) {
	return (y>=1 && y<=dimY && IsOk()==true) ? pV[y][x] : -1;
    }

    virtual bool IsOk () {
	return pV!=nil && gInt::IsOk();
    }

    virtual char* Str () ;
    virtual char* StrFormat (char* fmtOptions, char* fmtString) ;

    // Set methods
    virtual void Reset () {
	Zero();
    }

    virtual void Release () ;

    virtual void Resize (unsigned newY, unsigned newX) {
	thisNewDim( newY, newX );
    }

    void Zero () {
	Set( 0 );
    }

    void Set (int iVal) ;
    void Set (gVector& iVector) ;

    virtual void TransposeMe () ;
    virtual gArray& Transpose () ;

    virtual gArray& AddRow (gVector& aV) ;
    virtual gArray& AddEmptyRow (unsigned iter=1) ;
    virtual gArray& AddColumn (gVector& aV) ;
    virtual gArray& AddEmptyColumn (unsigned iter=1) ;
    gArray& JoinArrayRight (gArray& copy, bool doForceDim=false) ;
    gArray& JoinArrayDown (gArray& copy, bool doForceDim=false) ;

    // All operators
    virtual gArray& operator= (gArray& copy) ;
    virtual gArray& operator+ (gArray& copy) ;
    virtual gArray& operator- (gArray& copy) ;
    virtual gArray& operator* (gArray& copy) ;
    virtual gArray& operator| (gArray& copy) ;
    virtual gVector& operator[] (unsigned idx) ;
    virtual bool operator== (gArray& copy) ;
    virtual bool operator!= (gArray& copy) ;

protected:
    unsigned dimY, dimX;
    gVector* pV;

    int thisNewDim (unsigned y, unsigned x) ;
    int thisCopyArray (gArray& copy) ;
    char* thisFmtOptions (int nrFmtDigits, char fmtChar) ;
};
////////////////////////////////////////////////////////////
class gArrayCollect {
public:
    virtual ~gArrayCollect () ;

    static gArrayCollect& Self () {
        return myself;
    }

    // Get methods
    virtual int NumObjs () {
	return lGarbage.N();
    }

    virtual gStorage* GetObjectPtr (unsigned idx) {
	return lGarbage.GetObjectPtr( idx );
    }
    virtual gStorage* GetFirstObjectPtr () {
	return lGarbage.GetFirstObjectPtr();
    }
    virtual gStorage* GetLastObjectPtr () {
	return lGarbage.GetLastObjectPtr();
    }

    // Set methods
    int Garbage (gStorage* newA) ;
    int ReleaseAll () ;

protected:
    gArrayCollect () ;

    int thisAddIntoList (gStorage* newA, gList& aList) ;

private:
    static short iCtrl;
    static gArrayCollect myself;
    gList lGarbage;

    // Operators,empty
    gArrayCollect (gArrayCollect& ) ; //empty
    gArrayCollect& operator= (gArrayCollect& ) ; //empty
};
////////////////////////////////////////////////////////////
#endif //gARRAY_X_H

