// gcalendar.cpp -- Version 0.0

#include <string.h>
#include "gcalendar.h"

// Static members
int gDateTime::minimumYear=1900;
const t_uint8 gDateTime::tblCalDurMonth[13]={
    0,
    31,
    28,
    31,
    30,
    31,
    30,
    31,
    31,
    30,
    31,
    30,
    31};

////////////////////////////////////////////////////////////
gDateTime::gDateTime (t_stamp aStamp)
    : gStorage( e_StoreExtend ),
      lastOpError( 0 ),
      year( 0 ),
      month( 0 ),
      day( 0 ),
      hour( 0 ),
      minu( 0 ),
      sec( 0 ),
      wday( 0 ),
      yday( 0 ),
      isdst( -1 )
{
 lastOpError = thisConvertFromStamp( aStamp );
}

gDateTime::gDateTime (int iYear, int iMonth, int iDay,
		      t_uint8 uHour, t_uint8 uMin, t_uint8 uSec)
    : gStorage( e_StoreExtend ),
      lastOpError( 0 ),
      year( iYear ),
      month( iMonth ),
      day( iDay ),
      hour( uHour ),
      minu( uMin ),
      sec( uSec ),
      wday( 0 ),
      yday( 0 ),
      isdst( 0 )
{
 lastOpError = iMonth<1 || iDay<1;
 if ( lastOpError==0 ) lastOpError = thisCalCheck();
}

gDateTime::gDateTime (gDateTime::eKindTime aKind)
    : gStorage( e_StoreExtend ),
      lastOpError( 0 ),
      year( 0 ),
      month( 0 ),
      day( 0 ),
      hour( 0 ),
      minu( 0 ),
      sec( 0 ),
      wday( 0 ),
      yday( 0 ),
      isdst( -1 )
{
 lastOpError = thisConvertFromStamp( time(NULL) );
}

gDateTime::~gDateTime ()
{
}

bool gDateTime::IsOk ()
{
 // Note: gStorage::IsOk is always true...
 return lastOpError==0 && isdst!=-1 && thisCalCheck()==0;
}

char* gDateTime::StorageName ()
{
 return (char*)"DTTM";
}

void gDateTime::Reset ()
{
 lastOpError = 0;
 year = 0;
 month = 0;
 day = 0;
 hour = minu = sec = 0;
 wday = 0;
 yday = 0;
 isdst = -1;
}

int gDateTime::SetError (int opError)
{
 lastOpError = opError;
 return gStorage::SetError( lastOpError );
}

t_stamp gDateTime::GetTimeStamp ()
{
 t_stamp aStamp;
 SetError( thisConvertToStamp( *this, aStamp ) );
 return aStamp;
}

bool gDateTime::SetTimeStamp (t_stamp aStamp)
{
 return SetError( thisConvertFromStamp( aStamp ) )==0;
}

t_uint8 gDateTime::DaysOfMonth (int aYear, t_uint8 aMonth)
{
 t_uint8 nDays;
 if ( aMonth<1 || aMonth>12 ) return 0;
 nDays = tblCalDurMonth[aMonth];
 if ( aMonth==2 ) nDays += ((aYear%4)==0);
 return nDays;
}

gStorage* gDateTime::NewObject ()
{
 gDateTime* a = new gDateTime( GetTimeStamp() );
 return a;
}

t_uchar* gDateTime::ToString (t_uchar* uBuf)
{
 if ( uBuf==nil ) return nil;
 sprintf( (char*)uBuf, "%04d/%02u/%02u,%02u:%02u:%02u",
	  year, month, day,
	  hour, minu, sec );
 return uBuf;
}

bool gDateTime::SaveGuts (FILE* f)
{
 if ( CanSave( f )==false ) return false;
 return
     fprintf(f,"%04d/%02u/%02u,%02u:%02u:%02u",
	     year, month, day,
	     hour, minu, sec)>0;
}

bool gDateTime::RestoreGuts (FILE* f)
{
 t_stamp aStamp;
 int d1;
 unsigned d2, d3, h1, h2, h3;

 if ( CanRestore( f )==false ) return false;
 Reset();
 lastOpError =
     fscanf(f,"%d/%u/%u,%u:%u:%u",
	    &d1, &d2, &d3,
	    &h1, &h2, &h3);
 year = gRANGE0(d1,-65000,65000);
 month = gRANGE0(d2,1,12);
 day = gRANGE0(d3,1,31);
 hour = gRANGE0(h1,0,23);
 minu = gRANGE0(h2,0,59);
 sec = gRANGE0(h3,0,59);
 if ( lastOpError!=6 ) return SetError( 1 );
 lastOpError = 0;
 return SetError( thisConvertToStamp( *this, aStamp ) )==0;
}

int gDateTime::thisCalCheck ()
{
 return thisDateTimeCheck( *this, minimumYear );
}

int gDateTime::thisDateTimeCheck (gDateTime& aDtTm, int minYear)
{
 if ( thisDateCheck( aDtTm.year, aDtTm.month, aDtTm.day, minYear )!=0 )
     return 1;
 if ( thisTimeCheck( aDtTm.hour, aDtTm.minu, aDtTm.sec )!=0 )
     return 2;
 return 0;
}

int gDateTime::thisDateCheck (int aYear, t_uint8 aMonth, t_uint8 aDay, int minYear)
{
 t_uint8 nDays;
 if ( aYear<minYear ) return 1;
 if ( aYear>65000 ) return 2;
 nDays = DaysOfMonth( aYear, aMonth );
 if ( nDays==0 ) return 3;
 if ( aDay<1 ) return 4;
 return aDay>nDays ? 5 : 0;
}

int gDateTime::thisTimeCheck (t_uint8 aHour, t_uint8 aMin, t_uint8 aSec)
{
 return
     (aHour<24 &&
      aMin<60 && aSec<60)==false;
}

int gDateTime::thisConvertFromStamp (t_stamp aStamp)
{
 struct tm* pTM;
 time_t tStamp = (time_t)aStamp;

 pTM = gmtime( &tStamp );
 if ( pTM==nil ) return -1;
 year = (t_int16)pTM->tm_year+1900;
 month = (t_uint8)pTM->tm_mon+1;
 day = (t_uint8)pTM->tm_mday;
 hour = (t_uint8)pTM->tm_hour;
 minu = (t_uint8)pTM->tm_min;
 sec =(t_uint8)pTM->tm_sec;
 ;
 wday = (t_uint8)pTM->tm_wday;
 yday = (t_uint16)pTM->tm_yday;
 isdst = (t_int8)pTM->tm_isdst;
 return 0;
}

int gDateTime::thisConvertToStamp (gDateTime& aDtTm, t_stamp& aStamp)
{
 // Converts aDtTm into a timestamp (aStamp)
 time_t tStamp;
 struct tm aTM;

 // Within gobj lib, stamp 0 is invalid!, see below
 aStamp = 0;  // Default result already here, since mktime may fail, obviously!
 if ( thisConvertTo_libc_tm( aDtTm, &aTM )!=0 )
     return 4;  // Invalid date
 tStamp = mktime( &aTM );
 DBGPRINT_MIN("{[%d,%d,%d],dst?%d|%ld}",aTM.tm_year,aTM.tm_mon,aTM.tm_mday,aTM.tm_isdst,(long)tStamp);
 DBGPRINT_MIN("{tzname(%d)=%s:%s|%s}",daylight,tzname[daylight],tzname[0],tzname[1]);
 if ( tStamp==(time_t)-1 ) {
     // Cannot be represented as calendar time (a.k.a epoch time)
     return 8;
 }
 aStamp = (t_stamp)tStamp;
 // All ok, so returns 0
 return 0;
}

int gDateTime::thisConvertTo_libc_tm (gDateTime& aDtTm, struct tm* pTM)
{
 ASSERTION(pTM!=nil,"pTM!=nil");
 memset( pTM, 0x0, sizeof(struct tm) );
 // memset: not needed, but useful if tm is changed,
 // or if some checks fail
 if ( thisDateTimeCheck( aDtTm, 1900 )!=0 )
     return 1;
 // (Note: parameter 1900 depends clearly on libc functions!)
 // Here we are sure year>=1900, and month>=1, ...etc
 pTM->tm_year = aDtTm.year-1900;
 pTM->tm_mon = aDtTm.month-1;
 pTM->tm_mday = aDtTm.day;
 pTM->tm_hour = aDtTm.hour;
 pTM->tm_min = aDtTm.minu;
 pTM->tm_sec = aDtTm.sec;
 pTM->tm_wday = aDtTm.wday;
 pTM->tm_yday = (int)aDtTm.yday;
 pTM->tm_isdst = aDtTm.isdst;
 DBGPRINT_MIN("{tm_isdst=%d}",aDtTm.isdst);
 return aDtTm.isdst==-1 ? -1 : 0;
}
////////////////////////////////////////////////////////////
// gTimeStamp - Stamp class
// ---------------------------------------------------------
gTimeStamp::gTimeStamp (t_stamp aStamp)
    : gUInt( aStamp ),
      iValid( aStamp!=0 )  //iValid=1 means valid!
{
}

gTimeStamp::gTimeStamp (gDateTime& aDtTm)
    : iValid( aDtTm.IsOk()==true )
{
 SetStamp( aDtTm.GetTimeStamp() );
}

gTimeStamp::~gTimeStamp ()
{
}

bool gTimeStamp::IsOk ()
{
 return (bool)(iValid = c!=0);
}

bool gTimeStamp::SetStamp (t_stamp aStamp)
{
 return SetUInt( (unsigned)aStamp );
}

bool gTimeStamp::SetUInt (unsigned v)
{
 gUInt::SetUInt( v );
 return IsOk();
}
////////////////////////////////////////////////////////////

