// gconfig.cpp -- Version 0.2

#include <string.h>
#include "gconfig.h"

// Static members
// <>

////////////////////////////////////////////////////////////
gFileFetch::gFileFetch (int maxLines)
    : gFileText( NULL, false ),
      doEndNewLine( false ),
      maxNLines( maxLines ),
      isFetchBufferOk( true ),
      doResize( false ),
      doShowProgress( false ),
      fVRepErr( nil )
{
}

gFileFetch::gFileFetch (gString& sFName, int maxLines)
    : gFileText( sFName.Str(), true ),
      doEndNewLine( false ),
      maxNLines( maxLines ),
      isFetchBufferOk( true ),
      doResize( IsDevice()==false ),
      doShowProgress( false ),
      fVRepErr( nil )
{
 bool isOk;
 thisReadFile( isOk, aL );
}

gFileFetch::gFileFetch (char* fName, int maxLines, bool aShowProgress)
    : gFileText( fName, true ),
      doEndNewLine( false ),
      maxNLines( maxLines ),
      isFetchBufferOk( true ),
      doResize( IsDevice()==false ),
      doShowProgress( aShowProgress ),
      fVRepErr( nil )
{
 bool isOk;
 if ( fName==NULL || fName[0]==0 ) {
     OpenDevice( e_fStdin );
     ASSERTION(doShowProgress==false,"progress?");
 }
 else {
     if ( doShowProgress ) SetDeviceReport( e_fStderr );
 }
 thisReadFile( isOk, aL );
}

gFileFetch::gFileFetch (gString& sInput, bool aShowProgress)
    : gFileText( nil, false ),
      doEndNewLine( false ),
      maxNLines( -1 ),
      isFetchBufferOk( true ),
      doResize( false ),
      doShowProgress( aShowProgress ),
      fVRepErr( nil )
{
 if ( doShowProgress ) SetDeviceReport( e_fStderr );
 thisReadStringAsFile( sInput, aL );
}

gFileFetch::~gFileFetch ()
{
}

bool gFileFetch::SetFileReport (FILE* fRep)
{
 doShowProgress = fRep!=nil;
 fVRepErr = fRep;
 return doShowProgress;
}

bool gFileFetch::SetDeviceReport (eDeviceKind aDKind)
{
 fVRepErr = nil;
 switch ( aDKind ) {
 case e_fDevOther:
 case e_fStdin:
     return false;
 case e_fStdout:
     fVRepErr = stdout;
     break;
 case e_fStderr:
     fVRepErr = stderr;
     break;
 default:
     return false;
 }
 // doShowProgress is here updated as well...
 return SetFileReport( fVRepErr );
}

bool gFileFetch::Fetch (gString& sFName)
{
 bool isOk;
 int res;

 isOk = sFName.IsEmpty()==false && OpenToRead( sFName.Str() )==true;
 if ( isOk==false ) return false;
 res = thisReadFile( isOk, aL );
 ASSERTION(res!=2,"res!=2");
 return isOk;
}

int gFileFetch::thisReadFile (bool& isOk, gList& zL)
{
 int error = 0;
 t_uint32 nBytes;

 if ( IsOpened()==false ) return 2;

 error = thisReadAll( isOk, isFetchBufferOk, zL );
 if ( error!=0 ) return error;
 if ( doResize==false ) return error;

 ASSERTION(IsDevice()==false,"IsDevice()==false");
 // Check if there was a buffer overun
 if ( isFetchBufferOk ) return 0;
 // Empty list meanwhile used: zL
 zL.Delete();
 if ( Rewind()==false ) return 1;
 thisReadFileThrough( zL, nBytes );
 error = nBytes<Size() ? 4 : 0;
 //if ( error!=0 ) fprintf(stderr,"DBG::: nBytes=%ld, size()=%ld\n",(long)nBytes,(long)Size())
 isFetchBufferOk = error==0;
 return error;
}

int gFileFetch::thisReadAll (bool& isOk, bool& isBufOk, gList& zL)
{
 int iCount=0;

 isBufOk = true;

 while ( ReadLine( isOk, doEndNewLine ) ) {
     iCount++;
     if ( maxNLines>=0 && iCount>maxNLines ) {
	 doEndNewLine = true;
	 return lastOpError = -9;
     }

#ifdef TARFS_MICE
     static t_uint16 dummyCount;
     t_uint16 dummyRand = gStorageControl::Self().GetRandom(TARFS_MICE), dummyLimit;
     dummyLimit = dummyRand<1;
     if ( dummyLimit!=0 ) {
	if ( dummyCount++==0 ) fprintf(stderr,"@TARFS_MICE: skipped (%u:%u)\n",dummyRand,TARFS_MICE);
     }
     if ( dummyLimit!=0 )
#endif //TARFS_MICE
     zL.Add( Buffer() );
     if ( isBufferOk==false ) isBufOk = false;
 }
 isOk = true;

 return 0;
}

int gFileFetch::thisReadFileThrough (gList& zL, t_uint32& nBytes)
{
 int iChr;
 t_uchar c;
 gString s;
 t_uint32 aSize = Size();

 nBytes = 0;
 while ( (iChr = fgetc( f ))>=0 ) {
     c = (t_uchar)iChr;
     nBytes++;
     if ( c=='\r' ) continue;
     if ( c=='\n' ) {
	 zL.Add( s );
	 s.SetEmpty();
	 if ( doShowProgress ) {
	     fprintf(fVRepErr,"%ld of %ld\r",(long)nBytes,(long)aSize);
	     fflush( fVRepErr );
	 }
	 continue;
     }
     s.Add( c );
 }
 if ( doShowProgress ) {
     fprintf(fVRepErr,".%20s\n"," ");
 }
 if ( s.IsEmpty() ) return 0;
 zL.Add( s );
 return 0;
}

int gFileFetch::thisReadStringAsFile (gString& sInput, gList& zL)
{
 unsigned i, len=sInput.Length();
 char* str;  //for quickness
 char chr;

 gString sLine;
 for (i=0, str=sInput.Str(); i<len; i++) {
     chr = str[i];
     if ( chr=='\r' ) continue;
     if ( chr=='\n' ) {
	 zL.Add( sLine );
	 sLine.SetEmpty();
     }
     else {
	 sLine.Add( chr );
     }
 }
 doEndNewLine = sLine.IsEmpty();
 if ( doEndNewLine==false ) zL.Add( sLine );
 return 0;
}
////////////////////////////////////////////////////////////
gConfig::gConfig (eConfKind aKind)
    : lastOpError( 0 ),
      maxNLines( -1 )
{
 SetConfKind( aKind );
 thisConfigInit();
}

gConfig::gConfig (eConfKind aKind, char* fName, bool doOpenToRead)
    : lastOpError( 0 ),
      maxNLines( -1 )
{
 SetConfKind( aKind );
 thisConfigInit();
 if ( doOpenToRead )
     lastOpError = thisRead( fName, cfgL, linesInputL );
}

gConfig::~gConfig ()
{
 thisConfigDelete();
}

bool gConfig::HasSection (unsigned idxSection)
{
 unsigned nLines;
 return HasSection( idxSection, nLines );
}

bool gConfig::HasSection (unsigned idxSection, unsigned& nLines)
{
 if ( hasSections==false ) return false;
 if ( idxSection>=1 && idxSection<=nConf ) {
     nLines = confL[idxSection].N();
     return true;
 }
 return false;
}

unsigned gConfig::GetSectionFromName (char* sSection)
{
 unsigned i;
 if ( sSection==nil ) return 0;
 for (i=1; i<=nConf; i++) {
     gString sTemp( sSection );
     if ( sTemp.Match( sectL.Str(i) ) ) return i;
 }
 return 0;
}

gList* gConfig::GetSection (char* sSection)
{
 unsigned i = GetSectionFromName( sSection );
 if ( i==0 ) return nil;
 return &conpL[ i ];
}

bool gConfig::GetSectionLine (unsigned idxSection,
			      unsigned idxLine,
			      gString& sSep,
			      gString& sLeft,
			      gString& sRight,
			      bool& isAssign)
{
 unsigned nLines;
 gStorage* inObjL;
 gList* inL;

 sSep.SetEmpty();

 isAssign = false;
 if ( HasSection( idxSection, nLines )==false )
     return false;
 if ( idxLine<1 || idxLine>nLines )
     return false;

 inObjL = conpL[idxSection].GetObjectPtr(idxLine);
 isAssign = inObjL->Kind()==e_List;
 if ( isAssign==false ) {
     // Then it should be a string
     if ( inObjL->Kind()!=e_String ) return false;
     sLeft = inObjL->Str();
     // Returning ok, though 'isAssign' is false: it's not an assignment (e.g. Param=Value);
     // the exception is when there is no string (an empty line)
     return sLeft.IsEmpty()==false;
 }
 inL = (gList*)inObjL;
 isAssign = inL->N()==3;
 if ( isAssign ) {
     sSep = inL->Str( 1 );
     sLeft = inL->Str( 2 );
     sRight = inL->Str( 3 );
 }
 return isAssign;
}

bool gConfig::SetConfKind (eConfKind aKind)
{
 confKind = aKind;
 hasSections = false;

 switch ( aKind ) {
 case e_ConfLinearFree:
 case e_ConfLinear:
     break;
 case e_ConfSectionStrict:
 case e_ConfSectionFree:
     hasSections = true;
 default:
     return false;
 }
 return true;
}

bool gConfig::Process ()
{
 if ( sFileName.IsEmpty() ) return false;
 return Process( sFileName.Str() );
}

bool gConfig::Process (char* fName)
{
 lastOpError = thisRead( fName, cfgL, linesInputL );
 return lastOpError==0;
}

int gConfig::thisConfigInit ()
{
 lineStartCommentL.Add( (char*)"#" );
 lineStartPos = 0;
 doTrim = true;
 lineSepChrs = (char*)"=";
 lineSepStrs.Delete();
 //
 sectionStartPos = 1;
 sectionSep = (char*)"[]";
 //
 iLineError = 0;
 nConf = 0;
 confL = conpL = nil;
 return 0;
}

int gConfig::thisConfigDelete ()
{
 linesInputL.Delete();
 cfgL.Delete();
 nConf = 0;
 sectL.Delete();
 delete[] confL;
 confL = nil;
 delete[] conpL;
 conpL = nil;
 return 0;
}

int gConfig::thisRead (char* fName, gList& aL, gList& idxL)
{
 bool isOk;
 bool isComment;
 int lineCount=0;
 unsigned i, pos;

 ASSERTION(fName!=nil,"fName!=nil");
 sFileName = fName;

 thisConfigDelete();

 gFileText fCfg( fName );
 if ( fCfg.IsOpened()==false ) return 2;
 while ( fCfg.ReadLine(isOk) ) {
     lineCount++;
     if ( maxNLines>=0 && lineCount>maxNLines ) return 3;
     gString sTemp( fCfg.UBuffer() );
     gString sTrim( sTemp );
     if ( doTrim ) sTrim.Trim();
     // Get rid of comments...
     isComment = false;
     unsigned nCom = lineStartCommentL.N();
     for (i=1, pos=0; i<=nCom && isComment==false; i++) {
	 gString sLineStartComment( lineStartCommentL.Str(i) );
	 pos = sTrim.Find( sLineStartComment );
	 if ( lineStartPos==0 ) {
	     // Process rest of line from comment
	     // onwards, e.g.:
	     //      param=yes  ;comment
	     // Process
	     //      param=yes
	     if ( pos>0 ) isComment = true;
	 }
	 else {
	     if ( pos==lineStartPos ) isComment = true;
	 }
     }
     if ( isComment ) {
	 if ( pos<=1 ) continue;
	 sTrim.Delete( pos );
	 if ( doTrim ) sTrim.Trim();
     }
     aL.Add( sTrim );
     idxL.Add( lineCount );
 }
 return thisProcess( aL, linesInputL, iLineError );
}

int gConfig::thisProcess (gList& aL, gList& iLineL, int& lineError)
{
 // Returns 0 on success, on >20 for parse errors
 bool isSection;
 int line;
 unsigned i, k, n=aL.N();
 unsigned pos, posEnd, len;
 unsigned idxSection=0;
 t_uchar uSecChr=sectionSep[1];
 t_uchar uSecChr2=sectionSep[2];
 bool isChrStartEnd=thisSectionChrStartEnd( uSecChr );
 char* str;

 lineError = 0;

 // If has sections, count them first (nSections)
 if ( hasSections ) {
     if ( sectionSep.IsEmpty() ) return 21;
     for (i=1; i<=n; i++) {
	 line = iLineL.GetInt( i );
	 gString sTemp( aL.Str(i) );
	 len = sTemp.Length();
	 isSection = false;
	 pos = sTemp.Find( (char)uSecChr );
	 if ( sectionStartPos==0 ) {
	     isSection = pos>0;
	 }
	 else {
	     isSection = pos==sectionStartPos;
	 }
	 posEnd = sTemp.FindBack( (char)uSecChr2 );
	 if ( isSection==false ) continue;
	 ASSERTION(sectionSep.IsEmpty()==false,"sectionSep non empty");
	 gString sSection;
	 if ( isChrStartEnd ) {
	     if ( posEnd!=len ) {
		     lineError = line;
		     return 22;
	     }
	     sTemp[posEnd] = 0;
	     sSection.CopyFromTo( sTemp, pos+1 );
	 }
	 else {
	     if ( pos==len ) sTemp[pos] = 0;
	     if ( sectionStartPos==1 ) {
		 // e.g. ":abc" => "abc"
		 sSection.CopyFromTo( sTemp, sectionSep.Length()+1 );
	     }
	     else {
		 sSection = sTemp;
	     }
	 }

	 // Add section name
	 sSection.Trim();
	 sectL.Add( sSection );
	 nIdxSectL.Add( (int)i );
     }//end FOR (check sections)
     nConf = sectL.N();
 }//end IF (has sections)
 else {
     nConf = n;
 }

 confL = new gList[ nConf+1 ];
 conpL = new gList[ nConf+1 ];
 if ( nConf==0 ) return 0; // nothing to process

 // Process sections
 if ( hasSections ) {
     // Process 'header lines' before sections, if any
     k = nIdxSectL.GetInt( 1 );
     for (i=1; i<k; i++) noSectL.Add( aL.Str(i) );
     for (idxSection=1; idxSection<=nConf; idxSection++) {
	 unsigned nextK;
	 k = nIdxSectL.GetInt( idxSection );
	 nextK = idxSection<nConf ? nIdxSectL.GetInt( idxSection+1 )-1 : n;
	 for (i=k+1; i<=nextK; i++) {
	     str = aL.Str(i);
	     confL[idxSection].Add( str );
	     thisParseLine( str, conpL[idxSection] );
	 }
     }
 }
 return 0;
}

int gConfig::thisParseLine (char* str, gList& pL)
{
 // Returns the error-code.
 // pL contains: a list with
 //              -> the separating character (lineSepChrs)
 //              -> left arg
 //              -> right arg
 ;
 unsigned len, pos, posAny;

 ASSERTION(str!=nil,"str!=nil");
 len = lineSepStrs.N();
 if ( len>0 ) return -1; //TODO: separators as strings
 gString s( str );
 len = lineSepChrs.Length();
 pos = s.FindAnyChr( lineSepChrs, true, posAny );
 if ( pos==0 ) {
     pL.Add( str );
     return 0;
 }
 gString sLeft, sRight;
 if ( pos>1 ) sLeft.CopyFromTo( s, 1, pos-1 );
 sLeft.Trim();
 sRight.CopyFromTo( s, pos+1 );
 sRight.Trim();

 gList lineL;
 gString sSepThis( lineSepChrs[posAny] );
 lineL.Add( sSepThis );
 lineL.Add( sLeft );
 lineL.Add( sRight );

 return thisAddLineToList( lineL, pL );
}

int gConfig::thisAddLineToList (gList& lineL, gList& pL)
{
 // Add one element 'lineL' to list 'pL'
 unsigned i, n = lineL.N();
 gList* newObj;
 if ( n!=3 ) return -1;
 for (i=1; i<=n; i++) {
     if ( lineL.GetObjectPtr(i)->Kind()!=e_String )
	 return -2;
 }
 newObj = new gList;
 for (i=1; i<=n; i++) {
     newObj->Add( lineL.Str(i) );
 }
 pL.AppendObject( newObj );
 return 0;
}

bool gConfig::thisSectionChrStartEnd (t_uchar uSecChr)
{
 switch ( uSecChr ) {
 case '[':
 case '(':
 case '{':
 case '<':
     return true;
 default:
     break;
 }
 return false;
}
////////////////////////////////////////////////////////////

