#ifndef gCONFIG_X_H
#define gCONFIG_X_H

#include "glist.h"
#include "gfile.h"
#include "gfilestat.h"

#define CONF_CHR_LINE_ERR 0x255
////////////////////////////////////////////////////////////
class gFileFetch : public gFileText {
public:
    gFileFetch (int maxLines=-1) ;
    gFileFetch (gString& sFName, int maxLines=-1) ;
    gFileFetch (char* fName, int maxLines=-1, bool aShowProgress=false) ;
    gFileFetch (gString& sInput, bool aShowProgress) ;
    virtual ~gFileFetch () ;

    // Public data-members
    gList aL;
    bool doEndNewLine;

    // Get methods
    virtual bool IsBufferOk () {
	return isFetchBufferOk;
    }

    virtual char* Str (unsigned idx) {
	return aL.Str( idx );
    }

    virtual bool BufferAutoAdjust () {
	return doResize;
    }

    // Set methods
    bool Fetch (gString& sFName) ;
    void AdjustBuffer (bool doBufferResize) {
	doResize = doBufferResize;
    }

    virtual bool SetFileReport (FILE* fRep) ;
    virtual bool SetDeviceReport (eDeviceKind aDKind) ;

protected:
    int maxNLines;
    bool isFetchBufferOk;
    bool doResize;

    bool doShowProgress;
    FILE* fVRepErr;

    int thisReadFile (bool& isOk, gList& zL) ;
    int thisReadAll (bool& isOk, bool& isBufOk, gList& zL) ;
    int thisReadFileThrough (gList& zL, t_uint32& nBytes) ;
    int thisReadStringAsFile (gString& sInput, gList& zL) ;

private:
    // Operators,empty
    gFileFetch (gFileFetch& ) ; //empty
    gFileFetch& operator= (gFileFetch& ) ; //empty
};
////////////////////////////////////////////////////////////
class gConfig : public gControl {
public:
    // Public enums
    enum eConfKind {
	e_ConfLinearFree,   // e.g. /etc/host.conf
	                    // Each line one config.
	                    //
	e_ConfLinear,       // e.g. /etc/sysconfig/network
	                    // Each line one config, and
	                    // separated with specific char.
	                    //
	e_ConfSectionStrict,// e.g. win.ini or smb.conf
	                    // Section like [] or X:
	e_ConfSectionFree   // e.g. httpd.conf
	                    // Section like <A>&</A> and other
	                    // parts linear.
	                    //
    };

    gConfig (eConfKind aKind=e_ConfLinearFree) ;
    gConfig (eConfKind aKind, char* fName, bool doOpenToRead=true) ;
    virtual ~gConfig () ;

    // Public data-members
    int lastOpError;           // Errors
                               //  1-10 : pre-parse
                               //  21: No section specified
                               //  22: Invalid section separator
                               //
                               //
                               //
    int maxNLines;             // Max.number of lines allowed (-1:no limit)
    eConfKind confKind;
    gString sFileName;
    gList lineStartCommentL;   // Usually #
                               // This is a list with all
                               // strings considered as comments.
                               //
    unsigned lineStartPos;     // 0: any; 1: first chr, 2: ...
                               // Default is 0.
                               //
    bool doTrim;               // Usually true

    // => input control
    gList linesInputL;

    // => line parser configs
    gString lineSepChrs;       // By default '='
    gList lineSepStrs;         // By default no strings

    // => section parser configs
    unsigned sectionStartPos;  // By default 1 (first chr)
    gString sectionSep;        // By default "[]"

    // => parsed results
    //    If has n sections contains n * confL;
    //    If is linear, each list contains the separator,
    //    the left arg and the right arg (3 elements).
    int iLineError;   // Line index with error (0=no error)
    unsigned nConf;   // Number of 'confL'/'conpL'
    gList* confL;     // The raw configurations
    gList* conpL;     // The parsed configurations
                      // Both vectors of lists are [1..n] based.

    gList  sectL;     // Section raw list
    gList  nIdxSectL; // Index(line) of sections found
    gList  noSectL;   // Header lines before sections

    // Get methods
    bool HasSections () {
	return hasSections;
    }

    bool HasSection (unsigned idxSection) ;
    bool HasSection (unsigned idxSection, unsigned& nLines) ;

    unsigned GetSectionFromName (char* sSection) ;
    virtual gList* GetSection (char* sSection) ;

    bool GetSectionLine (unsigned idxSection,
			 unsigned idxLine,
			 gString& sSep,
			 gString& sLeft,
			 gString& sRight,
			 bool& isAssign) ;

    // Set methods
    bool SetConfKind (eConfKind aKind) ;
    bool Process () ;
    bool Process (char* fName) ;

protected:
    bool hasSections;
    gList cfgL;

    int thisConfigInit () ;
    int thisConfigDelete () ;

    int thisRead (char* fName, gList& aL, gList& idxL) ;

    int thisProcess (gList& aL, gList& iLineL, int& lineError) ;
    int thisParseLine (char* str, gList& pL) ;
    int thisAddLineToList (gList& lineL, gList& pL) ;
    bool thisSectionChrStartEnd (t_uchar uSecChr) ;

private:
    // Operators,empty
    gConfig (gConfig& ) ; //empty
    gConfig& operator= (gConfig& ) ; //empty
};
////////////////////////////////////////////////////////////
#endif //gCONFIG_X_H

