// gcontrol.cpp -- Version 0.2

#include <string.h>
#include <stdlib.h>
#include <stdarg.h>  // va_start...
#include "gcontrol.h"
////////////////////////////////////////////////////////////
// Static members
// ---------------------------------------------------------
int gControl::nErrors[ LOG_LOGMAX ];
gList gControl::lLog[ LOG_LOGMAX ];
int gControl::dbgLevelDefault=LOG_INFO;
char gControl::logBuf[4096];
short gControl::intCtrl=-1;
short gStorageControl::assertedNr=0;
sServiceRegister* gStorageControl::pServReg=nil;
int gStorageControl::rStaticStgObjs=GMIN_STAT_OBJ;
short gStorageControl::maxServices=10;  // Socket-serv, ... (up to 100)
gStorageControl gStorageControl::myself;
////////////////////////////////////////////////////////////
gControl::gControl (eStorage aKind)
    : gStorage( aKind, e_StgNoStore ),
      lastOpError( 0 ),
      dbgLevel( dbgLevelDefault )
{
 SetError( 0 );
 if ( intCtrl==-1 ) gStorageControl::Self().StaticAlloc( "gControl.lLog", LOG_LOGMAX );
 intCtrl = 0;
}

t_uint16 gControl::GetRandom (t_uint16 maxRange)
{
 return (t_uint16)(rand())%maxRange;
}

void gControl::Reset ()
{
 lastOpError = 0;
}

void gControl::ResetLog ()
{
 for (int errKind=LOG_ERROR; errKind<LOG_LOGMAX; errKind++) {
     nErrors[ errKind ] = 0;
     lLog[ errKind ].Delete();
 }
}

int gControl::SetError (int opError)
{
 char* str;
 memset( sStrError, 0x0, sizeof(sStrError) );
 lastOpError = opError;
 if ( opError==0 ) return 0;
 if ( opError==-1 ) {
     strcpy( sStrError, "Internal Error" );
     return -1;
 }
 str = strerror( opError );
 ASSERTION(str!=nil,"strerror");
 strcpy( sStrError, str );
 return opError;
}

int gControl::Log (FILE* dbgFile, int level, const char* formatStr, ...)
{
 // Return 0 if no log applied, or -1 if dbgFile stream is nil
 static char kindErr[24], kindStr[24];

 if ( level<=LOG_NONE ) return 0;
 if ( level>=dbgLevel ) return 1;
 ASSERTION(formatStr!=nil,"formatStr!=nil");
 memset( logBuf, 0x0, sizeof(logBuf) );

 // Or: va_start( ap, formatStr ); vfprintf( dbgFile, formatStr, ap ); va_end( ap );
 va_list ap;
 va_start( ap, formatStr );
 vsnprintf( logBuf, sizeof(logBuf)-1/*force null ended string*/, formatStr, ap );
 va_end( ap );
 switch ( level ) {
 case LOG_ERROR:
     strcpy( kindErr, "Error" );
     break;
 case LOG_WARNING:
 case LOG_NOTICE:
     strcpy( kindErr, "Warning" );
     break;
 default:
     level = LOG_LOGMAX-1;
     kindErr[0] = 0;
     break;
 }
 kindStr[0] = 0;
 if ( kindErr[0]!=0 ) snprintf( kindStr, sizeof(kindStr), "%s: ", kindErr );
 nErrors[ level ]++;
 if ( dbgFile==nil ) {
     lLog[ level ].Add( logBuf );
     DBGPRINT_LOG("DBG: LOG(%d) %s%s\n",level,kindStr,logBuf);
     return 1;
 }
 fprintf( dbgFile, "%s%s", kindStr, logBuf );
 return 1;
}

int gControl::ClearLogMem (int level)
{
 // Return 0 if o.k.
 if ( level <= LOG_NONE ) return -1;
 if ( level >= LOG_LOGMAX ) return 1;
 lLog[ level ].Delete();
 return 0;
}

int gControl::ClearLogMemAll ()
{
 int level;
 for (level=LOG_NONE+1; ; level++)
     if ( ClearLogMem( level )!=0 ) return 0;
 return -1;
}

int gControl::ClearAllLogs ()
{
 ClearLogMemAll();
 return 0;
}

gStorage* gControl::NewObject ()
{
 ASSERTION_FALSE("gControl::NewObject");
 return nil;
}
////////////////////////////////////////////////////////////
gStorageControl::gStorageControl ()
    : nServices( 0 )
{
 if ( pServReg==nil ) {
     pServReg = new sServiceRegister[ maxServices+1 ];
 }
 ASSERTION(pServReg!=nil,"pServReg!=nil");
}

gStorageControl::~gStorageControl ()
{
#ifdef DEBUG
 if ( assertedNr==0 ) fprintf(stderr,"@END not called.\n");
#endif //DEBUG
 Release();
}

int gStorageControl::NumObjs ()
{
 int deltaObjs = GetNumObjects()-rStaticStgObjs;
 //printf("DBG: deltaObjs=%d (%d, static-stg:%d)\n",deltaObjs,GetNumObjects(),rStaticStgObjs)
 return deltaObjs;
}

char* gStorageControl::StorageKindName (short refKind)
{
 short ref;
 char* desc;

 ASSERTION(refKind>=0 && refKind<e_UnusedStore,"refKind>=0...");
 ;
 desc = storeKindDesc[refKind].descStr;
 ref = storeKindDesc[refKind].refKind;
 ASSERTION(ref==refKind,"ref==refKind");
 ASSERTION(desc!=nil,"desc!=nil");
 return desc;
}

void gStorageControl::StaticAlloc (char* msgStr, int incrNrStaticStgObjs)
{
 ASSERTION(msgStr!=nil,"msgStr!=nil");
 DBGPRINT_MIN("DBG:StaticAlloc:%s:(cur=%d)+%d\n",msgStr,rStaticStgObjs,incrNrStaticStgObjs);
 rStaticStgObjs += incrNrStaticStgObjs;
}

bool gStorageControl::IsServiceUsed (short servIdx)
{
 return GetServiceDescriptor( servIdx ).refKind>0;
}

char* gStorageControl::GetServiceName (short servIdx)
{
 ASSERTION(servIdx>=0,"servIdx>=0");
 ASSERTION(servIdx<=maxServices,"servIdx<=maxServices");
 if ( servIdx==0 ) return (char*)"ROOT";
 return pServReg[ servIdx ].descriptor.descStr;
}

sStorageDesc& gStorageControl::GetServiceDescriptor (short servIdx)
{
 ASSERTION(servIdx>=0,"servIdx>=0");
 ASSERTION(servIdx<=maxServices,"servIdx<=maxServices");
 return pServReg[ servIdx ].descriptor;
}

sServiceRegister& gStorageControl::GetService (char* descStr)
{
 short refKind;
 short servIdx = thisFindService( descStr, refKind );
 ASSERTION(servIdx>=0,"servIdx>=0");
 return pServReg[ servIdx ];
}

sServiceRegister* gStorageControl::GetServicePtr (char* descStr)
{
 short refKind;
 short servIdx = thisFindService( descStr, refKind );
 if ( servIdx<0 ) return nil;
 return &pServReg[ servIdx ];
}

short gStorageControl::RegisterService (char* descStr, short refKind, short servIdx)
{
 // Returns -1 on error, or a positive index on success
 ;
 ASSERTION(descStr!=nil && descStr[0]!=0,"descStr!=nil");
 ASSERTION(refKind>0,"refKind>0");
 ASSERTION(servIdx==-1 || servIdx>0,"servIdx>0...");
 if ( servIdx==-1 ) {
     // Find available service
     for (servIdx=1; servIdx<=maxServices; maxServices++)
	 if ( IsServiceUsed( servIdx )==false ) break;
     if ( servIdx>maxServices )
	 return -1;
 }
 if ( IsServiceUsed( servIdx ) ) return -1;
 // Fill-up descriptor
 nServices++;
 ASSERTION(nServices<=maxServices,"nServices<=maxServices");
 pServReg[ servIdx ].descriptor.descStr = new char[ strlen(descStr)+1 ];
 strcpy( pServReg[ servIdx ].descriptor.descStr, descStr );
 pServReg[ servIdx ].descriptor.refKind = refKind;
 return servIdx;
}

void gStorageControl::ServiceAlloc (char* descStr, unsigned nNodes)
{
 if ( nNodes>0 ) GetService( descStr ).nMediumNodes = nNodes;
 GetService( descStr ).AllocateNodes();
}

bool gStorageControl::ServiceAdd (char* descStr, sServiceNode& aNode)
{
 sServiceRegister* regPtr = GetServicePtr( descStr );
 ASSERTION(regPtr!=nil,"ServiceAdd(1)"); //Must give a valid service name
 if ( regPtr->HasNodes()==false ) return false;
 ASSERTION(regPtr->data!=nil,"ServiceAdd(2)");
 regPtr->data->Add( aNode );
 // Replace 'used' attribute on referenced node as 'unused'
 aNode.used = -1;
 return true;
}

bool gStorageControl::ServiceDelete (char* descStr, gStorage* pStorage)
{
 sServiceRegister* regPtr = GetServicePtr( descStr );
 ASSERTION(regPtr!=nil,"ServiceDelete(1)"); //Must give a valid service name
 if ( regPtr->HasNodes()==false ) return false;
 ASSERTION(regPtr->data!=nil,"ServiceDelete(3)");
 return regPtr->data->Delete( pStorage );
}

void gStorageControl::Release ()
{
#ifdef DEBUG
 int deltaObjs = NumObjs();
 if ( deltaObjs>0 ) fprintf(stderr,"Uncleaned (nObjs=%d)\n",deltaObjs);
#endif //DEBUG
 ClearAllLogs();
 delete[] pServReg;
 pServReg = nil;
}

short gStorageControl::thisFindService (char* descStr, short& refKind)
{
 // Returns -1 on error, or a positive index on success
 ;
 short servIdx;
 ASSERTION(descStr!=nil && descStr[0]!=0,"descStr!=nil");
 refKind = 0;
 for (servIdx=0; servIdx<=maxServices; servIdx++) {
     char* str = pServReg[ servIdx ].descriptor.descStr;
     if ( str==nil ) continue;
     //printf("DBG:FIND(%s,%dof%d)\n",str,servIdx,maxServices);
     ASSERTION(str!=nil,"str!=nil");
     if ( strcmp( descStr, str )==0 ) {
	 refKind = pServReg[ servIdx ].descriptor.refKind;
	 return servIdx;
     }
 }
 fprintf(stderr,"@STG: Not found service: %s\n",descStr);
 return -1;
}
////////////////////////////////////////////////////////////

