#ifndef gDIR_X_H
#define gDIR_X_H

#include <dirent.h>
#include "glist.h"
#include "gfile.h"
#include "gfilestat.h"
////////////////////////////////////////////////////////////
enum eFileSystemName {
    e_FSysNoName,
    e_FSysFile,
    e_FSysDir
};
////////////////////////////////////////////////////////////
class gFileSysName : public gString {
public:
    virtual ~gFileSysName () ;

    // Get methods
    virtual bool IsOk () {
	return fsysName!=e_FSysNoName;
    }
    virtual bool IsStrOk () ;
    virtual bool IsDirectory () {
	return isDirectory;
    }
    virtual char* Str () ;
    virtual char* Name () {
	return Str();
    }
    virtual bool HasStat () {
	return myStat!=nil;
    }
    virtual gFileStat& GetStat () ;
    virtual gFileStat* GetStatPtr () {
	return myStat;
    }

    // Public data-members
    eFileSystemName fsysName;

protected:
    gFileSysName (char* s,
		  eFileSystemName aFSysName,
		  bool isADirectory) ;

    gFileSysName (char* s,
		  eFileSystemName aFSysName,
		  gFileStat& aStat) ;

    bool isDirectory;
    gFileStat* myStat;

private:
    // Operators,etc
    gFileSysName (gFileSysName& ) ; //empty
    gFileSysName& operator= (gFileSysName& ) ; //empty
};
////////////////////////////////////////////////////////////
class gFileName : public gFileSysName {
public:
    gFileName () ;
    gFileName (char* s) ;
    gFileName (gString& s) ;
    gFileName (gString& s, gFileStat& aStat) ;
    virtual ~gFileName () ;

private:
    // Operators,etc
    gFileName (gFileName& ) ; //empty
    gFileName& operator= (gFileName& ) ; //empty
};
////////////////////////////////////////////////////////////
class gDirName : public gFileSysName {
public:
    gDirName () ;
    gDirName (char* s) ;
    gDirName (gString& s) ;
    gDirName (gString& s, gFileStat& aStat) ;
    virtual ~gDirName () ;

    // Get methods
    virtual char* Str () ;
    virtual char* Name () ;

    // Set methods
    virtual void SetDirName (char* s) ;

    // Save/etc methods
    virtual t_uchar* ToString (t_uchar* uBuf) ;

protected:
    gString dirStr;

    // Protected methods
    void thisBuildDirStr (char* s) ;

private:
    // Operators,etc
    gDirName (gDirName& ) ; //empty
    gDirName& operator= (gDirName& ) ; //empty
};
////////////////////////////////////////////////////////////
class gDirStream : public gFile {
public:
    gDirStream (char* dName=nil) ;
    virtual ~gDirStream () ;

    // Public data-members
    DIR* pdir;
    gString dirName;  // Original directory name
    gFileStat dStat;

    // Get methods
    virtual bool IsOpened () ;
    virtual char* Name () {
	return dirName.Str();
    }

protected:
    bool isDirOpened;

    int doOpenDir (char* dName) ;
    int doCloseDir () ;

private:
    // Operators,etc
    gDirStream (gDirStream& ) ; //empty
    gDirStream& operator= (gDirStream& ) ; //empty
};
////////////////////////////////////////////////////////////
class gDirGeneric : public gList {
public:
    virtual ~gDirGeneric () ;

    // Public data-members
    int lastOpError;
    gDirName sDirName;
    static t_uchar slashChr;

    // Get methods
    virtual bool IsOk () {
	return lastOpError==0;
    }
    virtual bool IsNameOk (char* s) ;
    virtual bool AllNamesOk (int depthLevel) ;

    virtual gFileSysName* GetName (unsigned idx) ;

    // Set methods
    // .

    // Save/etc methods
    virtual t_uchar* ToString (t_uchar* uBuf) ;
    virtual bool SaveGuts (FILE* f) ;
    virtual bool RestoreGuts (FILE* f) ;

protected:
    gDirGeneric (eStorage kind) ;

    gString slashStr;

    // Protected methods
    bool thisNameOk (char* s,
		     unsigned& posBad) ;

    bool thisAllNamesOk (unsigned depthLevel,
			 unsigned& posBad,
			 int& notOkDepth) ;

private:
    // Operators,empty
    gDirGeneric (gDirGeneric& ) ;
    gDirGeneric& operator= (gDirGeneric& ) ;
};
////////////////////////////////////////////////////////////
class gDir : public gDirGeneric {
public:
    gDir (char* dName=nil) ;
    gDir (gString& dName) ;
    virtual ~gDir () ;

    // Public data-members
    gDirStream dirStream;
    gList errUnstatL;

    // Get methods
    virtual gFileStat* GetStat (unsigned idx) ;
    virtual bool GetNameDir (unsigned idx, gString& resName) ;
    virtual bool GetFullNameDir (unsigned idx, gString& resName, gString& resFullName) ;

    // Set methods
    virtual void AddDir (char* s) ;
    virtual void AddFile (char* s) ;

protected:
    int dirCount;

    gDir (char* dName, bool doReadDir) ;

    // Protected methods
    int thisScanDir () ;
    int thisAddStatName (eFileSystemName aFSysName, char* s, gFileStat& aStat) ;
    int thisAddSystemName (gFileSysName* aPtr) ;

private:
    // Operators,empty
    gDir (gDir& ) ;
    gDir& operator= (gDir& ) ;
};
////////////////////////////////////////////////////////////
#endif //gDIR_X_H

