#ifndef gFILE_X_H
#define gFILE_X_H

#include <stdio.h>

// OS-SPECIFIC
#ifdef gDOS_SPEC
	;
#else
#include <unistd.h>
#include <sys/types.h>      // For off_t, ...
#endif //gDOS_SPEC (~)
// *end* OS-SPECIFIC

#include "gstring.h"
#include "gcontrol.h"

#define DEF_FIL_BUFSIZE (4*4096)

#define FL_FILE_TO_READ     2
#define FL_FILE_TO_WRITE    4

#define FILENM_STRICT_LVL1  1  // Highly strict chrs on file names
#define FILENM_STRICT_LVL2  2
#define FILENM_STRICT_LVL3  3
#define FILENM_STRICT_LVL4  4  // Regular level for chrs...
#define FILENM_STRICT_LVL5  5  // may have blanks
#define FILENM_STRICT_LVL6  6
#define FILENM_STRICT_LVL7  7
#define FILENM_STRICT_LVL8  8  // Paths can have '\' or '/'
#define FILENM_STRICT_LVL9  9

#define GX_SIGHUP    1
#define GX_SIGINT    2
#define GX_SIGQUIT   3
#define GX_SIGTRAP   5
#define GX_SIGKILL   9
#define GX_SIGUSR1  10
#define GX_SIGSEGV  11
#define GX_SIGUSR2  12
#define GX_SIGPIPE  13
#define GX_SIGALRM  14
#define GX_SIGTERM  15
#define GX_SIGCHLD  17


struct sSignalTrap {
    t_int16 sigId;    // -1: end-of-table, or signal-id
    t_int16 handled;  // 0: not-signaled, 1: signaled
    char* shortName;
};

// Global var decl.
extern sSignalTrap global_signal_traps[];

// Global function decl.
extern int gglobal_EventTerminate (int v) ;


// Forward declarations
// ---------------------------------------------------------
class gList;
// --
////////////////////////////////////////////////////////////
class gFileOut : public gControl {
public:
    gFileOut (FILE* aFile=nil, bool isModeDOS=false) ;
    virtual ~gFileOut () ;

    FILE* f;

    // Get methods
    virtual char* NewLine () {
	return sNL.Str();
    }

    virtual int MinimumStorageSize () {
	return 0;
    }
    virtual int MaximumStorageSize () {
	return -1;
    }

    // Set methods
    void SetModeDOS (bool isDOS) ;
    virtual void SetNL (gString& aNL) {
	sNL = aNL;
    }

protected:
    gString sNL;

private:
    // Operators,empty
    gFileOut (gFileOut& ) ; //empty
    gFileOut& operator= (gFileOut& ) ; //empty
};
////////////////////////////////////////////////////////////
class gFileControl : public gControl {
public:
    ~gFileControl () ;

    static gFileControl& Self () {
	return myself;
    }

    // Public data-members
    int lastOpError;
    short isCharOnFile[256];
    short charOnFileRelax;
    char tmpPath[300];
    char tmpPrefix[10];
    char tmpSuffix[10];
    t_uint16 userId;
    FILE* fReport;
    int exitCodeOnSignal;
    gList* tempL;  // List of temporary files

    // Get methods
    bool IsOkChar (t_uchar c) {
	return isCharOnFile[c]!=0;
    }
    bool IsOkStrict (t_uchar c, bool isStrict) ;

    bool IsOkPathname (char* s) ;

    int CtrlGetTempPath (char* resPathStr) ;

    t_uint32 CtrlGetPid () ;

    char* CtrlGetUserNameStr () {
	return strUName;
    }
    t_uint32 GetCurrentEpoch () ;
    gString& GetUniqueName (char* s) ;

    char* ErrorStr (int aErrorNo) ;

    bool IsTerminated () {
	return isTerminated;
    }

    // Set methods
    void InitCharOnFile (t_uchar* s) ;
    bool AddTempFile (gString& s) ;

    // Specific methods
    int Init () ;
    int End (bool doReportObjs=true) ;
    int RemoveTemp (bool doReport=true) ;
    int RemoveTemp (unsigned& nTemp, bool doReport) ;
    int Release (bool doReport) ;

    t_int16 IsSignaled (t_int16 iSignalId) ;  // non-zero if yes
    t_int16 SetSignaled (t_int16 iSignalId, t_int16 handled) ;
    t_int16 SetSignaledByStr (char* shortStr, t_int16 handled) ;

    void SignalHandler (int signalId) ;
    int OnEventTerminate (int (*DoSomething)(int)) ;

    // I/O methods
    bool Write (int fHandle, t_uchar* uBuf, unsigned nBytes) ;

    // Utilities
    void GetCwd (gString& s) ; //Get current working directory
    int NanoSleep (t_uint32 uSec, t_uint32 nSec) ;
    int NanoSleep (t_uint32 nSec) {
	return NanoSleep( 0, nSec );
    }
    int SecSleep (t_uint32 aSec) ;
    int MiliSecSleep (t_uint32 aSec) ;

protected:
    gFileControl () ;

    gString sName;
    char strUName[ 32 ];

    int thisInit () ;
    int thisInitCharOnFile (bool doDefault) ;
    int thisGetUniqueName (char* s,
			   t_uint32& aStamp,
			   t_uint32& aRand,
			   gString& sRes) ;

private:
    static gFileControl myself;
    bool isTerminated;
    bool isInit;
    gString sErrorRef;

    // Operators,empty
    gFileControl (gFileControl& ) ; //empty
    gFileControl& operator= (gFileControl& ) ; //empty
};
////////////////////////////////////////////////////////////
class gFile {
public:
    // Public enums
    enum eFileKind {
	e_Text,
	e_Binary
    };

    enum eDeviceKind {
	e_fDevOther,
	e_fStdin,
	e_fStdout,
	e_fStderr
    };

    gFile (eFileKind aFKind, char* fName, bool doOpenToRead, bool isTmpFile=false) ;
    virtual ~gFile () ;

    // Public data-members
    int lastOpError;

    // Get methods
    virtual bool IsOk () {
	return lastOpError==0 && IsOpened();
    }

    eFileKind FileKind () {
	return fKind;
    }

    eDeviceKind Device () {
	return dKind;
    }

    virtual bool IsDevice () {
	return dKind!=e_fDevOther;
    }

    virtual bool IsOpened () {
	return f!=NULL;
    }
    FILE* Stream () {
	return f;
    }
    t_uint16 Mode () {
	return fMode;
    }
    char* LastErrorStr () ;
    char* ErrorStr (int aErrorNo) ;

    // IO-methods
    virtual bool OpenDevice (eDeviceKind aDKind) ;
    virtual bool OpenToRead (char* fName) ;
    virtual bool Overwrite (char* fName) ;
    virtual bool Close () ;

    bool ReadData (void* buf, t_uint16 bufSize) ;
    virtual bool ReadBuffer (void* buf, t_uint16 bufSize, t_uint16& nBytes) ;
    virtual bool Read (void* buf, t_uint16 bufSize, t_uint16& nBytes) ;

protected:
    FILE* f;
    eFileKind fKind;
    eDeviceKind dKind;
    t_uint16 fMode;
    char lastErrorMsg[1024];

    bool thisRead (int fd, void* buf, t_uint16 bufSize, t_uint16& nBytes) ;

private:
    // Operators,empty
    gFile (gFile& ) ; //empty
    gFile& operator= (gFile& ) ; //empty
};
////////////////////////////////////////////////////////////
class gFileStream : public gFile {
public:
    gFileStream (char* fName, bool doOpenToRead=true) ;
    gFileStream (gFile::eFileKind aFKind, char* fName, bool doOpenToRead=true, bool isTmpFile=false) ;
    virtual ~gFileStream () ;

    // Get methods
    virtual bool IsBufferOk () {
	return isBufferOk;
    }
    virtual t_uint16 BufferSize () {
	return bufferSize;
    }
    virtual t_uint32 SeekPos () {
	return (t_uint32)seekPos;
    }
    virtual t_uint32 Size () {
	return (t_uint32)seekEnd;
    }
    virtual char* Buffer () ;
    virtual t_uchar* UBuffer () ;

    // IO-methods
    virtual bool Overwrite (char* fName) ;

    virtual bool ReadBuffer (void* buf, t_uint16 bufSize, t_uint16& nBytes) ;
    virtual bool Read (void* buf, t_uint16 bufSize, t_uint16& nBytes) ;

    virtual bool Rewind () ;

protected:
    bool isOpOk, isFileChanged;
    bool isBufferOk;
    off_t seekPos, seekEnd;
    t_uint16 bufferSize;
    t_uchar* buffer;

    int thisAllocateBuffer (t_uint16 aBufferSize) ;

private:
    int thisGetStreamSize (int fd) ;

    // Operators,empty
    gFileStream (gFileStream& ) ; //empty
    gFileStream& operator= (gFileStream& ) ; //empty
};
////////////////////////////////////////////////////////////
class gFileText : public gFileStream {
public:
    gFileText (char* fName, bool doOpenToRead=true) ;
    virtual ~gFileText () ;

    // IO-methods
    bool ReadLine () ;
    bool ReadLine (bool& isOk) ;
    bool ReadLine (bool& isOk, bool& hasNewLine) ;

protected:
    bool isOpOk, isFileChanged;
    off_t seekPos, seekEnd;

private:
    // Operators,empty
    gFileText (gFileText& ) ; //empty
    gFileText& operator= (gFileText& ) ; //empty
};
////////////////////////////////////////////////////////////
class gFileTemp : public gFileStream {
public:
    // Public enums
    enum eTempMethod {
	e_NameStd,   // Via mkstemp
	e_NamePre    // Big string based on 'pid',...
    };

    gFileTemp (char* strTemplate, gFile::eFileKind aFKind=gFile::e_Binary) ;
    gFileTemp (gFile::eFileKind aFKind) ;
    virtual ~gFileTemp () ;

    // Public data-members
    int fHandle;
    gString sTempName;

    // Get methods
    virtual bool IsOpened () ;

    // Other methods
    virtual bool Rewind () ;

protected:
    eTempMethod tempMethod;
    bool isHandledFStream;

    int thisOverwrite (char* fName) ;

private:
    // Operators,empty
    gFileTemp (gFileTemp& ) ; //empty
    gFileTemp& operator= (gFileTemp& ) ; //empty
};
////////////////////////////////////////////////////////////
#endif //gFILE_X_H

