// gfilestat.cpp -- Version 0.8

#include <errno.h>
#include <string.h>  // memset,...
#include "gfilestat.h"
////////////////////////////////////////////////////////////
bool sFileStat::IsDirectory ()
{
 return S_ISDIR( mode );
}

bool sFileStat::IsLink ()
{
#ifdef linux
 return S_ISLNK( mode );
#else
 return false;
#endif //linux
}

long sFileStat::Size ()
{
 if ( IsDirectory() ) return 0L;
 return
     uid==0xFFFF ? -1L : (long)size;
}

t_uint32 sFileStat::USize ()
{
 long v = Size();
 return v<0 ? 0 : (t_uint32)v;
}

t_fs_perm sFileStat::Permission ()
{
 return mode & 0x1FF; // max mask ox777
}

t_fs_perm sFileStat::AllPermission ()
{
 ;
 // ox7777 (0xFFF) gives sticky-bit ('t', ox1000) and suid ('s', ox6000)
 return mode & 0xFFF;
}
////////////////////////////////////////////////////////////
// gFileStat - Generic file status
// ---------------------------------------------------------
gFileStat::gFileStat (char* fName)
    : lastOpError( 0 ),
      lastOpErrorL( 0 ),
      name( fName==NULL ? (char*)"\0" : fName ),
      hasStat( false ),
      fHandle( -1 )
{
 if ( name.IsEmpty()==false ) {
     thisStatName( fName, -1, status, statusL );
 }
}

gFileStat::gFileStat (gString& sName)
    : lastOpError( 0 ),
      lastOpErrorL( 0 ),
      name( sName ),
      hasStat( false ),
      fHandle( -1 )
{
 if ( name.IsEmpty()==false ) {
     thisStatName( name.Str(), -1, status, statusL );
 }
}

gFileStat::~gFileStat ()
{
}

bool gFileStat::Update (int fd)
{
 fHandle = fd;
 if ( name.IsEmpty() ) {
     if ( fd==-1 ) return false;
     return thisFileStat(fd,status,statusL)==0;
 }
 return Update( name.Str() );
}

bool gFileStat::Update (char* fName)
{
 ASSERTION(fName!=0 && fName[0]!=0,"gFileStat::Update");
 name.Set( fName );
 return thisStatName( fName, fHandle, status, statusL )==0;
}

gFileStat& gFileStat::CopyStat (gFileStat& copy)
{
 lastOpError = copy.lastOpError;
 lastOpErrorL = copy.lastOpErrorL;
 name.Copy( copy.name );
 status.Copy( copy.status );
 statusL.Copy( copy.statusL );
 hasStat = copy.hasStat;
 fHandle = -1;
 return *this;
}

int gFileStat::thisStatName (char* fName,
			     int fd,
			     sFileStat& st,
			     sFileStat& stL)
{
 int res, resL;
 struct stat zsStat;
 struct stat zsStatL;

 lastOpError = lastOpErrorL = 0;
 st.ToDefault();
 stL.ToDefault();

 hasStat = false;
 if ( fd!=-1 ) return thisFileStat(fd,st,stL);

 // zsStat must be zeroed, because 'stat' will
 // not overwrite the contents for symbolic links!!!
 memset( &zsStat, 0x0, sizeof(zsStat) );
 // Optionally, zeros on statL as well.
 memset( &zsStatL, 0x0, sizeof(zsStatL) );

 ASSERTION(fName!=NULL && fName[0]!=0,"gFileStat::thisStatName");
 res = stat( fName, &zsStat );
 if ( res==-1 ) {
     lastOpError = errno;
 }
#ifdef linux
 resL = lstat( fName, &zsStatL );
#else
 resL = stat( fName, &zsStatL );
#endif //linux
 if ( resL==-1 ) {
     lastOpErrorL = errno;
     return res;
 }
 DBGPRINT_MIN("stat(%s)=%d:%d (err=%d:%d)\n",fName,res,resL,lastOpError,lastOpErrorL);

 st.inode = (t_inode)/*(ino_t)*/zsStat.st_ino;
 st.mode = zsStat.st_mode;
 st.uid = zsStat.st_uid;
 st.gid = zsStat.st_gid;
 if ( res==0 ) st.size = zsStat.st_size;
#ifdef linux
 st.blocks = (t_uint32)zsStat.st_blocks;  // unsigned long<
#else
 st.blocks = 1;
#endif //linux
 st.nLinks = zsStat.st_nlink;
 st.aTime = zsStat.st_atime;
 st.mTime = zsStat.st_mtime;
 st.cTime = zsStat.st_ctime;

 stL.inode = (t_uint64)zsStatL.st_ino;
 stL.mode = zsStatL.st_mode;
 stL.uid = zsStatL.st_uid;
 stL.gid = zsStatL.st_gid;
 stL.size = zsStatL.st_size;
#ifdef linux
 stL.blocks = zsStatL.st_blocks;
#else
 st.blocks = 1;
#endif //linux
 st.nLinks = zsStat.st_nlink;  //Irrelevant, I think...
 stL.aTime = zsStatL.st_atime;
 stL.mTime = zsStatL.st_mtime;
 stL.cTime = zsStatL.st_ctime;

 hasStat = true;
 return 0;
}

int gFileStat::thisFileStat (int fd,
			     sFileStat& st,
			     sFileStat& stL)
{
 int res;
 struct stat zsStat;

 st.ToDefault();
 stL.ToDefault();
 lastOpError = lastOpErrorL = 0;
 hasStat = false;

 if ( fd==-1 ) return 0;
 res = fstat(fd,&zsStat);
 if ( res==-1 ) {
     lastOpError = errno;
     return -1;
 }

 st.mode = zsStat.st_mode;
 st.uid = zsStat.st_uid;
 st.gid = zsStat.st_gid;
 st.size = zsStat.st_size;
 st.aTime = zsStat.st_atime;
 st.mTime = zsStat.st_mtime;
 st.cTime = zsStat.st_ctime;

 hasStat = true;
 return 0;
}
////////////////////////////////////////////////////////////

