#ifndef gFILESTAT_X_H
#define gFILESTAT_X_H

#include <stdio.h>
#include <unistd.h>
#include <sys/stat.h>       // For fstat, ...
#include <sys/types.h>      // For off_t, ...
#include "gstring.h"

typedef t_uint16 t_fs_perm;
////////////////////////////////////////////////////////////
struct sFileStat {
    sFileStat ()
	: inode( 0 ),
	  mode( 0 ),
	  uid( 0xFFFF ),
	  gid( 0xFFFF ),
	  size( 0 ),
	  blocks( 0 ),
	  nLinks( 0 ),
	  aTime( 0 ),  // time of last access
	  mTime( 0 ),  // time of last modification
	  cTime( 0 ) { // time of last change
	               // cTime change: e.g. owner change
    }
    t_inode inode;
    t_uint32 mode;
    t_uint16 uid, gid;
    t_uint32 size;
    t_uint32 blocks;
    t_uint16 nLinks;
    time_t aTime, mTime, cTime;

    void ToDefault () {
	inode = 0;
	mode = 0;
	size = blocks = 0;
	nLinks = 0;
	uid = gid = 0xFFFF;
	aTime = mTime = cTime;
    }

    bool IsValid () {
	return uid!=0xFFFF && gid!=0xFFFF;
    }

    bool IsDirectory () ;
    bool IsLink () ;

    t_fs_perm Permission () ;
    t_fs_perm AllPermission () ;

    long Size () ;
    t_uint32 USize () ;

    void Copy (sFileStat& copy) {
	inode = copy.inode;
	mode = copy.mode;
	uid = copy.uid;
	gid = copy.gid;
	size = copy.size;
	blocks = copy.blocks;
	nLinks = copy.nLinks;
	aTime = copy.aTime;
	mTime = copy.mTime;
	cTime = copy.cTime;
    }
};
////////////////////////////////////////////////////////////
class gFileStat {
public:
    gFileStat (char* fName=NULL) ;
    gFileStat (gString& sName) ;
    gFileStat (int fd) ;
    virtual ~gFileStat () ;

    // Public data-members
    int lastOpError, lastOpErrorL;
    gString name;
    sFileStat status, statusL;

    // Get methods
    virtual bool IsOk () {
	return HasStat()==true && Error()==0;
    }

    virtual bool HasStat () {
	return hasStat==true;
    }

    int Error () {
	return lastOpError==0 ? lastOpErrorL : lastOpError;
    }

    bool IsDirectory () {
	return statusL.IsDirectory();
    }
    bool IsLink () {
	return statusL.IsLink();
    }

    // Set methods
    bool Update (int fd=-1) ;
    bool Update (char* fName) ;

    gFileStat& CopyStat (gFileStat& copy) ;

protected:
    // Protected methods
    int thisStatName (char* fName,
		      int fd,
		      sFileStat& st,
		      sFileStat& stL) ;
    int thisFileStat (int fd,
		      sFileStat& st,
		      sFileStat& stL) ;

private:
    bool hasStat;
    int fHandle;

    // Operators,empty
    gFileStat (gFileStat& ) ; //empty
    gFileStat& operator= (gFileStat& ) ; //empty
};
////////////////////////////////////////////////////////////
#endif //gFILESTAT_X_H

