// ghash.cpp -- Version 0.1

#include <string.h>
#include "ghash.h"
////////////////////////////////////////////////////////////
// Static members
// ---------------------------------------------------------
// void!

gKey::gKey (eKeyKind aKeyKind)
    : keyKind( aKeyKind ),
      pvalUInt( nil ),
      pvalDouble( nil )
{
}

gKey::gKey (int v)
    : gInt( v ),
      pvalUInt( nil ),
      pvalDouble( nil )
{
 keyKind = gKey::e_Int;
}

gKey::gKey (gInt& v)
    : gInt( v.GetInt() ),
      pvalUInt( nil ),
      pvalDouble( nil )
{
 keyKind = gKey::e_Int;
}

gKey::gKey (gString& s)
    : gInt( -1 ),
      pvalUInt( nil ),
      pvalDouble( nil ),
      pvalStr( s )
{
 keyKind = gKey::e_String;
}

gKey::gKey (char* s)
    : gInt( -1 ),
      pvalUInt( nil ),
      pvalDouble( nil )
{
 pvalStr.Set( s );
 keyKind = gKey::e_String;
}

gKey::~gKey ()
{
 delete pvalUInt;
 delete pvalDouble;
}

bool gKey::IsOk ()
{
 return true;
}

t_uint64 gKey::GetUInt64 ()
{
 ASSERTION(pvalUInt!=nil,"pvalUInt!=nil");
 return *pvalUInt;
}

double gKey::GetReal ()
{
 ASSERTION(pvalDouble!=nil,"pvalDouble!=nil");
 return *pvalDouble;
}

gString& gKey::GetString ()
{
 ASSERTION(pvalUInt==nil,"pvalUInt==nil");
 ASSERTION(pvalDouble==nil,"pvalDouble==nil");
 ASSERTION(GetInt()<=0,"GetInt()<=0");
 ASSERTION(keyKind==gKey::e_String,"keyKind=e_String");
 return pvalStr;
}

unsigned gKey::HashPos (unsigned hashSize)
{
#ifdef DEBUG_HASH
 return 1; //For debugging, it is easier...
#else
 unsigned uTemp;
 double dTemp;

 switch ( keyKind ) {
 case gKey::e_Int:
     return (unsigned)GetInt()%hashSize;
 case gKey::e_UInt64:
     return (unsigned)(*pvalUInt)%hashSize;
 case gKey::e_Real:
     dTemp = *pvalDouble;
     uTemp = dTemp > (double)MAX_LONG_L ? (unsigned)(dTemp/214748364.0) : (unsigned)(dTemp*101.0);
     return uTemp%hashSize;
 case gKey::e_String:
     return HashStrPos( hashSize, pvalStr );
 default:
     ASSERTION_FALSE("HashPos(1)");
 }
 return 0;
#endif //DEBUG_HASH
}

bool gKey::MatchKey (gKey& hKey)
{
 ; // Note: hKey.keyKind must match our keyKind
 if ( keyKind!=hKey.GetKeyKind() ) return false;
 if ( hKey.IsOk()==false ) return false; //Academic...
 switch ( keyKind ) {
 case gKey::e_Int:
     return GetInt()==hKey.GetInt();
 case gKey::e_UInt64:
     return GetUInt64()==hKey.GetUInt64();
 case gKey::e_Real:
     return GetReal()==hKey.GetReal();
 case gKey::e_String:
     return GetString().Match( hKey.GetString() );
 default:
     ASSERTION_FALSE("MatchKey(1)");
 }
 return false;
}

bool gKey::MatchStorage (gStorage* aObj)
{
 gHashElemGeneric* pHashElem;
 ASSERTION(aObj!=nil,"aObj!=nil");
 ASSERTION(aObj->Kind()==e_Control,"aObj->Kind()==e_Control");
 pHashElem = (gHashElemGeneric*)aObj;
 return MatchKey( pHashElem->hKey );
}

void gKey::Reset ()
{
 gInt::Reset();
 if ( pvalUInt!=nil ) *pvalUInt = 0;
 if ( pvalDouble!=nil ) *pvalDouble = 0.0;
 pvalStr.SetEmpty();
}

void gKey::Set (int v)
{
 thisNewKind( gKey::e_Int );
 SetInt( v );
}

void gKey::Set (t_uint64 v)
{
 thisNewKind( gKey::e_UInt64 );
 ASSERTION(pvalUInt!=nil,"pvalUInt!=nil");
 *pvalUInt = v;
}

void gKey::Set (double v)
{
 thisNewKind( gKey::e_Real );
 ASSERTION(pvalDouble!=nil,"pvalDouble!=nil");
 *pvalDouble = v;
}

void gKey::Set (gString& s)
{
 thisNewKind( gKey::e_String );
 pvalStr = s;
}

void gKey::Copy (gKey& copy)
{
 eKeyKind newKind = copy.GetKeyKind();
 switch ( newKind ) {
 case gKey::e_Int:
     Set( copy.GetInt() );
     break;
 case gKey::e_UInt64:
     Set( copy.GetUInt64() );
     break;
 case gKey::e_Real:
     Set( copy.GetReal() );
     break;
 case gKey::e_String:
     Set( copy.GetString() );
     break;
 default:
     ASSERTION_FALSE("gKey::Copy(1)");
 }
 ASSERTION(keyKind==newKind,"gKey::Copy(2)");
}

unsigned gKey::HashStrPos (unsigned hashSize, gString& s)
{
 unsigned i, n, uTemp=0, uMark;
 t_uchar uChr;
 // This method, in particular with hashSize=7591 (instead of default 971),
 // leaves only one element in each row for HTML tags.
 if ( hashSize==7591 && s.Match("FORM") ) return 1519;
 for (i=1, n=s.Length(); i<=n; i++) {
     uChr = s[i];
     uMark = uChr>='a' && uChr<='z' ? (unsigned)uChr-32 : (unsigned)uChr;
     uMark -= 65;
     if ( i==1 ) uMark *= hashSize / 26;
     uTemp += uMark+i*2+(uChr>='Y');
 }
 //printf("%u\tHASH_STR_POS(%s)\n",uTemp%hashSize,s.Str());
 return uTemp%hashSize;
}

gStorage* gKey::NewObject ()
{
 ASSERTION_FALSE("gKey::NewObject");
 return nil;
}

t_uchar* gKey::ToString (t_uchar* uBuf)
{
 char* sBuf;
 if ( uBuf==nil ) return nil;
 sBuf = (char*)uBuf;
 switch ( keyKind ) {
 case gKey::e_Int:
     sprintf(sBuf,"%d",GetInt());
     break;
 case gKey::e_UInt64:
     sprintf(sBuf,"%lx",(long)GetUInt64());
     break;
 case gKey::e_Real:
     sprintf(sBuf,"%f",(float)GetReal());
     break;
 case gKey::e_String:
     sprintf(sBuf,"%s",GetString().Str());
     break;
 default:
     ASSERTION_FALSE("gKey::ToString");
 }
 return (t_uchar*)sBuf;
}

void gKey::Show (bool doShowAll)
{
 t_uchar uBuf[50];
 printf("%s",ToString(uBuf));
}

void gKey::thisNewKind (eKeyKind newKind)
{
 Reset();
 delete pvalUInt;
 pvalUInt = nil;
 delete pvalDouble;
 pvalDouble = nil;
 switch ( keyKind = newKind ) {
 case gKey::e_Int:
     break;
 case gKey::e_UInt64:
     pvalUInt = new t_uint64;
     break;
 case gKey::e_Real:
     pvalDouble = new double;
     break;
 case gKey::e_String:
     break;
 default:
     ASSERTION_FALSE("thisNewKind");
 };
}
////////////////////////////////////////////////////////////
gHashElemGeneric::gHashElemGeneric (gStorage* aObj)
    : myObj( aObj )
{
}

gHashElemGeneric::~gHashElemGeneric ()
{
 // myObj is not allocated ever!
}

gHashElemTriple::gHashElemTriple (gKey& aKey, gString& s, gStorage* aObj)
    : gHashElemGeneric( aObj ),
      sDesc( s ),
      iVal( 0 )
{
 hKey.Copy( aKey );
}

gHashElemTriple::gHashElemTriple (gKey& aKey, int aVal, gStorage* aObj)
    : gHashElemGeneric( aObj ),
      iVal( aVal )
{
 hKey.Copy( aKey );
}

gHashElemTriple::~gHashElemTriple ()
{
}

void gHashElemTriple::Show (bool doShowAll)
{
 hKey.Show( doShowAll );
 if ( doShowAll ) {
     printf("{%s",sDesc.Str());
     /*if ( myObj!=nil ) {
	 printf("[");
	 myObj->Show( true );
	 printf("]\n");
     }
     NOTE: myObj is not copied!
     */
     printf("}");
 }
}
////////////////////////////////////////////////////////////
// gHashGeneric - Generic hash handling
// ---------------------------------------------------------
gHashGeneric::gHashGeneric (eStorage aKind, unsigned hashSize)
    : gStorage( aKind, e_StgDefault ),
      size( 0 ),
      pLst( nil ),
      usageCounter( 0 )
{
 thisPreAllocate( hashSize );
}

gHashGeneric::~gHashGeneric ()
{
 thisDelete();
}

bool gHashGeneric::IsValidIndex (unsigned idx)
{
 return idx<size;
}

gList* gHashGeneric::GetHash (unsigned idx)
{
 ASSERTION(pLst!=nil,"pLst!=nil");
 ASSERTION(IsValidIndex(idx),"GetHash(1)");
 return &pLst[idx];
}

unsigned gHashGeneric::Delete ()
{
 thisDelete();
 thisPreAllocate( size );
 return size;
}

gStorage* gHashGeneric::NewObject ()
{
 ASSERTION_FALSE("gHashGeneric::NewObject: Not used");
 return nil;
}

t_uchar* gHashGeneric::ToString (t_uchar* uBuf)
{
 ;
 // Output notation:
 //     Empty hash               ===>    ()
 //     Hash of strings          ===>    (me and you)
 //     Hash of strings and nums ===>    (I am '31 years old)
 //     Hash of sentences        ===>    ("I use a suit." "You use a dress.")
 //     Hash with some resvd chrs.==>    (123-a.com sold '100 domains)
 ;
 return nil;
}

bool gHashGeneric::SaveGuts (FILE* f)
{
 return CanSave( f );
}

bool gHashGeneric::RestoreGuts (FILE* f)
{
 return CanRestore( f );
}

void gHashGeneric::thisPreAllocate (unsigned toSize)
{
 ASSERTION(toSize>=LST_HASH_SIZE_MIN,"toSize>=101");
 ASSERTION(pLst==nil,"pLst==nil");
 pLst = new gList[ size = toSize ];
 ASSERTION(pLst!=nil,"pLst!=nil");
}

unsigned gHashGeneric::thisDelete ()
{
 ASSERTION(pLst!=nil,"pLst!=nil");
 delete[] pLst;
 pLst = nil;
 return size;
}
////////////////////////////////////////////////////////////
gHash::gHash (unsigned hashSize)
    : gHashGeneric( e_List, hashSize ),
      doFindBeforeInsert( false )
{
}

gHash::~gHash ()
{
}

bool gHash::FindKey (gKey& hKey)
{
 int x;
 int hashPos = thisFindKey( hKey, x );
 return hashPos>=0;
}

unsigned gHash::Add (int v)
{
 gInt aVal( v );
 return Add( aVal );
}

unsigned gHash::Add (gInt& v)
{
 gString sEmpty;
 return Add( v, sEmpty );
}

unsigned gHash::Add (gInt& v, gString& s)
{
 gKey hKey( v );
 return Add( hKey, s );
}

unsigned gHash::Add (gKey& hKey, gString& s)
{
 if ( doFindBeforeInsert ) {
     if ( FindKey( hKey ) ) return 0;
 }
 gHashElemTriple* pElem = new gHashElemTriple( hKey, s, nil/*No object needed*/ );
 ASSERTION(pElem!=nil,"pElem!=nil");
 thisAddElem( hKey.HashPos(size), pElem );
 return 0;
}

t_uchar* gHash::ToString (t_uchar* uBuf)
{
 return gHashGeneric::ToString( uBuf );
}

bool gHash::SaveGuts (FILE* f)
{
 return gHashGeneric::SaveGuts( f );
}

bool gHash::RestoreGuts (FILE* f)
{
 return gHashGeneric::RestoreGuts( f );
}

void gHash::Show (bool doShowAll)
{
 unsigned idx, k, n;
 unsigned count=0;
 gList* pRow;
 for (idx=0; idx<size; idx++) {
     pRow = GetHash( idx );
     n = pRow->N();
     if ( n==0 ) {
	 ;//if ( doShowAll ) printf("(%u)",idx);
     }
     else {
	 printf("%s%u:",count==0?"\0":" ",idx);
	 count++;
	 for (k=1; k<=n; k++) {
	     gStorage* aObj = pRow->GetObjectPtr(k);
	     ASSERTION(aObj!=nil,"aObj!=nil");
	     aObj->Show( doShowAll );
	 }
     }
 }
 if ( doShowAll ) printf("\n");
}

int gHash::thisAddElem (unsigned idx, gHashElemGeneric* pElem)
{
 gList* pRow = GetHash( idx );
 bool mustBeOk = pRow->AppendObject( pElem );
 ASSERTION(mustBeOk,"thisAddElem(1)");
 thisInsertElement();
 return 0;
}

int gHash::thisFindKey (gKey& hKey, int& x)
{
 unsigned idx = hKey.HashPos( size );
 gList* pRow = GetHash( idx );
 int n = (int)pRow->N();
 for (x=1; x<=n; x++) {
     if ( thisKeyMatch( hKey, pRow->GetObjectPtr(x) )==1 )
	 return (int)idx;
 }
 // ...here, no matches
 x = 0;
 return -1;
}

int gHash::thisKeyMatch (gKey& hKey, gStorage* hashElem)
{
 // Returns 1 if the 'hKey' matches the object 'hashElem'
 return (int)hKey.MatchStorage( hashElem );
}

#ifdef DEBUG_HASH
bool gHash::AddTriplePos_dbg (unsigned idx, gKey& hKey, char* str)
{
 gString s( str );
 gHashElemTriple* pElem = new gHashElemTriple( hKey, s, nil );
 thisAddElem( idx, pElem );
 return pElem!=nil;
}
#endif //DEBUG_HASH
////////////////////////////////////////////////////////////
gHashTriple::gHashTriple (unsigned hashSize)
    : gHash( hashSize )
{
}

gHashTriple::~gHashTriple ()
{
}

bool gHashTriple::FindKey (gKey& hKey)
{
 // Actually the same as in gHash: no need to redefine,
 // but here stated for easier expansion of double keys...
 return gHash::FindKey( hKey );
}

gHashElemTriple* gHashTriple::GetTriple (unsigned idx, int x)
{
 gHashElemTriple* aObj;
 ASSERTION(x>=1,"x>=1");
 aObj = (gHashElemTriple*)GetHash( idx )->GetObjectPtr( (unsigned)x );
 // No need for assertion here, already in gList::GetObjectPtr!
 return aObj;
}

gHashElemTriple* gHashTriple::Find (gKey& hKey, unsigned& idx, int& x)
{
 gHashElemTriple* pElem;
 idx = hKey.HashPos( size );
 gList* pRow = GetHash( idx );
 int n = (int)pRow->N();
 for (x=1; x<=n; x++) {
     pElem = (gHashElemTriple*)pRow->GetObjectPtr(x);
     if ( thisKeyMatch( hKey, pElem )==1 )
	 return pElem;
 }
 x = 0;
 return nil;
}

bool gHashTriple::AddTriple (gKey& hKey, char* str, gStorage* aObj)
{
 gString s( str );
 return AddTriple( hKey, s, aObj );
}

bool gHashTriple::AddTriple (gKey& hKey, gString& s, gStorage* aObj)
{
 if ( doFindBeforeInsert ) {
     if ( FindKey( hKey ) ) return false;
 }
 gHashElemTriple* pElem = new gHashElemTriple( hKey, s, aObj );
 ASSERTION(pElem!=nil,"pElem!=nil");
 thisAddElem( hKey.HashPos(size), pElem );
 return true;
}

bool gHashTriple::AddTriple (gKey& hKey, int iVal, gStorage* aObj)
{
 if ( doFindBeforeInsert ) {
     if ( FindKey( hKey ) ) return false;
 }
 gHashElemTriple* pElem = new gHashElemTriple( hKey, iVal, aObj );
 ASSERTION(pElem!=nil,"pElem!=nil");
 thisAddElem( hKey.HashPos(size), pElem );
 return true;
}
////////////////////////////////////////////////////////////

