#ifndef gHASH_X_H
#define gHASH_X_H

#include "glist.h"
#include "gcontrol.h"

#define LST_HASH_SIZE 971  // A prime number, default hash size
#ifdef DEBUG_HASH
#define LST_HASH_SIZE_MIN   3  // For easier debugging...
#else
#define LST_HASH_SIZE_MIN 101  // Minimum hash size
#endif //DEBUG_HASH

////////////////////////////////////////////////////////////
class gKey : public gInt {
public:
    // Public enums
    enum eKeyKind {
	e_Int,
	e_UInt64,
	e_Real,
	e_String
    };

    gKey (eKeyKind aKeyKind=e_Int) ;
    gKey (int v) ;
    gKey (gInt& v) ;
    gKey (gString& s) ;
    gKey (char* s) ;
    virtual ~gKey () ;

    // Get methods
    virtual bool IsOk () ;
    virtual eKeyKind GetKeyKind () {
	return keyKind;
    }
    virtual t_uint64 GetUInt64 () ;
    virtual double GetReal () ;
    virtual gString& GetString () ;

    virtual unsigned HashPos (unsigned hashSize) ;
    virtual bool MatchKey (gKey& hKey) ;
    virtual bool MatchStorage (gStorage* aObj) ;

    // Set methods
    virtual void Reset () ;
    void Set (int v) ;
    void Set (t_uint64 v) ;
    void Set (double v) ;
    void Set (gString& s) ;
    void Copy (gKey& copy) ;

    // Specific methods
    virtual unsigned HashStrPos (unsigned hashSize, gString& s) ;

    // Save/etc methods
    virtual gStorage* NewObject () ;
    virtual t_uchar* ToString (t_uchar* uBuf) ;

    // Show methods
    virtual void Show (bool doShowAll=true) ;

protected:
    eKeyKind keyKind;

    // Possible key values: int, t_uint64, double, ...
    t_uint64* pvalUInt;
    double* pvalDouble;
    gString pvalStr;

    void thisNewKind (eKeyKind newKind) ;

private:
    // Operators,empty
    gKey (gKey& ) ; //empty
    gKey& operator= (gKey& ) ; //empty
};
////////////////////////////////////////////////////////////
class gHashElemGeneric : public gControl {
public:
    virtual ~gHashElemGeneric () ;

    // Public members
    gKey hKey;
    gStorage* myObj;

protected:
    gHashElemGeneric (gStorage* aObj) ;

private:
    // Operators,empty
    gHashElemGeneric (gHashElemGeneric& ) ; //empty
    gHashElemGeneric& operator= (gHashElemGeneric& ) ; //empty
};

class gHashElemTriple : public gHashElemGeneric {
public:
    gHashElemTriple (gKey& aKey, gString& s, gStorage* aObj) ;
    gHashElemTriple (gKey& aKey, int aVal, gStorage* aObj) ;
    virtual ~gHashElemTriple () ;

    // Public data-members
    gString sDesc;
    int iVal;

    // Show methods
    virtual void Show (bool doShowAll=true) ;

private:
    // Operators,empty
    gHashElemTriple (gHashElemTriple& ) ; //empty
    gHashElemTriple& operator= (gHashElemTriple& ) ; //empty
};
////////////////////////////////////////////////////////////
class gHashGeneric : public gStorage {
public:
    virtual ~gHashGeneric () ;

    // Get methods
    virtual unsigned N () {
	return size;
    }
    virtual bool IsEmpty () {
	return usageCounter==0;
    }
    virtual bool IsValidIndex (unsigned idx) ;
    gList* GetHash (unsigned idx) ;
    virtual bool FindKey (gKey& hKey) = 0; //PURE

    // Set methods
    virtual unsigned Delete () ;

    // Save/etc methods
    virtual gStorage* NewObject () ;
    virtual t_uchar* ToString (t_uchar* uBuf) ;
    virtual bool SaveGuts (FILE* f) ;
    virtual bool RestoreGuts (FILE* f) ;

protected:
    gHashGeneric (eStorage aKind, unsigned hashSize) ;

    unsigned size;
    gList* pLst;

    void thisPreAllocate (unsigned toSize) ;
    unsigned thisDelete () ;
    unsigned thisInsertElement () {
	usageCounter++;
	return usageCounter;
    }

private:
    unsigned usageCounter;

    // Operators,empty
    gHashGeneric (gHashGeneric& ) ;
    gHashGeneric& operator= (gHashGeneric& ) ;
};
////////////////////////////////////////////////////////////
class gHash : public gHashGeneric {
public:
    gHash (unsigned hashSize=LST_HASH_SIZE) ;
    virtual ~gHash () ;

    // Public data-members
    bool doFindBeforeInsert;

    // Get methods
    virtual bool FindKey (gKey& hKey) ;
    // .

    // Set methods
    unsigned Add (int v) ;
    unsigned Add (gInt& v) ;
    unsigned Add (gInt& v, gString& s) ;
    unsigned Add (gKey& hKey, gString& s) ;
    // .

    // Save/etc methods
    virtual t_uchar* ToString (t_uchar* uBuf) ;
    virtual bool SaveGuts (FILE* f) ;
    virtual bool RestoreGuts (FILE* f) ;

    // Show methods
    virtual void Show (bool doShowAll=true) ;

    // Debugging methods
    bool AddTriplePos_dbg (unsigned idx, gKey& hKey, char* str) ;

protected:
    int thisAddElem (unsigned idx, gHashElemGeneric* pElem) ;
    int thisFindKey (gKey& hKey, int& x) ;
    int thisKeyMatch (gKey& hKey, gStorage* hashElem) ;

private:
    // Operators,empty
    gHash (gHash& ) ;
    gHash& operator= (gHash& ) ;
};
////////////////////////////////////////////////////////////
class gHashTriple : public gHash {
public:
    gHashTriple (unsigned hashSize=LST_HASH_SIZE) ;
    virtual ~gHashTriple () ;

    // Get methods
    virtual bool FindKey (gKey& hKey) ;

    gHashElemTriple* Find (gKey& hKey) {
	unsigned idx;
	int x;
	return Find( hKey, idx, x );
    }
    gHashElemTriple* Find (gKey& hKey, unsigned& idx) {
	int x;
	return Find( hKey, idx, x );
    }
    gHashElemTriple* Find (gKey& hKey, unsigned& idx, int& x) ;

    virtual gHashElemTriple* GetTriple (unsigned idx, int x) ;
    // .

    // Set methods
    bool AddTriple (gKey& hKey, char* str, gStorage* aObj=nil) ;
    bool AddTriple (gKey& hKey, gString& s, gStorage* aObj=nil) ;
    bool AddTriple (gKey& hKey, int iVal, gStorage* aObj=nil) ;
    // .

private:
    // Operators,empty
    gHashTriple (gHashTriple& ) ; //empty
    gHashTriple& operator= (gHashTriple& ) ; //empty
};
////////////////////////////////////////////////////////////
#endif //gHASH_X_H

