#ifndef gHTTP_X_H
#define gHTTP_X_H

#include "gnet.h"
#include "gfile.h"
#include "gtime.h"  // For stats

#define MAX_HTTP_CHUNK_SIZE ((t_uint32)MAX_UINT16_U)
#define MAX_HTTP_HUGE_CHUNK_SIZE ((t_uint32)MAX_UINT16_U*2UL)

////////////////////////////////////////////////////////////
class gVersion : public gString {
public:
    gVersion (short major=-1, short minor=-1) ;
    virtual ~gVersion () {
    }

    // Public data-members
    short vMajor;
    short vMinor;

    // Set methods
    virtual bool SetVersion (gString& s) ;

    // Operators,valid
    bool operator> (gVersion& version) ;
    bool operator< (gVersion& version) ;

protected:
    int thisSetMajorOrMinor (char* s, short& ver) ;

private:
    // Operators,etc
    gVersion (gVersion& ) ; //empty
    gVersion& operator= (gVersion& ) ; //empty
};
////////////////////////////////////////////////////////////
// CLASS gBigBuffer
////////////////////////////////////////////////////////////
class gBigBuffer : public gControl {
public:
    gBigBuffer (t_uint32 bufSize=0) ;
    virtual ~gBigBuffer () ;

    // Public enums
    enum eBufferKind {
	e_NormalRAM,
	e_NodesRAM,
	e_InFile
    };

    // Public data-members
    gUCharBuffer* baseBuf;
    bool isDynamic;
    t_uint64 usedSize;

    // Get methods
    eBufferKind GetBufferKind () ;
    unsigned GetBaseSize () {
	return baseSizeBuf;
    }
    virtual unsigned GetNumberNodes () {
	return nNodes;
    }

    virtual gUCharBuffer* GetNode (unsigned idxNode) ;
    virtual gFileTemp& GetFile () ;

    // Special methods
    virtual bool NewFile () ;
    virtual bool WriteBuf (int fHandle, gUCharBuffer& sBuf, unsigned nBytes) ;
    virtual bool Flush (unsigned nBytes) ;

protected:
    unsigned nNodes, maxNodes;
    gUCharBuffer** bNode;  //ranges: [0..maxNodes-1]
    gFileTemp* fTemp;

    bool thisAllocateBuf (t_uint32 bufSize) ;

private:
    bool hasOverbuffer;
    unsigned baseSizeBuf;

    int thisOverbuffer (unsigned nBytes) ;
    int thisWrite (int fHandle, gUCharBuffer& sBuf, unsigned nBytes) ;

    // Operators,empty
    gBigBuffer (gBigBuffer& ) ; //empty
    gBigBuffer& operator= (gBigBuffer& ) ; //empty
};

////////////////////////////////////////////////////////////
// CLASS gURI
////////////////////////////////////////////////////////////
class gURI : public gString {
public:
    // Public enums
    enum eScheme {
	// Schemes as signed by IANA: http://www.iana.org/assignments/uri-schemes
	e_ftp,     // File Transfer Protocol
	e_http,    // Hypertext Transfer Protocol
	e_gopher,  // The Gopher Protocol
	e_mailto,  // Electronical mail address
	e_news,    // USENET news
	e_nntp,    // USENET news using NNTP
	e_telnet,  // Reference to interactive sessions
	e_wais,    // Wide Area Information
	e_file,    // Host-specific file names
	e_prospero,// Prospero Directory Service
	e_z39_50s, // z39.50 Session
	e_z39_50r, // z39.50 Retrieval
	e_cid,     // content identifier
	e_mid,     // message identifier
	e_vemmi,   // versatile multimedia interface
	e_service, // service location
	e_imap,    // internet message access protocol
	e_nfs,     // network file system protocol
	e_acap,    // application configuration access protocol
	e_rtsp,    // real time streaming protocol
	e_tip,     // Transaction Internet Protocol
	e_pop,     // Post Office Protocol
	e_data,    // data
	e_dav,     // dav
	e_opaquelocktoken,// opaquelocktoken             [RFC2518]
	e_sip,           // session initiation protocol  [RFC3261]
	e_sips,          // secure session intitiaion p. [RFC3261]
	e_tel,           // telephone                    [RFC2806]
	e_fax,           // fax                          [RFC2806]
	e_modem,         // modem                        [RFC2806]
	e_ldap,          // Lightweight Directory Access Protocol
	e_https,         // Hypertext Transfer Protocol Secure [RFC2818]

	e_soap_beep,     // soap.beep
	e_soap_beeps,    // soap.beeps

	xmlrpc_beep,     // xmlrpc.beep
	xmlrpc_beeps,    // xmlrpc.beeps

	e_urn,           // Uniform Resource Names       [RFC2141]
                         // (http://www.iana.org/assignments/urn-namespaces)
	e_go,            // go                           [RFC3368]
	e_h323,          // H.323                        [RFC3508]
	e_ipp,           // Internet Printing Protocol   [RFC3510]
	e_tftp,          // Trivial File Transfer p.     [RFC3617]
	e_mupdate,       // Mailbox Update (MUPDATE) Protocol
	e_pres,          // Presence                     [RFC-ietf-impp-pres-04.txt]
	e_im,            // Instant Messaging            [RFC-ietf-impp-im-04.txt]
	e_mtqp,          // Message Tracking Query p.    [RFC-ietf-msgtrk-mtqp-12.txt]

	e_RSVD_reserved, // MARK START OF RESERVED URIs

	e_afs,           // Andrew File System global file names
	e_tn3270,        // Interactive 3270 emulation sessions
	e_mailserver,    // Access to data available from mail servers

	e_invalid
    };

    enum ePathType {
	e_NoType,
	e_Absolute,
	e_Relative,
	e_RelativeQualified  // single-slash
    };

    // Structs
    struct sSchemeUse {
	eScheme theScheme;
	char* name;
	short rfcCode;
    };

    // Public data-members
    int lastOpError;

    gURI () ;
    gURI (eScheme aScheme) ;
    virtual ~gURI () ;

    // Get methods
    virtual bool IsOk () ;

    virtual bool IsOkScheme () {
	return isOkScheme;
    }

    virtual eScheme GetScheme () {
	return scheme;
    }

    virtual char* Str () ;

    ePathType GetPathType () {
	return pathType;
    }
    virtual char* GetOriginalStr () {
	return sOriginal.Str();
    }
    virtual char* GetDomainStr () {
	return sDomain.Str();
    }
    virtual char* GetPathStr () {
	return sPath.Str();
    }

    virtual bool IsValidDomain () ;

    // Set methods
    virtual bool SetString (char* s) ;

protected:
    ePathType pathType;
    gString sOriginal;
    gString sDomain;
    gString sPath;
    static sSchemeUse schemeTbl[];

    int thisParseString (char* s, eScheme aScheme) ;
    int thisSetDomainPath (char* s, bool hasSpecScheme, eScheme aScheme) ;
    int thisSetPath (char* s) ;
    gURI::eScheme thisGetSchemeFromString (char* s, int& error) ;
    int thisFetchScheme (char* s, eScheme& foundScheme) ;

private:
    eScheme scheme;
    bool isOkScheme;

    // Operators,empty
    gURI (gURI& ) ; //empty
    gURI& operator= (gURI& ) ; //empty
};
////////////////////////////////////////////////////////////
class gHttpGeneric : public gControl {
public:
    virtual ~gHttpGeneric () ;

    // Public enums
    enum eContentImport {
	e_Binary,
	e_TextPlain
    };

    enum eContentMethod {
	e_Get,
	e_Post
    };

    // Get methods
    virtual gBigBuffer& GetBuffer () ;
    virtual gUCharBuffer& GetBaseBuffer () ;

    virtual bool GetContent (eContentMethod method,
			     gString& sHost,
			     gString& sPath,
			     gString& sReply) = 0;  //PURE

protected:
    gHttpGeneric (gTcpConnect& connection) ;

    gTcpConnect* pConnection;

private:
    gBigBuffer biffBuffer;
};
////////////////////////////////////////////////////////////
#endif //gHTTP_X_H

