// glist.cpp -- Version 0.2

#include <string.h>
#include "glist.h"
#include "gstring.h"
////////////////////////////////////////////////////////////
// Static members
// ---------------------------------------------------------
// void!

////////////////////////////////////////////////////////////
// gElem - Generic elements of lists
// ---------------------------------------------------------
gElem::gElem ()
    : gStorage( e_ResvdStore ),
      prev( nil ),
      next( nil ),
      me( nil )
{
}

gElem::gElem (gStorage* newObj)
    : gStorage( e_ResvdStore ),
      prev( nil ),
      next( nil ),
      me( newObj )
{
 ASSERTION(me!=nil,"me!=nil");
}

gElem::~gElem ()
{
 DBGPRINT("%s",me==nil?"DBG: BOGUS: ~:gElem->me...\n":"\0");
 //if ( me->Kind()==e_String ) {
 //    printf("ELEM(%u|%s):[%s]\n",me->Kind(),me->storeKindDesc[me->Kind()].descStr,((gString*)me)->Str());
 //}

 delete me;
 me = nil;
}

gStorage* gElem::NewObject ()
{
 ASSERTION_FALSE("gElem::NewObject: Not used");
 return nil;
}

t_uchar* gElem::ToString (t_uchar* uBuf)
{
 ASSERTION(me!=nil && uBuf!=nil,"me!=nil");
 return me->ToString( uBuf );
}

char* gElem::Str ()
{
 ASSERTION(me!=nil,"me!=nil");
 return me->Str();
}

bool gElem::SaveGuts (FILE* f)
{
 ASSERTION(me!=nil,"gElem::SaveGuts");
 return me->SaveGuts( f );
}

bool gElem::RestoreGuts (FILE* f)
{
 ASSERTION(me!=nil,"gElem::SaveGuts");
 return me->SaveGuts( f );
}
////////////////////////////////////////////////////////////
// gListGeneric - Generic list handling
// ---------------------------------------------------------
gListGeneric::gListGeneric (eStorage aKind,
			    eListStrategy aStgy,
			    eListIndexes aLstIdxs)
    : gStorage( aKind, e_StgDefault ),
      size( 0 ),
      preMaxSize( 0/*aStgy==e_LstPointer ? 0 : LST_LIST_SIZE*/ ),
      maxSize( 0 ),
      lstIdxs( aLstIdxs ),
      stgy( aStgy ),
      pStart( nil ),
      pEnd( nil ),
      pCurrent( nil ),
      doIgnoreCase( false )
{
 thisPreAllocate( preMaxSize );
 ASSERTION(lstIdxs==e_LstIdx1toN || lstIdxs==e_LstIdx1toNstrict,"gListGeneric:...");
}

gListGeneric::~gListGeneric ()
{
 thisDelete();
}

bool gListGeneric::IsValidIndex (unsigned idx)
{
 return thisIndex(idx);
}

char* gListGeneric::Str (unsigned idx)
{
 thisIndex(idx);
 genUCharBuf.Clear();
 if ( pCurrent==nil ) return (char*)genUCharBuf.uBuf;
 return pCurrent->Str();
}

t_uchar* gListGeneric::UStr (unsigned idx)
{
 /*
 if ( me->Kind()==e_String ) {
     char* dbgStr = me->Str();
     if ( dbgStr!=nil && strlen(dbgStr)>=genUCharBuf.size ) {
	 printf("DBG: ToString([%s]), size-of-genUChar=%u, wanting %u chars!!!\n",dbgStr,genUCharBuf.size,strlen(dbgStr));
     }
 }
 return pCurrent->ToString( genUCharBuf.uBuf )  ==> Not used because of inexisting size check
 */
 return (t_uchar*)Str( idx );
}

gStorage::eStorage gListGeneric::ElementsKind ()
{
 eStorage thisKind=e_NoStorage, lastKind=e_NoStorage;
 unsigned idx=1, nDiffs=0;
 if ( thisIndex( idx )==false ) return e_List;
 ASSERTION(size>0,"size>0");
 // Here, list is not empty:
 // it can be all-of-the-same-kind, or variously filled.
 for (pCurrent=pStart; pCurrent!=nil; pCurrent=pCurrent->next) {
     thisKind = pCurrent->me->Kind();
     if ( thisKind!=lastKind && lastKind!=e_NoStorage ) nDiffs++;
     lastKind = thisKind;
 }
 if ( nDiffs>0 ) return e_List;
 return thisKind;
}

gElem& gListGeneric::GetElement (unsigned idx)
{
 static gElem tempElem;
 if ( thisIndex( idx ) ) return *pCurrent;
 return tempElem;
}

gElem* gListGeneric::GetElementPtr (unsigned idx)
{
 if ( thisIndex( idx ) ) return pCurrent;
 return nil;
}

gStorage* gListGeneric::GetObjectPtr (unsigned idx)
{
 gElem* pElem = GetElementPtr( idx );
 ASSERTION(pElem!=nil,"GetObjectPtr(1)");
 ASSERTION(pElem->me!=nil,"GetObjectPtr(2)");
 return pElem->me;
}

gStorage* gListGeneric::GetFirstObjectPtr ()
{
 if ( pStart==nil ) return nil;
 return pStart->me;
}

gStorage* gListGeneric::GetLastObjectPtr ()
{
 if ( pEnd==nil ) return nil;
 return pEnd->me;
}

unsigned gListGeneric::Delete (unsigned startPos, unsigned endPos)
{
 unsigned
     counter=0,
     oldSize=N(),
     pIter,
     p0=startPos==0?1:startPos,
     p1=endPos==0?oldSize:gMIN(endPos,oldSize);
 gElem* pTemp;
 gElem* pNext;

 if ( startPos==0 && endPos==0 ) {
     thisDelete();
     return oldSize;
 }

 if ( p0>p1 ) return 0;

 // Delete only some positions

 // Place current pointer on first element to delete.
 thisIndex( p0 );
 ASSERTION(p0>=1 && p1<=oldSize,"Delete(1)");

 for (pIter=p0, pTemp=GetElementPtr(p0); pIter<=p1; pIter++) {
     ASSERTION(size>0,"Delete(2)");
     size--;
     counter++;
     pCurrent = pTemp;
     ASSERTION(pCurrent!=nil,"pCurrent!=nil");
     pNext = pTemp->next;
     if ( pTemp->prev==nil ) {
	 // Start of list
	 pStart = pNext;
	 pTemp = pNext;
	 if ( pTemp!=nil ) pTemp->prev = nil;
     }
     else {
	 if ( pNext==nil ) {
	     // At end of list
	     pTemp = pTemp->prev;
	     pTemp->next = nil;
	     pEnd = pTemp;
	 }
	 else {
	     // At middle of list
	     // (pTemp->prev is not nil!)
	     pTemp = pTemp->prev;
	     pTemp->next = pNext;
	     pNext->prev = pTemp;
	     pTemp = pNext;  // Placed iterator in next position
	 }
     }
     ASSERTION(pCurrent!=nil,"pCurrent!=nil");
     delete pCurrent;
     if ( pTemp==nil ) {
	 pStart = pEnd = nil;
     }
     else {
	 if ( pTemp->next==nil ) pEnd = pTemp;
	 if ( pTemp->prev==nil ) pStart = pTemp;
     }
 }//end FOR
 pCurrent = nil;
 return counter;
}

gStorage* gListGeneric::NewObject ()
{
 ASSERTION_FALSE("gListGeneric::NewObject: Not used");
 return nil;
}

t_uchar* gListGeneric::ToString (t_uchar* uBuf)
{
 ;
 // Output notation:
 //     Empty list               ===>    ()
 //     List of strings          ===>    (me and you)
 //     List of strings and nums ===>    (I am '31 years old)
 //     List of sentences        ===>    ("I use a suit." "You use a dress.")
 //     List with some resvd chrs.==>    (123-a.com sold '100 domains)
 ;
 return nil;
}

bool gListGeneric::SaveGuts (FILE* f)
{
 unsigned i, n=N();
 gString sBlank( ' ' );

 if ( CanSave( f )==false ) return false;
 for (i=1; i<=n; i++) {
     if ( i>1 ) sBlank.SaveGuts( f );
     GetObjectPtr( i )->SaveGuts( f );
 }
 return true;
}

bool gListGeneric::RestoreGuts (FILE* f)
{
 return CanRestore( f );
}

void gListGeneric::thisPreAllocate (unsigned toSize)
{
 if ( stgy==e_LstDefault ) stgy = e_LstPointer;
 ASSERTION(stgy==e_LstPointer,"stgy==e_LstPointer: TODO other");

 pCurrent = nil;
 //if ( preMaxSize==0 ) return;
}

unsigned gListGeneric::thisDelete ()
{
 unsigned countCheck=0;
 gElem* pNext;

 for (pCurrent=pStart; pCurrent!=nil; )
 {
     pNext = pCurrent->next;
     delete pCurrent;
     pCurrent = pNext;
     countCheck++;
 }
 ASSERTION(countCheck==size,"countCheck==size");
 size = 0;
 pStart = pEnd = pCurrent = nil;
 return countCheck;
}

bool gListGeneric::thisIndex (unsigned& idx)
{
 bool isOk = idx>=1 && idx<=size;
 unsigned i;

 pCurrent = nil;
 if ( lstIdxs==e_LstIdx1toNstrict ) {
     ASSERTION_USE(isOk,"gListGeneric::thisIndex");
     return true;
 }
 if ( idx<1 ) idx = 1;
 isOk = idx<=size;
 if ( idx>size ) idx = size;
 if ( isOk==false ) return false;

 for (i=1, pCurrent=pStart; i<idx; i++) {
     pCurrent = pCurrent->next;
     ASSERTION(pCurrent!=nil,"pCurrent!=nil");
 }
 ASSERTION(pCurrent!=nil,"pCurrent!=nil");
 ASSERTION(pCurrent->me!=nil,"pCurrent->me!=nil");
 return true;
}

bool gListGeneric::thisAppend (gStorage* newObj)
{
 // Returns true iff new re-allocation performed.
 // Since no vector functions are present (yet),
 // returns always false.
 ;
 gElem* newElem;

 ASSERTION(newObj!=nil,"newObj!=nil");
 newElem = new gElem( newObj );
 ASSERTION(newElem!=nil,"newElem!=nil");

 if ( size==0 ) {
     ASSERTION(pStart==nil && pEnd==nil,"::thisAppend(1)");
     pStart = newElem;
 }
 else {
     newElem->prev = pEnd;
     pEnd->next = newElem;
 }

 pCurrent = pEnd = newElem;
 size++;

 return false;
}
////////////////////////////////////////////////////////////
gList::gList ()
    : gListGeneric(
	e_List,
	e_LstPointer,
	e_LstIdx1toN)
{
}

gList::~gList ()
{
}

int gList::GetInt (unsigned idx)
{
 if ( thisIndex(idx)==false ) return 0;
 if ( pCurrent->me->Kind()!=e_SInt ) return 0;
 return ((gInt*)(pCurrent->me))->GetInt();
}

unsigned gList::GetUInt (unsigned idx)
{
 if ( thisIndex(idx)==false ) return 0;
 if ( pCurrent->me->Kind()!=e_UInt ) return 0;
 return ((gUInt*)(pCurrent->me))->GetUInt();
}

unsigned gList::Match (char* s)
{
 unsigned i, n=N();
 char* thisStr;
 for (i=1; i<=n; i++) {
     thisStr = Str( i );
     if ( thisStr==nil ) continue;
     gString sTemp( thisStr );
     if ( sTemp.Match( s, doIgnoreCase ) ) return i;
 }
 return 0;
}

unsigned gList::FindFirst (char* s,
			   unsigned strPos,
			   eFindCriteria findCriteria)
{
 ;
 // Returns the index of first occurrence obeying 'findCriteria'
 // for the string 's' in list (on position strPos).
 ;
 gList posL;
 return thisFind( s, strPos, findCriteria, true, posL );
}

unsigned gList::Find (char* s,
		      unsigned strPos,
		      eFindCriteria findCriteria)
{
 gList posL;
 return thisFind( s, strPos, findCriteria, false, posL );
}

unsigned gList::FindAny (char* s,
			 unsigned strPos,
			 eFindCriteria findCriteria,
			 gList& posL)
{
 return thisFind( s, strPos, findCriteria, false, posL );
}

unsigned gList::Add (int v)
{
 gInt* newObj;
 newObj = new gInt(v);
 thisAppend( newObj );
 return size;
}

unsigned gList::Add (unsigned v)
{
 gUInt* newObj;
 newObj = new gUInt(v);
 thisAppend( newObj );
 return size;
}

unsigned gList::Add (gString& copy)
{
 return Add( copy.Str() );
}

unsigned gList::Add (t_uchar* s)
{
 gString* newObj;
 newObj = new gString( s );
 thisAppend( newObj );
 return size;
}

unsigned gList::Add (char* s)
{
 return Add( (t_uchar*)s );
}

gList& gList::CopyList (gList& aL)
{
 bool isOk;
 gStorage* aObj;
 gStorage* newObj;
 gElem* thisElem;
 unsigned i, n=aL.N();

 thisDelete();
 thisElem = aL.GetElementPtr(1);
 for (i=1; i<=n; i++) {
     ASSERTION(thisElem!=nil,"thisElem!=nil");
     aObj = thisElem->me;
     ASSERTION(aObj!=nil,"aObj!=nil");
     newObj = aObj->NewObject();
     isOk = AppendObject( newObj );
     ASSERTION(isOk,"isOk");
     thisElem = thisElem->next;
 }
 return *this;
}

unsigned gList::DeleteString (gString& s)
{
 ;
 // Returns 0 if no string found, 1 otherwise
 // Delete the first string found.
 ;
 unsigned pos = Match( s.Str() );
 if ( pos==0 ) return 0;
 Delete( pos, pos );
 return 1;
}

unsigned gList::DeleteStrings (gString& s)
{
 ;
 // Returns number of elements deleted.
 // Delete all strings found
 ;
 unsigned count=0, pos;
 while ( (pos = Match( s.Str() ))>0 ) {
     Delete( pos, pos );
     count++;
 }
 return count;
}

t_uchar* gList::ToString (t_uchar* uBuf)
{
 return gListGeneric::ToString( uBuf );
}

bool gList::SaveGuts (FILE* f)
{
 return gListGeneric::SaveGuts( f );
}

bool gList::RestoreGuts (FILE* f)
{
 return gListGeneric::RestoreGuts( f );
}

void gList::Show (bool doShowAll)
{
 unsigned i, n=N();

 if ( doShowAll ) printf("(");
 for (i=1; i<=n; i++) {
     gStorage* aObj;
     aObj = GetElementPtr(i)->me;
     ASSERTION(aObj!=nil,"aObj!=nil");
     if ( doShowAll ) {
	 aObj->Show( true );
	 printf("%s",i==n?"\0":" ");
     }
     else {
	 printf("%s%s",Str(i),i==n?"\0":" ");
     }
 }
 if ( doShowAll ) printf(")\n");
}

unsigned gList::thisFind (char* s,
			  unsigned strPos,
			  eFindCriteria findCriteria,
			  bool doStopOnFirst,
			  gList& posL)
{
 bool isOk;
 unsigned i, n=N(), pos, posFirst=0;

 for (i=1; i<=n; i++) {
     gString sTemp( Str( i ) );
     pos = sTemp.Find( s, doIgnoreCase );
     if ( pos==0 ) continue;
     switch ( findCriteria ) {
     case e_FindExactPosition:
	 isOk = pos==strPos;
	 break;
     case e_FindFromPosition:
	 isOk = pos>=strPos;
	 break;
     case e_FindBeforePosition:
	 isOk = pos<strPos;
	 break;
     default:
	 return 0;
     }
     if ( isOk==false ) continue;
     if ( posFirst==0 ) posFirst = i;
     posL.Add( i );
     if ( doStopOnFirst ) return posFirst;
 }
 return posFirst;
}
////////////////////////////////////////////////////////////
gListInt::gListInt ()
{
}

gListInt::~gListInt ()
{
}

unsigned gListInt::FindInt (int v)
{
 unsigned i, n = N();
 for (i=1; i<=n; i++) {
     if ( v==GetInt( i ) ) return i;
 }
 return 0;
}

bool gListInt::Append (int v)
{
 gInt* newObj = new gInt( v );
 return AppendObject( newObj );
}
////////////////////////////////////////////////////////////

