#ifndef gLIST_X_H
#define gLIST_X_H

#include "gstorage.h"
#include "gstring.h"
////////////////////////////////////////////////////////////
// Common enums
// ---------------------------------------------------------
enum eFindCriteria {
    e_FindExactPosition,
    e_FindFromPosition,
    e_FindBeforePosition
};

////////////////////////////////////////////////////////////
// Arguments and their corresponding support classes
// ---------------------------------------------------------

////////////////////////////////////////////////////////////
class gElem : public gStorage {
public:
    gElem () ;
    gElem (gStorage* newObj) ;
    virtual ~gElem () ;

    // Public data-members
    gElem* prev;
    gElem* next;
    gStorage* me;

    // Get methods
    virtual int MaximumStorageSize () {
	return 1;
    }

    virtual t_uchar* ToString (t_uchar* uBuf) ;
    virtual char* Str () ;

    // Save/etc methods
    virtual gStorage* NewObject () ;
    virtual bool SaveGuts (FILE* f) ;
    virtual bool RestoreGuts (FILE* f) ;

private:
    //Operators,empty
    gElem (gElem& ) ; //empty
    gElem& operator= (gElem& ) ; //empty
};
////////////////////////////////////////////////////////////
class gListGeneric : public gStorage {
public:
    // Public enums
    enum eListStrategy {
	e_LstDefault,
	e_LstPointer,
	e_LstHash
    };
    enum eListIndexes {
	e_LstIdx1toN,
	e_LstIdx1toNstrict,
	e_LstIdx0Unused
    };

    virtual ~gListGeneric () ;

    // Get methods
    virtual bool IsEmpty () {
	return size==0;
    }
    virtual unsigned N () {
	return size;
    }
    virtual bool IsValidIndex (unsigned idx) ;
    virtual char* Str (unsigned idx) ;
    virtual t_uchar* UStr (unsigned idx) ;

    virtual gElem& GetElement (unsigned idx) ;
    virtual gElem* GetElementPtr (unsigned idx) ;

    virtual gStorage* GetObjectPtr (unsigned idx) ;
    virtual gStorage* GetFirstObjectPtr () ;
    virtual gStorage* GetLastObjectPtr () ;

    virtual eStorage ElementsKind () ;

    virtual int MinimumStorageSize () {
	return -1;
    }
    virtual int MaximumStorageSize () {
	return -1;
    }

    // Set methods
    virtual void Reset () {
	Delete();
    }

    virtual unsigned Delete (unsigned startPos=0, unsigned endPos=0) ;

    virtual bool SetCaseSense (bool doCaseSense) {
	doIgnoreCase = doCaseSense==false;
	return doCaseSense;
    }
    virtual bool AppendObject (gStorage* newObj) {
	if ( newObj==nil ) return false;
	thisAppend( newObj );
	return true;
    }

    // Specific methods
    virtual gElem* StartPtr () {
	return pStart;
    }
    virtual gElem* EndPtr () {
	return pEnd;
    }

    // Save/etc methods
    virtual gStorage* NewObject () ;
    virtual t_uchar* ToString (t_uchar* uBuf) ;
    virtual bool SaveGuts (FILE* f) ;
    virtual bool RestoreGuts (FILE* f) ;

protected:
    gListGeneric (eStorage aKind,
		  eListStrategy aStgy,
		  eListIndexes aLstIdxs) ;

    unsigned size, preMaxSize, maxSize;
    eListIndexes lstIdxs;
    eListStrategy stgy;
    gUCharBuffer genUCharBuf;
    gElem* pStart;
    gElem* pEnd;
    gElem* pCurrent;
    bool doIgnoreCase;

    void thisPreAllocate (unsigned toSize) ;
    unsigned thisDelete () ;
    bool thisIndex (unsigned& idx) ;
    bool thisAppend (gStorage* newObj) ;

private:
    // Operators,empty
    gListGeneric (gListGeneric& ) ;
    gListGeneric& operator= (gListGeneric& ) ;
};
////////////////////////////////////////////////////////////
class gList : public gListGeneric {
public:
    gList () ;
    virtual ~gList () ;

    // Get methods
    virtual int GetInt (unsigned idx) ;
    virtual unsigned GetUInt (unsigned idx) ;

    // Find methods
    virtual unsigned Match (char* s) ;
    virtual unsigned FindFirst (char* s,
				unsigned strPos,
				eFindCriteria findCriteria) ;
    virtual unsigned Find (char* s,
			   unsigned strPos,
			   eFindCriteria findCriteria) ;

    virtual unsigned FindAny (char* s,
			      unsigned strPos,
			      eFindCriteria findCriteria,
			      gList& posL) ;

    // Set methods
    unsigned Add (int v) ;
    unsigned Add (unsigned v) ;
    unsigned Add (gString& copy) ;
    unsigned Add (t_uchar* s) ;
    unsigned Add (char* s) ;

    virtual gList& CopyList (gList& aL) ;

    virtual unsigned DeleteString (gString& s) ;
    virtual unsigned DeleteStrings (gString& s) ;

    // Save/etc methods
    virtual t_uchar* ToString (t_uchar* uBuf) ;
    virtual bool SaveGuts (FILE* f) ;
    virtual bool RestoreGuts (FILE* f) ;

    // Show methods
    virtual void Show (bool doShowAll=true) ;

protected:
    unsigned thisFind (char* s,
		       unsigned strPos,
		       eFindCriteria findCriteria,
		       bool doStopOnFirst,
		       gList& posL) ;

private:
    // Operators,empty
    gList (gList& ) ;
    gList& operator= (gList& ) ;
};
////////////////////////////////////////////////////////////
class gListInt : public gList {
public:
    gListInt () ;
    virtual ~gListInt () ;

    // Get methods
    virtual unsigned FindInt (int v) ;

    // Set methods
    bool Append (gInt& v) {
	return Append( v.GetInt() );
    }
    bool Append (int v) ;

    // Valid operators
    int operator[] (unsigned idx) {
	return GetInt( idx );
    }

private:
    // Operators,empty
    gListInt (gListInt& ) ; //empty
    gListInt& operator= (gListInt& ) ; //empty
};
////////////////////////////////////////////////////////////
#endif //gLIST_X_H

