#ifndef gMAIN_X_H
#define gMAIN_X_H

#define LIB_VERSION_GLIB_MAJOR	1
#define LIB_VERSION_GLIB_MINOR	0

#include <stdlib.h>
#include "gstdtypes.h"

#define GMIN_STAT_OBJ 5


#define false 0
#define true  1

typedef unsigned char t_uchar;
typedef t_uchar t_uint8;
typedef char t_int8;
typedef unsigned short t_uint16;
typedef short t_int16;

typedef unsigned long long t_uint64;

typedef t_uint16 t_gPort;

typedef t_uint64 t_inode;


#ifdef gUINT_IS_32bit
// 32bit word microprocessor specific
typedef unsigned t_uint32;
typedef int t_int32;
#define MAX_UINT_UL 4294967295UL
#else
#error TODO: doing 16bit OSs
#endif //gUINT_IS_32bit

typedef t_uint32 t_uint4B;
typedef t_uint16 t_uint2B;

typedef t_uint32 t_gTicElapsed;
typedef t_uint32 t_stamp;  // traditional Unix epoch-time (1970...)


#ifdef SWAP_16bit16
// Not usual: ab|cd => cd|ab
#define gENDLITTLE4(x) ((x<<16) | (x>>16))
#else
// Usual on Intel based OSs: ab|cd => dc|ba
#define gENDLITTLE2(x) ((x<<8) | (x>>8))
#define gENDLITTLE4(x) ( (x<<24) | ((x&0xFF0000)>>8) | ((x&0xFF00)<<8) | ((x&0xFF000000)>>24) )
#endif //SWAP_16bit16

// OS-Specific (or Machine-Specific)
#define TO_UD2(x) ((t_uint16)gENDLITTLE2(x))
#define TO_UD4(x) ((t_uint32)gENDLITTLE4(x))
#define FROM_UD2(x) TO_UD2(x)   // (De-)Coding is symetric
#define FROM_UD4(x) TO_UD4(x)   // (De-)Coding is symetric

#define MAX_INT16_I 32767  // 2^15-1
#define MIN_INT16_I -32768
#define MAX_UINT16_U 65535U  // 2^16-1

#define MAX_DLINT32 2147483647L
#define MIN_DLINT32 -2147483647L
#define MAX_DUINT32 2147483646UL

#define MAX_LONG_L 2147483648LL
#define MIN_LONG_L -2147483648LL
#define MAX_U4B_ULL 4294967295ULL

#define MAX_STAMP_UL MAX_DUINT32

#ifdef gDOS_SPEC
#define gEND_CLEANUP_WSA WSACleanup()
#else
#define gEND_CLEANUP_WSA //Nothing needed for non-Win32/DOS
#endif //~gDOS_SPEC

#ifdef DEBUG
#define gEND_STATEMENT gFileControl::Self().End()
#else
#define gEND_STATEMENT gFileControl::Self().End( false )
#endif

// File macros
#ifdef gDOS_SPEC
#ifndef __S_IFIFO
#define __S_IFIFO       0010000 /* FIFO.  */
#endif //__S_IFIFO
#endif //gDOS_SPEC
#define gFILE_S_IFIFO __S_IFIFO
// <end of File macros>

// ---------------------------------------------------
// Singletons control
// ---------------------------------------------------
#define gEND gEND_STATEMENT;gEND_CLEANUP_WSA

#define gINIT gStrControl::Self().Run(); gFileControl::Self().Init()

// ---------------------------------------------------
// Utilities
// ---------------------------------------------------
#define gMAX(a,b) ((a>b)?a:b)
#define gMIN(a,b) ((a<b)?a:b)

#define gRANGEX(v,a,b,x) ((v<a)?x:(v>b?x:v))
#define gRANGE0(v,a,b) gRANGEX(v,a,b,0)
// e.g.: gRANGE0(v,a,b) ((v<a)?0:(v>b?0:v))

#define ISyORn(x) (x ? 'Y' : 'N')

#ifdef gDOS_SPEC
#define gSLASHSTR "\\"
#define gSLASHCHR '\\'
#define gDOS_LIB_CRT_DLL // e.g. getpid on libcrtdll.a
#undef gDOS_LIB_XIO // e.g. mkstemp on libiberty.a
#else
#define gSLASHSTR "/"
#define gSLASHCHR '/'
#ifdef linux
#define gGX_SIG_SUPPORT
#define gGX_USR_SUPPORT
	#ifdef FREESCO
		// ZipSlack does not have this no unistd.h
		typedef unsigned int __socklen_t;
		typedef __socklen_t socklen_t;
	#endif
#endif //linux (specific self-defines)
#endif //gDOS_SPEC


#ifdef gX_KERN2_0
// If not regularly use: netinet/in.h:typedef uint32_t in_addr_t;
typedef t_uint32 in_addr_t;
#endif //gX_KERN2_0

#ifdef gX_KERN2_4
;
#endif //gX_KERN2_4


// ---------------------------------------------------
// Debugging
// ---------------------------------------------------

#ifdef DEBUG
#define DBGPRINT(args...) printf(args)
#else
#define DBGPRINT(args...) ;
#endif //DEBUG

#ifdef DEBUG_MIN
#define DBGPRINT_MIN(args...) printf(args)
#else
#define DBGPRINT_MIN(args...) ;
#endif //DEBUG

#ifdef DEBUG_MEA  //Memory all...highly verbose
#define DBGPRINT_MEA(args...) printf(args)
#else
#define DBGPRINT_MEA(args...) ;
#endif //DEBUG

#ifdef DEBUG_LOG
#define DBGPRINT_LOG(args...) printf(args)
#else
#define DBGPRINT_LOG(args...) ;
#endif //DEBUG_LOG

// ---------------------------------------------------
// Assertions
// ---------------------------------------------------

#define ASSERTION_FALSE(y) \
            { \
               /* gStorageControl::Self().assertedNr = -1*/; \
               fprintf(stderr,"ASSERTION[v%d.%d]:%s:%u:%s\n", \
			LIB_VERSION_GLIB_MAJOR, LIB_VERSION_GLIB_MINOR, \
			__FILE__,__LINE__,y); \
               exit(1); \
            }
#define ASSERTION(x,y) \
            { \
            if ( (x)==0 ) { \
               /* gStorageControl::Self().assertedNr = -2*/; \
               fprintf(stderr,"ASSERTION[v%d.%d]:%s:%u:%s\n", \
			LIB_VERSION_GLIB_MAJOR, LIB_VERSION_GLIB_MINOR, \
			__FILE__,__LINE__,y); \
               exit(1); \
            } \
            }

#define ASSERTION_USE(x,y) ASSERTION(x,y)


#endif //gMAIN_X_H

