// gnet.cpp -- Version 0.3

#include <string.h>
#include <unistd.h>      //close...etc
#include "gmain.h"

#ifdef gDOS_SPEC
#include <winsock2.h>
#else
#include <sys/socket.h>  //socket... (sys/types.h)
#include <netdb.h>       //gethostbyname...etc
#include <arpa/inet.h>   //inet_addr...
#endif //gDOS_SPEC (~)

#include <errno.h>

#include "gnet.h"

// Static members
int gSocket::sockCount=0;

////////////////////////////////////////////////////////////
gIpAddr::gIpAddr (char* ipStr)
    : b1( 0 ),
      b2( 0 ),
      b3( 0 ),
      b4( 0 )
{
 DBGPRINT_MIN("DBG: gIpAddr '%s'\n",ipStr);
 if ( ipStr!=nil ) SetAddrFromStr( ipStr );
}

gIpAddr::gIpAddr (t_uint8 c1, t_uint8 c2, t_uint8 c3, t_uint8 c4)
    : b1( c1 ),
      b2( c2 ),
      b3( c3 ),
      b4( c4 )
{
}

gIpAddr::~gIpAddr ()
{
}

bool gIpAddr::IsOk ()
{
 return (b1 | b2 | b3 | b4)!=0;
}

t_uchar* gIpAddr::String ()
{
 gString sTemp( 30, ' ' );
 addrStr = sTemp;
 sprintf( addrStr.Str(), "%u.%u.%u.%u", b1, b2, b3, b4 );
 return (t_uchar*)addrStr.Str();
}

t_uchar* gIpAddr::ToString (t_uchar* uBuf)
{
 t_uchar* uStr = String();
 if ( uBuf==nil ) return uStr;
 strcpy( (char*)uBuf, (char*)uStr );
 return uBuf;
}

void gIpAddr::Reset ()
{
 gControl::Reset();
 b1 = b2 = b3 = b4 = 0;
}

void gIpAddr::SetAddr (t_gIpAddr nboAddr)
{
 b4 = nboAddr & 0xFF;
 nboAddr >>= 8;
 b3 = nboAddr & 0xFF;
 nboAddr >>= 8;
 b2 = nboAddr & 0xFF;
 nboAddr >>= 8;
 b1 = nboAddr & 0xFF;
 nboAddr >>= 8;
 ASSERTION(nboAddr==0,"nboAddr>32bit?");
}


bool gIpAddr::SetAddrFromStr (char* str)
{
 int a1=-1, a2=-1, a3=-1, a4=-1;
 ASSERTION(str!=nil,"str!=nil");
 Reset();
 lastOpError = sscanf(str,"%d.%d.%d.%d",&a1,&a2,&a3,&a4);
 lastOpError = -1*(lastOpError!=4);
 if ( lastOpError!=0 ) return false;
 lastOpError = a1<0 || a1>255;
 lastOpError += a2<0 || a2>255;
 lastOpError += a3<0 || a3>255;
 lastOpError += a4<0 || a4>255;
 lastOpError = -lastOpError;
 if ( lastOpError!=0 ) return false;
 b1 = (t_uint8)a1;
 b2 = (t_uint8)a2;
 b3 = (t_uint8)a3;
 b4 = (t_uint8)a4;
 return true;
}

t_gIpAddr gIpAddr::GetNetworkAddress ()
{
 unsigned long nBO = thisGetIP( b1, b2, b3, b4 );
 return (t_gIpAddr)nBO;
}

int gIpAddr::GetHostByName (char* hostname)
{
 struct hostent* hp=nil;
 ASSERTION(hostname!=nil,"hostname!=nil");
 Reset();
#ifdef gSUPPRESS_INET
 ASSERTION_FALSE("gSUPPRESS_INET: gethostbyname");
#else
 hp = gethostbyname( hostname );
#endif
 lastOpError = hp==nil;
 if ( lastOpError!=0 ) {
     SetError( errno );
     return -1;
 }
 char** hAddr;
 hAddr = &hp->h_addr;
 thisSetIPfromHostEnt( hAddr, hp->h_length );
 return 0;
}

int gIpAddr::GetHostByAddr (gString& sRes)
{
 gList lRes;
 return thisGetHostByAddr( (char*)String(), sRes, lRes );
}

gIpAddr& gIpAddr::operator= (gIpAddr& copy)
{
 b1 = copy.b1;
 b2 = copy.b2;
 b3 = copy.b3;
 b4 = copy.b4;
 return *this;
}

void gIpAddr::Show (bool doShowAll)
{
 printf("%s%s",String(),doShowAll?"\n":"\0");
}

unsigned long gIpAddr::thisGetIP (t_uint8 a1, t_uint8 a2, t_uint8 a3, t_uint8 a4)
{
 unsigned long result;
 result = a1;
 result <<= 8;
 result += a2;
 result <<= 8;
 result += a3;
 result <<= 8;
 result += a4;
 return result;
}

unsigned long gIpAddr::thisGetHostByteOrder (t_gIpAddr netByteOrder)
{
#ifdef gSUPPRESS_INET
 ASSERTION_FALSE("gSUPPRESS_INET: ntohl");
 return 0UL;
#else
 unsigned long result = ntohl( (unsigned long)netByteOrder );
 return result;
#endif
}

int gIpAddr::thisSetIPfromHostByteOrder (in_addr_t hostByteOrder)
{
#ifdef gSUPPRESS_INET
 ASSERTION_FALSE("gSUPPRESS_INET: htonl");
#else
 unsigned long nBO = htonl( hostByteOrder );
 // nBO is the network-byte-order
 SetAddr( (t_gIpAddr)nBO );
#endif
 return 0;
}

int gIpAddr::thisSetIPfromHostEnt (char* h_addr, int h_length)
{
 struct sockaddr_in hostAddressData;

 ASSERTION(h_addr!=nil,"h_addr!=nil");
 memset( (char*)&hostAddressData, 0, sizeof(hostAddressData) );
 memcpy( &hostAddressData.sin_addr, h_addr, h_length );
 // Acc. netinet/in.h;
 // typedef uint32_t in_addr_t;
 // struct in_addr {
 //     in_addr_t s_addr;
 // };
 in_addr_t hostByteOrder = hostAddressData.sin_addr.s_addr;
 return thisSetIPfromHostByteOrder( hostByteOrder );
}

int gIpAddr::thisGetHostByAddr (char* ipStr, gString& sRes, gList& lRes)
{
 struct hostent* hp=nil;
 char* inAddrTmp;
 char* str;
 char** pStr;
 struct in_addr hardInAddr;

 ASSERTION(ipStr!=nil,"ipStr!=nil");
 memset( &hardInAddr, 0x0, sizeof(hardInAddr) );
 inAddrTmp = (char*)&hardInAddr;

#ifdef gSUPPRESS_INET
 ASSERTION_FALSE("gSUPPRESS_INET: inet_addr");
#else
 hardInAddr.s_addr = inet_addr( ipStr );
 if ( hardInAddr.s_addr==INADDR_NONE ) {
     return SetError( 2 );
 }
 hp = gethostbyaddr( inAddrTmp, sizeof(hardInAddr), AF_INET );
 if ( hp==nil ) {
     return SetError( -1 );
 }
 sRes.Set( (char*)hp->h_name );
 // Aliases for host
 pStr = hp->h_aliases;
 while ( (str = *pStr)!=nil ) {
     pStr++;
     lRes.Add( str );
 }
#endif //gSUPPRESS_INET (~)
 return 0;
}
////////////////////////////////////////////////////////////
gIpV4Mask::gIpV4Mask (t_uint32 aMask)
    :  nBitsCIDR( 0 ), mask( aMask )
{
 thisSetNBitsFromMask( aMask );
}

gIpV4Mask::gIpV4Mask (t_uint8 bitsCIDR)
    : nBitsCIDR( bitsCIDR ),
      mask( 0xFFFFFFFF )
{
 SetBitsCIDR( bitsCIDR );
}

gIpV4Mask::gIpV4Mask (gIpV4Mask::eClass addrClass)
    : mask( 0 )
{
 switch ( addrClass ) {
 case e_ClassA:
     SetBitsCIDR( 8 );
     break;
 case e_ClassB:
     SetBitsCIDR( 16 );
     break;
 case e_ClassC:
     SetBitsCIDR( 24 );
     break;
 default:
     ASSERTION_FALSE("gIpV4Mask(addrClass)");
     break;
 }
}

gIpV4Mask::~gIpV4Mask ()
{
}

void gIpV4Mask::SetBitsCIDR (t_uint8 bitsCIDR)
{
 ASSERTION(bitsCIDR<=32,"bitsCIDR<=32");
 nBitsCIDR = bitsCIDR;
 mask = 0xFFFFFFFF;
 if ( bitsCIDR<32 ) {
     mask >>= bitsCIDR;
     mask ^= 0xFFFFFFFF;
 }
}

int gIpV4Mask::thisSetNBitsFromMask (t_uint32 aMask)
{
 t_uint16 shiftCount( 0 ), maxCount( 0 );
 t_uint32 uMask( aMask );

 nBitsCIDR = 0;
 mask = 0xFFFFFFFF;
 for ( ; shiftCount<32 && (uMask & 1)==0; shiftCount++) {
     uMask >>= 1;
 }

 if ( shiftCount>=32 ) return 1;

 for (maxCount=shiftCount; maxCount<32; maxCount++) {
     //printf("DBG: 0x%08lX, shiftCount=%u\n",(unsigned long)uMask,maxCount);
     if ( (uMask & 1)==0 ) return 2;  // Invalid mask!
     uMask >>= 1;
 }

 nBitsCIDR = 32-(t_uint8)shiftCount;
 mask = aMask;

 return 0;
}
////////////////////////////////////////////////////////////
gHostAddr::gHostAddr (char* hostStr)
    : destHostStr( hostStr )
{
 GetHostByName( hostStr );
}

gHostAddr::gHostAddr (t_uint8 c1, t_uint8 c2, t_uint8 c3, t_uint8 c4)
    : gIpAddr( c1, c2, c3, c4 )
{
}

gHostAddr::gHostAddr (gIpAddr& ipAddr)
    : gIpAddr( ipAddr.GetB1(), ipAddr.GetB2(), ipAddr.GetB3(), ipAddr.GetB4() )
{
}

gHostAddr::~gHostAddr ()
{
}

bool gHostAddr::IsOk ()
{
 return gIpAddr::IsOk();
}

t_uchar* gHostAddr::String ()
{
 if ( gIpAddr::IsOk() ) return gIpAddr::String();
 return (t_uchar*)destHostStr.Str();
}

bool gHostAddr::SetHostName (char* hostStr)
{
 destHostStr.SetEmpty();
 if ( hostStr==nil || hostStr[0]==0 ) return false;
 destHostStr.Set( hostStr );
 return true;
}

void gHostAddr::Show (bool doShowAll)
{
 if ( doShowAll==false ) {
     gIpAddr::Show( false );
 }
 printf("%s:%s\n",destHostStr.Str(),String());
}
////////////////////////////////////////////////////////////
gSocket::gSocket (int socketHandle, eSocketKind socketKind)
    : gInt( socketHandle ),
      sockKind( socketKind ),
      hostConnected( nil )
{
 bool isFirst = sockCount==0;

 if ( isFirst ) {
     gStorageControl::Self().RegisterService( "socket", STG_SVC_KND_SOCK, STG_SVC_IDX_SOCK );
     gStorageControl::Self().ServiceAlloc( "socket" );
 }

 sockCount++;
}

gSocket::~gSocket ()
{
 delete hostConnected;
 hostConnected = nil;
}

bool gSocket::IsOk ()
{
 return IsOpened() && IsConnected();
}

bool gSocket::IsOpened ()
{
 return c>0;
}

int gSocket::Handle ()
{
 ASSERTION(IsOpened(),"IsOpened()");
 return c;
}

bool gSocket::IsConnected ()
{
 return hostConnected!=nil;
}

bool gSocket::Open (t_uchar* uName)
{
 char sBuf[50];

 if ( uName==nil ) {
     // 'Invent' one string
     sprintf(sBuf,"%s#%d",sockKind==e_TCP?"TCP":"UDP",sockCount);
 }
 else {
     // The code below induces problems
     //strncpy( sBuf, (char*)uName, 49 );
     //uName[49] = 0;
     // (not used due to bug of glibc)
     ASSERTION(strlen((char*)uName)+1<50,"Node-name too long");
     strcpy( sBuf, (char*)uName );
 }

#ifdef DEBUG
 {
     sServiceRegister* pSvc = gStorageControl::Self().GetServicePtr("socket");
     printf("DBG: [%s|%s] index=%d, used=%d\n",
	    uName, "<>",
	    pSvc->data->pNodes[0].index,
	    pSvc->data->pNodes[0].used);
 }
#endif //DEBUG

#ifdef gSUPPRESS_INET
 c = -1;
 ASSERTION_FALSE("gSUPPRESS_INET: socket");
#else
 c = socket( AF_INET, SOCK_STREAM, 0 );
#endif
 if ( c<0 ) return false;

 sServiceNode aNode;
 aNode.used = (short)sockKind;
 aNode.handle = c;
 strcpy( aNode.shortName, sBuf );
 aNode.pStorage = this;
 gStorageControl::Self().ServiceAdd( "socket", aNode );

 return true;
}

bool gSocket::Close ()
{
 if ( c<=0 ) return false;
 close( c );
 c = -1;
 return gStorageControl::Self().ServiceDelete( "socket", this );
}

bool gSocket::SetConnection (gIpAddr& ipAddr)
{
 // Set connection twice? Return false
 if ( IsOk() ) return false;
 hostConnected = new gHostAddr( ipAddr );
 ASSERTION(hostConnected!=nil,"hostConnected!=nil");
 return hostConnected->IsOk();
}

bool gSocket::SetConnection (gHostAddr& hostAddr)
{
 bool isOk;
 gIpAddr ipAddr( hostAddr.GetB1(), hostAddr.GetB2(), hostAddr.GetB3(), hostAddr.GetB4() );
 isOk = SetConnection( ipAddr );
 if ( isOk==false ) return false;
 return hostConnected->SetHostName( hostAddr.GetHostName() );
}
////////////////////////////////////////////////////////////
gNetConnect::gNetConnect ()
{
}

gNetConnect::~gNetConnect ()
{
}

bool gNetConnect::Read (t_uchar& c)
{
 static t_uchar uChr;
 ssize_t nOctetsRead;
 // In DOS, 'recv' is used often
#ifdef gDOS_SPEC
 static char cChr;
 nOctetsRead = recv( Handle(), &cChr, 1, 0 );
 uChr = (t_uchar)cChr;
#else
 nOctetsRead = read( Handle(), &uChr, 1 );
#endif //gDOS_SPEC (~)
 if ( nOctetsRead==1 ) {
     c = uChr;
     lastOpError = 0;
     return true;
 }
 SetError( errno );
 return false;
}

bool gNetConnect::Read (t_uchar* uBuf, unsigned nBytes)
{
 ssize_t nOctetsRead;
 lastOpError = 0;
 if ( nBytes==0 ) return true;
 // Always initialize memory
 memset( uBuf, 0x0, (size_t)nBytes );
 // Read from socket
#ifdef gDOS_SPEC
 nOctetsRead = recv( Handle(), (char*)uBuf, (size_t)nBytes, 0 );
#else
 nOctetsRead = read( Handle(), uBuf, (size_t)nBytes );
 DBGPRINT("DBG: read(%d)=%u\n",(unsigned)nBytes,(int)nOctetsRead);
#endif //gDOS_SPEC (~)
 if ( nOctetsRead==(ssize_t)nBytes ) return true;
 SetError( errno );
 return false;
}

bool gNetConnect::ReadLine (gString& s)
{
 static t_uchar uChr;

 s.SetEmpty();
 for ( ; Read( uChr )==true; ) {
     if ( uChr=='\r' ) continue;
     if ( uChr=='\n' ) return true;
     s.Add( uChr );
 }
 return false;
}

int gNetConnect::Write (char* s)
{
 return Write( s, strlen(s) );
}

int gNetConnect::Write (gString& s)
{
 return Write( s.Str(), s.Length() );
}

int gNetConnect::Write (char* s, unsigned nOctets)
{
 ssize_t nOctetsWritten, nOctetsToWrite = (ssize_t)nOctets;;
 ASSERTION(s!=nil,"s!=nil");
 if ( nOctets==0 ) return 0;
 // In DOS, 'send' is used often
#ifdef gDOS_SPEC
 nOctetsWritten = send( Handle(), s, nOctets, 0 );
#else
 nOctetsWritten = write( Handle(), s, nOctets );
#endif //gDOS_SPEC (~)
 if ( nOctetsWritten==nOctetsToWrite ) {
     return lastOpError = 0;
 }
 return SetError( errno );
}
////////////////////////////////////////////////////////////
gTcpConnect::gTcpConnect (char* destHostname, t_gPort cPort, bool doConnect)
    : destAddr( destHostname ),
      destPort( cPort ),
      socket( 0, gSocket::e_TCP )
{
 SetError( destAddr.lastOpError );
 if ( lastOpError==0 && doConnect==true ) Connect();
}

gTcpConnect::gTcpConnect (t_uint8 c1, t_uint8 c2, t_uint8 c3, t_uint8 c4, t_gPort cPort, bool doConnect)
    : destAddr( c1, c2, c3, c4 ),
      destPort( cPort ),
      socket( 0, gSocket::e_TCP )
{
 lastOpError = destAddr.IsOk();
 if ( lastOpError==0 && doConnect==true ) Connect();
}

gTcpConnect::~gTcpConnect ()
{
 socket.Close();
}

bool gTcpConnect::IsOk ()
{
 return socket.IsOpened();
}

int gTcpConnect::Handle ()
{
 int fd;
 // CHECK: assert: socket must be opened&valid
 fd = socket.Handle();
 ASSERTION(IsOk(),"gTcpConnect::Handle (1)");
 ASSERTION(fd>=0,"gTcpConnect::Handle (2)");
 return fd;
}

t_uchar* gTcpConnect::String ()
{
 return destAddr.String();
}

t_uchar* gTcpConnect::ToString (t_uchar* uBuf)
{
 t_uchar* uStr = String();
 if ( uBuf==nil ) return uStr;
 strcpy( (char*)uBuf, (char*)uStr );
 return uBuf;
}

bool gTcpConnect::Connect ()
{
 bool isOk;
 int error;
 isOk = socket.IsOpened()==false;
 if ( isOk==false ) {
     SetError( -1 );
     return false;
 }
 isOk = socket.Open( String() );
 if ( isOk==false ) return false;
 error = thisConnect( socket );
 isOk = error==0;
 return isOk;
}

bool gTcpConnect::Close ()
{
 return socket.Close();
}

void gTcpConnect::Show (bool doShowAll)
{
 printf("IP:%s%s\n",
	String(),
	doShowAll ?
	(socket.IsOpened() ? "[opened]" : "[closed]")
	:
	"\0");
}

int gTcpConnect::thisConnect (gSocket& aSocket)
{
 int socketId = aSocket.Handle();
 struct sockaddr_in toHostAddressData;
 int sizeofAddrData = sizeof(toHostAddressData);

#ifdef gSUPPRESS_INET
 lastOpError = -1;
 ASSERTION_FALSE("gSUPPRESS_INET: connect");
#else
 memset( (char*)&toHostAddressData, 0, sizeofAddrData );
 toHostAddressData.sin_family = AF_INET;
 toHostAddressData.sin_addr.s_addr = destAddr.GetHostAddress();
 toHostAddressData.sin_port = htons( destPort );

 lastOpError = connect( socketId, (struct sockaddr*)&toHostAddressData, sizeofAddrData )!=0;
#endif //gSUPPRESS_INET (~)
 if ( lastOpError!=0 ) return SetError( errno );

 // Inform the socket to whom the connection was established
 aSocket.SetConnection( destAddr );

 return 0;
}
////////////////////////////////////////////////////////////

