// gstorage.cpp -- Version 0.2

#include <stdio.h>
#include <string.h>
#include "gstorage.h"
////////////////////////////////////////////////////////////
// Static members
// ---------------------------------------------------------
short gTop::nObjHistogram[MAX_INTSTGKIND];
int gTop::nObjs=0;
gUCharBuffer gStorage::oneCharBuf( 20 );
gStorage::eStoreMethod gStorage::globalStoreMethod=gStorage::e_StgGlobalString;

const sStorageDesc gStorage::storeKindDesc[]={
    {"NIL",     0},
    {"RSVD",    1},       //e_ResvdStore  = 1
    {"UCHAR",   2},       //e_UChar       = 2
    {"SCHAR",   3},       //e_SChar       = 3
    {"UINT",    4},       //e_UInt        = 4
    {"SINT",    5},       //e_SInt        = 5
    {"ULINT",   6},       //e_ULongInt    = 6
    {"SLINT",   7},       //e_SLongInt    = 7
    {"RSVD08",  0},       // = 8
    {"RSVD09",  0},       // = 9
    {"REAL",   10},       //e_Real        = 10
    {"LIST",   11},       //e_List        = 11
    {"RSVD12",  0},       // = 12
    {"RSVD13",  0},       // = 13
    {"RSVD14",  0},       // = 14
    {"RSVD15",  0},       // = 15
    {"RSVD16",  0},       // = 16
    {"RSVD17",  0},       // = 17
    {"RSVD18",  0},       // = 18
    {"RSVD19",  0},       // = 19
    {"STR",    20},       //e_String      = 20
    {"UNISTR", 21},       //e_StrUnicode  = 21
    {"RSVD22",  0},       // = 22
    {"RSVD23",  0},       // = 23
    {"RSVD24",  0},       // = 24
    {"RSVD25",  0},       // = 25
    {"RSVD26",  0},       // = 26
    {"RSVD27",  0},       // = 27
    {"RSVD28",  0},       // = 28
    {"RSVD29",  0},       // = 29
    {"RSVD30",  0},       // = 30
    {"EXT___", 31},       //e_StoreExtend  = 31
    {"CTRL",   32},       //e_Control
};
////////////////////////////////////////////////////////////
// gTop - Generic control for objects
// ---------------------------------------------------------
gTop::gTop (t_uint16 intStgKind)
    : stgKind( intStgKind )
{
#ifdef DEBUG_MIN
 static int dbgInst;
#endif //DEBUG_MIN

 nObjs++;
 ASSERTION(stgKind>=1 && stgKind<MAX_INTSTGKIND,"gTop::gTop");
 nObjHistogram[stgKind]++;
 ///if ( stgKind==gStorage::e_List ) printf("DBG:::+ %d lists (%d)\n",nObjHistogram[stgKind],nObjs)
 DBGPRINT_MIN("DBG:::+ (#%d)\tKIND=%u KIND#%d  (%d)\n",nObjs,stgKind,nObjHistogram[stgKind],dbgInst++);
}

gTop::~gTop ()
{
 nObjs--;
 DBGPRINT_MIN("DBG: gTop::~gTop: nObjs=%d, stgKind=%d\n",nObjs,stgKind);
#ifdef DEBUG
 ASSERTION(stgKind>=1 && stgKind<MAX_INTSTGKIND,"gTop::~gTop");  // only debug: too much consuming
#endif //DEBUG
 nObjHistogram[stgKind]--;
 ///if ( stgKind==gStorage::e_List ) printf("DBG:::- %d lists (%d)\n",nObjHistogram[stgKind],nObjs)
 //printf("DBG:::- (#%d)\tKIND=%u KIND#%d\n",nObjs,stgKind,nObjHistogram[stgKind]);
}
////////////////////////////////////////////////////////////
gUCharBuffer::gUCharBuffer (t_uint16 bufSize)
    : size( bufSize ),
      uBuf( nil )
{
 if ( size==0 ) size = GENUCHAR_USU_BUFSIZE;
 thisAllocate( size );
}

gUCharBuffer::~gUCharBuffer ()
{
 delete[] uBuf;
}

void gUCharBuffer::Copy (t_uchar* s)
{
 thisCopy( s, strlen((char*)s) );
}

void gUCharBuffer::Copy (char* s)
{
 thisCopy( (t_uchar*)s, strlen(s) );
}

void gUCharBuffer::Clear ()
{
 memset( uBuf, 0x0, (size_t)size );
}

void gUCharBuffer::thisAllocate (t_uint16 bufSize)
{
 delete[] uBuf;
 uBuf = new t_uchar[size=bufSize];
 ASSERTION(uBuf!=nil,"gUCharBuffer::thisAllocate");
 Clear();
}

void gUCharBuffer::thisCopy (t_uchar* s, t_uint16 len)
{
 if ( len+1>=size ) {
     // Buffer somehow short, rebuilding it
     thisAllocate( len*2 );
 }
 Clear();
 for (t_uint16 iter=0; iter<len; iter++) uBuf[iter] = s[iter];
}
////////////////////////////////////////////////////////////
// gStorage - Generic storage handling
// ---------------------------------------------------------
gStorage::gStorage (gStorage::eStorage aKind,
		    gStorage::eStoreMethod aMethod)
    : gTop( (t_uint16)aKind ),
      kind( aKind ),
      storeMethod( aMethod )
{
}

gStorage::~gStorage ()
{
}

char* gStorage::StorageName ()
{
 ASSERTION(kind>=0 && kind<e_UnusedStore,"StorageName(1)");
 return storeKindDesc[ kind ].descStr;
}

char* gStorage::Str ()
{
 return (char*)ToString( oneCharBuf.uBuf );
}

int gStorage::SetError (int opError)
{
 if ( opError==0 ) return 0;
 //printf("DBG: SETERROR(%d:%s)\n",opError,StorageName());
 //fprintf(stderr,"@STG: opError=%d\n",opError);
 return opError;
}

void gStorage::SetStoreMethod (gStorage::eStoreMethod aMethod)
{
 eStoreMethod resMethod( e_StgDefault );
 bool isOk = thisSetStoreMethod(aMethod,storeMethod);
 ASSERTION(isOk,"gStorage::SetStoreMethod");
 storeMethod = resMethod;
}

bool gStorage::EndGuts ()
{
 ;
 // Return true if this object is 'active'
 return false;
}

void gStorage::Show (bool doShowAll)
{
 short refKind;
 if ( kind>=e_UnusedStore ) {
     printf("[unknown]");
     return;
 }
 printf("[%s]",storeKindDesc[kind].descStr);
 refKind = storeKindDesc[kind].refKind;
#ifdef DEBUG_IO
 if ( refKind==0 ) fprintf(stderr,"DBG: refKind=0, kind=%d (%s)\n",(int)kind,storeKindDesc[kind].descStr);
#endif

 ASSERTION(refKind==0 || refKind==kind,"...refKind==kind");
}

bool gStorage::thisSetStoreMethod (gStorage::eStoreMethod aMethod,
				   gStorage::eStoreMethod& resMethod)
{
 resMethod = aMethod;

 switch ( aMethod ) {
 case gStorage::e_StgNoStore:
     break;
 case gStorage::e_StgGlobalString:
     globalStoreMethod = aMethod;
     break;
 case gStorage::e_StgGlobalFlat:
     globalStoreMethod = aMethod;
     break;
 case gStorage::e_StgString:
 case gStorage::e_StgFlat:
 case gStorage::e_StgDefault:
     break;
 default:
     return false;
 }
 return true;
}

bool gStorage::CanSave (FILE* f)
{
 return CanRestore( f );
}

bool gStorage::CanRestore (FILE* f)
{
 if ( globalStoreMethod==e_StgNoStore ) return false;
 return f!=NULL;
}
////////////////////////////////////////////////////////////
sServiceNode::~sServiceNode ()
{
 // This is anoying for forked programs: uncomment yourself! (ugly currently)
 //if ( used>=0 ) {
     //fprintf(stderr,"@STG: Undeleted node (%s#%d): used=%d, handle=%d\n",shortName,(int)index,(int)used,handle);
 //}
}

void sServiceNode::Copy (sServiceNode& copy)
{
 ASSERTION(copy.shortName[0]!=0,"invalid-copy(1)"); //Must copy a valid name
 used = copy.used;
 handle = copy.handle;
 strcpy( shortName, copy.shortName );
 pStorage = copy.pStorage;
}

void sServiceNode::Delete ()
{
 used = -1;
 index = 0;
 handle = -1;
 memset( shortName, 0x0, sizeof(shortName) );
 pStorage = nil;
}

sServiceArray::sServiceArray (unsigned max)
    : n( 0 ),
      maxN( max ),
      pNodes( nil )
{
 Allocate( maxN );
}

sServiceArray::~sServiceArray ()
{
 Release();
}

void sServiceArray::Release ()
{
 if ( pNodes!=nil ) delete[] pNodes;
 pNodes = nil;
 n = 0;
}

void sServiceArray::Allocate (unsigned nAlloc)
{
// Release();  :: CHECK THIS, use Release to allow multiple allocation
 ASSERTION(nAlloc>0,"nAlloc>0");
 // .
 ASSERTION(pNodes==nil,"pNodes==nil");
 pNodes = new sServiceNode[ (maxN = nAlloc)+1 ];
 ASSERTION(pNodes!=nil,"pNodes!=nil");
}

bool sServiceArray::Add (sServiceNode& aNode)
{
 unsigned idx;
 n++;
 idx = n;
 if ( n>maxN ) {
     idx = FindFreeNode();
     if ( idx==0 ) return false;
     n = maxN;
 }
 ASSERTION(idx>=1 && idx<=maxN,"idx<=maxN");
 ASSERTION(pNodes!=nil,"pNodes!=nil");
 pNodes[ idx ].Copy( aNode );
 pNodes[ idx ].index = idx;
 return true;
}

bool sServiceArray::Delete (gStorage* pStorage)
{
 unsigned idx;
 ASSERTION(pStorage!=nil,"pStorage!=nil");
 ASSERTION(n>0,"n>0");
 // Find the storage in the nodes
 for (idx=1; idx<=maxN; idx++) {
     if ( pNodes[idx].pStorage==pStorage ) {
	 pStorage->EndGuts();
	 pNodes[idx].Delete();
	 return true;
     }
 }
 return false;
}

unsigned sServiceArray::FindFreeNode ()
{
 unsigned idx;
 for (idx=1; idx<=maxN; idx++) {
     if ( pNodes[idx].used==-1 ) return idx;
 }
 return 0;
}
////////////////////////////////////////////////////////////
sServiceRegister::~sServiceRegister ()
{
 ReleaseDescriptor();
 if ( data!=nil ) ReleaseData();
}

int sServiceRegister::AllocateNodes ()
{
 unsigned n = nMediumNodes;
 ASSERTION(data==nil,"data==nil");
 ASSERTION(n>1,"n>1"); // ...more than 1 node
 data = new sServiceArray( n );
 ASSERTION(data!=nil,"data!=nil");
 return (int)n;
}

void sServiceRegister::ReleaseDescriptor ()
{
 delete[] descriptor.descStr;
 descriptor.descStr = nil;
}

void sServiceRegister::ReleaseData ()
{
 ASSERTION(data!=nil,"data!=nil");
 delete data;
 data = nil;
}
////////////////////////////////////////////////////////////
void gUChar::Reset ()
{
 gStorage::Reset();
 c = 0;
}

gStorage* gUChar::NewObject ()
{
 gUChar* a = new gUChar( c );
 return a;
}

t_uchar* gUChar::ToString (t_uchar* uBuf)
{
 if ( uBuf==nil ) return nil;
 sprintf( (char*)uBuf, "%c", c );
 return uBuf;
}

bool gUChar::SaveGuts (FILE* f)
{
 if ( CanSave( f )==false ) return false;
 return fprintf( f,"%c", c );
}

bool gUChar::RestoreGuts (FILE* f)
{
 if ( CanRestore( f )==false ) return false;
 return fscanf( f, "%c", &c )>=1;
}
////////////////////////////////////////////////////////////
void gInt::Reset ()
{
 gStorage::Reset();
 c = 0;
}

gStorage* gInt::NewObject ()
{
 gInt* a = new gInt( c );
 return a;
}

t_uchar* gInt::ToString (t_uchar* uBuf)
{
 if ( uBuf==nil ) return nil;
 sprintf( (char*)uBuf, "%d", c);
 return uBuf;
}

bool gInt::SaveGuts (FILE* f)
{
 if ( CanSave( f )==false ) return false;
 return fprintf( f, "%d", c );
}

bool gInt::RestoreGuts (FILE* f)
{
 if ( CanRestore( f )==false ) return false;
 return fscanf( f, "%d", &c )>=1;
}

void gInt::Show (bool doShowAll)
{
 printf("%s%d",doShowAll?"'":"\0",c);
}
////////////////////////////////////////////////////////////
void gUInt::Reset ()
{
 gStorage::Reset();
 c = 0;
}

gStorage* gUInt::NewObject ()
{
 gUInt* a = new gUInt( c );
 return a;
}

t_uchar* gUInt::ToString (t_uchar* uBuf)
{
 if ( uBuf==nil ) return nil;
 sprintf( (char*)uBuf, "%u", c);
 return uBuf;
}

bool gUInt::SaveGuts (FILE* f)
{
 if ( CanSave( f )==false ) return false;
 return fprintf( f, "%u", c );
}

bool gUInt::RestoreGuts (FILE* f)
{
 if ( CanRestore( f )==false ) return false;
 return fscanf( f, "%u", &c )>=1;
}

void gUInt::Show (bool doShowAll)
{
 printf("%s%u",doShowAll?"'":"\0",c);
}
////////////////////////////////////////////////////////////
void gReal::Reset ()
{
 gStorage::Reset();
 c = 0.0;
}

gStorage* gReal::NewObject ()
{
 gReal* a = new gReal( (float)c );
 return a;
}

t_uchar* gReal::ToString (t_uchar* uBuf)
{
 if ( uBuf==nil ) return nil;
 sprintf( (char*)uBuf, "%f", c);
 return uBuf;
}

bool gReal::SaveGuts (FILE* f)
{
 if ( CanSave( f )==false ) return false;
 return fprintf( f, "%f", c );
}

bool gReal::RestoreGuts (FILE* f)
{
 bool isOk;
 float val;
 if ( CanRestore( f )==false ) return false;
 isOk = fscanf( f, "%f", &val );
 if ( isOk ) c = val;
 return isOk;
}

void gReal::Show (bool doShowAll)
{
 printf("%s%f",doShowAll?"`":"\0",c);
}
////////////////////////////////////////////////////////////

