#ifndef gSTORAGE_X_H
#define gSTORAGE_X_H

#include <stdio.h>
#include "gmain.h"

#define nil 0
#define MAX_INTSTGKIND 1000
#ifdef DEBUG
#define GENUCHAR_USU_BUFSIZE 20    // for debug, only (a small buffer)
#else
#define GENUCHAR_USU_BUFSIZE 4096  // UCharBuffer...
#endif //DEBUG

#ifndef LOG_NONE
#define LOG_NONE -1
#endif
#ifndef LOG_ERROR
#define LOG_ERROR 0
#endif
#ifndef LOG_WARNING
#define LOG_WARNING 1
#endif
#ifndef LOG_NOTICE
#define LOG_NOTICE 2
#endif
#ifndef LOG_INFO
#define LOG_INFO 3
#endif
#ifndef LOG_LOGMAX
#define LOG_LOGMAX (LOG_INFO+1)
#endif
#ifndef LOG_NOCLASS
#define LOG_NOCLASS 99
#endif
////////////////////////////////////////////////////////////
struct sStorageDesc {
    char* descStr;
    short refKind;
};
////////////////////////////////////////////////////////////
class gTop {
public:
    virtual ~gTop () ;

    // Public data-members
    t_uint16 stgKind;
    static short nObjHistogram[MAX_INTSTGKIND];

    int GetNumObjects () {
	return nObjs;
    }

protected:
    gTop (t_uint16 intStgKind) ;

private:
    static int nObjs;
};
////////////////////////////////////////////////////////////
class gUCharBuffer {
public:
    gUCharBuffer (t_uint16 bufSize=0) ;
    ~gUCharBuffer () ;

    // Public data-members
    t_uint16 size;
    t_uchar* uBuf;

    // Get methods
    char* Str () {
	return (char*)uBuf;
    }

    // Set methods
    void Copy (t_uchar* s) ;
    void Copy (char* s) ;
    void Clear () ;

private:
    void thisAllocate (t_uint16 bufSize) ;
    void thisCopy (t_uchar* s, t_uint16 len) ;

    // Operators,empty
    gUCharBuffer (gUCharBuffer& ) ; //empty
    gUCharBuffer& operator= (gUCharBuffer& ) ; //empty
};
////////////////////////////////////////////////////////////
class gStorage : public gTop {
public:
    // Public enums
    enum eStorage {
	e_NoStorage,
	e_ResvdStore  = 1,
	e_UChar       = 2,
	e_SChar       = 3,
	e_UInt        = 4,
	e_SInt        = 5,
	e_ULongInt    = 6,
	e_SLongInt    = 7,
	e_ResvdStore8 = 8,
	e_ResvdStore9 = 9,
	e_Real        = 10,
	e_List        = 11,
	e_ResvdStore12= 12,
	e_ResvdStore13= 13,
	e_ResvdStore14= 14,
	e_ResvdStore15= 15,
	e_ResvdStore16= 16,
	e_ResvdStore17= 17,
	e_ResvdStore18= 18,
	e_ResvdStore19= 19,
	e_String      = 20,
	e_StrUnicode  = 21,
	e_ResvdStore22= 22,
	e_ResvdStore23= 23,
	e_ResvdStore24= 24,
	e_ResvdStore25= 25,
	e_ResvdStore26= 26,
	e_ResvdStore27= 27,
	e_ResvdStore28= 28,
	e_ResvdStore29= 29,
	e_ResvdStore30= 30,
	e_StoreExtend = 31,
	e_Control,
	e_UnusedStore
    };
    enum eStoreMethod {
	e_StgNoStore      = 0,
	e_StgGlobalString = 1,
	e_StgGlobalFlat   = 2,
	e_StgString       = 3,
	e_StgFlat         = 4,
	e_StgDefault      = 64
    };
    //
    virtual ~gStorage () ;

    // Public data-members
    static const sStorageDesc storeKindDesc[];

    // Get methods
    virtual bool IsOk () {
	return true;
    }
    eStorage Kind () {
	return kind;
    }
    virtual eStoreMethod GetStoreMethod () {
	return storeMethod;
    }
    virtual char* StorageName () ;

    virtual bool IsString () {
	return MinimumStorageSize()==-1;
    }

    virtual char* Str () ;

    virtual int MinimumStorageSize () {   // >=0; -1 means it is a string-kind
	return 1;
    }
    virtual int MaximumStorageSize () {   // -1 means arbitrary size, like in strings...
#ifdef gUINT_IS_32bit
	return 10;  // ten octets for decimal representation of an unsigned int;
#else
	return 5;   // only five octets for representation of 2^16
#endif
    }

    // Set methods
    virtual void Reset () {
	SetError( 0 );
    }
    virtual int SetError (int opError) ;
    virtual void SetStoreMethod (eStoreMethod aMethod) ;

    // Special methods
    virtual gStorage* NewObject () = 0; //PURE

    // Save/etc methods
    virtual t_uchar* ToString (t_uchar* uBuf) = 0; //PURE
    virtual bool EndGuts () ;
    virtual bool SaveGuts (FILE* f) = 0; //PURE
    virtual bool RestoreGuts (FILE* f) = 0; //PURE

    // Show methods
    virtual void Show (bool doShowAll=true) ;

protected:
    gStorage (eStorage aKind,
	      eStoreMethod aMethod=e_StgDefault) ;

    eStorage kind;
    eStoreMethod storeMethod;
    static gUCharBuffer oneCharBuf;

    bool thisSetStoreMethod (eStoreMethod aMethod,
			     eStoreMethod& resMethod) ;

    bool CanSave (FILE* f) ;
    bool CanRestore (FILE* f) ;

private:
    static eStoreMethod globalStoreMethod;

    // Operators,empty
    gStorage (gStorage& ) ; //empty
    gStorage& operator= (gStorage& ) ; //empty
};
////////////////////////////////////////////////////////////
struct sServiceNode {
    sServiceNode ()
	: used( -1 ),
	  index( 0 ),
	  handle( -1 ),
	  pStorage( nil ) {
	Delete(); //Clean-up name
    }
    ~sServiceNode () ;

    // Members
    short used;
    short index;
    int handle;
    char shortName[60];
    gStorage* pStorage;

    // Methods
    void Copy (sServiceNode& copy) ;
    void Delete () ;

    // Operators,empty
    sServiceNode (sServiceNode& ) ; //empty
    sServiceNode& operator= (sServiceNode& ) ; //empty
};

struct sServiceArray {
    sServiceArray (unsigned max) ;
    ~sServiceArray () ;

    unsigned n, maxN;
    sServiceNode* pNodes;  // 1..maxN (0th not used)

    // Methods
    void Allocate (unsigned nAlloc) ;
    void Release () ;
    bool Add (sServiceNode& aNode) ;
    bool Delete (gStorage* pStorage) ;
    unsigned FindFreeNode () ;

    // Operators,empty
    sServiceArray (sServiceArray& ) ; //empty
    sServiceArray& operator= (sServiceArray& ) ; //empty
};
////////////////////////////////////////////////////////////
struct sServiceRegister {
    sServiceRegister (unsigned aMediumNodes=5)
	: nMediumNodes( aMediumNodes ),
	  data( nil ) {
	descriptor.descStr = nil;
	descriptor.refKind = 0;
    }
    ~sServiceRegister () ;

    // Members
    sStorageDesc descriptor;
    unsigned nMediumNodes;
    sServiceArray* data;  // Ptr to one instance, only

    // Methods
    bool HasNodes () {
	return data!=nil;
    }
    unsigned NumberOfNodes () {
	return nMediumNodes;
    }
    int AllocateNodes () ;
    void ReleaseDescriptor () ;
    void ReleaseData () ;

    // Operators,empty
    sServiceRegister (sServiceRegister& ) ; //empty
    sServiceRegister& operator= (sServiceRegister& ) ; //empty
};
////////////////////////////////////////////////////////////
class gUChar : public gStorage {
public:
    gUChar (t_uchar v='\0')
	: gStorage( e_UChar ),
	  c( v ) {
    }
    virtual ~gUChar () {
    }

    // Get methods
    virtual t_uchar GetUChar () {
	return c;
    }

    virtual int MaximumStorageSize () {
	return 1;
    }

    // Set methods
    virtual void Reset () ;
    virtual bool SetUChar (t_uchar v) {
	c = v;
	return true;
    }

    virtual gStorage* NewObject () ;

    // Save/etc methods
    virtual t_uchar* ToString (t_uchar* uBuf) ;
    virtual bool SaveGuts (FILE* f) ;
    virtual bool RestoreGuts (FILE* f) ;

protected:
    t_uchar c;

private:
    // Operators,empty
    gUChar (gUChar& ) ; //empty
    gUChar& operator= (gUChar& ) ; //empty
};
////////////////////////////////////////////////////////////
class gSwitch : public gUChar {
public:
    gSwitch (bool isOn=false)
	: gUChar( isOn ? 'T' : '\0' ) {
    }
    virtual ~gSwitch () {
    }

    virtual bool IsOn () {
	return c!=0;
    }
    virtual bool SetOn (bool isOn) {
	return (c = isOn ? 'T' : '\0')!=0;
    }

private:
    // Operators,empty
    gSwitch (gSwitch& ) ; //empty
    gSwitch& operator= (gSwitch& ) ; //empty
};
////////////////////////////////////////////////////////////
class gInt : public gStorage {
public:
    gInt (int v=0)
	: gStorage( e_SInt ),
	  c( v ) {
    }
    virtual ~gInt () {
    }

    // Get methods
    virtual int GetInt () {
	return c;
    }

    int GetMinInt () {
#ifdef gUINT_IS_32bit
	return (int)MIN_DLINT32;
#else
	return -32768;
#endif
    }

    int GetMaxInt () {
#ifdef gUINT_IS_32bit
	return (int)MAX_DLINT32;
#else
	return MAX_INT16_I;
#endif
    }

    // Set methods
    virtual void Reset () ;
    virtual bool SetInt (int v) {
	c = v;
	return true;
    }

    virtual int Incr (int v=1) {
	c += v;
	return c;
    }

    virtual gStorage* NewObject () ;

    // Save/etc methods
    virtual t_uchar* ToString (t_uchar* uBuf) ;
    virtual bool SaveGuts (FILE* f) ;
    virtual bool RestoreGuts (FILE* f) ;

    // Show methods
    virtual void Show (bool doShowAll=true) ;

protected:
    int c;

private:
    // Operators,empty
    gInt (gInt& ) ; //empty
    gInt& operator= (gInt& ) ; //empty
};
////////////////////////////////////////////////////////////
class gUInt : public gStorage {
public:
    gUInt (unsigned v=0)
	: gStorage( e_UInt ),
	  c( v ) {
    }
    virtual ~gUInt () {
    }

    // Get methods
    virtual unsigned GetUInt () {
	return c;
    }

    int GetMaxUInt () {
#ifdef gUINT_IS_32bit
	return (int)MAX_DUINT32;
#else
	return MAX_UINT16_U;
#endif
    }

    // Set methods
    virtual void Reset () ;
    virtual bool SetUInt (unsigned v) {
	c = v;
	return true;
    }

    virtual gStorage* NewObject () ;

    // Save/etc methods
    virtual t_uchar* ToString (t_uchar* uBuf) ;
    virtual bool SaveGuts (FILE* f) ;
    virtual bool RestoreGuts (FILE* f) ;

    // Show methods
    virtual void Show (bool doShowAll=true) ;

protected:
    unsigned c;

private:
    // Operators,empty
    gUInt (gUInt& ) ; //empty
    gUInt& operator= (gUInt& ) ; //empty
};
////////////////////////////////////////////////////////////
class gReal : public gStorage {
public:
    gReal (float v)
	: gStorage( e_Real ),
	  c( (double)v ) {
    }
    gReal (unsigned v=0)
	: gStorage( e_Real ),
	  c( (double)v ) {
    }
    virtual ~gReal () {
    }

    // Get methods
    virtual int GetInt () {
	return (int)c;
    }

    virtual int MaximumStorageSize () {
	return oneCharBuf.size;
    }

    // Set methods
    virtual void Reset () ;
    virtual bool SetInt (int v) {
	c = (double)v;
	return true;
    }

    virtual gStorage* NewObject () ;

    // Save/etc methods
    virtual t_uchar* ToString (t_uchar* uBuf) ;
    virtual bool SaveGuts (FILE* f) ;
    virtual bool RestoreGuts (FILE* f) ;

    // Show methods
    virtual void Show (bool doShowAll=true) ;

protected:
    double c;

private:
    // Operators,empty
    gReal (gReal& ) ; //empty
    gReal& operator= (gReal& ) ; //empty
};
////////////////////////////////////////////////////////////
#endif //gSTORAGE_X_H

