// gstring.cpp -- Version 1.0

#include <string.h>
#include "gstring.h"
////////////////////////////////////////////////////////////
// Static members
// ---------------------------------------------------------
t_uchar gStringGeneric::myChrNul=0;

////////////////////////////////////////////////////////////
// gStringGeneric - Generic string handling
// ---------------------------------------------------------
gStringGeneric::gStringGeneric (eStorage kind,
				eStrStrategy aStgy,
				t_uchar* s,
				unsigned aMaxSize)
    : gStorage( kind, e_StgDefault ),
      size( 0 ),
      preMaxSize( aStgy==e_StrAlloc ? 0 : GENUCHAR_USU_BUFSIZE ),
      maxSize( aMaxSize ),
      stgy( aStgy ),
      str( nil ),
      doRevalidateSize( false )
{
 thisPreAllocate( s, preMaxSize );
}

gStringGeneric::~gStringGeneric ()
{
 thisDelete();
}

char* gStringGeneric::Str ()
{
 return (char*)UStr();
}

t_uchar* gStringGeneric::UStr ()
{
 return str;
}

t_uchar gStringGeneric::GetUChar (unsigned idx)
{
 if ( thisIndex(idx)==false ) return myChrNul;
 return str[idx-1];
}

void gStringGeneric::Reset ()
{
 gStorage::Reset();
 SetEmpty();
}

void gStringGeneric::SetEmpty ()
{
 thisDelete();
 thisPreAllocate( (t_uchar*)"\0", 0 );
}

void gStringGeneric::Set (char* s)
{
 ASSERTION(s!=nil,"gStringGeneric::Set");
 Set( (t_uchar*)s );
}

void gStringGeneric::Set (t_uchar* s)
{
 ASSERTION(s!=nil,"gStringGeneric::Set");
 thisDelete();
 thisPreAllocate( s, preMaxSize );
}

void gStringGeneric::Set (int v)
{
 thisDelete();
 Add( v );
}

void gStringGeneric::Set (unsigned v)
{
 thisDelete();
 Add( v );
}


unsigned gStringGeneric::Add (char c)
{
 char s[2];
 s[0] = c;
 s[1] = 0;
 return Add( s );
}

unsigned gStringGeneric::Add (t_uchar c)
{
 return Add( (char)c );
}

unsigned gStringGeneric::Add (char* s)
{
 return Add( (t_uchar*)s );
}

unsigned gStringGeneric::Add (t_uchar* s)
{
 unsigned len, totalLen;
 int copyCode;

 if ( s==nil ) return size;
 len = strlen( (char*)s );
 if ( size==0 ) {
     thisCopy( s, len );
     return len;
 }
 // Re-calculate existing string length
 size = strlen( (char*)str );
 totalLen = size+len;
 // Keep the old string
 gString tempStr( str );
 copyCode = thisCopy( tempStr.UStr(), totalLen );
 if ( copyCode<0 ) {
     copyCode = thisCopy( tempStr.UStr(), tempStr.Length() );
     ASSERTION(copyCode>=0,"copyCode>=0");
     return size;
 }
 strcat( (char*)str, (char*)s );
 return size;
}

unsigned gStringGeneric::Add (int v)
{
 gUCharBuffer s( 30 );
 sprintf( (char*)s.uBuf, "%d", v );
 Add( s.uBuf );
 return size;
}

unsigned gStringGeneric::Add (unsigned v)
{
 gUCharBuffer s( 30 );
 sprintf( (char*)s.uBuf, "%u", v );
 Add( s.uBuf );
 return size;
}

void gStringGeneric::UpString ()
{
 unsigned k;
 t_uchar chr;

 thisSize();
 for (k=0; k<size; k++) {
     chr = str[k];
     if ( chr>='a' && chr<='z' ) {
	 chr = (t_uchar)(chr - 32);
	 str[k] = chr;
     }
 }
}

void gStringGeneric::DownString ()
{
 unsigned k;
 t_uchar chr;

 thisSize();
 for (k=0; k<size; k++) {
     chr = str[k];
     if ( chr>='A' && chr<='Z' ) {
	 chr = (t_uchar)(chr + 32);
	 str[k] = chr;
     }
 }
}

bool gString::Trim ()
{
 bool didTrimL = TrimLeft();
 bool didTrimR = TrimRight();
 return didTrimL || didTrimR;
}

bool gString::TrimLeft ()
{
 char* s = Str();
 int i, k=0, n = (int)Length();
 gString tempS;
 for (i=0; i<n; i++) {
     if ( s[i]!=' ' && s[i]!='\t' ) {
	 break;
     }
     k++;
 }
 for ( ; i<n; i++) tempS.Add( s[i] );
 Copy( tempS );
 return k>0;
}

bool gString::TrimRight ()
{
 char* s = Str();
 int i, k, n = (int)Length();

 for (i=n-1, k=n; i>=0; i--) {
     if ( s[i]!=' ' && s[i]!='\t' ) {
	 break;
     }
     s[i] = 0;
     k--;
 }
 gString tempS( s );
 Copy( tempS );
 return k<n;
}

t_uchar& gStringGeneric::operator[] (int index)
{
 if ( thisIsValidIndex(index)==false ) return myChrNul;
 doRevalidateSize = true;
 return str[index-1];
}

t_uchar* gStringGeneric::ToString (t_uchar* uBuf)
{
 if ( uBuf!=nil ) {
     strcpy( (char*)uBuf, (char*)str );
 }
 return str;
}

bool gStringGeneric::SaveGuts (FILE* f)
{
 if ( CanSave( f )==false ) return false;
 if ( str==nil ) return true;
 return fprintf(f,"%s",str)==1;
}

bool gStringGeneric::RestoreGuts (FILE* f)
{
 unsigned n=0;
 t_uchar uChr;
 gUCharBuffer sTemp;

 if ( CanRestore( f )==false ) return false;
 thisDelete();

 while ( fscanf(f,"%c",&uChr) ) {
     if ( uChr=='\n' || uChr==0 ) break;
     sTemp.uBuf[n++] = uChr;
     sTemp.uBuf[n] = 0;
     if ( n>=sTemp.size ) return false;
 }
 thisCopy( sTemp.uBuf, n );
 return true;
}

void gStringGeneric::thisPreAllocate (t_uchar* s, unsigned toSize)
{
 if ( stgy==e_StrDefault ) stgy = e_StrAlloc;
 ASSERTION(stgy==e_StrAlloc,"stgy==e_StrAlloc: TODO other");
 ASSERTION(str==nil,"str==nil");
 if ( s==nil ) return;
 thisCopy( s, strlen( (char*)s ) );
}

void gStringGeneric::thisDelete ()
{
 size = 0;
 if ( str!=nil ) delete[] str;
 str = nil;
}

int gStringGeneric::thisCopy (t_uchar* s, unsigned len)
{
 thisDelete();
 ASSERTION(preMaxSize==0,"preMaxSize==0: TODO other");
 if ( s==nil ) return -1;
 if ( maxSize>0 && len>maxSize ) return -2;
 str = new t_uchar[len+1];
 ASSERTION(str!=nil,"str!=nil");
 strcpy( (char*)str, (char*)s );
 size = len;
 return (int)size;
}

bool gStringGeneric::thisIndex (unsigned& idx)
{
 bool isOk = idx>=1 && idx<=size;

 if ( idx<1 ) idx = 1;
 if ( idx>size ) idx = size;
 isOk = idx>=1 && idx<=size;

 return isOk;
}

bool gStringGeneric::thisIsValidIndex (int index)
{
 bool isOk = index>0;
 unsigned idx = (unsigned)index;

 if ( isOk==false ) return false;

 return idx>=1 && idx<=size;
}

unsigned gStringGeneric::thisSize ()
{
 if ( doRevalidateSize==false ) return size;
 size = str==nil?0:(strlen((char*)str));
 return size;
}
////////////////////////////////////////////////////////////
gString::gString ()
    : gStringGeneric(
	e_String,
	e_StrAlloc,
	(t_uchar*)"\0",
	0)
{
}

gString::gString (gString& copy)
    : gStringGeneric(
	e_String,
	e_StrAlloc,
	(t_uchar*)(copy.Str()),
	0)
{
 ASSERTION(str!=nil,"str!=nil");
}

gString::gString (char* s)
    : gStringGeneric(
	e_String,
	e_StrAlloc,
	(t_uchar*)s,
	0)
{
 ASSERTION(s!=nil,"s!=nil");
}

gString::gString (t_uchar* s)
    : gStringGeneric(
	e_String,
	e_StrAlloc,
	s,
	0)
{
 ASSERTION(s!=nil,"s!=nil");
}

gString::gString (char c)
    : gStringGeneric(
	e_String,
	e_StrAlloc,
	(t_uchar*)"\0",
	0)
{
 Add( c );
}

gString::gString (unsigned nBytes, char c)
    : gStringGeneric(
	e_String,
	e_StrAlloc,
	(t_uchar*)"\0",
	nBytes)
{
 if ( nBytes ) {
     thisCopy( (t_uchar*)"\0", nBytes );
     //for (i=0; i<nBytes; i++) str[i] = (t_uchar)c ==> too slow
     memset( str, (int)c, (size_t)nBytes );
 }
}

gString::~gString ()
{
}

bool gString::Match (gString& copy, bool doIgnoreCase)
{
 return thisMatch( Str(), copy.Str(), doIgnoreCase )==0;
}

bool gString::Match (char* s, bool doIgnoreCase)
{
 return thisMatch( Str(), s, doIgnoreCase )==0;
}

unsigned gString::Find (gString& sSub, bool doIgnoreCase)
{
 return thisFind( Str(), sSub.Str(), 1, doIgnoreCase );
}

unsigned gString::Find (char* s, bool doIgnoreCase)
{
 return thisFind( Str(), s, 1, doIgnoreCase );
}

unsigned gString::Find (char c, bool doIgnoreCase)
{
 char s[2];
 s[0] = c;
 s[1] = 0;
 if ( c==0 ) return 0;
 return thisFind( Str(), s, 1, doIgnoreCase );
}

unsigned gString::Find (gString& sSub, unsigned& nOcc, bool doIgnoreCase)
{
 return thisFindFwd( Str(), sSub.Str(), 1, doIgnoreCase, nOcc );
}

unsigned gString::Find (char* s, unsigned& nOcc, bool doIgnoreCase)
{
 return thisFindFwd( Str(), s, 1, doIgnoreCase, nOcc );
}

unsigned gString::Find (char c, unsigned& nOcc, bool doIgnoreCase)
{
 char s[2];
 if ( c==0 ) return 0;
 s[0] = c;
 s[1] = 0;
 return thisFindFwd( Str(), s, 1, doIgnoreCase, nOcc );
}

unsigned gString::FindBack (gString& sSub, bool doIgnoreCase)
{
 unsigned nOcc;
 return thisFindBack( Str(), sSub.Str(), 1, doIgnoreCase, nOcc );
}

unsigned gString::FindBack (char* s, bool doIgnoreCase)
{
 unsigned nOcc;
 return thisFindBack( Str(), s, 1, doIgnoreCase, nOcc );
}

unsigned gString::FindBack (char c, bool doIgnoreCase)
{
 unsigned nOcc;
 char s[2];
 if ( c==0 ) return 0;
 s[0] = c;
 s[1] = 0;
 return thisFindBack( Str(), s, 1, doIgnoreCase, nOcc );
}

unsigned gString::FindAnyChr (gString& b, bool doIgnoreCase)
{
 unsigned posAny;
 return thisFindAny( (char*)str, b.Str(), doIgnoreCase, posAny );
}

unsigned gString::FindAnyChr (char* s, bool doIgnoreCase)
{
 unsigned posAny;
 return thisFindAny( (char*)str, s, doIgnoreCase, posAny );
}

unsigned gString::FindAnyChr (gString& b, bool doIgnoreCase, unsigned& posAny)
{
 return thisFindAny( (char*)str, b.Str(), doIgnoreCase, posAny );
}

unsigned gString::FindAnyChr (char* s, bool doIgnoreCase, unsigned& posAny)
{
 return thisFindAny( (char*)str, s, doIgnoreCase, posAny );
}

unsigned gString::FindExcept (gString& sExcept, bool doIgnoreCase)
{
 ;
 // Finds any character except those in string 'sExcept'

 gString sTemp( sExcept );
 if ( doIgnoreCase ) sTemp.UpString();
 return thisFindExcept( Str(), sTemp.Str(), 1, doIgnoreCase );
}

unsigned gString::FindExcept (char* s, bool doIgnoreCase)
{
 gString sTemp( s );
 if ( doIgnoreCase ) sTemp.UpString();
 return thisFindExcept( Str(), sTemp.Str(), 1, doIgnoreCase );
}

unsigned gString::AddString (gString& a)
{
 if ( a.IsEmpty() ) return 0;
 return Add( a.Str() );
}

void gString::Copy (gString& copy)
{
 Set( copy.Str() );
}

gString& gString::CopyFromTo (gString& copy, unsigned startPos, unsigned endPos)
{
 unsigned i, k, len = copy.Length();

 SetEmpty();  // (previously we used thisDelete : but the result should be an empty string, not really null!)
 if ( startPos==0 ) startPos = 1;
 if ( endPos==0 ) endPos = len;
 if ( endPos>len ) endPos = len;
 if ( startPos>endPos ) return *this;

 thisCopy( (t_uchar*)"\0", 1+endPos-startPos );

 for (i=startPos, k=0; i<=endPos; i++) {
     str[k++] = copy[i];
 }
 str[k] = 0;
 doRevalidateSize = true;

 return *this;
}

unsigned gString::Delete (unsigned startPos, unsigned endPos)
{
 unsigned
     oldSize=size,
     i, k,
     p0=startPos==0?1:startPos,
     p1=endPos==0?size:gMIN(endPos,size);

 if ( startPos==0 && endPos==0 ) {
     SetEmpty();
     return oldSize;
 }

 if ( p0>size ) return 0;

 gString sTemp;
 for (i=1, k=0; i<p0; i++, k++) sTemp.Add( (char)str[k] );
 for (i=p1+1, k=p1; i<=size; i++, k++) sTemp.Add( (char)str[k] );

 Copy( sTemp );
 ASSERTION(oldSize>=size,"oldSize>=size");
 oldSize -= size;

 return oldSize;
}

unsigned gString::Insert (gString& copy, unsigned startPos)
{
 unsigned p0=startPos==0?1:startPos;
 gString sTemp;
 gString sCopy( Str() );

 if ( p0>1 ) {
     sTemp = CopyFromTo( sCopy, 1, p0-1 );
 }
 sTemp.AddString( copy );
 sTemp += CopyFromTo( sCopy, p0 );
 *this = sTemp;
 return 0;
}

gString& gString::operator= (gString& copy)
{
 Copy( copy );
 return *this;
}

gString& gString::operator= (char*s)
{
 ASSERTION(s!=nil,"s!=nil");
 gString copy( s );
 Copy( copy );
 return *this;
}

gString& gString::operator= (char c)
{
 char s[2];
 s[0] = c;
 s[1] = 0;
 Set( s );
 return *this;
}

gString& gString::operator= (int v)
{
 char s[40];
 snprintf(s,40,"%d",v);
 Set( s );
 return *this;
}

gString& gString::operator+ (gString& copy)
{
 char* s = copy.Str();
 Add( s );
 return *this;
}

gString& gString::operator+= (gString& copy)
{
 AddString( copy );
 return *this;
}

gStorage* gString::NewObject ()
{
 gString* a = new gString( str );
 return a;
}

t_uchar* gString::ToString (t_uchar* uBuf)
{
 return gStringGeneric::ToString( uBuf );
}

bool gString::SaveGuts (FILE* f)
{
 return gStringGeneric::SaveGuts( f );
}

bool gString::RestoreGuts (FILE* f)
{
 return gStringGeneric::RestoreGuts( f );
}

void gString::Show (bool doShowAll)
{
 printf("%s%s%s",
	doShowAll?"\"":"\0",
	str==NULL?"\0":(char*)str,
	doShowAll?"\"":"\0");
}
////////////////////////////////////////////////////////////
// gString Protected methods
////////////////////////////////////////////////////////////
int gString::thisMatch (char* s1, char* s2, bool doIgnoreCase)
{
 if ( s1==nil || s2==nil ) return -2;
 if ( doIgnoreCase==false ) {
     return strcmp(s1,s2);
 }
 gString uStr1( s1 );
 gString uStr2( s2 );
 uStr1.UpString();
 uStr2.UpString();
 return strcmp( uStr1.Str(), uStr2.Str() );
}

unsigned gString::thisFind (char* s,
			    char* sub,
			    unsigned startPos,
			    bool doIgnoreCase)
{
 char* ptr;

 if ( s==nil || sub==nil ) return 0;
 ASSERTION(startPos>=1,"startPos>=1");

 ptr = strstr( s+startPos-1, sub );

 if ( doIgnoreCase==false ) {
     if ( ptr!=0 ) return ptr-s+startPos;
     return 0;
 }

 gString uStr( s );
 s = uStr.Str();  //Important for using ptr-s+startPos formula!
 gString uSub( sub );
 uStr.UpString();
 uSub.UpString();

 ptr = strstr( s, uSub.Str() );
 if ( ptr!=0 ) return ptr-s+startPos;
 return 0;
}

unsigned gString::thisFindAny (char* s,
			       char* strAny,
			       bool doIgnoreCase,
			       unsigned& posAny)
{
 t_uchar uChr, fChr;
 unsigned i, k, uLen, aLen;
 char* uStr;
 char* aStr;

 posAny = 0;
 if ( s==nil || strAny==nil ) return 0;
 gString sAll( s );
 gString sAny( strAny );
 if ( doIgnoreCase ) {
     sAll.UpString();
     sAny.UpString();
 }
 uLen = sAll.Length();
 uStr = sAll.Str();
 aLen = sAny.Length();
 aStr = sAny.Str();
 for (k=0; k<uLen; k++) {
     uChr = (t_uchar)uStr[k];
     for (i=0; i<aLen; ) {
	 fChr = (t_uchar)aStr[i];
	 i++;
	 if ( uChr==fChr ) {
	     posAny = i;
	     return k+1;
	 }
     }
 }
 return 0;
}

unsigned gString::thisFindFwd (char* s,
			       char* sub,
			       unsigned startPos,
			       bool doIgnoreCase,
			       unsigned& nOcc)
{
 if ( s==nil || sub==nil ) return 0;
 ASSERTION(startPos==1,"startPos==1");

 gString uStr( s );
 gString uSub( sub );
 if ( doIgnoreCase ) {
     uStr.UpString();
     uSub.UpString();
 }
 return thisFindFwdOcc( uStr.Str(), uSub.Str(), nOcc );
}

unsigned gString::thisFindFwdOcc (char* s,
				  char* sub,
				  unsigned& nOcc)
{
 unsigned i, k;
 unsigned posFirst=0;
 unsigned pos=0, aPos, endPos=Length();

 nOcc = 0;
 for (i=1, k=0; i<=endPos; i++) {
     aPos = thisFind( s+k, sub, 1, false );
     if ( aPos>0 ) {
	 pos = k + aPos;
	 if ( posFirst==0 ) posFirst = pos;
         nOcc++;
         k = pos;
         i = k+1;
     }
 }
 return posFirst;
}

unsigned gString::thisFindBack (char* s,
				char* sub,
				unsigned startPos,
				bool doIgnoreCase,
				unsigned& nOcc)
{
 if ( s==nil || sub==nil ) return 0;
 ASSERTION(startPos==1,"startPos==1");

 gString uStr( s );
 gString uSub( sub );
 if ( doIgnoreCase ) {
     uStr.UpString();
     uSub.UpString();
 }
 return thisFindBackOcc( uStr.Str(), uSub.Str(), nOcc );
}

unsigned gString::thisFindBackOcc (char* s,
				   char* sub,
				   unsigned& nOcc)
{
 unsigned i, k;
 unsigned pos=0, aPos, endPos=Length();

 nOcc = 0;
 for (i=1, k=0; i<=endPos; i++) {
     aPos = thisFind( s+k, sub, 1, false );
     //printf("DBG: Check: F(%s,'%s')=%u, k=%u\n",s+k,sub,aPos,k);
     if ( aPos==0 ) return pos;
     pos = k + aPos;
     nOcc++;
     k = pos;
     i = k;
 }
 //printf("DBG: Check: F(%s,'%s')=%u, k=%u (FINAL, i=%u, pos=%u, [%s])\n",s+k,sub,aPos,k,i,pos,s+pos-1);
 return pos;
}

unsigned gString::thisFindExcept (char* s,
				  char* exceptStr,
				  unsigned startPos,
				  bool doIgnoreCase)
{
 // Finds any character except those in 'exceptStr'
 // Returns 0 if only characters in 'exceptStr' were found
 bool didFound;
 unsigned i=0, k, excLen=strlen(exceptStr);
 char c;

 gString sUp( s );
 if ( doIgnoreCase ) sUp.UpString();
 char* sAll = sUp.Str();

 ASSERTION(startPos==1,"startPos==1");

 for (i=startPos-1; (c = sAll[i])!=0; ) {
     didFound = false;
     for (k=0; k<excLen && didFound==false; k++) {
	 didFound = c==exceptStr[k];
     }
     i++;
     if ( didFound==false ) return i;
 }
 return 0;
}
////////////////////////////////////////////////////////////

