#ifndef gSTRING_X_H
#define gSTRING_X_H

#include "gstorage.h"
////////////////////////////////////////////////////////////
class gStringGeneric : public gStorage {
public:
    // Public enums
    enum eStrStrategy {
	e_StrDefault,
	e_StrAlloc,
	e_StrPreAlloc
    };

    virtual ~gStringGeneric () ;

    // Get methods
    virtual bool IsEmpty () {
	return thisSize()==0;
    }

    virtual unsigned Length () {
	return thisSize();
    }

    virtual t_uchar GetUChar (unsigned idx) ;
    virtual char GetChar (unsigned idx) {
	return (char)GetUChar(idx);
    }

    virtual char* Str () ;
    virtual t_uchar* UStr () ;

    virtual int MinimumStorageSize () {
	return -1;
    }
    virtual int MaximumStorageSize () {
	return -1;
    }

    // Set methods
    virtual void Reset () ;
    virtual void SetEmpty () ;
    void Set (char* s) ;
    void Set (t_uchar* s) ;
    void Set (int v) ;
    void Set (unsigned v) ;

    unsigned Add (char c) ;
    unsigned Add (t_uchar c) ;
    unsigned Add (char* s) ;
    unsigned Add (t_uchar* s) ;
    unsigned Add (int v) ;
    unsigned Add (unsigned v) ;

    virtual void UpString () ;
    virtual void DownString () ;

    // Operators
    t_uchar& operator[] (int index) ;

    // Save/etc methods
    virtual t_uchar* ToString (t_uchar* uBuf) ;
    virtual bool SaveGuts (FILE* f) ;
    virtual bool RestoreGuts (FILE* f) ;

protected:
    gStringGeneric (eStorage kind,
		    eStrStrategy aStgy,
		    t_uchar* s,
		    unsigned aMaxSize) ;

    unsigned size, preMaxSize, maxSize;
    eStrStrategy stgy;
    t_uchar* str;
    bool doRevalidateSize;
    static t_uchar myChrNul;

    // Protected methods
    void thisPreAllocate (t_uchar* s, unsigned toSize) ;
    void thisDelete () ;
    int thisCopy (t_uchar* s, unsigned len) ;
    bool thisIndex (unsigned& idx) ;
    bool thisIsValidIndex (int index) ;
    virtual unsigned thisSize () ;

private:
    // Operators,empty
    gStringGeneric (gStringGeneric& ) ;
    gStringGeneric& operator= (gStringGeneric& ) ;
};
////////////////////////////////////////////////////////////
class gString : public gStringGeneric {
public:
    gString () ;
    gString (gString& copy) ;
    gString (char* s) ;
    gString (t_uchar* s) ;
    gString (char c) ;
    gString (unsigned nBytes, char c) ;
    virtual ~gString () ;

    // Get methods
    bool Match (gString& copy, bool doIgnoreCase=false) ;
    bool Match (char* s, bool doIgnoreCase=false) ;

    unsigned Find (gString& sSub, bool doIgnoreCase=false) ;
    unsigned Find (char* s, bool doIgnoreCase=false) ;
    unsigned Find (char c, bool doIgnoreCase=false) ;

    unsigned Find (gString& sSub, unsigned& nOcc, bool doIgnoreCase=false) ;
    unsigned Find (char* s, unsigned& nOcc, bool doIgnoreCase=false) ;
    unsigned Find (char c, unsigned& nOcc, bool doIgnoreCase=false) ;

    unsigned FindBack (gString& sSub, bool doIgnoreCase=false) ;
    unsigned FindBack (char* s, bool doIgnoreCase=false) ;
    unsigned FindBack (char c, bool doIgnoreCase=false) ;

    unsigned FindAnyChr (gString& b, bool doIgnoreCase=false) ;
    unsigned FindAnyChr (char* s, bool doIgnoreCase=false) ;
    unsigned FindAnyChr (gString& b, bool doIgnoreCase, unsigned& posAny) ;
    unsigned FindAnyChr (char* s, bool doIgnoreCase, unsigned& posAny) ;

    unsigned FindExcept (gString& sExcept, bool doIgnoreCase=false) ;
    unsigned FindExcept (char* s, bool doIgnoreCase=false) ;

    // Set methods
    virtual unsigned AddString (gString& a) ;
    virtual void Copy (gString& copy) ;

    virtual gString& CopyFromTo (gString& copy, unsigned startPos=0, unsigned endPos=0) ;
    virtual gString& CopyFromToStr (char* s, unsigned startPos=0, unsigned endPos=0) {
	gString copy( s );
	return CopyFromTo( copy, startPos, endPos );
    }

    virtual unsigned Delete (unsigned startPos=0, unsigned endPos=0) ;

    virtual unsigned Insert (gString& copy, unsigned startPos=0) ;
    virtual unsigned InsertStr (char* s, unsigned startPos=0) {
	gString copy( s );
	return Insert( copy, startPos );
    }

    virtual bool Trim () ;
    virtual bool TrimLeft () ;
    virtual bool TrimRight () ;

    // Operators
    gString& operator= (gString& copy) ;
    gString& operator= (char* s) ;
    gString& operator= (char c) ;
    gString& operator= (int v) ;

    gString& operator+ (gString& copy) ;

    gString& operator+= (gString& copy) ;

    // Save/etc methods
    virtual gStorage* NewObject () ;
    virtual t_uchar* ToString (t_uchar* uBuf) ;
    virtual bool SaveGuts (FILE* f) ;
    virtual bool RestoreGuts (FILE* f) ;

    // Show methods
    virtual void Show (bool doShowAll=true) ;

protected:
    // Other methods
    int thisMatch (char* s1, char* s2, bool doIgnoreCase) ;

    unsigned thisFind (char* s,
		       char* sub,
		       unsigned startPos,
		       bool doIgnoreCase) ;

    unsigned thisFindAny (char* s,
			  char* strAny,
			  bool doIgnoreCase,
			  unsigned& posAny) ;

    unsigned thisFindFwd (char* s,
			  char* sub,
			  unsigned startPos,
			  bool doIgnoreCase,
			  unsigned& nOcc) ;

    unsigned thisFindFwdOcc (char* s,
			     char* sub,
			     unsigned& nOcc) ;

    unsigned thisFindBack (char* s,
			   char* sub,
			   unsigned startPos,
			   bool doIgnoreCase,
			   unsigned& nOcc) ;

    unsigned thisFindBackOcc (char* s,
			      char* sub,
			      unsigned& nOcc) ;

    unsigned thisFindExcept (char* s,
			     char* exceptStr,
			     unsigned startPos,
			     bool doIgnoreCase) ;

private:
    // Operators,empty
    gString (gStringGeneric& ) ; //empty
};
////////////////////////////////////////////////////////////
#endif //gSTRING_X_H

