// gstringext.cpp -- Version 1.0

#include <string.h>
#include "gstringext.h"
#include "garg.h"

////////////////////////////////////////////////////////////
// Static members
// ---------------------------------------------------------
char sHtmlSymbol::sBuf[50];
////short sUnCodeSymbol::mySymNameLen=60;
char sUnCodeSymbol::symBufExt[257];
char gStrControl::strPathUnicode[ 512 ];
sUnCodeSymbol gStrControl::mySym;
gStrControl gStrControl::myself;

////////////////////////////////////////////////////////////
gDateDay::gDateDay (t_uint32 aDd)
    : dd( aDd )
{
}

gDateDay::~gDateDay ()
{
}

t_uint16 gDateDay::GetYear ()
{
 return dd / 10000;
}

t_uint16 gDateDay::GetMonth ()
{
 return (dd % 10000)/100;
}

t_uint16 gDateDay::GetDay ()
{
 return dd%100;
}

bool gDateDay::SetDay (t_uint32 aDd)
{
 dd = aDd;
 return IsValid();
}

gStorage* gDateDay::NewObject ()
{
 gDateDay* a = new gDateDay( dd );
 return a;
}

bool gDateDay::SaveGuts (FILE* f)
{
 if ( CanSave( f )==false ) return false;
 // e.g.: 20031231 (last day of 2003)
 return fprintf( f, "%08lu", (unsigned long)dd );
}

bool gDateDay::RestoreGuts (FILE* f)
{
 unsigned long lVal;
 if ( CanRestore( f )==false ) return false;
 if ( fscanf( f, "%lu", &lVal )<1 ) return false;
 dd = lVal;
 return true;
}
////////////////////////////////////////////////////////////
sTcbBin::sTcbBin ()
    : cLevel( '0' ),
      revisionRef( 1 ),  // Revision 1, draft
      cNormRef( 'D' ),
      cpCode( 0 ),
      cLinearInd( 'a' ),
      nSymb( 0 ),
      linearTbl( nil ),
      bigTbl( nil )
{
 unsigned i;
 linearTbl = new t_gUniPage[ 256 ];
 for (i=0; i<256; i++) linearTbl[i] = 0;
 strcpy(verMajor,"TCB1A");
}

sTcbBin::~sTcbBin ()
{
 delete[] linearTbl;
 delete[] bigTbl;
}

bool sTcbBin::IsLinear ()
{
 return cLinearInd=='a' || cLinearInd=='b';
}

int sTcbBin::ReadFromStream (FILE* f)
{
 int error;
 t_uint32 aUD4;
 t_uint32 sIdx, sMax;
 t_uint16 aMagicOffset;
 t_uchar aBuf[100];

 if ( f==nil ) return -1;

 error = fread(verMajor,6,1,f);
 if ( error<1 ) return 1;
 error = fread(&aUD4,4,1,f);
 if ( error<1 ) return 1;
 aMagicOffset = (t_uint16)( FROM_UD4(aUD4) );
 if ( aMagicOffset!=72 ) return 2;
 error = fscanf(f,"%c",&cLevel);
 if ( error<1 ) return 1;

 error = fread(&aUD4,4,1,f);
 if ( error<1 ) return 1;
 revisionDate.SetDay( FROM_UD4(aUD4) );

 error = fread(&revisionRef,1,1,f);
 if ( error<1 ) return 1;

 error = fread(&cNormRef,1,1,f);
 if ( error<1 ) return 1;

 error = fread(&sDescTag,13,1,f);
 if ( error<1 ) return 1;

 memset(aBuf,0x0,100);
 error = fread(aBuf,34,1,f);
 if ( error<1 ) return 1;
 sDescStr = (char*)aBuf;

 error = fread(&aUD4,4,1,f);
 if ( error<1 ) return 1;
 cpCode = FROM_UD4(aUD4);

 error = fread(&cLinearInd,1,1,f);
 if ( error<1 ) return 1;

 memset(aBuf,0x00,10);
 error = fread(aBuf,3,1,f);
 if ( error<1 ) return 1;
 if ( strncmp((char*)(aBuf+1),"  ",2)!=0 ) return 2;

 error = fread(&aUD4,4,1,f);
 if ( error<1 ) return 1;
 nSymb = FROM_UD4(aUD4);

 ASSERTION(linearTbl!=nil,"linearTbl!=nil");
 sMax = nSymb>256 ? 256 : nSymb;
 for (sIdx=1; sIdx<sMax; sIdx++) {
     error = fread(&aUD4,4,1,f);
     if ( error<1 ) return 1;
     linearTbl[sIdx] = FROM_UD4( aUD4 );
 }

 return 0;
}

int sTcbBin::ReadFromFile (char* fName)
{
 int error;
 FILE* f;
 if ( fName==nil || fName[0]==0 ) return -1;
 f = fopen(fName,"rb");
 error = ReadFromStream( f );
 printf("sTcbBin::ReadFromFile: %s (error=%d)\n",
	fName,
	error);
 if ( f!=nil ) fclose( f );
 return error;
}
////////////////////////////////////////////////////////////
sUnCodeSymbol::~sUnCodeSymbol ()
{
 delete symbolName;
 if ( strNumber ) free( strNumber );
 if ( strNote ) free( strNote );
 if ( strLocale ) free( strLocale );
}

bool sUnCodeSymbol::IsValid ()
{
 return
     (cIsOpenClose=='Y' || cIsOpenClose=='N') &&
     (twoLetterCat[0]>='A' && twoLetterCat[0]<='Z');
}

void sUnCodeSymbol::SetCode (t_gUniPage aCode)
{
 bool isAlpha;

 uniCode = aCode;
 extUtMask = 0;
 if ( aCode>0xFF ) return;
 if ( aCode>='0' && aCode<='9' ) {
     extUtMask = EXTUMASK_ISDIGIT | EXTUMASK_ISPRINT;
     return;
 }
 if ( aCode==' ' || aCode=='\t' || aCode=='\f' ||
      aCode=='\n' || aCode=='\r' || aCode=='\v' ) {
     extUtMask = EXTUMASK_ISSPACE | EXTUMASK_ISPRINT;
     return;
 }
 ASSERTION(twoLetterCat[0]!=0,"twoLetterCat[0]!=0");
 isAlpha = twoLetterCat[0]=='L';
 extUtMask = isAlpha;
 extUtMask |= (aCode>=' ' && aCode<127) ? EXTUMASK_ISPRINT : 0x0;
 // Note: other values above ASCII 127 decimal are not
 // parsed since it is deterministic, as follow:
 if ( aCode>=0x00A1 ) extUtMask |= EXTUMASK_ISPRINT;
}

int sUnCodeSymbol::SaveBin (FILE* f)
{
 unsigned i;
 t_uint16 aUD2;
 t_uint32 aUD4;

 if ( f==nil ) return -1;
 if ( IsValid()==false ) return -1;

 aUD4 = TO_UD4(uniCode);
 fwrite(&aUD4,4,1,f);

 //for (i=0; i<257; i++) symBufExt[i] = 0;
 memset(symBufExt,0x00,257);

 //if ( myStrSymNameLen>0 ) {
 ASSERTION(symbolName!=nil,"symbolName!=nil");
 strncpy( symBufExt, symbolName->Str(), myStrSymNameLen-1 );
 fwrite(symBufExt,myStrSymNameLen,1,f);
 //}
 fwrite(twoLetterCat,2,1,f);
 fprintf(f,"0");
 fwrite(thrLetterStyle,4,1,f); // 3octets+0x00

 aUD4 = TO_UD4(ucCompat);
 fwrite(&aUD4,4,1,f);

 for (i=0; i<10; i++) symBufExt[i] = 0;
 if ( strNumber ) {
     strncpy( symBufExt, strNumber, 9 );
 }
 fwrite(symBufExt,10,1,f);
 fprintf(f,"%c",cIsOpenClose);

 aUD4 = TO_UD4(extEqClose);
 fwrite(&aUD4,4,1,f);

 for (i=0; i<18; i++) symBufExt[i] = 0;
 if ( strNote ) {
     strncpy( symBufExt, strNote, 16 );
 }
 fwrite(symBufExt,18,1,f); // 16octets+0x00+0x00

 aUD4 = TO_UD4(ucUpper);
 fwrite(&aUD4,4,1,f);

 aUD4 = TO_UD4(ucDown);
 fwrite(&aUD4,4,1,f);

 for (i=0; i<17; i++) symBufExt[i] = 0;
 fwrite(symBufExt,4,1,f);
 fwrite(symBufExt,2,1,f);

 aUD4 = 0xFFFFFFFF;
 fwrite(&aUD4,4,1,f);
 ASSERTION((aUD4>>30)==0x3,"dummy-test");

 for (i=0; i<61; i++) symBufExt[i] = 0;
 if ( strLocale ) {
     strncpy( symBufExt, strLocale, sizeof(symBufExt)-1 );
 }
 fwrite(symBufExt,61,1,f); // 60octets+0x00

 fprintf(f,"%c",refVersLetter);

 aUD2 = TO_UD2(extUtMask);
 return fwrite(&aUD2,2,1,f)<1;
}

int sUnCodeSymbol::ReadBin (FILE* f)
{
 int error;
 t_uint16 aUD2;
 t_uint32 aUD4;

 if ( f==nil ) return -1;

 error = fread(&aUD4,4,1,f);
 DBGPRINT_MIN("DBG: sUnCodeSymbol::ReadBin error=%d (network-order UD4: %u)\n",error,FROM_UD4(aUD4));
 if ( error<1 ) return error;
 uniCode = FROM_UD4(aUD4);

 memset(symBufExt,0x00,257);

 delete symbolName;
 symbolName = nil;

 error = fread(symBufExt,myStrSymNameLen,1,f);
 if ( error<1 ) return 1;
 if ( symBufExt[myStrSymNameLen-1]!=0x00 ) return 2;
 symbolName = new gString( (char*)symBufExt );

 error = fread(twoLetterCat,3,1,f); // 3octet+0x00
 if ( error<1 ) return 1;
 if ( twoLetterCat[2]!='0' ) return 2;
 twoLetterCat[2] = 0;

 error = fread(thrLetterStyle,4,1,f); // 4octet+0x00
 if ( error<1 ) return 1;
 if ( thrLetterStyle[3]!=0x00 ) return 2;

 error = fread(&aUD4,4,1,f);
 if ( error<1 ) return 1;
 ucCompat = FROM_UD4(aUD4);

 memset(symBufExt,0x00,11);
 error = fread(symBufExt,10,1,f);
 if ( error<1 ) return 1;
 if ( symBufExt[9]!=0x00 ) return 2;
 strNumber = strdup( (char*)symBufExt );

 error = fscanf(f,"%c",&cIsOpenClose );
 if ( error<1 ) return 1;

 error = fread(&aUD4,4,1,f);
 if ( error<1 ) return 1;
 extEqClose = FROM_UD4(aUD4);

 memset(symBufExt,0x00,19);
 error = fread(symBufExt,18,1,f); // 16octet+0x00+0x00
 strNote = strdup( (char*)symBufExt );

 error = fread(&aUD4,4,1,f);
 if ( error<1 ) return 1;
 ucUpper = FROM_UD4(aUD4);

 error = fread(&aUD4,4,1,f);
 if ( error<1 ) return 1;
 ucDown = FROM_UD4(aUD4);

 error = fread(symBufExt,4,1,f); // 4 pad nulls
 if ( error<1 ) return 1;
 error = fread(symBufExt,2,1,f); // 2 pad nulls
 if ( error<1 ) return 1;

 error = fread(&aUD4,4,1,f);
 if ( error<1 ) return 1;
 if ( aUD4!=0xFFFFFFFF ) return 2;

 memset(symBufExt,0x00,61);
 error = fread(symBufExt,61,1,f); // 60octet+0x00
 if ( error<1 || symBufExt[60]!=0x00 ) return 1;
 // String symbol name Locale
 strLocale = symBufExt[ 0 ] ? strdup( (char*)symBufExt ) : nil;

 error = fread(&refVersLetter,1,1,f);
 if ( error<1 ) return 1;

 error = fread(&aUD2,2,1,f);
 if ( error<1 ) return 1;
 extUtMask = FROM_UD2(aUD2);

 return IsValid() ? 0 : 2;
}
////////////////////////////////////////////////////////////
char* sHtmlSymbol::GetHtmlStr ()
{
 memset(sBuf,0x00,50);
 if ( IsOk()==false ) return sBuf;
 if ( IsNormal() ) {
     strcpy( sBuf, sTransHtm.Str() );
     return sBuf;
 }
 sprintf( sBuf, "&%s;", sTransHtm.Str() );
 return sBuf;
}
////////////////////////////////////////////////////////////
sUnCodeBin::sUnCodeBin ()
    : n( 0 ),
      mapTab( nil ),
      verHtml( '3' ),
      descStg( e_Desc60ByteDef ),
      revisionRef( THB_VERS_REVISION ),
      cExtTbl( '2' ),
      nHtmlTrans( 0 ),
      mapHtm( nil ),
      cHtmAsciiStt( 0x0 )
{
 short ix;
 for (ix=0; ix<=16; ix++) revReferenceStr[ix] = 0;
 sprintf( verMajor, "THB_Z" );
 strcpy( revReferenceStr, "Henrique Moreira" );
 strcpy( strHtmlUnmap, "[hs." );
}

sUnCodeBin::~sUnCodeBin ()
{
 delete[] mapTab;
 delete[] mapHtm;
}

int sUnCodeBin::GetMajorVersion ()
{
 if ( verMajor[3]<'1' ) return -1;
 return (int)(verMajor[3]-'1');
}

char sUnCodeBin::GetDescStgChr ()
{
 switch ( descStg ) {
 case e_DescNone:
     return 'a';
 case e_Desc40Byte:
     return 'b';
 case e_Desc60ByteDef:
     return 'c';
 case e_Desc80Byte:
     return 'd';
 case e_Desc255Byte:
     return 'e';
 default:
     break;
 }
 ASSERTION_FALSE("descStg");
 return '\0';
}

bool sUnCodeBin::IsFlat (t_uchar& chr)
{
 chr = ';';
 return chr==';'; // Fixedly Flat!
}

bool sUnCodeBin::IsFlatOk ()
{
 unsigned i, nElem=n<256?n:256;
 t_uchar chr;
 t_gUniPage lastCode=0x0, thisCode;

 if ( IsFlat(chr)==false ) return false;
 // Check:
 //  a) sequential symbols in 'mapTab'
 //  b) no gaps between 0x00 and 0xFF
 ASSERTION(mapTab!=nil,"mapTab!=nil");
 for (i=2; i<=nElem; i++) {
     thisCode = mapTab[i].uniCode;
     if ( thisCode<=lastCode ) return false;
     lastCode = thisCode;
 }
 return true;
}

t_uint32 sUnCodeBin::FindCode (t_gUniPage aCode)
{
 unsigned i;
 for (i=1; i<=n; i++) {
     if ( aCode==mapTab[i].uniCode ) return i;
 }
 return 0;
}

char sUnCodeBin::GetLetterCompat (t_gUniPage aCode)
{
 unsigned i;
 t_gUniPage ucCompat;
 ASSERTION(mapTab!=NULL,"mapTab!=NULL");
 ASSERTION(aCode<N(),"aCode<N()");
 i = FindCode(aCode);
 ASSERTION(i==aCode,"revise assert"); // Note: this can be not-linear!
 ucCompat = mapTab[i].ucCompat;
 if ( ucCompat==0 ) {
     return mapTab[i].IsLetter() && i<127 ? i : '*';
 }
 if ( ucCompat>=127 ) return 0;
 return ucCompat;
}

bool sUnCodeBin::HasHtml ()
{
 bool isOk = verHtml>'0' && cExtTbl>'2';
 DBGPRINT("DBG:HasHtml?%d,cHtmAsciiStt=%03u\n",isOk,(unsigned)cHtmAsciiStt);
 return isOk==true && cHtmAsciiStt>'\0';
}

int sUnCodeBin::AllocateSymbols (int nSymbols)
{
 if ( nSymbols<=0 ) {
     n = 0;
     delete[] mapTab;
     mapTab = nil;
     return 0;
 }
 ASSERTION(mapTab==nil,"mapTab==nil");
 n = (unsigned)nSymbols;
 mapTab = new sUnCodeSymbol[ n+1 ];
 ASSERTION(mapTab!=nil,"mapTab!=nil");
 return nSymbols;
}

int sUnCodeBin::SetVersion (t_uchar aMajorVersion, t_uchar aMinorAlpha)
{
 if ( aMajorVersion<'1' || aMajorVersion>'9' ) return -1;
 if ( aMinorAlpha<'A' || aMinorAlpha>'Z' ) return -1;
 verMajor[3] = aMajorVersion;
 verMajor[4] = aMinorAlpha;
 return 0;
}

int sUnCodeBin::SetNTables (bool hasHtmlTrans, int nExtTbl)
{
 cExtTbl = hasHtmlTrans?'3':'2';
 // Other values for cExtTbl are for future use
 return nExtTbl>=0;
}

int sUnCodeBin::SaveToFile (char* fName)
{
 FILE* f;
 int error;

 if ( verMajor[4]<'A' ) return -2;
 if ( mapTab==nil ) return -3;
 if ( IsFlatOk()==false ) return -4;

 if ( fName==NULL || fName[0]==0 ) return -1;
 f = fopen(fName,"wb");
 error = SaveToStream( f );
 if ( f!=NULL ) fclose( f );
 return error;
}

int sUnCodeBin::SaveToStream (FILE* f)
{
 int error;
 int scanCode;
 unsigned i;
 t_uint16 aUD2;
 t_uint32 aUD4;
 t_uint32 nStrCompat=0;
 t_uchar chr;

 if ( f==NULL ) return -1;

 if ( fprintf(f,"%s",verMajor)<1 ||
      fprintf(f,"%c%c0",0x00,verHtml)<1 ) {
     return -1;
 }

 fprintf(f,"%c %c",GetDescStgChr(),0x00);
 aUD4 = TO_UD4(revisionDate.GetValue32());
 fwrite(&aUD4,4,1,f);

 fwrite(&revisionRef,1,1,f);
 fwrite(&revReferenceStr,17,1,f);
 IsFlat( chr );
 fprintf(f,"Si.%c%c",0x00,chr); // Usually "Si._;"

 aUD2 = TO_UD2(3); // Fixedly with HTML-TRANS
 fwrite(&aUD2,2,1,f);

 aUD4 = TO_UD4(n);
 error = fwrite(&aUD4,4,1,f);
 if ( error<1 ) return 1;

 for (i=1, error=0; i<=n; i++) {
     if ( mapTab[i].SaveBin( f )!=0 ) error = 1;
     ASSERTION(mapTab[i].strComptb,"mapTab[i].strCompat");
     nStrCompat += mapTab[i].strComptb[ 0 ]!=0;
 }
 if ( error!=0 ) return 1;

 // TABLE OF COMPATIBILITY NOTES (section four)
 // The number of non empty 'strComptb' is nStrCompat
 aUD4 = TO_UD4(nStrCompat);
 fwrite(&aUD4,4,1,f);

 for (i=1, error=0; i<=n; i++) {
     if ( mapTab[i].strComptb[ 0 ] ) {
	 // Write that strComptb
	 aUD4 = TO_UD4(mapTab[i].uniCode);
	 fwrite(&aUD4,4,1,f);
	 gString sNote( mapTab[i].strComptb );
	 fprintf(f,sNote[1]=='<'?"<>":"  ");
	 memset(mapTab[0].symBufExt,0x00,257);
	 strncpy(mapTab[0].symBufExt,sNote.Str(),36);
	 scanCode = fwrite(mapTab[0].symBufExt,36,1,f);
	 if ( scanCode<1 ) error = 1;
     }
 }
 if ( error!=0 ) return 1;

 // TABLE OF HTML-TRANS (section five)
 aUD4 = TO_UD4( nHtmlTrans );
 error = fwrite(&aUD4,4,1,f);
 if ( error<1 ) return 1;

 if ( HasHtml() ) {
     error = fprintf(f,"HTML%-2s;%c%c%c%c%c%c~",
		     strHtmlUnmap,
		     0x20, // 'Regular field'
		     0x00, // Pad null
		     cHtmAsciiStt,
		     '%',
		     '%',
		     '%');
     if ( error<1 ) return 1;
     unsigned count, aLen;
     i = (unsigned)cHtmAsciiStt;
     for (count=0; count<nHtmlTrans; count++, i++) {
	 aLen = mapHtm[i].sTransHtm.Length();
	 if ( aLen>9 ) return 2;
	 memset(mapTab[0].symBufExt,0x00,10);
	 strcpy( mapTab[0].symBufExt, mapHtm[i].sTransHtm.Str() );
	 DBGPRINT("DBG: write[%u]: {%s}\n",i,mapTab[0].symBufExt);
	 error = fwrite(mapTab[0].symBufExt,9,1,f);
	 if ( error<1 ) return 1;
	 error = fwrite(&mapHtm[i].cCompNote,1,1,f);
	 if ( error<1 ) return 1;
     }
 }

 return 0;
}

int sUnCodeBin::ReadFromFile (char* fName)
{
 // Returns 0 on success,
 // -3 if mapTab is not null
 // -1 if name or file not opened,
 //  2 if file is not THB or has invalid fields,
 //  1 if file too short or other.
 ;
 FILE* f;
 int error;

 if ( mapTab!=nil ) return -3;
 if ( fName==NULL || fName[0]==0 ) return -1;
 f = fopen(fName,"rb");
 error = ReadFromStream( f );
 //printf("sUnCodeBin::ReadFromFile: %s (error=%d)\n",fName,error) : DBG
 if ( f!=NULL ) fclose( f );
 return error;
}

int sUnCodeBin::ReadFromStream (FILE* f)
{
 int error;
 bool isThisFlat;
 unsigned i;
 t_uint16 nrExtTables;
 t_uchar chr, chr2;
 char bufMe[257];
 t_uint16 aUD2;
 t_uint32 aUD4;
 t_uint32 idx, nStrCompat( 0 );
 t_gUniPage aCode;

 if ( f==NULL ) return -1;

 memset(bufMe,0x00,257);

 error = fread(bufMe,6,1,f);
 if ( error<1 ) return 1;
 if ( strncmp(bufMe,verMajor,3)!=0 ||
      bufMe[5]!=0x00 ) return 2;

 strcpy( verMajor, bufMe );

 error = fscanf(f,"%c%c%c",&verHtml,&cUniLevel,&chr);
 if ( error<3 ) return 1;

 switch ( chr ) {
 case 'a':
     descStg = e_DescNone;
     break;
 case 'b':
     descStg = e_Desc40Byte;
     break;
 case 'c':
     descStg = e_Desc60ByteDef;
     break;
 case 'd':
     descStg = e_Desc80Byte;
     break;
 case 'e':
     descStg = e_Desc255Byte;
     break;
 default:
     return 2;
 }

 error = fscanf(f,"%c%c",&chr,&chr);
 if ( error<2 ) return 2;

 error = fread(&aUD4,4,1,f);
 if ( error<1 ) return 1;

 revisionDate.SetDay( FROM_UD4(aUD4) );

 error = fscanf(f,"%c",&revisionRef);
 if ( error<1 || revisionRef==0xFF ) return 1;

 error = fread(revReferenceStr,17,1,f); // 16octet+0x00

 error = fread(bufMe,4,1,f); // "Si."+0x00
 if ( error<1 ) return 1;
 if ( strncmp(bufMe,"Si.",3)!=0 ) return 2;

 fscanf(f,"%c",&chr);
 IsFlat( chr2 );
 isThisFlat = chr==chr2;

 error = fread(&aUD2,2,1,f);
 nrExtTables = FROM_UD2(aUD2);
 if ( error<1 || nrExtTables<2 ) return 2;
 cExtTbl = nrExtTables;
 ASSERTION(cExtTbl<4,"cExtTbl<4: >=4 for future use");

 error = fread(&aUD4,4,1,f);
 if ( error<1 ) return 1;

 n = FROM_UD4(aUD4);

 // Now read the table
 mapTab = new sUnCodeSymbol[ n+1 ];
 ASSERTION(mapTab!=nil,"mapTab!=nil");

 ////mapTab[0].strSymNameLen = myStrSymNameLen;  //// (short)descStg; // Usually 60

 for (i=0, error=0; i<n; i++) {
     if ( mapTab[i].ReadBin( f )!=0 ) return 1;
     DBGPRINT("DBG:%03ud(0x%02X):%04X\n",i,i,mapTab[i].uniCode);
 }

 error = fread(&aUD4,4,1,f);
 if ( error>=1 && n>1 ) {
     nStrCompat = FROM_UD4(aUD4);
     for (i=1; i<=nStrCompat; i++) {
	 error = fread(&aUD4,4,1,f);
	 if ( error<1 ) continue;
	 aCode = FROM_UD4(aUD4);
	 error = fread(&aUD2,2,1,f); // 2 chars
	 if ( error<1 ) continue;
	 error = fread(mapTab[0].symBufExt,36,1,f);
	 if ( error<1 ) return 1;
	 idx = FindCode( aCode );
	 if ( idx>1 ) {
	     mapTab[idx].strComptb = strdup( (char*)mapTab[0].symBufExt );
	 }
     }
 }

 error = fread(&aUD4,4,1,f);
 if ( error<1 ) return 1;

 nHtmlTrans = FROM_UD4(aUD4);
 ASSERTION(mapHtm==nil,"mapHtm==nil");
 mapHtm = new sHtmlSymbol[ 256 ];
 ASSERTION(mapHtm!=nil,"mapHtm!=nil");

 if ( nHtmlTrans>0 ) {
     memset(bufMe,0x00,257);
     error = fread(bufMe,4,1,f);
     if ( error<1 ) return 1;
     if ( strncmp(bufMe,"HTML",4)!=0 ) return 2;
     error = fread(bufMe,4,1,f);
     if ( error<1 ) return 1;
     bufMe[4] = 0;
     strncpy( strHtmlUnmap, (char*)bufMe, sizeof(strHtmlUnmap) );
     // Semicolon, regular, and null field
     error = fread(bufMe,3,1,f);
     if ( strcmp(bufMe,"; ")!=0 ) return 2;
     fscanf(f,"%c",&cHtmAsciiStt);
     if ( cHtmAsciiStt<0x80 ) return 2;
     error = fread(bufMe,4,1,f);
     if ( error<1 || strncmp(bufMe,"%%%~",4)!=0 ) return 2;
     // Iterate through all 'mapHtm'
     unsigned count;
     i = (unsigned)cHtmAsciiStt;
     for (count=0; count<nHtmlTrans; count++, i++) {
	 error = fread(bufMe,10,1,f);
	 if ( error<1 ) return 1;
	 mapHtm[i].cCompNote = bufMe[9];
	 bufMe[9] = 0;
	 mapHtm[i].sTransHtm = (char*)bufMe;
	 DBGPRINT("DBG:::%u:%-9s!%c\n",i,mapHtm[i].sTransHtm.Str(),mapHtm[i].cCompNote);
     }
     for (i=' '; i<127; i++) {
	 mapHtm[i].SetCompatibleZero();
	 mapHtm[i].sTransHtm.Add( (char)i );
     }
     mapHtm[' '].sTransHtm = (char*)"nbsp";
     mapHtm[' '].SetCompatible( true );
     mapHtm['"'].sTransHtm = (char*)"quot"; // &#34;
     mapHtm['"'].SetCompatible( true );
     mapHtm['&'].sTransHtm = (char*)"amp";  // &#38
     mapHtm['&'].SetCompatible( true );
     mapHtm['<'].sTransHtm = (char*)"lt";   // &#60
     mapHtm['<'].SetCompatible( true );
     mapHtm['>'].sTransHtm = (char*)"gt";   // &#62
     mapHtm['>'].SetCompatible( true );
 }

 error = isThisFlat ? 0 : 16;
 return error;
}
////////////////////////////////////////////////////////////
gStrControl::gStrControl ()
    : lastOpError( 0 ),
      fReport( stderr ),
      isInit( false ),
      useCaseCp( false ),
      doDigConvertRelaxed( false ),
      defIsoPage( 1 ),
      strPathnameThb( nil )
{
 char* strEnv( getenv( "CommonProgramFiles" ) );
 char* strPath( strEnv ? strEnv : (char*)"/usr/share" );

#ifdef gDOS_SPEC
 if ( strEnv==nil || strPath[0]==0 ) {
     // Invent a reasonable Win32 path
     strPath = (char*)"C:\\Program Files\\Common Files";
 }
#endif //gDOS_SPEC
 snprintf( strPathUnicode, sizeof(strPathUnicode)-1, "%s%cunicodeplus", strPath, gSLASHCHR );
 gStorageControl::Self().StaticAlloc( "gStrControl", 3 );
}

gStrControl::~gStrControl ()
{
 delete[] strPathnameThb;
 strPathnameThb = nil;
}

bool gStrControl::IsOkChar (t_uchar c)
{
 return qckValidChr[c]!=0;
}

char* gStrControl::GetPathNameTHB ()
{
 ASSERTION(strPathnameThb,"strPathnameThb");
 return strPathnameThb;
}

int gStrControl::GetFilenameCode (t_gCpCode aCpCode, gString& sCodepage)
{
 unsigned uLen( (strPathUnicode ? strlen( strPathUnicode ) : 0) + 30 );
 gString sTemp( uLen, '\0' );
 if ( aCpCode>899999 ) return -1;
 snprintf( sTemp.Str(), uLen, "%s%ccp%06u.tcb",
	   GetPathUnicode(),
	   gSLASHCHR,
	   aCpCode );
 sCodepage = sTemp;
 return 0;
}

int gStrControl::Compare (char* s1, char* s2, bool doIgnoreCase)
{
 ASSERTION(s1!=nil,"s1!=nil");
 ASSERTION(s2!=nil,"s2!=nil");
 if ( doIgnoreCase==false ) return strcmp( s1, s2 );
 gString uS1( s1 );
 gString uS2( s2 );
 uS1.UpString();
 uS2.UpString();
 return strcmp( uS1.Str(), uS2.Str() );
}

int gStrControl::Run ()
{
 if ( isInit ) {
     DBGPRINT_MIN("DBG: gStrControl isInit'ed: %s\n",strPathUnicode);
 }
 else {
     isInit = Init( strPathUnicode, defIsoPage )==0;
     DBGPRINT_MIN("DBG: gStrControl isInit'ed?%c (%ld): %s\n",ISyORn(isInit),(long)defIsoPage,strPathUnicode);
 }
 return isInit==false;
}

int gStrControl::Init (char* strPath, t_gCpCode aIsoPage)
{
 int lastOpError;

 if ( aIsoPage<1 ) return -1;
 ASSERTION(strPath!=nil,"gStrControl(0)");
 lastOpError = thisInit( strPath, aIsoPage );
 return lastOpError;
}

bool gStrControl::UseCaseCodepage (bool doUseCaseCp)
{
 useCaseCp = doUseCaseCp;
 // if ( doUseCaseCp==false ) return false
 // Read-out codepages, etc...
 return doUseCaseCp;
}

int gStrControl::InitTab (t_gCpCode aIsoPage)
{
 ASSERTION_FALSE("InitTab");
 return 0;
}

int gStrControl::ConvertToUInt32 (char* s, t_uint32& vRes)
{
 // Return 0 on success
 unsigned posErr;
 return
     ConvertToUInt32( s,
		      10,
		      e_DigConvAny,
		      vRes,
		      posErr );
}

int gStrControl::ConvertToInt32 (char* s, t_int32& vRes)
{
 // Return 0 on success
 unsigned posErr;
 return
     ConvertToInt32( s,
		     10,
		     e_DigConvAny,
		     vRes,
		     posErr );
}

int gStrControl::ConvertToUInt32 (char* s,
				  unsigned base,
				  eDigitConv caseSense,
				  t_uint32& vRes,
				  unsigned& posErr)
{
 // Return 0 on success
 unsigned i, len, kU, kD;
 t_uchar chr;
 static t_uchar convTblUp[17]="0123456789ABCDEF";
 static t_uchar convTblDn[17]="0123456789abcdef";
 t_uint64 uRes=0;
 bool doRelax = IsDigConvRelaxed();

 posErr = 0;
 if ( s==nil || base>16 ) return -1;
 gString sVal( s );
 sVal.Trim();
 len = sVal.Length();
 if ( len==0 ) {
     if ( caseSense!=e_DigConvAnyEmpty0 ) return -1;
     vRes = 0;
     return 0;
 }
 uRes = 0;
 for (i=1; i<=len; i++) {
     chr = sVal[i];
     posErr = i;
     for (kU=0; kU<base; kU++)
	 if ( chr==convTblUp[kU] ) break;
     for (kD=0; kD<base; kD++)
	 if ( chr==convTblDn[kD] ) break;

     switch ( caseSense ) {
     case e_DigConvLower:
	 if ( kD>=base ) {
	     if ( doRelax ) return ReturnAndAssignUInt32( -1, uRes, vRes );
	     return -1;
	 }
	 uRes *= base;
	 uRes += (t_uint32)kD;
	 break;
     case e_DigConvUpper:
	 if ( kU>=base ) {
	     if ( doRelax ) return ReturnAndAssignUInt32( -1, uRes, vRes );
	     return -1;
	 }
	 uRes *= base;
	 uRes += (t_uint32)kU;
	 break;
     case e_DigConvAny:
     case e_DigConvAnyEmpty0:
	 if ( kU>=base && kD>=base ) {
	     if ( doRelax ) return ReturnAndAssignUInt32( -1, uRes, vRes );
	     return -1;
	 }
	 uRes *= base;
	 if ( kU<base )
	     uRes += (t_uint32)kU;
	 else
	     uRes += (t_uint32)kD;
	 break;
     default:
	 ASSERTION_FALSE("ConvertToUInt32(1)");
	 break;
     }
 }
 posErr = 0;
 ReturnAndAssignUInt32( 0, uRes, vRes );
 return 0;
}

int gStrControl::ConvertToInt32 (char* s,
				 unsigned base,
				 eDigitConv caseSense,
				 t_int32& vRes,
				 unsigned& posErr)
{
 // Return 0 on success
 int error;
 t_uint32 vAux = 0;
 t_int32 mySign;
 gString sTrim;
 bool isSigned;

 if ( s==nil ) return -1;
 sTrim.Set( s );
 if ( caseSense==e_DigConvAny ) sTrim.Trim();
 char* str = sTrim.Str();
 isSigned = str[0]=='-';
 mySign = isSigned ? -1 : 1;
 error = ConvertToUInt32( str+isSigned, base, caseSense, vAux, posErr );
 if ( error!=0 ) return error;
 // Check ranges
 if ( (long long)vAux >= MAX_LONG_L ) return 1;

 // All seem o.k., set the value for output: 'vRes'
 vRes = (t_int32)vAux * mySign;

 return 0;
}

void gStrControl::ConvertBinToStr (t_uint32 v, t_int16 places, gString& sRes)
{
 unsigned i, n;
 char c;
 gString sReverse;

 sRes.SetEmpty();

 for (n=0; v>0; n++) {
     c = (v & 1) ? '1' : '0';
     sReverse.Add( c );
     v >>= 1;
 }
 if ( n==0 ) {
     sReverse.Set( "0" );
     n++;
 }

 for (i=n; (int)i<places; i++) {
     sRes.Add( "0" );
 }
 for (i=n; i>0; i--) {
     sRes.Add( sReverse[i] );
 }
}

int gStrControl::ReturnAndAssignUInt32 (int returnValue, t_uint64 value, t_uint32& vRes)
{
 vRes = 0;
 if ( value > MAX_U4B_ULL ) return -2;
 vRes = value;
 return returnValue;
}

int gStrControl::thisInit (char* thbFilename, t_gCpCode aIsoPage)
{
 short qckIdx;
 DBGPRINT_MIN("DBG:gStrControl::thisInit: aIsoPage=%ld, thb=%s\n",(long)aIsoPage,thbFilename);
 memset( qckValidChr, 0, 256 );
 for (qckIdx=(short)' '; qckIdx<256; qckIdx++) {
     qckValidChr[qckIdx] = qckIdx>=' ' && qckIdx<127;
 }
 if ( thbFilename ) {
     strPathnameThb = new char[ strlen( thbFilename ) + 32 ];
     if ( strPathnameThb ) {
	 sprintf( strPathnameThb, "%s%cunicode256.thb",
		  thbFilename,
		  gSLASHCHR );
	 return 0;
     }
 }
 return -1;
}


unsigned gStrControl::thisFind (char* s,
				char* sub,
				unsigned startPos,
				bool doIgnoreCase,
				bool doUseCaseCp)
{
 char* ptr;

 if ( s==nil || sub==nil ) return 0;
 ASSERTION(startPos>=1,"startPos>=1");

 gString uStr( s );
 gString uSub( sub );

 if ( doIgnoreCase ) {
     uStr.UpString();
     uSub.UpString();
     s = uStr.Str();
     sub = uSub.Str();
 }

 ptr = strstr( s+startPos-1, sub );

 ASSERTION(doUseCaseCp==false,"doUseCaseCp==false: TODO");

 if ( ptr!=0 ) return ptr+startPos-s;
 return 0;
}

unsigned gStrControl::thisFindChr (char* s,
				   t_uchar subChr,
				   unsigned startPos,
				   bool doIgnoreCase,
				   bool doUseCaseCp)
{
 char sAux[2];
 sAux[0] = (char)subChr;
 sAux[1] = 0;
 return thisFind( s, sAux, startPos, doIgnoreCase, doUseCaseCp );
}
////////////////////////////////////////////////////////////
sFileKind::sFileKind ()
{
}

sFileKind::sFileKind (gString& s)
{
 Process( s );
}

sFileKind::~sFileKind ()
{
}

int sFileKind::Process (gString& s)
{
 unsigned i, n, pos;

 sAll = s;
 gParam aParam( s, ";" );
 n = aParam.N();
 for (i=1; i<=n; i++) compL.Add( aParam.Str(i) );
 if ( n==0 ) return -1;
 gString sLeft( compL.Str(1) );
 gString sRight( n<2 ? (char*)"\0" : aParam.Str(2) );
 pos = sLeft.Find(',');
 if ( pos>0 ) sLeft.Delete(1,pos);
 sLeft.Trim();
 gParam leftParam( sLeft, "/" );
 n = leftParam.N();
 if ( n<2 ) return 1;
 sMain = leftParam.Str( 1 );
 sSub = leftParam.Str( 2 );
 mainL.Add( sMain );
 mainL.Add( sSub );

 if ( sRight.Find("charset=")>0 ) {
     sCharset.CopyFromTo( sRight, 9 );
 }

 return 0;
}

void sFileKind::Show (bool doShowAll)
{
 if ( doShowAll ) {
     printf("All:%s\n",sAll.Str());
     printf("Main:"); mainL.Show(true);
     printf("Comp:"); compL.Show(true);
 }
 printf("[%s|%s]%s!\n",sMain.Str(),sSub.Str(),sCharset.Str());
}
////////////////////////////////////////////////////////////
gUniString::gUniString ()
    : gStringGeneric(
	e_String,
	gStringGeneric::e_StrAlloc,
	(t_uchar*)"\0",
	0)
{
}

gUniString::gUniString (char* s)
    : gStringGeneric(
	e_String,
	gStringGeneric::e_StrAlloc,
	(t_uchar*)s,
	0)
{
 ASSERTION(s!=nil,"s!=nil");
}

gUniString::gUniString (t_uchar* s)
    : gStringGeneric(
	e_String,
	gStringGeneric::e_StrAlloc,
	s,
	0)
{
 ASSERTION(s!=nil,"s!=nil");
}

t_uint32 gUniString::GetHexa (eDigitConv caseSense)
{
 unsigned posErr, vRes=0;
 gStrControl::Self().ConvertToUInt32( (char*)str,
				      16,
				      caseSense,
				      vRes,
				      posErr );
 return vRes;
}

t_uint32 gUniString::GetHexa0x (eDigitConv caseSense)
{
 unsigned posErr, vRes=0;
 if ( str[0]!='0' || str[1]!='x' ) return 0;
 gStrControl::Self().ConvertToUInt32( (char*)(str+2),
				      16,
				      caseSense,
				      vRes,
				      posErr );
 return vRes;
}

void gUniString::SetEmpty ()
{
 gStringGeneric::SetEmpty();
 thisUniUpdate();
}

void gUniString::Set (char* s)
{
 ASSERTION(s!=nil,"gUniStringGeneric::Set");
 Set( (t_uchar*)s );
}

void gUniString::Set (t_uchar* s)
{
 ASSERTION(s!=nil,"gUniStringGeneric::Set");
 gStringGeneric::Set( s );
 thisUniUpdate();
}

unsigned gUniString::Add (char c)
{
 return Add( (t_uchar)c );
}

unsigned gUniString::Add (t_uchar c)
{
 unsigned vRes;
 vRes = Add( (t_uchar)c );
 thisUniUpdate();
 return vRes;
}

unsigned gUniString::Add (char* s)
{
 return Add( (t_uchar*)s );
}

unsigned gUniString::Add (t_uchar* s)
{
 unsigned vRes;
 vRes = gStringGeneric::Add( s );
 thisUniUpdate();
 return vRes;
}

void gUniString::UpString ()
{
 gStringGeneric::UpString();
}

void gUniString::DownString ()
{
 gStringGeneric::DownString();
}

gUniString& gUniString::CopyFromTo (gUniString& copy, unsigned startPos, unsigned endPos)
{
 thisUniCopyFromTo( copy, startPos, endPos );
 thisUniUpdate();
 return *this;
}

unsigned gUniString::Delete (unsigned startPos, unsigned endPos)
{
 unsigned vRes = thisUniDelete( startPos, endPos );
 thisUniUpdate();
 return vRes;
}

t_uchar& gUniString::operator[] (int index)
{
 if ( thisIsValidIndex(index)==false ) return myChrNul;
 thisUniUpdate();
 doRevalidateSize = true;
 return str[index-1];
}

void gUniString::Show (bool doShowAll)
{
 gStringGeneric::Show(doShowAll);
}

gStorage* gUniString::NewObject ()
{
 gUniString* a = new gUniString( str );
 return a;
}

t_uchar* gUniString::ToString (t_uchar* uBuf)
{
 return str;
}

bool gUniString::SaveGuts (FILE* f)
{
 if ( CanSave( f )==false ) return false;
 if ( str==nil ) return true;
 return fprintf(f,"%s",str)==1;
}

bool gUniString::RestoreGuts (FILE* f)
{
 unsigned n=0;
 t_uchar uChr;
 gUCharBuffer sTemp;

 if ( CanRestore( f )==false ) return false;
 thisDelete();

 while ( fscanf(f,"%c",&uChr) ) {
     if ( uChr=='\n' || uChr==0 ) break;
     sTemp.uBuf[n++] = uChr;
     sTemp.uBuf[n] = 0;
     if ( n>=sTemp.size ) return false;
 }
 thisCopy( sTemp.uBuf, n );
 return true;
}

void gUniString::thisUniUpdate()
{
}

unsigned gUniString::thisUniDelete (unsigned startPos, unsigned endPos)
{
 // This method is copied from gString::Delete !
 ;
 unsigned
     oldSize=size,
     i, k,
     p0=startPos==0?1:startPos,
     p1=endPos==0?size:gMIN(endPos,size);

 if ( startPos==0 && endPos==0 ) {
     SetEmpty();
     return oldSize;
 }

 if ( p0>size ) return 0;

 gString sTemp;
 for (i=1, k=0; i<p0; i++, k++) sTemp.Add( (char)str[k] );
 for (i=p1+1, k=p1; i<=size; i++, k++) sTemp.Add( (char)str[k] );

 Set( sTemp.Str() );
 ASSERTION(oldSize>=size,"oldSize>=size");
 oldSize -= size;
 return oldSize;
}

gUniString& gUniString::ConvertChrFromTo (t_uchar chrFrom, t_uchar chrTo)
{
 unsigned i, len = Length();

 ASSERTION(chrFrom!=0,"chrFrom!=0");
 ASSERTION(chrTo!=0,"chrTo!=0");
 for (i=0; i<len; i++) {
     if ( str[ i ]==chrFrom ) str[ i ] = chrTo;
 }

 return *this;
}

int gUniString::thisUniCopyFromTo (gUniString& copy, unsigned startPos, unsigned endPos)
{
 // This method is copied from gString::CopyFromTo !
 ;
 unsigned i, k, len = copy.Length();

 thisDelete();
 if ( startPos==0 ) startPos = 1;
 if ( endPos==0 ) endPos = len;
 if ( endPos>len ) endPos = len;
 if ( startPos>endPos ) return -1;

 thisCopy( (t_uchar*)"\0", 1+endPos-startPos );

 for (i=startPos, k=0; i<=endPos; i++) {
     str[k++] = copy[i];
 }
 str[k] = 0;
 doRevalidateSize = true;
 return 0;
}
////////////////////////////////////////////////////////////

