#ifndef gSTRINGEXT_X_H
#define gSTRINGEXT_X_H

#include "gstring.h"
#include "glist.h"

#define THB_MAJOR_VERSION     '1'
#define THB_MINOR_VERSION_APH 'A'
#define THB_VERS_REVISION     0
#define SYM_REFVERS_UNICODE   'D'

#define EXTUMASK_ISALPHA      1
#define EXTUMASK_ISDIGIT      2
#define EXTUMASK_ISSPACE      4
#define EXTUMASK_ISPRINT      8


typedef t_uint16 t_gPageCode;
typedef t_uint32 t_gUniPage;
typedef t_uint32 t_gCpCode;

enum eDigitConv {
    e_DigConvLower,
    e_DigConvUpper,
    e_DigConvAny,
    e_DigConvAnyEmpty0
};

#define myStrSymNameLen 60
////////////////////////////////////////////////////////////
class gDateDay : public gUInt {
public:
    gDateDay (t_uint32 aDd=0) ;
    virtual ~gDateDay () ;

    // Get methods
    virtual bool IsValid () {
	return dd>0;
    }
    virtual t_uint32 GetValue32 () {
	return dd;
    }
    t_uint16 GetYear () ;
    t_uint16 GetMonth () ;
    t_uint16 GetDay () ;

    virtual int MinimumStorageSize () {
	return MaximumStorageSize();
    }

    // Set methods
    bool SetDay (t_uint32 aDd) ;

    // Save/etc methods
    virtual gStorage* NewObject () ;
    virtual bool SaveGuts (FILE* f) ;
    virtual bool RestoreGuts (FILE* f) ;

protected:
    t_uint32 dd;

private:
    // Operators,empty
    gDateDay (gDateDay& ) ;
    gDateDay& operator= (gDateDay& ) ;
};
////////////////////////////////////////////////////////////
struct sIanaCountry2 {
    char twoLetter[3];
    char* strCountry;  // english designation of country
};
////////////////////////////////////////////////////////////
struct sCodeToUni {
    sCodeToUni ()
	: lCode( 0x0000 ),
	  uniCode( 0 ) {
    }
    t_gPageCode lCode;
    t_gUniPage uniCode;
};

struct sTcbBin {
    sTcbBin () ;
    ~sTcbBin () ;

    char verMajor[6];        // 5 octet
    char cLevel;             // '0'
    gDateDay revisionDate;
    t_uchar revisionRef;
    char cNormRef;           // 'I'=ISO, 'C'=codepage, 'D' other
    char sDescTag[13];       // Short description ("ISO8859-1")
    gString sDescStr;
    t_gCpCode cpCode;        // e.g. 885901 (ISO8859-1)
    char cLinearInd;         // 'a'/'b' or 'B'
    t_uint32 nSymb;          // number of symbols,
                             // including 0x00
    t_gUniPage* linearTbl;   // 0-based, for 'a' linear
    sCodeToUni* bigTbl;      // sorted by lCode, for 'B'

    bool IsLinear () ;

    int ReadFromStream (FILE* f) ;
    int ReadFromFile (char* fName) ;
};
////////////////////////////////////////////////////////////
struct sUnCodeSymbol {
    sUnCodeSymbol ()
	: uniCode( 0x7FFFF ),
	  symbolName( nil ),
	  statZero( '0' ),
	  strComptb( nil ),
	  ucCompat( 0 ),
	  strNumber( nil ),
	  cIsOpenClose( ' ' ),
	  extEqClose( 0 ),
	  strNote( nil ),
	  ucUpper( 0 ),
	  ucDown( 0 ),
	  ucUpRngExt( 0 ),
	  strLocale( nil ),
	  refVersLetter( (t_uchar)SYM_REFVERS_UNICODE ),
	  extUtMask( 0x0000 ) {
	twoLetterCat[0] = twoLetterCat[1] = twoLetterCat[2] = 0;
	thrLetterStyle[0] = thrLetterStyle[1] = 0;
	thrLetterStyle[2] = thrLetterStyle[3] = 0;
    }

    ~sUnCodeSymbol () ;

    t_gUniPage uniCode;
    gString* symbolName;
    char twoLetterCat[3]; //Category (e.g. Lu=letter, uppercase)
    char statZero;        // Not relevant, always '0'
    char thrLetterStyle[4];
    char* strComptb;      // e.g.: <fraction> 0031 2044 0034
    t_gUniPage ucCompat;  // Not in UniCode: uniCode char equivalent, 0x0 if not applicable
    char* strNumber;      // e.g.: 1/2
    char cIsOpenClose;    // Either Y or N; e.g. '(' is Y.
    t_gUniPage extEqClose;// Not in UniCode: '(' points to ')', not vice-versa
    char* strNote;        // LATIN SMALL LETTER SHARP S has note "German"
    t_gUniPage ucUpper;   // Uppercase equivalent code
    t_gUniPage ucDown;    // Downcase equivalent code
    t_gUniPage ucUpRngExt;// Uppercase extended range (not used)
    char* strLocale;      // Usually nil/empty
    t_uchar refVersLetter;// 'D' stands for UniCode version 4
    t_uint16 extUtMask;   // Not in UniCode: utilization mask

    ////static short strSymNameLen;  // 60 octet, including 0x00
    static char symBufExt[257];

    // Get methods
    bool IsValid () ;
    bool HasName () {
	return symbolName!=nil;
    }
    bool IsLetter () {
	return twoLetterCat[0]=='L';
    }
    bool IsLetterUpcase () {
	return IsLetter()==true && twoLetterCat[1]=='u';
    }
    bool IsLetterDowncase () {
	return IsLetter()==true && twoLetterCat[1]=='l';
    }

    // Set methods
    void SetCode (t_gUniPage aCode) ;

    // Save/etc methods
    int SaveBin (FILE* f) ;
    int ReadBin (FILE* f) ;
};

struct sHtmlSymbol {
    sHtmlSymbol ()
	: cCompNote( (t_uchar)'@' ) {
    }
    gString sTransHtm;
    t_uchar cCompNote;
    // Aux members
    static char sBuf[50];
    // Get methods
    bool IsNormal () {
	return cCompNote=='`';
    }
    bool IsOk () {
	return cCompNote>='`' && cCompNote!='b' && cCompNote!='c' && cCompNote<='m';
    }
    int GetCompatLevel () {
	// @=invalid; a=1; b,c=-1; d=4; e=5, ...m
	return IsOk() ? (int)cCompNote-(int)'a'+1 : -1;
    }
    char* GetHtmlStr () ;
    // Set methods
    bool SetCompatible (bool isOk) {
	cCompNote = isOk ? 'a' : '@';
	return isOk;
    }
    bool SetCompatibleZero () {
	cCompNote = (t_uchar)'`';
	return true;
    }
};

struct sUnCodeBin {
    enum eDescStg {
	e_DescNone = 0,
	e_Desc40Byte = 40,
	e_Desc60ByteDef = 60,
	e_Desc80Byte = 80,
	e_Desc255Byte = 255
    };

    sUnCodeBin () ;
    ~sUnCodeBin () ;

    unsigned n;
    sUnCodeSymbol* mapTab;    // [1...]based index
    char verMajor[6];         // 5 octet
    t_uchar verHtml;
    t_uchar cUniLevel;        // '0'
    eDescStg descStg;
    gDateDay revisionDate;
    t_uchar revisionRef;
    char revReferenceStr[17]; // 16 octet
    t_uchar cExtTbl;          // '2':no HTML; 3:with
    // HTML-TRANS
    unsigned nHtmlTrans;      // number of symbols
    char strHtmlUnmap[ 32 ];  // default "[hs."
    sHtmlSymbol* mapHtm;      // HTML-TRANS
    t_uchar cHtmAsciiStt;     // usually 0xA0

    // Get methods
    unsigned N () {
	return n;
    }
    int GetMajorVersion () ;
    bool HasHtml () ;
    char GetDescStgChr () ;
    bool IsFlat (t_uchar& chr) ;
    bool IsFlatOk () ;
    t_uint32 FindCode (t_gUniPage aCode) ;

    char GetLetterCompat (t_gUniPage aCode) ;

    // Set methods
    int AllocateSymbols (int nSymbols) ;
    int SetVersion (t_uchar aMajorVersion, t_uchar aMinorAlpha) ;
    int SetNTables (bool hasHtmlTrans, int nExtTbl) ;

    int SaveToFile (char* fName) ;
    int SaveToStream (FILE* f) ;
    int ReadFromFile (char* fName) ;
    int ReadFromStream (FILE* f) ;
};
////////////////////////////////////////////////////////////
class gStrControl {
public:
    ~gStrControl () ;

    static gStrControl& Self () {
	return myself;
    }

    // Public data-members
    int lastOpError;
    FILE* fReport;

    // Get methods
    bool DoUseCaseCp () {
	return useCaseCp;
    }

    bool IsOkChar (t_uchar c) ;

    t_gCpCode GetDefaultIsoPage () {
	return defIsoPage;
    }

    bool IsDigConvRelaxed () {
	return doDigConvertRelaxed;
    }

    char* GetPathUnicode () {
	return strPathUnicode;
    }

    char* GetPathNameTHB () ;

    int GetFilenameCode (t_gCpCode aCpCode, gString& sCodepage) ;

    // Other get methods
    bool Match (char* s1, char* s2, bool doIgnoreCase=false) {
	return Compare( s1, s2, doIgnoreCase )==0;
    }

    int Compare (char* s1, char* s2, bool doIgnoreCase=false) ;

    unsigned Find (char* s, char* sub, bool doIgnoreCase=false) {
	return thisFind( s, sub, 1, doIgnoreCase, DoUseCaseCp() );
    }

    unsigned Find (char* s, t_uchar subChr, bool doIgnoreCase=false) {
	return thisFindChr( s, subChr, 1, doIgnoreCase, DoUseCaseCp() );
    }

    // Set methods
    int Run () ;

    int Init (char* strPath=nil, t_gCpCode aIsoPage=1) ;
    int InitTab (t_gCpCode aIsoPage) ;
    bool UseCaseCodepage (bool doUseCaseCp) ;

    void SetConvertRelax (bool doRelax) {
	doDigConvertRelaxed = doRelax;
    }

    // Specific methods
    int ConvertToUInt32 (char* s,
			 t_uint32& vRes) ;

    int ConvertToInt32 (char* s,
			t_int32& vRes) ;

    int ConvertToUInt32 (char* s,
			 unsigned base,
			 eDigitConv caseSense,
			 t_uint32& vRes,
			 unsigned& posErr) ;

    int ConvertToInt32 (char* s,
			 unsigned base,
			 eDigitConv caseSense,
			 t_int32& vRes,
			 unsigned& posErr) ;

    int ConvertHexToUInt32 (char* s,
			    eDigitConv caseSense,
			    t_uint32& vRes) {
	unsigned posErr;
	return ConvertToUInt32( s, 16, caseSense, vRes, posErr );
    }

    void ConvertBinToStr (t_uint32 v,
			  t_int16 places,  // <=0 for no justify
			  gString& sRes) ;

    void ConvertBinToStr (t_uint32 v,
			  gString& sRes) {
	ConvertBinToStr( v, 0, sRes );
    }

    int ReturnAndAssignUInt32 (int returnValue,
			       t_uint64 value,
			       t_uint32& vRes) ;

protected:
    gStrControl () ;

    static char strPathUnicode[ 512 ];

    int thisInit (char* thbFilename, t_gCpCode aIsoPage) ;

    unsigned thisFind (char* s,
		       char* sub,
		       unsigned startPos,
		       bool doIgnoreCase,
		       bool doUseCaseCp) ;

    unsigned thisFindChr (char* s,
			  t_uchar subChr,
			  unsigned startPos,
			  bool doIgnoreCase,
			  bool doUseCaseCp) ;

private:
    static sUnCodeSymbol mySym;
    static gStrControl myself;
    bool isInit;
    bool useCaseCp;
    bool doDigConvertRelaxed;
    t_gCpCode defIsoPage;
    t_uchar qckValidChr[256];
    char* strPathnameThb;

    // Operators,empty
    gStrControl (gStrControl& ) ; //empty
    gStrControl& operator= (gStrControl& ) ; //empty
};
////////////////////////////////////////////////////////////
struct sFileKind {
    sFileKind () ;
    sFileKind (gString& s) ;
    ~sFileKind () ;
    gString sAll;    // Complete string
    gString sMain;   // e.g.: text (of text/plain)
    gString sSub;    // e.g.: plain
    gString sCharset;// e.g.: iso-8859-1
    gList mainL;     // e.g.: (text plain)
    gList compL;     // e.g.: (text/plain charset=iso-8859-1)
    // Set methods
    int Process (gString& s) ;
    // Show methods
    void Show (bool doShowAll=true) ;
};
////////////////////////////////////////////////////////////
class gUniString : public gStringGeneric {
public:
    gUniString () ;
    gUniString (char* s) ;
    gUniString (t_uchar* s) ;

    // Get methods
    t_uint32 GetHexa (eDigitConv caseSense=e_DigConvAny) ;
    t_uint32 GetHexa0x (eDigitConv caseSense=e_DigConvAny) ;

    // Set methods
    virtual void SetEmpty () ;
    void Set (char* s) ;
    void Set (t_uchar* s) ;

    unsigned Add (char c) ;
    unsigned Add (t_uchar c) ;
    unsigned Add (char* s) ;
    unsigned Add (t_uchar* s) ;
    virtual void UpString () ;
    virtual void DownString () ;

    virtual gUniString& CopyFromTo (gUniString& copy, unsigned startPos, unsigned endPos) ;

    virtual unsigned Delete (unsigned startPos=0, unsigned endPos=0) ;

    virtual gUniString& ConvertChrFromTo (t_uchar chrFrom, t_uchar chrTo) ;

    // Operators
    t_uchar& operator[] (int index) ;

    // Save/etc methods
    gStorage* NewObject () ;
    virtual t_uchar* ToString (t_uchar* uBuf) ;
    virtual bool SaveGuts (FILE* f) ;
    virtual bool RestoreGuts (FILE* f) ;

    // Show methods
    void Show (bool doShowAll) ;

protected:
    void thisUniUpdate() ;
    unsigned thisUniDelete (unsigned startPos, unsigned endPos) ;
    int thisUniCopyFromTo (gUniString& copy, unsigned startPos, unsigned endPos) ;

private:
    // Operators,empty
    gUniString (gUniString& ) ;
    gUniString& operator= (gUniString& ) ;
};
////////////////////////////////////////////////////////////
#endif //gSTRINGEXT_X_H

