// gweb.cpp -- Version 0.2

#include <string.h>
#include "gweb.h"
#include "gstringext.h"

////////////////////////////////////////////////////////////
// Static members
// ---------------------------------------------------------
// void!

////////////////////////////////////////////////////////////
gIntCoord::gIntCoord ()
    : iAux( 0 ),
      y( 0 ),
      z( 0 )
{
}

gIntCoord::gIntCoord (int ax, int ay)
    : gInt( ax ),
      iAux( 0 ),
      y( ay ),
      z( 0 )
{
}

gIntCoord::gIntCoord (int ax, int ay, int az)
    : gInt( ax ),
      iAux( 0 ),
      y( ay ),
      z( az )
{
}

gStorage* gIntCoord::NewObject ()
{
 gIntCoord* a = new gIntCoord( c, y, z );
 return a;
}

t_uchar* gIntCoord::ToString (t_uchar* uBuf)
{
 if ( uBuf==nil ) return nil;
 if ( z!=0 )
     sprintf( (char*)uBuf, "[%d,%d,%d]", c, y, z );
 else
     sprintf( (char*)uBuf, "[%d,%d]", c, y );
 return uBuf;
}

gIntCoord& gIntCoord::operator= (gIntCoord& copy)
{
 iAux = copy.iAux;
 c = copy.GetX();
 y = copy.GetY();
 z = copy.GetZ();
 return *this;
}

void gIntCoord::Show (bool doShowAll)
{
 printf("[%d,%d]%s",c,y,doShowAll?"\n":"\0");
}
////////////////////////////////////////////////////////////
gTagCoord::gTagCoord ()
    : opId( 0 ),
      depth( 0 )
{
}

gTagCoord::~gTagCoord ()
{
}

void gTagCoord::Reset ()
{
 opId = 0;
 depth = 0;
 gIntCoord::Reset();
}

bool gTagCoord::SetZ (int az)
{
 bool assignOk = az>GetY();
 ASSERTION(assignOk,"assignOk");
 return gIntCoord::SetZ( az ) && assignOk;
}
////////////////////////////////////////////////////////////
bool sCoordText::IsOk ()
{
 bool isOk = coordL.N()==textL.N();
 ASSERTION(isOk,"coordL.N()==textL.N()");
 return isOk;
}

unsigned sCoordText::N ()
{
 if ( IsOk()==false ) return 0;
 return coordL.N();
}

char* sCoordText::Str (unsigned idx)
{
 return textL.Str( idx );
}

gIntCoord& sCoordText::GetCoord (unsigned idx)
{
 gStorage* inObj = coordL.GetObjectPtr( idx );
 gIntCoord* pairPtr = (gIntCoord*)inObj;
 ASSERTION(pairPtr!=nil,"pairPtr!=nil");
 return *pairPtr;
}

unsigned sCoordText::FindCoordX (int x)
{
 gList foundL, foundTxtL;
 return FindCoordX( x, foundL, foundTxtL );
}

unsigned sCoordText::FindCoordX (int x, gList& foundL, gList& foundTxtL)
{
 unsigned idx, firstIdx=0, n = N();
 int thisX, y, z;

 for (idx=1; idx<=n; idx++) {
     thisX = GetCoord( idx ).GetX();
     if ( thisX==x ) {
	 if ( firstIdx==0 ) firstIdx = idx;
	 y = GetCoord( idx ).GetY();
	 z = GetCoord( idx ).GetZ();
	 AddCoordToList( x, y, z, foundL );
	 foundTxtL.Add( textL.Str( idx ) );
     }
 }
 return firstIdx;
}

bool sCoordText::Add (int x, int y, char* s)
{
 ASSERTION(s!=nil,"s!=nil");
 gString aS( s );
 return Add( x, y, 0, aS );
}

bool sCoordText::Add (int x, int y, int z, gString& s)
{
 gString sTrim( s );
 sTrim.Trim();
 if ( sTrim.IsEmpty() ) return false;

#ifdef DEBUG
 if ( z==0 ) {
     printf("DBG:AddTxt(%d,%d):'%s' as '%s'\n",x,y,s.Str(),sTrim.Str());
 }
 else {
     printf("DBG:AddCmd(%d,%d):%s!\n",x,y,s.Str());
 }
#endif //DEBUG

 AddCoord( x, y, z );
 AddText( sTrim.Str() );
 return true;
}

bool sCoordText::Add (int x, int y, int z, char* s)
{
 ASSERTION(s!=nil,"s!=nil");
 gString aS( s );
 return Add( x, y, z, aS );
}

bool sCoordText::AddCoord (int x, int y, int z)
{
 return AddCoordToList( x, y, z, coordL );
}

bool sCoordText::AddCoordToList (int x, int y, int z, gList& resL)
{
 bool isOk;
 gIntCoord* newObj = new gIntCoord( x, y, z );
 ASSERTION(newObj!=nil,"newObj!=nil");
 isOk = IsOk();
 resL.AppendObject( newObj );
 return isOk;
}

bool sCoordText::AddText (gString& s)
{
 return AddText( s.Str() );
}

bool sCoordText::AddText (char* s)
{
 if ( s==nil ) return false;
 textL.Add( s );
 return true;
}

void sCoordText::Delete ()
{
 coordL.Delete();
 textL.Delete();
}

void sCoordText::Report (FILE* fRep, bool doShowAll)
{
 unsigned i, n = N();

 if ( fRep==nil ) fRep = stdout;
 for (i=1; i<=n; i++) {
     fprintf(fRep,"[%s%d,%s%d]:%s\n",
	     doShowAll?"Line:":"\0",
	     GetCoord( i ).GetX(),
	     doShowAll?"Col:":":",
	     GetCoord( i ).GetY(),
	     Str( i ));
 }
}
////////////////////////////////////////////////////////////
gUnweb::gUnweb (char* fName, bool doVerbose)
    : gFileFetch( fName, -1, doVerbose ),
      firstHtmlError( 0 ),
      firstHtmlErrLNr( 0 ),
      lastHtmlError( 0 ),
      lastHtmlErrLNr( 0 ),
      nHtmlErrors( 0 ),
      showKind( e_ShowAll ),
      isVerbose( doVerbose ),
      tagCaseCh( e_CaseUnchange ),
      scriptLevel( -9 ),
      cCHR_HTM_PARSED( cCHR_HTM_PARSED_DEFAULT ),
      cCHR_HTM_INFO( cCHR_HTM_INFO_DEFAULT )
{
 // gFileFetch calls SetDeviceReport( e_fStderr ) whenever doVerbose=True
 SetChrHtmParsed( cCHR_HTM_PARSED );
}

gUnweb::gUnweb (gString& sInput, bool doVerbose)
    : gFileFetch( sInput, doVerbose ),
      firstHtmlError( 0 ),
      firstHtmlErrLNr( 0 ),
      lastHtmlError( 0 ),
      lastHtmlErrLNr( 0 ),
      nHtmlErrors( 0 ),
      showKind( e_ShowAll ),
      isVerbose( doVerbose ),
      tagCaseCh( e_CaseUnchange ),
      scriptLevel( -9 ),
      cCHR_HTM_PARSED( cCHR_HTM_PARSED_DEFAULT ),
      cCHR_HTM_INFO( cCHR_HTM_INFO_DEFAULT )
{
 SetChrHtmParsed( cCHR_HTM_PARSED );
}

gUnweb::~gUnweb ()
{
}

FILE* gUnweb::VerboseStream ()
{
 if ( isVerbose==false ) return nil;
 ASSERTION(fVRepErr!=nil,"fVRepErr!=nil");
 return fVRepErr;
}

bool gUnweb::IsBufferOk ()
{
 return IsOpened() && gFileFetch::IsBufferOk();
}

char* gUnweb::Str (unsigned idx)
{
 return
     HasProcessed() ? coordSerial.Str( idx ) : aL.Str( idx );
}

void gUnweb::SetChrHtmParsed (t_uchar c)
{
 cCHR_HTM_PARSED = c;
 sCHR_HTM_PARSED[0] = (char)c;
 sCHR_HTM_PARSED[1] = 0;
}

unsigned gUnweb::FindStringKey (gString& s, char* sub, unsigned startPos, unsigned endPos)
{
 unsigned k, uLen = s.Length();
 short quoting = 0;
 t_uchar uChr;
 char* str;

 ASSERTION(sub!=nil,"sub!=nil");
 if ( uLen==0 ) return 0;

 ASSERTION(startPos>0,"startPos>0");
 ASSERTION(endPos>0,"endPos>0");
 ASSERTION(endPos<=uLen,"endPos<=uLen");

 if ( sub[0]==0 ) return 0;  // Nothing to find

 for (k=startPos, str=s.Str()+k-1; k<=endPos; k++, str++) {
     uChr = s[k];
     if ( uChr=='"' ) {
	 quoting = quoting==0;
     }
     if ( quoting!=0 ) continue;
     if ( gStrControl::Self().Find( str, sub )==1 ) {
	 return k;
     }
 }
 return 0;
}

unsigned gUnweb::FindStringKey (gString& s, char* sub)
{
 return FindStringKey( s, sub, 1, s.Length() );
}

unsigned gUnweb::FindStringKey (char* s, char* sub, unsigned startPos, unsigned endPos)
{
 ASSERTION(s!=nil,"s!=nil");
 gString aS( s );
 return FindStringKey( aS, sub, startPos, endPos );
}

unsigned gUnweb::FindStringKey (char* s, char* sub)
{
 ASSERTION(s!=nil,"s!=nil");
 gString aS( s );
 return FindStringKey( aS, sub );
}

int gUnweb::Dump (FILE* fRepErr)
{
 int error;
 gList headL;

 error = thisPreProcess( fRepErr, aL, headL );
 if ( error<0 ) return 0;  // Not re-parsing...just return.
 if ( error!=0 ) {
     if ( isVerbose ) fprintf(VerboseStream(),"Error pre-processing HTML: error-code: %d.\n",error);
     return error;
 }

 gList oL;
 error = thisPostProcess( fRepErr, headL, stripL, oL, coordSerial );
 if ( error!=0 ) return error;

 DBGPRINT_WEB("thisPostProcess: DONE\n");

#ifdef DEBUG_WEB
 unsigned ix;
 for (ix=1; ix<=coordSerial.N(); ix++) {
     unsigned eX=coordSerial.GetCoord( ix ).GetX();
     printf("WEB: ix=%u (%u): %s\n",ix,eX,Str(ix));
 }
 // One bug found in POST-PROCESS: </SCRIPT> tags disapear if xyz<SCRIPT>abc</SCRIPT> exists (same line).
 // Corrected, ok now.
#endif //DEBUG_WEB

 // At this point, 'coordSerial' contains the serialized HTML, i.e.,
 // one line per command or text.

 thisConsolidateData( stripL, 1, stripL.N(), 2 );

 if ( isVerbose ) fprintf(VerboseStream(),"gUnweb::Post-Consolidate (%u lines)\n",coordSerial.N());

 thisConsolidateDataCoord( coordParseErr, coordSerial );

 if ( isVerbose ) fprintf(VerboseStream(),"gUnweb::Post-Consolidate-check (%u error(s))\n",coordParseErr.N());

#ifdef DEBUG
 if ( 1 ) {
     printf("DBG:{\n");
     for (short dbgIter=1, dbgAll=1; dbgIter<=(short)coordSerial.N(); dbgIter++) {
	 printf("[%d,%d:%3d]:%s%s%s\n",
		coordSerial.GetCoord( dbgIter ).GetX(),
		coordSerial.GetCoord( dbgIter ).GetY(),
		coordSerial.GetCoord( dbgIter ).GetZ(),
		dbgAll?"'":"\0",
		coordSerial.Str( dbgIter ),
		dbgAll?"'":"\0");
     }
     printf("DBG:}\n");
     if ( coordParseErr.N()>0 ) {
	 printf("DBG:TXT-ERRORS:START:");
	 coordParseErr.Report(stdout,true);
	 printf("DBG:TXT-ERRORS:END.\n");
     }
 }
#endif //DEBUG

 thisCutCommentsInTxt( coordComments, coordSerial );

 return nHtmlErrors>0;
}

bool gUnweb::SetVerboseStream (FILE* aFRepErr)
{
 fVRepErr = aFRepErr;
 isVerbose = fVRepErr!=nil;
 return isVerbose;
}

int gUnweb::thisConsolidateData (gList& ioL, unsigned lowL, unsigned highL, short step)
{
 unsigned i;
 unsigned kIter, kCode;
 unsigned kLen;
 char* str;

 for (i=lowL; i<=highL; i++) {
     str = ioL.Str( i );
     kLen = (unsigned)strlen( str );
     for (kIter=0; kIter<kLen; kIter++) {
	 kCode = (t_uchar)str[kIter];
	 switch ( step ) {
	 case 1:
	 case 2:
	     if ( kCode==cCHR_HTM_PARSED )
		 str[kIter] = ' '; // a blank or cCHR_HTM_INFO
	     break;
	 default:
	     break;
	 }
     }
 }

 switch ( step ) {
 case 2:  //Use 1 or 2 for debug (1st or 2nd parse-step)
     // Under normal input and post-processing, this check is not useful!
#ifdef DEBUG
     for (i=lowL; i<=highL; i++) {
	 gString sTemp( ioL.Str( i ) );
	 sTemp.TrimRight();
	 if ( sTemp.Length()>0 ) printf("DBG %3d:@@ %s\n",i,sTemp.Str());
     }
#endif //DEBUG
     break;
 default:
     break;
 }
 return 0;
}

int gUnweb::thisConsolidateDataCoord (sCoordText& cErr, sCoordText& coord)
{
 unsigned i, k, n=coord.N();
 unsigned kLen;
 unsigned pos, posEnd;
 unsigned x, y;
 char* str;
 short levelScript=0;
 bool hasSomeScript;

 for (i=1; i<=n; i++) {
     short quoting = 0, quoting2 = 0;
     short notOkCount = 0, doesEnd = 0;
     t_uchar uChr;
     gString sCmd( str = coord.Str( i ) );
     gString sCmdUp( str );
     sCmdUp.UpString();
     x = coord.GetCoord(i).GetX();
     y = coord.GetCoord(i).GetY();

     DBGPRINT_WEB("WEB: DataCoord(%u,%u): %s\n",x,y,str);
     // Scripting skip
     hasSomeScript = false;
     if ( (pos = sCmdUp.Find("<SCRIPT"))>0 ) {
	 if ( levelScript!=0 ) {
	     cErr.Add( x, y+pos-1, 126, "Script-Script tag, invalid nesting" ); //HTML:ERR
	 }
	 levelScript = 1;
	 hasSomeScript = true;
     }
     posEnd = sCmdUp.Find("</SCRIPT");
     if ( posEnd>pos ) {
	 levelScript--;
	 hasSomeScript = true;
     }
     if ( levelScript<0 ) {
	 cErr.Add( x, y+posEnd-1, 127, "EndScript-EndScript tag, invalid nesting" ); //HTML:ERR
	 levelScript = 0;
     }
     hasSomeScript = hasSomeScript==true || levelScript>0;
     ///printf("DBG:%s:%d:[lvl%d]: [%s]\n",hasSomeScript?"SCR":"nor",x,levelScript,str);
     if ( hasSomeScript ) continue;  // Avoid script checking

     if ( sCmd[1]!='<' ) continue;  // Not a command

     kLen = sCmd.Length();
     uChr = sCmdUp[2];
     doesEnd = uChr=='/';
     if ( ( (uChr>='A' && uChr<='Z') || uChr=='!' || uChr=='?' || doesEnd!=0 )==false ) {
	 if ( kLen>2 )
	     cErr.Add( x, y+1, 123, "Strictly not a tag." ); //HTML:ERR
	 else
	     cErr.Add( x, y+1, 133, "Strictly not a tag (empty)." ); //HTML:ERR
	 notOkCount++;
     }

     if ( sCmd[ kLen ]!='>' ) {
	 // Weird command, add it to 'cErr'
	 cErr.Add( x, y, 116, "Unterminated tag." ); //HTML:ERR
	 continue;
     }

     if ( notOkCount==0 ) {
	 // Check remaining tag validity
	 gString sCmdRem;
	 sCmdRem.CopyFromTo( sCmdUp, 2+doesEnd, kLen );
	 k = sCmdRem.FindExcept("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!?");  // <?xml> is a valid tag!
	 ASSERTION(k>0,"k>0"); //At least will find '>'
	 uChr = sCmdRem[k];
	 k += (unsigned)doesEnd;
	 if ( k+1<kLen ) {
	     if ( uChr!=' ' && uChr!='=' ) {
		 cErr.Add( x, y+k, 124, "Not strictly valid character(s) in tag." ); //HTML:ERR
		 notOkCount++;
	     }
	     else {
		 if ( doesEnd!=0 ) {
		     cErr.Add( x, y+k, 125, "Not strictly valid character(s) in end-tag." ); //HTML:ERR
		     notOkCount++;
		 }
	     }
	 }
	 // Only the first word was checked here, for simplicity.
     }

     for (k=2; k<=kLen && (uChr = sCmd[ k ])!=0; k++) {
	 switch ( uChr ) {
	 case '"':
	     if ( quoting2==0 ) quoting = (quoting==0)*((short)k);
	     break;
	 case '\'':
	     if ( quoting==0 ) quoting2 = (quoting2==0)*((short)k);
	     break;
	 default:
	     if ( quoting!=0 || quoting2!=0 ) continue;
	     // If there is an error, add-it to 'cErr'
	     if ( uChr>=127 ) {
		 // Limited charset allowed for commands.
		 // 127..255 is non-7bit.
		 // To be fully strict, DTD compliance, tilde (126: '~') would not be allowed.
		 cErr.Add( x, y+k-1, 115, "Non 7bit symbol" ); //HTML:ERR
		 continue;
	     }
	     ; // Now check '<' cannot appear twice
	     if ( uChr=='<' ) {
		 cErr.Add( x, y+k-1, 121, "Too many '<'" ); //HTML:ERR
		 continue;
	     }
	     if ( k<kLen && uChr=='>' ) {
		 cErr.Add( x, y+k-1, 122, "Too many '>'" ); //HTML:ERR
		 continue;
	     }
	     //
	     // Now beautify commands (except between quotes)
	     //
	     switch ( tagCaseCh ) {
	     case e_CaseUnchange:
		 break;
	     case e_CaseConvUpper:
		 if ( uChr>='a' && uChr<='z' ) {
		     uChr -= 32;
		 }
		 str[ k-1 ] = uChr;
		 break;
	     case e_CaseConvDown:
		 break; //Todo.
	     default:
		 break;
	     }//end CASE (tagCaseCh)
	     break;
	 }//end CASE (uChr)
     }//end FOR k
     if ( quoting!=0 ) {
	 cErr.Add( x, y+quoting-1, 117, "Unterminated double-quote" ); //HTML:ERR
     }
     if ( quoting2!=0 ) {
	 cErr.Add( x, y+quoting-1, 118, "Unterminated single-quote" ); //HTML:ERR
     }
 }//end FOR i

 // Update nErrorsXXX data-members
 n = cErr.N();
 nHtmlErrors += n;
 if ( n>0 ) {
     firstHtmlErrLNr = cErr.GetCoord( 1 ).GetX();
     lastHtmlErrLNr = cErr.GetCoord( n ).GetX();
 }
 return 0;
}

int gUnweb::thisCutCommentsInTxt (sCoordText& coordCmt, sCoordText& zRes)
{
 unsigned i, nPost = zRes.N();
 unsigned idx;
 int iLen;
 int x, y, z;
 int y1;
 char* str;

 if ( coordCmt.N()==0 ) return 0;  // Nothing to do

 for (i=1; i<=nPost; i++) {
     z = zRes.GetCoord( i ).GetZ();
     if ( z!=0 ) continue; // Not a command/tag, ignore
     gList foundL, foundTxtL;
     x = zRes.GetCoord( i ).GetX();
     idx = coordCmt.FindCoordX( x, foundL, foundTxtL );
     if ( idx==0 ) continue; // No comment in line x
     y = zRes.GetCoord( i ).GetY();
     unsigned ilIdx, ilN;
     int strIter, cmtIter, cmtIterMax;
     str = zRes.Str( i );
     iLen = strlen(str);
     for (ilIdx=1, ilN=foundL.N(); ilIdx<=ilN; ilIdx++) {
	 gIntCoord* pCoord;
	 pCoord = (gIntCoord*)foundL.GetObjectPtr( ilIdx );
	 y1 = pCoord->GetY();
	 if ( y1 < y+iLen ) {
	     cmtIterMax = strlen( foundTxtL.Str( ilIdx ) );
	     strIter = y1-y;
	     if ( strIter<0 ) continue;  // Comment at begining of line, already suppressed.
	     for (cmtIter=0; cmtIter<cmtIterMax; cmtIter++) {
		 ASSERTION(strIter<iLen,"strIter...");
		 str[ strIter++ ] = cCHR_HTM_PARSED;;
		 //fprintf(stderr,"DBG:STR=%s|y=%d,y1=%d (ilIdx=%d, strIter=%d)\n",str,y,y1,ilIdx,strIter);
	     }
	 }
     }//end FOR ilIdx
     gString sTemp;
     for (strIter=0; strIter<iLen; strIter++) {
	 t_uchar uChr( (t_uchar)str[ strIter ] );
	 if ( uChr!=cCHR_HTM_PARSED ) sTemp.Add( uChr );
     }
     strcpy( str, sTemp.Str() );
 }//end FOR i (line loop)
 return 0;
}

int gUnweb::thisPreProcess (FILE* fRepErr, gList& iL, gList& oL)
{
 ;
 // Pre-processes, if not already
 // Returns -1 if already processed;
 //         0 if all ok
 //         >0 for an error parsing HTML
 //
 int error;
 int len;
 unsigned i, n;
 char* str;

 if ( HasProcessed()==false ) return -1;

 // cCHR_HTM_PARSED is never changed.
 firstHtmlError = firstHtmlErrLNr = 0;
 lastHtmlError = lastHtmlErrLNr = 0;
 nHtmlErrors = 0;

 // Clean-up output lst
 oL.Delete();

 // Build stripL the first time
 error = 0;
 for (i=1, n=iL.N(); i<=n; i++) {
     str = iL.Str( i );
     // Check tailing '\r', and cut it!
     while ( (len = strlen(str))>0 && str[--len]=='\r' ) str[len] = 0;
     // Check if used HTM_PARSED character will collide with existing input
     if ( gStrControl::Self().Find( str, cCHR_HTM_PARSED )>0 ) {
	 fprintf(fRepErr,"Found chr 0x%02X on line: %u\n",cCHR_HTM_PARSED,i);
	 error++;
     }
     if ( gStrControl::Self().Find( str, '\r' )>0 ) {
	 fprintf(fRepErr,"Found chr 0x%02X on line: %u\n",'\r',i);
	 error++;
     }
     gString sTrim( str );
     sTrim.TrimRight();
     stripL.Add( sTrim );
 }

 if ( error!=0 ) return 1;

 if ( isVerbose ) fprintf(VerboseStream(),"gUnweb:PreJoin (%u lines)\n",n);

 error = thisPreJoin( fRepErr, iL, stripL, oL );
 if ( error!=0 ) return error;

 // Just checking iL cardinality is the same as stripL
 ASSERTION(iL.N()==stripL.N(),"iL.N()==stripL.N()");

 // Now we want to knock-out comments,
 // and document-it at 'coordComments'
 error = thisPreStripComments( fRepErr, stripL, coordComments );
 if ( error!=0 ) return error;

 if ( isVerbose ) fprintf(VerboseStream(),"gUnweb:StripComments (%u lines)\n",coordComments.N());

 // So far the hardest parsing part is done.

 // Now consolidate data...
 thisConsolidateData( stripL, 1, n, 1 );

 if ( isVerbose ) fprintf(VerboseStream(),"gUnweb:Post\n");

 return error;
}

int gUnweb::thisPreJoin (FILE* fRepErr, gList& iL, gList& sL, gList& oL)
{
 unsigned i, n;
 unsigned uLen, pos, pos2;
 unsigned nOcc = 0;
 char* stripStr;

 // Step 1:
 // * Find '<!DOCTYPE', i.e. '<!' and join-in

 for (i=1, n=iL.N(); i<=n; i++) {
     stripStr = sL.Str( i );
     gString s( stripStr );
     uLen = s.Length();

     while ( (pos = s.Find("<!DOCTYPE ",nOcc,false))>0 ) {
	 // Pass begining of string (if any chrs before '<!') to oL: thisStrMove( s, 1, pos-1, oL )

	 // Find '>' on this line, or in the remaining lines
	 gString sEnd;
	 sEnd.CopyFromTo( s, pos, uLen );
	 pos2 = sEnd.Find( ">" );
	 if ( pos2>0 ) {
	     pos2 += pos - 1;
	     // Explanation follows for the first usage of 'thisStrMove' method:
	     // - Found XXX<!DOCTYPE...YYY>*
	     // - Replace it by XXX@@@@@*
	     //   where @ here referred represents CHR_HTM_PARSED (inverted excl.mark)
	     thisStrMove( s, pos, pos2, oL );
	     // Note: 'oL' has just has the parsed '<!DOCTYPE...>' instructions.

	     // Next sequence:
	     // 'thisStripReplace' copies the current changed string (with @s)
	     // to the stripStr memory location (no fancy copy here!)
	     // Basically this allows to keep tracking of the characters
	     // already parsed from the original text, in the exact (line,col) locations.
	     thisStripReplace( s, stripStr );
	     continue;
	 }
	 // line extends beyond: <!xxx ... \n ...> ?
	 // => '<!DOCTYPE' beyond newline
	 //
	 // First step is to change the current line XXX<!DOCTYPE...YYY
	 // to:  XXX@@@
	 gString sCat, sLimp;
	 thisStrSetParse( s, pos, uLen, sCat );
	 thisStripReplace( s, stripStr );
	 for (i=i+1, pos=0; i<=n && pos==0; i++) {
	     stripStr = sL.Str( i );
	     s = stripStr;
	     uLen = s.Length();
	     if ( uLen==0 ) continue;
	     sCat.Add( " " );
	     pos = s.Find( ">" );
	     if ( pos==0 ) {
		 gString sTrim( s );
		 sTrim.Trim();
		 sCat.AddString( sTrim );
		 thisStrSetParse( s, pos, uLen, sLimp );
		 thisStripReplace( s, stripStr );
	     }
	 }//end FOR inner-i
	 if ( pos==0 ) return 101;  // HTML:ERR: DocType not ended
	 // ...YYY>ZZZ =>sLimp-becomes=> YYY>
	 thisStrSetParse( s, 1, pos, sLimp );
	 // ...now stripStr becomes: @@@@ZZZ
	 thisStripReplace( s, stripStr );
	 // ...and sCat must contain YYY>, e.g.: <!DOCTYPE AAA\nBBB\nYYY>
	 sCat.AddString( sLimp );
	 // Finally adding the command to 'oL'
	 thisStrMove( sCat, 1, sCat.Length(), oL );
     }
 }//end FOR (i)

 return 0;
}

int gUnweb::thisPreStripComments (FILE* fRepErr, gList& iL, sCoordText& zRes)
{
 unsigned i, n;
 unsigned k;
 unsigned uLen, pos, pos2;
 char* stripStr;

 // Note: rewrites 'iL' strings

 for (i=1, n=iL.N(); i<=n; ) {
     stripStr = iL.Str( i );
     uLen = (unsigned)strlen( stripStr );
     pos = gStrControl::Self().Find( stripStr, "<!--" );
     if ( pos==0 ) {
	 i++;
	 continue;
     }
     zRes.AddCoord( i, pos );
     // 'sCom' will get bigger for the comments in, including '\n'
     gString sCom;
     // 0-based index for stripStr, thus decrement posXXX
     pos--;
     // One or more occurrences of a comment here.
     // The comment may end in this line or below.
     pos2 = gStrControl::Self().Find( stripStr+pos, "-->" );
     if ( pos2>0 ) {
	 pos2--;
	 pos2 += pos;
	 // The comment ends at this line
	 for (k=pos; k<pos2+3; k++) {
	     // ...why pos2+3? Because '-->' has 3 chars.
	     sCom.Add( stripStr[k] );
	     stripStr[k] = cCHR_HTM_PARSED;
	 }
	 zRes.AddText( sCom );
	 // ...and try again the same line
	 continue;
     }
     // Here we know there was a '<!--',
     // ...knock-out comments here
     for (k=pos; k<uLen; k++) {
	 sCom.Add( stripStr[k] );
	 stripStr[k] = cCHR_HTM_PARSED;
     }
     sCom.Add( '\n' );
     // ...and find in the lines below: '-->'
     i++;
     for (pos2=0; i<=n && pos2==0; i++) {
	 stripStr = iL.Str( i );
	 uLen = (unsigned)strlen( stripStr );
	 pos2 = gStrControl::Self().Find( stripStr, "-->" );
	 if ( pos2==0 ) {
	     sCom.Add( stripStr );
	     sCom.Add( '\n' );
	     for (k=0; k<uLen; k++) stripStr[k] = cCHR_HTM_PARSED;
	     continue;
	 }
	 for (k=0; k+1<pos2+3; k++) {
	     // ...why pos2+3? Because '-->' has 3 chars.
	     sCom.Add( stripStr[k] );
	     stripStr[k] = cCHR_HTM_PARSED;
	 }
     }// end FOR inner-i...
     zRes.AddText( sCom );
     if ( pos2==0 ) return 102;  // HTML:ERR: Comment did not end
 }
 return 0;
}

int gUnweb::thisStrMove (gString& s, unsigned startPos, unsigned endPos, gList& oL)
{
 unsigned x;
 int count=0;
 gString sOut;
 bool anyScript=false;

 sOut.CopyFromTo( s, startPos, endPos );
 sOut.Trim();
 gString sUp( sOut );
 sUp.UpString();

 unsigned pos = sUp.Find("<SCRIPT");
 if ( (anyScript = (pos>0))==true ) {
     scriptLevel++;
 }
 else {
     pos = sUp.Find("</SCRIPT");
     scriptLevel -= pos>0;
     if ( scriptLevel<0 ) scriptLevel = 0;  // No error report, here.
     anyScript = scriptLevel>0;
 }

 //printf("DBG:IN_:'%s' (scriptLevel=%d)\n",sOut.Str(),scriptLevel);

 // Fill-up string 's' with HTM_PARSED chars
 for (x=startPos; x<=endPos; x++, count++) s[x] = cCHR_HTM_PARSED;

 //if ( anyScript ) printf("DBG:OUT:'%s'\n",sOut.Str());
 //printf("DBG:OUT:'%s'\nP:%s\n\n",s.Str(),sOut.Str());

 if ( sOut.Length()==0 ) return 0;
 oL.Add( sOut );

 return count;
}

int gUnweb::thisStripReplace (gString& s, char* resStr)
{
 unsigned uLen = s.Length();
 unsigned rLen;

 if ( resStr==nil ) return -1;
 rLen = (unsigned)strlen( resStr );
 ASSERTION(uLen==rLen,"uLen==rLen");
 strcpy( resStr, s.Str() );
 return 0;
}

int gUnweb::thisStrSetParse (gString& s, unsigned startPos, unsigned endPos, gString& sResult)
{
 gList tempL;
 ASSERTION(endPos>0,"endPos>0");
 if ( thisStrMove( s, startPos, endPos, tempL )==0 ) return 0;
 sResult = tempL.Str( 1 );
 return 1;
}

////////////////////////////////////////////////////////////
// continuing... gUnweb post-processing
////////////////////////////////////////////////////////////
int gUnweb::thisPostProcess (FILE* fRepErr, gList& headL, gList& iL, gList& oL, sCoordText& coord)
{
 unsigned i, n;
 unsigned uLen, pos, pos2;
 unsigned posStart;
 char* stripStr;
 gString sLimp;

 ASSERTION(coord.N()==0,"coord.N()==0");
 scriptLevel = 0;  // Nesting level of <SCRIPT...> keywords.

 // Fancy adjust of output
 for (i=1, n=headL.N(); i<=n; i++) {
     oL.Add( stripStr = headL.Str( i ) );
     thisAddCmd( 1, 1, stripStr, coord );
 }
 // .

 // Serialize keywords and text found (into 'coord'; iL rewritten)

 for (i=1, n=iL.N(), posStart=1; i<=n; ) {
     stripStr = iL.Str( i );
     gString s( stripStr );
     // stripStr is already trimmed
     uLen = s.Length();
     if ( posStart>uLen ) {
	 i++; posStart = 1; // Parse next line
	 continue;
     }

     ASSERTION(posStart>0,"posStart>0");
     pos = FindStringKey( s, "<", posStart, uLen );
     if ( pos==0 ) {
	 thisAddTxt( i, posStart, stripStr+posStart-1, coord );
	 thisStrMove( s, posStart, uLen, oL );
	 thisStripReplace( s, stripStr );
	 i++; posStart = 1;  // Parse next line
	 continue;
     }

     ASSERTION(pos>=posStart,"pos>=posStart");

     // Strip before keyword, e.g.: XXX<BODY>YYY becomes @@@<BODY>YYY
     if ( pos>posStart ) {
	 gString sTemp;
	 unsigned aKeepPos;
	 if ( scriptLevel>0 ) {
	     DBGPRINT_WEB("WEB: Post: XXX<tag>YYY [%s] scriptLevel=%d pos=%u,posStart=%u,len=%d\n",stripStr,scriptLevel,pos,posStart,strlen(stripStr));
	     // Check if something like
	     // "@@@<SCRIPT>"+"xyz"+!"</SCRIPT>"
	     char* strTagEtc = stripStr+pos-1;
	     if ( gStrControl::Self().Find( strTagEtc, "</SCRIPT", true )==1 ) {
		 sTemp.CopyFromTo( s, posStart, pos+gStrControl::Self().Find( strTagEtc, ">" ) );
		 DBGPRINT_WEB("WEB: Post: scriptLevel_now=%d [strTagEtc=%s] [%s]\n",scriptLevel,strTagEtc,sTemp.Str());
		 pos--;
		 thisAddTxt( i, posStart, sTemp, coord );
		 thisStrMove( s, posStart, pos, oL );
		 thisStripReplace( s, stripStr );
		 DBGPRINT_WEB("WEB: Post: after1 <tag>YYY [%s] scriptLevel=%d pos=%u,posStart=%u\n",stripStr,scriptLevel,pos,posStart);
		 posStart = pos+1;
		 continue;
	     }
	     i++; posStart = 1;  // SCR_IGN
	     continue;
	 }
	 aKeepPos = pos;
	 pos--;
	 sTemp.CopyFromTo( s, posStart, pos );
	 thisAddTxt( i, posStart, sTemp, coord );
	 thisStrMove( s, posStart, pos, oL );
	 thisStripReplace( s, stripStr );
	 posStart = aKeepPos;  // Keep seeking, in the above e.g.: <BODY>YYY
	 continue;
     }

     //
     // At least one keyword: either all in this line or below, e.g.:
     //    <BODY bgcolor="#FFEECC">xyz</HEAD>
     // or <BODY bgcolor="#FFE>ECC">xyz</HEAD> (misleading '<', belongs to string)
     //
     pos2 = FindStringKey( s, ">", pos+1, uLen );
     DBGPRINT("FIND(%s):posStart=%d,uLen=%d:POS=%d, POS2=%d\n",stripStr,posStart,uLen,pos,pos2);
     if ( pos2>0 ) {
	 gString sTemp;
	 ASSERTION(pos2>pos,"pos2>pos");
	 sTemp.CopyFromTo( s, pos, pos2 );
	 if ( gStrControl::Self().Find( sTemp.Str(), "</SCRIPT", true ) ||
	      scriptLevel>0 ) {
	     DBGPRINT_WEB("DBG:/SCRIPT:sTemp=%s!\n",sTemp.Str());
	     thisAddCmd( i, pos, sTemp, coord );  // SCR_IGN
	     thisStrMove( s, pos, pos2, oL );
	 }
	 else {
	     thisAddCmd( i, pos, sTemp, coord );
	     thisStrMove( s, pos, pos2, oL );
	     thisStripReplace( s, stripStr );
	 }
	 posStart = pos2+1;
	 continue;
     }

     // Something like e.g.: <BODY ... (no ending '>')
     ASSERTION(pos2==0,"pos2==0");
     gString sCmd;
     sCmd.CopyFromTo( s, pos, uLen );

     DBGPRINT("DBG: sCmd=[%s]\n",sCmd.Str());

     int iKeep = i;
     // Find in the lines below the trailing '>'
     i++; posStart = 1;  // ...next line
     for (pos2=0; i<=n && pos2==0; ) {
	 char* stripStr = iL.Str( i );
	 pos2 = FindStringKey( stripStr, ">" );
	 gString s( stripStr );
	 s.Trim();
	 if ( s.IsEmpty() ) {
	     i++;
	     continue;
	 }
	 if ( pos2==0 ) {
	     sCmd.Add( ' ' );
	     sCmd.AddString( s );
	     s.Set( stripStr );
	     thisStrSetParse( s, 1, s.Length(), sLimp );
	     thisStripReplace( s, stripStr );
	     i++;
	 }
     }
     if ( pos2==0 ) return 111;  // ...command did not end (till EOF)

     // Update stripStr, not yet striked
     thisStrSetParse( s, pos, uLen, sLimp );
     thisStripReplace( s, stripStr );
     // Update sCmd, which has not the latter line 'i'
     s.Set( stripStr = iL.Str( i ) );
     sLimp.CopyFromTo( s, 1, pos2 );
     sLimp.Trim();
     // ...just add a blank if necessary...
     if ( FindStringKey( sLimp, ">" )>1 ) sCmd.Add(' ');
     sCmd.AddString( sLimp );
     // Add command to coord!
     thisAddCmd( iKeep, pos, sCmd, coord );
     oL.Add( sCmd );
     // ...and strike chars before end of command, i.e. before '>'
     thisStrSetParse( s, 1, pos2, sLimp );
     thisStripReplace( s, stripStr );
     // Update indexing vars, and keep parsing the current line
     posStart = pos2+1;
 }
 return 0;
}

int gUnweb::thisAddTxt (int x, int y, char* s, sCoordText& coord)
{
 if ( s==nil ) return -1;
 gString aS( s );
 return thisAddTxt( x, y, aS, coord );
}

int gUnweb::thisAddTxt (int x, int y, gString& sTxt, sCoordText& coord)
{
 // Return 0 if o.k.
 bool isOk;
 int iPos = (int)sTxt.FindExcept( " \t" ), iDelta;

 iDelta = ( iPos>0 ? iPos-1 : 0 );
 y += iDelta;

 if ( scriptLevel>0 && (showKind & e_ShowScript)==0 ) return 0;
 if ( scriptLevel<=0 && (showKind & e_ShowNonscript)==0 ) return 0;

 isOk = coord.Add( x, y, sTxt );
 if ( isOk==false ) return -1;
 isOk = thisTxtCheck( x, y, sTxt, coordParseErr )==0;
 if ( isOk==false ) return -1;
 // When using <PRE>, the exact string location must be held
 unsigned n = coord.N();
 coord.GetCoord(n).iAux = iDelta;
 // .
 return 0;
}

int gUnweb::thisAddCmd (int x, int y, char* s, sCoordText& coord)
{
 if ( s==nil ) return -1;
 gString aS( s );
 return thisAddCmd( x, y, aS, coord );
}

int gUnweb::thisAddCmd (int x, int y, gString& sCmd, sCoordText& coord)
{
 int z=-1;
 coord.Add( x, y, z, sCmd );
 //printf("DBG:CMD:'%s'\n",sCmd.Str());
 return z;
}

int gUnweb::thisTxtCheck (int x, int y, gString& s, sCoordText& cErr)
{
 // Return 0 if o.k.
 unsigned pos, posCount = 0;

 if ( scriptLevel>0 ) return 0;  // Using scripts, no checking performed
 pos = FindStringKey( s, "<" );
 posCount += pos;
 if ( pos>0 ) {
     // At least one invalid symbol found
     cErr.Add( x, y+pos-1, 105, "Invalid symbol: '<'" ); //HTML:ERR
 }
 pos = FindStringKey( s, ">" );
 posCount += pos;
 if ( pos>0 ) {
     // At least one invalid symbol found
     cErr.Add( x, y+pos-1, 106, "Invalid symbol: '>'" ); //HTML:ERR
 }
 return posCount==0 ? 0 : 1;
}
////////////////////////////////////////////////////////////

