#ifndef gWEB_X_H
#define gWEB_X_H

#include "gconfig.h"
#include "gnet.h"

#if defined(DEBUG) || defined(DEBUG_WEB)
#define cCHR_HTM_PARSED_DEFAULT 0xA1  // Acc.UniCode4: 00A1;INVERTED EXCLAMATION MARK
// When debugging, 0xA1 character is easier to dump
#else
#define cCHR_HTM_PARSED_DEFAULT 0x7F  // must not be dumped
#endif //DEBUG
#define cCHR_HTM_INFO_DEFAULT 0xB7    // Acc.UniCode4: 00B7;MIDDLE DOT

#ifdef DEBUG_WEB
#define DBGPRINT_WEB(args...) printf(args)
#else
#define DBGPRINT_WEB(args...) ;
#endif //DEBUG_WEB

////////////////////////////////////////////////////////////
// Generic classes
////////////////////////////////////////////////////////////
class gIntCoord : public gInt {
public:
    gIntCoord () ;
    gIntCoord (int ax, int ay) ;
    gIntCoord (int ax, int ay, int az) ;
    virtual ~gIntCoord () {
    }

    // Public data-members
    short iAux;

    // Get methods
    virtual int GetX () {
	return c;
    }

    virtual int GetY () {
	return y;
    }

    virtual int GetZ () {
	return z;
    }

    // Set methods
    virtual void Reset () {
	c = 0;
	y = z = 0;
    }

    virtual bool SetX (int ax) {
	return SetInt( ax );
    }

    virtual bool SetY (int ay) {
	y = ay;
	return true;
    }

    virtual bool SetZ (int az) {
	z = az;
	return true;
    }

    // Operators,valid
    gIntCoord& operator= (gIntCoord& copy) ;

    // Save/etc methods
    gStorage* NewObject () ;
    virtual t_uchar* ToString (t_uchar* uBuf) ;

    // Show methods
    virtual void Show (bool doShowAll=true) ;

protected:
    int y, z;

private:
    // Operators,empty
    gIntCoord (gIntCoord& ) ;
};
////////////////////////////////////////////////////////////
class gTagCoord : public gIntCoord {
public:
    gTagCoord () ;
    virtual ~gTagCoord () ;

    // Public data-members
    t_int16 opId;
    int depth;

    // Get methods
    virtual bool IsOk () {
	return c>=0 && GetY()>0 && GetZ()>0;
    }
    virtual t_int16 GetTagId () {
	return (t_int16)c;
    }

    // Set methods
    virtual void Reset () ;

    virtual bool SetTagId (t_int16 id) {
	SetX( id );
	return id>=0;
    }
    virtual bool SetZ (int az) ;

private:
    // Operators,empty
    gTagCoord (gTagCoord& ) ; //empty
    gTagCoord& operator= (gTagCoord& ) ; //empty
};
////////////////////////////////////////////////////////////
// Generic structures
////////////////////////////////////////////////////////////
struct sCoordText {
    sCoordText () {
    }
    ~sCoordText () {
    }

    gList coordL;
    gList textL;

    // Get methods
    bool IsOk () ;
    unsigned N () ;
    char* Str (unsigned idx) ;
    gIntCoord& GetCoord (unsigned idx) ;
    unsigned FindCoordX (int x) ;
    unsigned FindCoordX (int x, gList& foundL, gList& foundTxtL) ;

    // Set methods
    bool Add (int x, int y, gString& s) {
	return Add( x, y, s.Str() );
    }
    bool Add (int x, int y, char* s);
    bool Add (int x, int y, int z, gString& s) ;
    bool Add (int x, int y, int z, char* s);

    bool AddCoord (int x, int y, int z=0) ;
    bool AddCoordToList (int x, int y, int z, gList& resL) ;

    bool AddText (gString& s) ;
    bool AddText (char* s) ;

    void Delete () ;

    void Report (FILE* fRep, bool doShowAll=true) ;
};
////////////////////////////////////////////////////////////
// CLASS gUnweb
////////////////////////////////////////////////////////////
class gUnweb : public gFileFetch {
public:
    gUnweb (char* fName, bool doVerbose) ;
    gUnweb (gString& sInput, bool doVerbose) ;
    virtual ~gUnweb () ;

    // Public enums
    enum eTagCaseChange {
	e_CaseUnchange,
	e_CaseConvUpper,
	e_CaseConvDown
    };

    enum eDumpMask {
	e_ShowNonscript = 1,   // Usual also
	e_ShowScript = 2,
	e_ShowAll = e_ShowNonscript | e_ShowScript
    };

    // Public data-members
    int firstHtmlError, firstHtmlErrLNr;
    int lastHtmlError, lastHtmlErrLNr;
    int nHtmlErrors;
    eDumpMask showKind;  //Default eDumpMask is e_ShowAll
    // ...handy structs...
    sCoordText coordComments;
    sCoordText coordSerial;
    sCoordText coordParseErr;

    // Get methods
    virtual bool IsVerbose () {
	return isVerbose;
    }
    virtual FILE* VerboseStream () ;

    virtual bool IsBufferOk () ;
    virtual char* Str (unsigned idx) ;

    virtual bool HasProcessed () {
	return nHtmlErrors!=-1;
    }
    virtual bool HasErrors () {
	return nHtmlErrors>0;
    }
    // .

    // Specific methods
    unsigned FindStringKey (gString& s, char* sub, unsigned startPos, unsigned endPos) ;
    unsigned FindStringKey (gString& s, char* sub) ;
    unsigned FindStringKey (char* s, char* sub, unsigned startPos, unsigned endPos) ;
    unsigned FindStringKey (char* s, char* sub) ;

    // Set methods
    bool SetVerboseStream (FILE* aFRepErr) ;

    void SetTagCaseChange (eTagCaseChange aTagCaseCh) {
	tagCaseCh = aTagCaseCh;
    }
    void SetChrHtmParsed (t_uchar c) ;

    int Dump (FILE* fRepErr) ;
    // .

protected:
    bool isVerbose;
    eTagCaseChange tagCaseCh;

    int thisConsolidateData (gList& iL, unsigned lowL, unsigned highL, short step) ;
    int thisConsolidateDataCoord (sCoordText& cErr, sCoordText& coord) ;

    int thisCutCommentsInTxt (sCoordText& coordCmt, sCoordText& zRes) ;

    int thisPreProcess (FILE* fRepErr, gList& iL, gList& oL) ;
    int thisPreJoin (FILE* fRepErr, gList& iL, gList& sL, gList& oL) ;
    int thisPreStripComments (FILE* fRepErr, gList& iL, sCoordText& zRes) ;

    int thisStrMove (gString& s, unsigned startPos, unsigned endPos, gList& oL) ;
    int thisStripReplace (gString& s, char* resStr) ;
    int thisStrSetParse (gString& s, unsigned startPos, unsigned endPos, gString& sResult) ;

    int thisPostProcess (FILE* fRepErr, gList& headL, gList& iL, gList& oL, sCoordText& coord) ;

    int thisAddTxt (int x, int y, char* s, sCoordText& coord) ;
    int thisAddTxt (int x, int y, gString& sTxt, sCoordText& coord) ;
    int thisAddCmd (int x, int y, char* s, sCoordText& coord) ;
    int thisAddCmd (int x, int y, gString& sCmd, sCoordText& coord) ;

    int thisTxtCheck (int x, int y, gString& s, sCoordText& cErr) ;

private:
    short scriptLevel;
    t_uchar cCHR_HTM_PARSED;
    t_uchar cCHR_HTM_INFO;
    char sCHR_HTM_PARSED[10];

    // Internal aux-lists
    gList stripL;   // as aL, but with stripped parsed chars
                    // (parsed-chr: see cCHR_HTM_PARSED)

    // Operators,empty
    gUnweb (gUnweb& ) ; //empty
    gUnweb& operator= (gUnweb& ) ; //empty
};
////////////////////////////////////////////////////////////
#endif //gWEB_X_H

