// gNetinfoHttp.cpp -- Version 2.9

#include <string.h>
#include "gNetinfoHttp.h"
////////////////////////////////////////////////////////////
gReworkedHttpCont::gReworkedHttpCont (gTcpConnect& connection)
    : gXHttpCont( connection )
{
}

gReworkedHttpCont::~gReworkedHttpCont ()
{
}
////////////////////////////////////////////////////////////
gNetHttpCont::gNetHttpCont (gTcpConnect& connection)
    : gReworkedHttpCont( connection ),
      doPace( true )
{
}

gNetHttpCont::~gNetHttpCont ()
{
}

int gNetHttpCont::ReadReplyChunked (gTcpConnect& connection,
				    gBigBuffer& bBuf,
				    unsigned& count)
{
 // Returns 0 on success, 1 if chunk-size is invalid, 2 if read failed, -1 on writting error.
 // 4 indicates no CR-LF sequence before chunk
 unsigned iter=0;
 int error;
 t_uint32 chunkSize=0;
 t_uint32 timeBefore=vStat[1].GetUInt();
 t_uint32 timeNew=0, timeChunk=0, waited=0;
 bool isFlushOk;
 bool isHugeBuffer;
 gBigBuffer* ptrBuf;

 bBuf.NewFile();

 for (count=0, error=0, isFlushOk=true;
      error==0 && isFlushOk==true && (chunkSize = thisReadChunkSize( connection, error ))>0;
      count+=chunkSize, iter++) {
     ASSERTION(chunkSize<MAX_HTTP_HUGE_CHUNK_SIZE,"ReadReplyChunked(0)");
     isHugeBuffer = chunkSize>=MAX_HTTP_CHUNK_SIZE;
     DBGPRINT_time("DBG: ReadReplyChunked()[%u]=%d chunkSize=%ld%s\n",iter,error,(long)chunkSize,isHugeBuffer?" (HUGE-CHUNK)":"\0");
     if ( error!=0 ) break;
     ptrBuf = new gBigBuffer( isHugeBuffer ? 10 : chunkSize );
     ASSERTION(ptrBuf!=nil,"ReadReplyChunked(1)");
     if ( isHugeBuffer ) {
	 delete[] ptrBuf->baseBuf->uBuf;
	 ptrBuf->baseBuf->uBuf = new t_uchar[ chunkSize+1 ];
	 ASSERTION(ptrBuf->baseBuf->uBuf!=nil,"Not enough memory to read big chunks");
     }

     if ( doPace ) {
	 // Pace of server respected (acc.RFC2616, ch.8.2.4)
#ifdef gDOS_SPEC
	 timeBefore /= 1000;  // (CHECK THIS)
#else
	 timeBefore /= 100;  // 1/10th seconds before...
#endif //gDOS_SPEC

	 timeChunk = chunkSize>8192;
	 timeChunk += 10*(chunkSize>MAX_HTTP_CHUNK_SIZE);

#ifdef gDOS_SPEC
	 timeNew = timeBefore + 2*(chunkSize>3000) + (chunkSize>1024)/*+1 would be an ugly try*/;
	 waited = timeNew * 1000;
	 DBGPRINT_time("DBG: [gXHttpCont] (DOS) Pace: %u s\n",(unsigned)timeNew);
	 SecSleep( timeNew );
#else
	 timeNew = timeBefore + timeChunk;
	 waited = timeNew*10+chunkSize/8;
	 DBGPRINT_time("DBG: [gXHttpCont] (non-DOS) Pace: %u ms\n",(unsigned)waited);
	 MiliSecSleep( waited );
#endif

	 if ( fRepErr )
	     fprintf(fRepErr,"Pace (ms): %u = (time=%u/%u [chunk=%ld, timeChunk=%u])\n",
		     waited,
		     (unsigned)timeNew,
		     (unsigned)timeBefore,
		     (long)chunkSize,
		     (unsigned)timeChunk);

	 // Just increment the vStat[3] (pace)
	 vStat[3].SetUInt( vStat[3].GetUInt() + waited );  // Miliseconds made (pace) on our own HTTP client
     }

     DBGPRINT_time("DBG: (iter=%u) stat=%u, waited=%lu (total_wait:%u)\n",
		   iter,
		   vStat[1].GetUInt(),
		   (unsigned long)waited,
		   vStat[3].GetUInt());
     // Read from socket...
     error = connection.Read( ptrBuf->baseBuf->uBuf, chunkSize ) ? 0 : 2;

     DBGPRINT_time("DBG: timeNew=%u = (%u + %u [chunk=%ld]),error=%d\n",
	    (unsigned)timeNew,
	    (unsigned)timeBefore,
	    (unsigned)timeChunk,
	    (long)chunkSize,
	    error);

     {
	int dosCount=0;
	for ( ; error!=0 && dosCount<3/* 200 ??*/; dosCount++) {
	    DBGPRINT("DBG: READ(iter=%u,dosCount=%d):err=%d, size=%u\n",iter,dosCount,error,(unsigned)chunkSize);
	    error = connection.Read( ptrBuf->baseBuf->uBuf, chunkSize ) ? 0 : 2;
	    vStat[3].SetUInt( vStat[3].GetUInt() + 1 );  // Just loose some time, and increment statistical pace
	}
#ifdef DEBUG
	if ( dosCount>0 ) printf("DBG: READ(iter=%u,dosCount=%d):err=%d (FINAL)\n",iter,dosCount,error);
#endif //DEBUG
     }

     if ( error==0 ) {
	 // ...and write to file
	 isFlushOk = bBuf.WriteBuf( bBuf.GetFile().fHandle, *(ptrBuf->baseBuf), chunkSize );
	 // Read CR-LF
	 memset( ptrBuf->baseBuf->uBuf, 0x0, 10 );
	 error = connection.Read( ptrBuf->baseBuf->uBuf, 2 )==false;
	 if ( error==0 ) error = ptrBuf->baseBuf->uBuf[1]=='\n' ? 0 : 4;
     }
     delete ptrBuf;
     ptrBuf = nil;
 }//end FOR

 if ( isFlushOk==false ) return -1;
 if ( error==-1 ) return 1;
 if ( error>=2 ) return error;
 return count==0;
}
////////////////////////////////////////////////////////////

