// gUnCoder.cpp -- Version 0.5


#define thisProgramVersion "Version 0.5"
#define thisProgramCopyright "Prized Season & Sons"
#define thisProgramYear 2008
#define thisProgramCopy "This is free software (GPL)\n\
There is no warranty, not even for MERCHANTABILITY or\n\
FITNESS FOR A PARTICULAR PURPOSE."

#include <stdio.h>
#include <string.h>
#include "garg.h"
#include "gconfig.h"
#include "gcalendar.h"
#include "gstringext.h"

#define DEFAULT_CP_CODEPAGE 437

#ifdef gDOS_SPEC
#include <windows.h>
#endif //gDOS_SPEC
////////////////////////////////////////////////////////////
struct sTxtHtmlTrans {
    sTxtHtmlTrans ()
	: firstAscii( 0x0 ) {
	// First ASCII usually 0xA0
    }
    unsigned firstAscii;
    gList transL;
    gList compL;
    unsigned N () {
	return transL.N();
    }
};
////////////////////////////////////////////////////////////
struct sOptUnCoder {
    sOptUnCoder ()
	: isVerbose( false ),
	  isVeryVerbose( false ),
	  aDateDay( 0 ),  // 0 means today
	  defCpCode( DEFAULT_CP_CODEPAGE ) {
	sTblPath = gStrControl::Self().GetPathUnicode();
	sUnTblTextName = "unicode256.tbt";
    }
    ~sOptUnCoder () {
    }
    bool isVerbose, isVeryVerbose;
    gString sTblPath;
    gString sUnTblTextName;
    gString sHtmlFilename;   // not relative to sTblPath
    gString sCpTxtFilename;
    sTxtHtmlTrans tHtmlTrans;
    // Opt-ins...
    t_uint32 aDateDay;  // e.g. last day of 2003: 20031231
    t_gCpCode defCpCode;
    gString sCodePage;
    gString sTemp;
    // Get methods
    gString& GetTblTextName () {
	sTemp = sTblPath;
	if ( sTemp.IsEmpty()==false ) sTemp.Add( gSLASHCHR );
	sTemp.Add( sUnTblTextName.Str() );
	return sTemp;
    }
    gString& GetTblBinName () {
	return GetTblTextName();
    }
};
////////////////////////////////////////////////////////////
struct sUnCodeTxt {
    sUnCodeTxt ()
	: error( 0 ),
	  linesProcessed( 0 ),
	  maxLen( 0 ) {
    }
    ~sUnCodeTxt () {
    }
    int error;
    int linesProcessed;
    unsigned maxLen;
    gList sL;
    gList failedNLinesL;
};

////////////////////////////////////////////////////////////
// glib required functions
////////////////////////////////////////////////////////////
int gglobal_EventTerminate (int v)
{
 return 0;
}
////////////////////////////////////////////////////////////
t_uint32 convert_DateTime2DateDay (gDateTime& aDtTm)
{
 t_uint32 result;

 DBGPRINT("DAY: %u, MONTH: %u, YEAR: %u, aDtTm.IsOk?%c\n",
	  aDtTm.day, aDtTm.month, aDtTm.year,
	  ISyORn(aDtTm.IsOk()));
 result = aDtTm.year;
 result *= 100;
 result += aDtTm.month;
 result *= 100;
 result += aDtTm.day;
 return result;
}

long convert_DateDay2DateTime (long val, t_uint32& aDateDay, gDateTime& oDtTm)
{
 int error;
 long iResult;
 t_uint32 keepDateDay = (t_uint32)val;

 if ( val<=19000000L ) return 0;

 oDtTm.Reset();
 oDtTm.isdst = 0;  // Not assuming date is valid, yet 'DST' should be zero!
 iResult = val % 100;
 val /= 100;
 oDtTm.day = iResult;
 iResult = val % 100;
 val /= 100;
 oDtTm.month = iResult;
 iResult = val;
 oDtTm.year = iResult;
 error = oDtTm.IsOk()==false;
 DBGPRINT("DBG: error=%d, DaysOfMonth(year=iResult=%ld,%u)=%u, isdst=%d\n",
	  error,
	  iResult,
	  (unsigned)oDtTm.month,
	  (unsigned)oDtTm.DaysOfMonth( oDtTm.year, oDtTm.month ),
	  (int)oDtTm.isdst);
 if ( error!=0 ) return -1;

 // Result is only set to 'aDateDay' if everything is correct
 aDateDay = keepDateDay;

 return iResult;  // actually this is the year (>=1900)
}
////////////////////////////////////////////////////////////
int read_ucbin (char* thbFilename, sUnCodeBin& ucBin)
{
 int error;
 error = ucBin.ReadFromFile( thbFilename );
 DBGPRINT("DBG: ucBin.ReadFromFile( %s ) returned %d\n",thbFilename,error);
 if ( error!=0 ) {
     fprintf(stderr,"Could not read thb file: %s\n",thbFilename);
 }
 return error;
}

int read_opt_ucbin (sOptUnCoder& opt, sUnCodeBin& ucBin)
{
 int error;
 gString ucFilename( opt.GetTblTextName() );
 error = ucBin.ReadFromFile( ucFilename.Str() );
 DBGPRINT_MIN("ReadFromFile: %s: %d\n",ucFilename.Str(),error);
 return error;
}
////////////////////////////////////////////////////////////
int print_ucbin (sUnCodeBin& ucBin, bool isVerbose, FILE* fRep)
{
 unsigned i, n=ucBin.N();
 sUnCodeSymbol* symPtr;

 fprintf(fRep,"ucBin:::START:n=%u:%s\n",n,ucBin.verMajor);
 for (i=1; i<=n; i++) {
     symPtr = &ucBin.mapTab[i];
     if ( isVerbose ) {
	 fprintf(fRep,"%03d\t",
		 symPtr->uniCode);
     }
     fprintf(fRep,"%04X%c",
	     symPtr->uniCode,
	     symPtr->IsValid()?':':'?');
     fprintf(fRep,"%-2s,%-3s,",
	     symPtr->twoLetterCat,
	     symPtr->thrLetterStyle);
     fprintf(fRep,"'%s',",
	     symPtr->strNumber ? symPtr->strNumber : "\0");
     fprintf(fRep,"OPCL?%c<->%c:",
	     symPtr->cIsOpenClose,
	     symPtr->extEqClose==0?'.':symPtr->extEqClose);
     fprintf(fRep,"'%s',",
	     symPtr->strNote);
     if ( symPtr->ucUpper )
	 fprintf(fRep,"(UP=%04X:%c)",
		 symPtr->ucUpper,
		 symPtr->ucUpper);
     if ( symPtr->ucDown )
	 fprintf(fRep,"(DN=%04X:%c)",
		 symPtr->ucDown,
		 symPtr->ucDown);
     if ( symPtr->symbolName!=nil )
	 fprintf(fRep,":%s",
		 symPtr->symbolName->Str());
     fprintf(fRep,"\n");
     if ( isVerbose ) {
	 fprintf(fRep,"\tucCompat=%04X(%c),",
		 symPtr->ucCompat,
		 symPtr->ucCompat==0?' ':symPtr->ucCompat);
	 fprintf(fRep,"[%s]\n",
		 symPtr->strComptb ? symPtr->strComptb : "\0" );
     }
 }
 fprintf(fRep,"ucBin:::END\n");
 return 0;
}

int process_symbol (char* str,
		    bool doDescription,
		    sUnCodeSymbol& sym)
{
 ;
 // Returns 0 on successful parsing
 // or other internal error number.
 ;
 unsigned n, pos;
 char c;
 t_gUniPage aCode;

 if ( str==nil ) return -1;

 gString sTrim( str );
 // UniCode recommends that trailing #
 // should be also considered as comment.
 pos = sTrim.Find('#');
 if ( pos>0 ) {
     sTrim[pos] = 0;
     sTrim.Trim();
 }

 gParam aParam( sTrim, ";" );
 n = aParam.N();
 if ( n<14 ) return 1;

 if ( n==14 ) {
     aParam.Add( (char*)"\0" );
     n++;
 }

 gUniString sCode( aParam.Str(1) );
 aCode = sym.uniCode = sCode.GetHexa( e_DigConvUpper );

 //printf("<%s>,n=%u:",sTrim.Str(),n);
 //printf("UniCode=%04X\n",sym.uniCode);

 gString sIsoDesc( aParam.Str(2) );
 gString sUniDesc( aParam.Str(11) );

 if ( sIsoDesc.Find("<control>")==1 )
     sIsoDesc.Delete();

 ASSERTION(sym.symbolName==nil,"symbolName==nil");
 if ( doDescription ) {
     if ( sUniDesc.IsEmpty() )
	 sym.symbolName = new gString( sIsoDesc );
     else
	 sym.symbolName = new gString( sUniDesc );
 }

 gString sTemp;

 sTemp = aParam.Str(3);
 if ( sTemp.Length()>2 ) return 1;
 strcpy( sym.twoLetterCat, sTemp.Str() );

 sTemp = aParam.Str(5);
 if ( sTemp.Length()>3 ) return 1;
 strcpy( sym.thrLetterStyle, sTemp.Str() );

 sym.strComptb = strdup( aParam.Str(6) );
 gString sMyCompat( sym.strComptb );
 pos = sMyCompat.Find(" ");
 if ( sMyCompat[1]!='<' && pos>0 ) {
     gUniString sUcCompat( sMyCompat.Str() );
     sUcCompat.Delete( pos );
     t_uint32 uCompat = sUcCompat.GetHexa();
     if ( (uCompat & 0xF0000)==0 ) {
	 sym.ucCompat = uCompat;
     }
     else {
	 if ( sym.symbolName!=nil ) {
	     fprintf(stderr,"Warning: suppressed compatibility %04X for: %s\n",
		     uCompat,
		     sym.symbolName->Str());
	 }
     }
 }

 c = aParam.Str(10)[0];
 if ( c!=0 ) sym.cIsOpenClose = c;

 sym.extEqClose = 0;  //To be filled later

 sym.strNote = strdup( aParam.Str(12) );

 gUniString sHi( aParam.Str(13) );
 sym.ucUpper = sHi.GetHexa( e_DigConvUpper );

 gUniString sLo( aParam.Str(14) );
 sym.ucDown = sLo.GetHexa( e_DigConvUpper );

 sym.SetCode( aCode );

 return 0;
}

int report_ucbin (sOptUnCoder& opt,
		  sUnCodeBin& ucBin)
{
 // Returns 0 if binary file is ok;
 // 1 if input(s) failed;
 // 2 if is not a THB file.
 ;
 int error=0;
 bool isVerbose = opt.isVerbose;
 unsigned uIdx, uCount;
 ;
 sHtmlSymbol* pHtm;
 t_uchar cCompNote;
 t_gUniPage thisUcCompat;

 error = read_opt_ucbin( opt, ucBin );
 if ( error!=0 ) return 2;

 printf("THB version:%s, HTML:%c(%01u:%c), revision:%02u.%02u.%04u,%u\n",
	ucBin.verMajor,
	ucBin.verHtml, ucBin.cExtTbl, ucBin.HasHtml()?'H':'*',
	ucBin.revisionDate.GetDay(),
	ucBin.revisionDate.GetMonth(),
	ucBin.revisionDate.GetYear(),
	ucBin.revisionRef);
 if ( isVerbose ) {
     printf("Reference string (author): '%s'\n",ucBin.revReferenceStr);
     printf("Number of Symbols: %u\n",ucBin.n);
 }

 ASSERTION(ucBin.N()==ucBin.n,"ucBin.N()==ucBin.n");

 // Show: sequential numbering 128..255
 for (uIdx=0x80; uIdx<256 && uIdx<ucBin.n; uIdx++) {
     if ( uIdx!=ucBin.mapTab[uIdx].uniCode )
	 printf(":\tIndex 0x%02X has UniCode=0x%04X\n",
		uIdx,
		ucBin.mapTab[uIdx].uniCode);
 }
 printf(":::Map-Down2Uppercase\n");
 for (uIdx=0x80; uIdx<256 && uIdx<ucBin.n; uIdx++) {
     t_gUniPage thisUcCompat = ucBin.mapTab[uIdx].ucCompat;
     t_gUniPage ucUp = ucBin.mapTab[uIdx].ucUpper;
     if ( ucUp!=0 ) {
	 printf("ASCII 0x%02X (%s):%-2s: upper:%c, ucCompat:%c\n",
		uIdx,
		ucBin.mapTab[uIdx].twoLetterCat,
		ucBin.mapTab[uIdx].symbolName->Str(),
		ucUp>256 ? '?' : (char)ucUp,
		thisUcCompat==0 ? '*' : (char)thisUcCompat);
	 // ucBin.mapTab[uIdx].refVersLetter (D=68dec is UniCode 4.0)
     }
 }
 printf(":::Map-Upper2Downcase\n");
 for (uIdx=0x80; uIdx<256 && uIdx<ucBin.n; uIdx++) {
     t_gUniPage thisUcCompat = ucBin.mapTab[uIdx].ucCompat;
     t_gUniPage ucDn = ucBin.mapTab[uIdx].ucDown;
     if ( ucDn!=0 ) {
	 printf("ASCII 0x%02X (%s):%-2s: down:%c, ucCompat:%c\n",
		uIdx,
		ucBin.mapTab[uIdx].twoLetterCat,
		ucBin.mapTab[uIdx].symbolName->Str(),
		(char)ucDn,
		thisUcCompat==0 ? '*' : (char)thisUcCompat);
     }
 }

 if ( isVerbose ) {
     printf(":::Map-7bit\n");
     for (uIdx=0x20; uIdx<0x80 && uIdx<ucBin.n; uIdx++) {
	 t_gUniPage thisExtEqClose = ucBin.mapTab[uIdx].extEqClose;
	 thisUcCompat = ucBin.mapTab[uIdx].ucCompat;
	 printf("ASCII 0x%02X(%3u) %c[%c|%c]: %s\n",
		uIdx, uIdx,
		thisUcCompat==0 ? '*' : (thisUcCompat>0xFF ? '@' : (char)thisUcCompat),
		thisExtEqClose==0 ? '*' : (char)thisExtEqClose,
		ucBin.GetLetterCompat(uIdx),
		ucBin.mapTab[uIdx].symbolName->Str());
	 ASSERTION(ucBin.GetLetterCompat(uIdx)!=0,"...7bit");
     }
 }

 // Show: TABLE OF HTML-TRANS
 printf(":::HTML: %u symbols, ASCII-first-map:0x%02X\n",
	ucBin.nHtmlTrans,
	ucBin.cHtmAsciiStt);
 if ( isVerbose ) {
     for (uIdx=ucBin.cHtmAsciiStt, uCount=0; uCount<ucBin.nHtmlTrans; uIdx++, uCount++) {
	 if ( uCount==0 ) printf("Special HTML symbols:\n");
	 pHtm = &ucBin.mapHtm[uIdx];
	 cCompNote = pHtm->cCompNote;
	 switch ( cCompNote ) {
	 case 'd':
	     printf("::ASCII 0x%02X: %s (Distinct notations)\n",uIdx,pHtm->sTransHtm.Str());
	     break;
	 case 'e':
	     printf("::ASCII 0x%02X: %s (Non compatibility across browsers)\n",uIdx,pHtm->sTransHtm.Str());
	     break;
	 case 'a':
	     break;
	 default:
	     printf("::ASCII 0x%02X: invalid (%02udec)\n",uIdx,(unsigned)cCompNote);
	     break;
	 }
     }
     if ( uIdx==256 ) {
	 uIdx--;
	 pHtm = &ucBin.mapHtm[uIdx];
	 printf("E.g., ASCII 0x%02X(%3u): '%s', GetCompatLevel=%d, as '%s'\n",
		uIdx, uIdx,
		pHtm->sTransHtm.Str(),
		pHtm->GetCompatLevel(),
		pHtm->GetHtmlStr());
	 uIdx = 162;  // 'cent'
	 pHtm = &ucBin.mapHtm[uIdx];
	 printf("E.g., ASCII 0x%02X(%3u): '%s', GetCompatLevel=%d, as '%s'\n",
		uIdx, uIdx,
		pHtm->sTransHtm.Str(),
		pHtm->GetCompatLevel(),
		pHtm->GetHtmlStr());
	 uIdx = (unsigned)' ';
	 pHtm = &ucBin.mapHtm[uIdx];
	 printf("E.g., ASCII 0x%02X(%3u): '%s', GetCompatLevel=%d, as '%s'\n",
		uIdx, uIdx,
		pHtm->sTransHtm.Str(),
		pHtm->GetCompatLevel(),
		pHtm->GetHtmlStr());
	 uIdx = (unsigned)'<';
	 pHtm = &ucBin.mapHtm[uIdx];
	 printf("E.g., ASCII 0x%02X(%3u): '%s', GetCompatLevel=%d, as '%s'\n",
		uIdx, uIdx,
		pHtm->sTransHtm.Str(),
		pHtm->GetCompatLevel(),
		pHtm->GetHtmlStr());
	 uIdx = (unsigned)'b';
	 pHtm = &ucBin.mapHtm[uIdx];
	 printf("E.g., ASCII 0x%02X(%3u): '%s', GetCompatLevel=%d, as '%s'\n",
		uIdx, uIdx,
		pHtm->sTransHtm.Str(),
		pHtm->GetCompatLevel(),
		pHtm->GetHtmlStr());
     }
 }

 return error;
}

int process_uctext (sOptUnCoder& opt,
		    sUnCodeTxt& ucTxt,
		    sUnCodeBin& ucBin)
{
 // Returns 0 if files are ok and converted;
 // 1 if input(s) failed;
 // 2 if conversion had defects.
 ;
 int error=0;
 int iLine;
 unsigned i, n, len, maxLen=0;
 bool doVerbose = opt.isVeryVerbose;

 gString ucFilename( opt.GetTblTextName() );
 gFileFetch aF( ucFilename );
 if ( doVerbose ) {
     printf("Opening text input: %s (error=%d)\n",ucFilename.Str(),aF.lastOpError);
 }
 if ( aF.lastOpError!=0 ) return 1;

 n = aF.aL.N();
 for (i=1; i<=n; i++) {
     gString s( aF.aL.Str( i ) );
     s.Trim();
     if ( doVerbose ) printf("Processing: %u(%u) [%s]\n",i,ucTxt.linesProcessed,s.Str());
     if ( s.Find('#')==1 || s.IsEmpty()==true ) continue;
     ucTxt.linesProcessed++;
     ucTxt.sL.Add( s );
 }

 error = ucBin.AllocateSymbols( ucTxt.linesProcessed )==0;
 if ( error!=0 ) return error;

 for (iLine=1, i=1; iLine<=ucTxt.linesProcessed; iLine++, i++) {
     int thisError = process_symbol( ucTxt.sL.Str(i), true, ucBin.mapTab[i] );
     if ( thisError==0 ) {
	 if ( ucBin.mapTab[i].symbolName!=nil ) {
	     len = ucBin.mapTab[i].symbolName->Length();
	     if ( len>maxLen ) maxLen = len;
	 }
     }
     else {
	 ucTxt.failedNLinesL.Add( iLine );
	 error = 2;
     }
 }
 ucTxt.maxLen = maxLen;

 return error;
}

int create_ucbin (sUnCodeBin& ucBin,
		  sTxtHtmlTrans& tHtmlTrans,
		  char* thbFilename)
{
 int error;
 unsigned i, count, n;

 if ( thbFilename==0 || thbFilename[0]==0 ) return -1;

 // Convert fields (if any) from the HTML-TRANS into THB
 n = tHtmlTrans.N();
 if ( n>0 ) {
     ASSERTION(ucBin.mapHtm==nil,"create_ucbin(1)");
     ASSERTION(n<256 || tHtmlTrans.firstAscii<0x80,"create_ucbin(2)");
     ucBin.mapHtm = new sHtmlSymbol[ 256 ];
     ASSERTION(ucBin.mapHtm!=nil,"create_ucbin(3)");
     i = tHtmlTrans.firstAscii;
     DBGPRINT("DBG: tHtmlTrans.firstAscii=%u\n",i);
     ucBin.cExtTbl = '3'; //Has HTML-TRANS Table
     ucBin.nHtmlTrans = n;
     ucBin.cHtmAsciiStt = (t_uchar)i;
     for (count=1; count<=n; count++, i++) {
	 ASSERTION(i<256,"i<256");
	 ucBin.mapHtm[i].cCompNote = (t_uchar)tHtmlTrans.compL.GetInt(count);
	 ucBin.mapHtm[i].sTransHtm = (char*)tHtmlTrans.transL.Str(count);
	 DBGPRINT("DBG: HTML, ASCII: %u (%s)\n",i,ucBin.mapHtm[i].sTransHtm.Str());
     }
 }
 error = ucBin.SaveToFile( thbFilename );
 return error;
}

int parse_html_trans (gString& sFilename, sTxtHtmlTrans& res)
{
 ;
 // Returns 0 on success, and fills 'res'
 ;
 bool isOk, isComment, hasStarted=false;
 unsigned i, n, aLen, uAscii=0;
 char c;
 gFileFetch aF( sFilename );

 if ( aF.lastOpError!=0 ) return 1;
 n = aF.aL.N();
 for (i=1; i<=n; i++) {
     gString s( aF.aL.Str(i) );
     s.Trim();
     gParam aParam( s, ";" );
     aLen = aParam.N();
     isComment = s[1]=='#';
     isOk = s[1]=='&' && aLen>=2;
     if ( s.IsEmpty() ) continue;
     if ( hasStarted ) {
	 if ( isComment==true || isOk==false ) return 2;
	 gString sHtml( aParam.Str(1)+1 );
	 c = aLen==2 ? 'a' : aParam.Str(3)[0];
	 res.transL.Add( sHtml );
	 res.compL.Add( (int)c );
	 uAscii++;
     }
     else {
	 if ( isComment ) continue;
	 gParam oParam( s, ":" );
	 aLen = oParam.N();
	 if ( aLen<2 ) return 2;
	 if ( sscanf(oParam.Str(2),"%X",&uAscii)!=1 )
	     return 2;
	 if ( uAscii<=0x80 )
	     return 2;
	 res.firstAscii = uAscii;
	 hasStarted = true;
     }
 }
 DBGPRINT_MIN("Processed uAscii=%u, firstAscii=%u, tot=%u(would be:%u)\n",
	      uAscii,
	      res.firstAscii,
	      res.N(),
	      256-res.firstAscii);
 return 0;
}

int create_cpbin (sUnCodeBin& ucBin,
		  bool isVerbose,
		  char* cptxtFilename,
		  char* cpBinFilename)
{
 int error;
 unsigned i, n, pos, uCount;
 unsigned long ulVal;
 char cBuf[100];
 short usageCount[256];
 t_gUniPage aCode;
 //t_uint16 aUD2;
 t_uint32 aUD4;

 gList sL, commentsL;
 sTcbBin tcBin;
 FILE* f;
 FILE* fIn=stdin;
 FILE* fRep=isVerbose?stdout:nil;

 tcBin.revisionDate.SetDay( 20031231 );

 ASSERTION(cptxtFilename!=nil,"cptxtFilename!=nil");
 ASSERTION(cpBinFilename!=nil,"cpBinFilename!=nil");
 // Input: cptxtFilename
 gFileFetch aF( cptxtFilename );
 error = aF.lastOpError;
 if ( error!=0 ) {
     fprintf(stderr,"Cannot open text source: %s\n",cptxtFilename);
     return 1;
 }

 n = aF.aL.N();
 for (i=1; i<=n; i++) {
     gString sTemp( aF.aL.Str(i) );
     gString sComment;
     sTemp.TrimRight();
     // Discard any comments
     pos = sTemp.Find('#');
     if ( sTemp.Length()<4 || pos==1 ) continue;
     if ( pos>0 ) {
	 sComment.CopyFromTo(sTemp,pos+1);
	 sComment.Trim();
	 sTemp.Delete(pos);
     }
     sTemp.Trim();
     sL.Add( sTemp );
     commentsL.Add( sComment );
 }

 n = sL.N();
 if ( n<1 ) return 1;

 // sL contains all relevant convertions
 // Check whether it is a linear table or not
 bool isLinear, isSeq=true;
 t_gPageCode lastV=0;
 sCodeToUni conv;
 long val1, val2;
 for (i=1; i<=n; i++) {
     aF.Buffer()[0] = 0;
     val2 = 0;
     error =
	 sscanf(sL.Str(i),"%lx%lx%s",
		&val1,
		&val2,
		aF.Buffer());
     if ( error>2 && aF.Buffer()[0]!=0 )
	 return 2;
     if ( error<1 ) {
	 fprintf(stderr,"Invalid input: '%s', '%s'\n",
		 sL.Str(i),
		 commentsL.Str(i));
	 return 3;
     }
     if ( error<2 ) {
	 fprintf(stderr,"Ignored input: '%s', '%s'\n",
		 sL.Str(i),
		 commentsL.Str(i));
     }
     conv.lCode = (t_gPageCode)val1;
     conv.uniCode = (t_gUniPage)val2;
     if ( (lastV>0 && conv.lCode<=lastV) || val1<0L || val2<0L ) {
	 fprintf(stderr,"Error: This code=%ld, last code=%ld\n",val1,(long)lastV);
	 return 4;
     }
     if ( lastV>0 && conv.lCode!=lastV+1 ) {
	 if ( isSeq ) {
	     printf("lastV=%ld(%lXh), thisV=%ld(%lXh)\n",
		    (long)lastV, (long)lastV,
		    (long)conv.lCode, (long)conv.lCode);
	 }
	 isSeq = false;
     }
     lastV = conv.lCode;
 }

 isLinear = lastV<256;
 if ( isLinear ) {
     tcBin.cLinearInd = 'a'+(isSeq==false);
     tcBin.nSymb = 256;
 }
 else {
     tcBin.cLinearInd = 'B';
     tcBin.bigTbl = new sCodeToUni[ n ];
     tcBin.nSymb = n;
 }

 for (i=1, uCount=0; i<=n; i++, uCount++) {
     sscanf(sL.Str(i),"%lx%lx",
	    &val1,
	    &val2);
     conv.lCode = val1;
     conv.uniCode = val2;
     lastV = conv.lCode;
     if ( lastV<256 ) {
	 tcBin.linearTbl[lastV] = conv.uniCode;
     }
     if ( isLinear==false ) {
	 tcBin.bigTbl[uCount].lCode = lastV;
	 tcBin.bigTbl[uCount].uniCode = conv.uniCode;
     }
 }

 // Accept no translation on characters before ASCII 0x20 (blank)
 lastV = 0;
 for (i=1; i<0x20; i++) {
     lastV += tcBin.linearTbl[i]>0;
 }
 if ( lastV==0 ) {
     // Fill-up characters linearly translated before ASCII 0x20
     for (i=1; i<0x20; i++) {
	 tcBin.linearTbl[i] = i;
     }
     if ( fRep!=nil ) fprintf(fRep,"::: Linearly translated 0x00 through 0x20.\n");
 }

 // Create cpBinFilename (*.tcb)
 f = fopen(cpBinFilename,"wb");
 if ( f==NULL ) return -1;

 fprintf(stderr,"I:ISO, C:Code-page D:other:\n> ");
 error = fscanf( fIn, "%c", &tcBin.cNormRef );
 if ( error<1 || tcBin.cNormRef<'A' || tcBin.cNormRef>'Z' )
     return -1;

 fprintf(stderr,"Short description tag:\n> ");
 memset(cBuf,0x00,20);
 error = fscanf( fIn, "%s", cBuf );
 if ( error<1 ) return -1;
 strncpy( tcBin.sDescTag, cBuf, 12 );

 fprintf(stderr,"cpCode (e.g. 885902 for ISO8859-2):\n> ");
 error = fscanf( fIn, "%lu", &ulVal );
 if ( error<1 ) return -1;
 tcBin.cpCode = (t_gCpCode)ulVal;

 fwrite(tcBin.verMajor,6,1,f); // 5octet+0x00
 aUD4 = TO_UD4(72);
 fwrite(&aUD4,4,1,f);
 fwrite(&tcBin.cLevel,1,1,f);
 aUD4 = TO_UD4(tcBin.revisionDate.GetValue32());
 fwrite(&aUD4,4,1,f);
 fwrite(&tcBin.revisionRef,1,1,f);
 fwrite(&tcBin.cNormRef,1,1,f);
 fwrite(tcBin.sDescTag,13,1,f); // 12octet+0x00

 memset(cBuf,0x20,100);
 fwrite(cBuf,34,1,f); // sDescStr

 aUD4 = TO_UD4(tcBin.cpCode);
 fwrite(&aUD4,4,1,f);

 fwrite(&tcBin.cLinearInd,1,1,f);

 memset(cBuf,0x00,20);
 cBuf[1] = cBuf[2] = 0x20;
 fwrite(cBuf,3,1,f);

 aUD4 = TO_UD4(tcBin.nSymb);
 fwrite(&aUD4,4,1,f);

 for (i=0; i<256; i++) usageCount[i] = 0;

 if ( tcBin.IsLinear() ) {
     // :TABLE OF SYMBOLS
     ASSERTION(tcBin.nSymb==256,"tcBin.nSymb==256");
     for (i=1; i<256; i++) {
	 aCode = tcBin.linearTbl[i];
	 aUD4 = TO_UD4(aCode);
	 fwrite(&aUD4,4,1,f);
	 if ( aCode<256 ) usageCount[aCode]++;
     }
 }
 else {
     printf("#TODO\n");
 }

 // :TABLE OF SYMB-BEST-MAP
 aUD4 = 0;
 error = fwrite(&aUD4,4,1,f);
 // :TABLE OF COUNTRIES-SERVED
 aUD4 = 0;
 error = fwrite(&aUD4,4,1,f);

 // Close file
 fclose( f );

 // Some reports, if required

 if ( fRep!=nil ) {
     // Report higher-than-0xFF mapped symbols
     fprintf(fRep,"::: Non UniCode lower-part code page conversion\n");
     for (i=1; i<256; i++) {
	 aCode = tcBin.linearTbl[i];
	 if ( aCode==0 || aCode>256 ) {
	     fprintf(fRep,"cpCode=%02X: mapped %04lX {%s}\n",
		     i,
		     (unsigned long)aCode,
		     commentsL.Str(i+1));
	 }
     }

     // Report more than one match in UniCode (lower-part)
     fprintf(fRep,"::: Multiple UniCode lower-part code page conversion\n");
     for (i=1; i<256; i++) {
	 short thisCounter = usageCount[i];
	 if ( thisCounter>1 ) {
	     fprintf(fRep,"uniCode=%04X: occ=%d\n",i,thisCounter);
	 }
     }

     // Report UniCode (lower-part) names unmatched
     fprintf(fRep,"::: Umatched UniCode lower-part code page conversion\n");
     for (i=1; i<256; i++) {
	 if ( usageCount[i]==0 ) {
	     ASSERTION(ucBin.mapTab!=nil,"ucBin.mapTab!=nil");
	     fprintf(fRep,"uniCode=%04X unmapped: %s\n",
		     i,
		     ucBin.mapTab[i].symbolName==nil ? "?" : ucBin.mapTab[i].symbolName->Str());
	 }
     }
 }// IF rep!=nil (informative output)

 return error<1 ? -1 : 0;
}

int report_cpbin (bool isVerbose,
		  FILE* fRep,
		  FILE* fRepErr,
		  char* cpBinFilename)
{
 int error;
 FILE* fIn;
 sTcbBin tcBin;

 if ( cpBinFilename==nil || cpBinFilename[0]==0 ) return -1;
 fIn = fopen(cpBinFilename,"rb");
 if ( fIn==NULL ) {
     fprintf(fRepErr,"Cannot access codepage file (tcb): %s\n",cpBinFilename);
     return 2;
 }

 error = tcBin.ReadFromStream( fIn );
 if ( error!=0 ) {
     fclose( fIn );
     if ( isVerbose ) fprintf(fRepErr,"tcBin error code: %d\n",error);
     return 1;
 }

 fprintf(fRep,"verMajor:%s\n",tcBin.verMajor);
 fprintf(fRep,"Year=%u,Month=%02u,Day=%02u\n",
	 tcBin.revisionDate.GetYear(),
	 tcBin.revisionDate.GetMonth(),
	 tcBin.revisionDate.GetDay());
 fprintf(fRep,"revisionRef=%u\n",tcBin.revisionRef);
 fprintf(fRep,"cNormRef:'%c'\n",tcBin.cNormRef);
 fprintf(fRep,"sDescTag:%s\n",tcBin.sDescTag);
 fprintf(fRep,"sDescStr:%s\n",tcBin.sDescStr.Str());
 fprintf(fRep,"cpCode=%lu\n",(unsigned long)tcBin.cpCode);
 fprintf(fRep,"cLinearInd='%c'\n",tcBin.cLinearInd);
 fprintf(fRep,"nSymb:%lu\n",(unsigned long)tcBin.nSymb);
 fprintf(fRep,"linearTbl==nil?%c\n",ISyORn(tcBin.linearTbl==nil));
 fprintf(fRep,"bigTbl==nil?%c\n",ISyORn(tcBin.bigTbl==nil));
 fprintf(fRep,"---END tcb---\n");

 fclose( fIn );

 return 0;
}

int dump_file_use_cp (sTcbBin& tcBin,
		      char* inFilename,
		      FILE* fOut,
		      int level)
{
 unsigned i, k, len, n;
 char* str;
 t_uchar c1, c2;
 t_gUniPage* linearTbl = tcBin.linearTbl;
 t_gUniPage uniConv;

 ASSERTION(inFilename!=nil,"inFilename!=nil");
 ASSERTION(linearTbl!=nil,"linearTbl!=nil");

 gFileFetch aF( inFilename );
 if ( aF.IsOpened()==false ) {
     fprintf(stderr,"Cannot open file: %s\n",inFilename);
     return 2;
 }

 n = aF.aL.N();
 for (i=1; i<=n; i++) {
     str = aF.aL.Str(i);
     gString s( str );
     len = s.Length();
     for (k=1; k<=len; k++) {
	 c1 = s[k];
	 uniConv = linearTbl[c1];
	 c2 = uniConv<127 ? (t_uchar)uniConv : 0;
	 ASSERTION(c1!='\n',"line-feed here?");
	 switch ( level ) {
	 case 1:
	     if ( c2==0 )
		 printf("[%04X]",(unsigned)uniConv);
	     else
		 printf("%c",c2);
	     break;
	 case 2:
	     fprintf(fOut,"0x%04X\n",uniConv);
	     break;
	 default:
	     fprintf(fOut,"%c",c2==0?'?':c2);
	 }
     }
     if ( level<=1 ) fprintf(fOut,"\n");
 }
 return 0;
}
////////////////////////////////////////////////////////////
int print_help (char* progStr)
{
 char
     *msgHelp = "%s - %s\n\n\
Usage:\n\
        %s command [options] [file ...]\n\
\n\
Commands are:\n\
   report       Parses and prints UniCode file\n\
   create       Creates thb or tcb binary files.\n\
   dump         Dumps a file coded using a given code page.\n\
\n\
Options are:\n\
   -h           This help (or --help / --version)\n\
   -v           Verbose (use twice, more verbose)\n\
\n\
   -p X         Use base path as X (or --path)\n\
   --html X     Use X as HTML-trans text\n\
   -c X         Use X as the code page (or --codepage)\n\
   -t X         Use X as the code page text (or --codepage-txt)\n\
   -d N         Use date-day, not today (or --date-day)\n\
\n";

 printf(msgHelp,
	progStr,
	thisProgramVersion,
	progStr);
 return 0;
}
////////////////////////////////////////////////////////////
int print_version (char* progStr)
{
 char
     *msgVersion = "%s - %s\n\
Written by Henrique Moreira.\n\
\n\
Copyright (C) %u %s.\n\
%s\n";

 printf(msgVersion,
	progStr,
	thisProgramVersion,
	thisProgramYear,thisProgramCopyright,
	thisProgramCopy);
 return 0;
}
////////////////////////////////////////////////////////////
int adapt_def_table (sOptUnCoder& opt, char* str)
{
 if ( str==NULL ) {
     opt.sUnTblTextName.Set( "unicode256.thb" );
     return 0;
 }
 opt.sUnTblTextName = str;
 if ( opt.sUnTblTextName.Find( gSLASHCHR )>0 ) {
     // If no path was given, then assume default path;
     // otherwise set the default path empty
     opt.sTblPath.SetEmpty();
 }
 return 0;
}

int intext_strings (char* fName, sOptUnCoder& opt, sUnCodeBin& ucBin)
{
 unsigned i, n;
 gFileFetch iF( fName );

 n = iF.aL.N();
 for (i=1; i<=n; i++) {
     fprintf(stdout,"LINE:TODO:%u/%u:<%s>\n",i,n,iF.aL.Str(i));
 }
 return 0;
}

int intext_strings_many (gArg& arg, sOptUnCoder& opt, sUnCodeBin& ucBin)
{
 int error, nErrors=0;
 unsigned i, n=arg.N();
 char* str;

 ASSERTION(n>1,"n>1");
 n--;
 for (i=1; i<=n; i++) {
     str = arg.Str( i+1 );
     error = intext_strings( str, opt, ucBin );
     if ( error!=0 ) {
	 nErrors++;
	 fprintf(stderr,"Error code %d reading file: %s\n",error,str);
	 //TODO:;#;
     }
 }
 return 0;
}
////////////////////////////////////////////////////////////
int go (char** argv, char** envp, int& result)
{
 int error;
 int level;
 unsigned i, n, nL;
 long val;
 unsigned long possiblyRelevantCodePage=0;
 bool isVerbose;
 bool isCpBinCreate;
 gArg arg( argv, envp );
 sOptUnCoder opt;
 char* str = nil;
 FILE* fOut = stdout;
 FILE* fRepErr = stdout;

 result = 0;

 arg.AddParams( 1, "-vv|--verbose -h|--help --version\
 -p:%s|--path\
 --html:%s\
 -c:%s|--codepage:%s\
 -t:%s|--codepage-txt:%s\
 -d:%d|--date-day:%d\
");
 error = arg.FlushParams();
 n = arg.N();
 arg.FindOptionOccurr( "verbose", opt.isVerbose, opt.isVeryVerbose );
 isVerbose = opt.isVerbose;

 if ( arg.FindOption('h') || error!=0 ) {
     print_help( arg.Program() );
     return 0;
 }
 if ( arg.FindOption("version") ) {
     print_version( arg.Program() );
     return 0;
 }

 arg.FindOption('p',opt.sTblPath);

 if ( arg.FindOption("html",opt.sHtmlFilename) ) {
     // Parse sHtmlFilename
     error = parse_html_trans( opt.sHtmlFilename, opt.tHtmlTrans );
     if ( error!=0 ) {
	 fprintf(fRepErr,"Error-code %d in tht file: %s\n",
		 error,
		 opt.sHtmlFilename.Str());
	 return 1;
     }
 }

 isCpBinCreate = arg.FindOption("codepage-txt",opt.sCpTxtFilename);

 if ( arg.FindOption("codepage",opt.sCodePage) ) {
     if ( opt.sCodePage.IsEmpty() ) return 2;  // Just bail-out if '--codepage "" was entered'
 }
 else {
     // No '--codepage cpNNN...tcb' specified:
     // finding default file name for 'tcb'.
     error = gStrControl::Self().GetFilenameCode( opt.defCpCode, opt.sCodePage );
     if ( error!=0 ) {
	 fprintf(fRepErr,"No suitable code-page tcb file for cp-code: %lu\n",(unsigned long)opt.defCpCode);
	 return -1;
     }
     else {
	 possiblyRelevantCodePage = (unsigned long)opt.defCpCode;
     }
 }

 gDateTime oDtTm;
 gDateTime aDtTm( gDateTime::e_Now );
 t_uint32 aDdNow = convert_DateTime2DateDay( aDtTm );

 DBGPRINT("DBG: aDdNow=%lu\n",(unsigned long)aDdNow);
 opt.aDateDay = aDdNow;

 if ( arg.FindOption("date-day",val) ) {
     DBGPRINT("DBG: aDtTm timestamp(e_Now): %lu (val=%ld)\n",(unsigned long)aDtTm.GetTimeStamp(),val);
     val = convert_DateDay2DateTime( val, opt.aDateDay, oDtTm );
     // Check if date is invalid
     if ( val<=0 ) {
	 fprintf(fRepErr,"Invalid date-day: %s: %lu.\n",
		 val==0 ? "do you mean now? Then suppress this option or use" : "Current date-day is",
		 (unsigned long)aDdNow);
	 return 1;
     }
 }

 if ( n==0 ) {
     fprintf(fRepErr,"No command.\n");
     return 0;
 }
 char* cmdStr = arg.Str( 1 );
 // parsing nearly ends...

 sUnCodeTxt ucTxt;
 sUnCodeBin ucBin;

 ucBin.SetVersion( '1', 'A' );

 if ( strcmp(cmdStr,"report")==0 ) {
     if ( n>=2 ) {
	 adapt_def_table( opt, arg.Str(2) );
     }
     // If it's a binary file, 'report' it and return!
     if ( (error = report_ucbin( opt, ucBin ))!=2 ) return error;
     error = process_uctext( opt, ucTxt, ucBin );
     nL = ucTxt.failedNLinesL.N();
     switch ( error ) {
     case 0:
	 print_ucbin( ucBin, isVerbose, stdout );
	 //printf("Max-desc-length=%u\n",ucTxt.maxLen)
	 break;
     case 1:
	 fprintf(fRepErr,"Unable to use tbt file(s), install-path '%s', %s\n",
		 opt.sTblPath.Str(),
		 opt.sUnTblTextName.Str());
	 break;
     default:
	 fprintf(fRepErr,"Error %d: Invalid conversion.\n",error);
	 fprintf(fRepErr,"Processed %d lines.\n",ucTxt.linesProcessed);
	 fprintf(fRepErr,"The following lines failed:\n");
	 for (i=1; i<=nL; i++) {
	     fprintf(fRepErr,"%d:%s\n",
		     ucTxt.failedNLinesL.GetInt( i ),
		     ucTxt.sL.Str( i ));
	 }
     }
     return error;
 }

 if ( strcmp(cmdStr,"create")==0 ) {
     if ( n<3 ) {
	 const char* strOutputThbFix="new_binary_table-pt.thb";
	 fprintf(fRepErr,"'Create' must have two files.\n");
	 printf("Example:\n\
\trm /usr/share/unicodeplus/%s\n\
\t%s create -p /usr/share/unicodeplus \\\n\
\t\tunicode-pt.tbt %s\n\
# -p %s stands for the local installation in use.\n",
		strOutputThbFix,
		arg.Program(),
		strOutputThbFix,
		opt.sTblPath.Str());
	 return 1;
     }
     opt.sUnTblTextName = arg.Str(2);
     str = arg.Str(3);
     gString sThbName( opt.sTblPath );
     if ( sThbName.IsEmpty()==false ) sThbName.Add( gSLASHCHR );
     sThbName.Add( str );
     str = sThbName.Str();
     gFileText fTest( str );
     if ( fTest.IsOpened() ) {
	 fprintf(fRepErr,"'Create' will not overwrite file: %s\n",str);
	 return 1;
     }
     if ( fTest.Overwrite( str )==false ) {
	 fprintf(fRepErr,"'Create' cannot write file: %s\n",str);
	 return 1;
     }
     fTest.Close();
     if ( isCpBinCreate ) {
	 error = read_ucbin( opt.sUnTblTextName.Str(), ucBin );
	 if ( error!=0 ) return error;
	 error = create_cpbin( ucBin, isVerbose, opt.sCpTxtFilename.Str(), str );
     }
     else {
	 error = process_uctext( opt, ucTxt, ucBin );
	 if ( error!=0 ) return error;
	 ucBin.revisionDate.SetDay( opt.aDateDay );  // e.g.20031231
	 if ( isVerbose ) {
	     printf("n=%u,",ucBin.N());
	     printf("Year=%u,Month=%02u,Day=%02u (0x%08X)\n",
		    ucBin.revisionDate.GetYear(),
		    ucBin.revisionDate.GetMonth(),
		    ucBin.revisionDate.GetDay(),
		    ucBin.revisionDate.GetValue32());
	 }

	 // Filling-up the name: I know this code is anoying for the programmer
	 // and for the author, but I want real names in authors of THBs.
	 fprintf(stderr,"Please type your name (16 characters allowed, minimum of 4): ");
	 char keyPress, sAuthor[ 52 ];
	 int nameLen, iterName=-1;
	 do {
	     do {
		 do {
		     memset( sAuthor, 0x0, 51 ); // Absolutely clear all buffer
		 } while ( fgets( sAuthor, 50, stdin )==nil || sAuthor[0]=='\n' || sAuthor[0]==0 ) ;
		 if ( strcmp(sAuthor,".\n")==0 ) {
		     strcpy( sAuthor, ucBin.revReferenceStr );
		     fprintf(stderr,"Are you sure you are: %s ? (Y/N) ",sAuthor);
		     while ( fscanf(stdin,"%c",&keyPress)!=1 || (keyPress!='Y' && keyPress!='N') ) ;
		     if ( keyPress!='Y' ) return result = 8;  // Not sure..., bail out
		 }
		 else {
		     sAuthor[ strlen(sAuthor)-1 ] = 0;  // Remove '\n'
		 }
		 nameLen = strlen( sAuthor );
	     } while ( nameLen<4 || nameLen>16 );
	     for (iterName=0; iterName>=0 && iterName<nameLen; iterName++) {
		 if ( sAuthor[iterName]<' ' || sAuthor[iterName]==127 ) iterName = -1;
	     }
	 } while ( iterName<=0 );
	 strcpy( ucBin.revReferenceStr, sAuthor );

	 // We are done, finally the real stuff, below:
	 error = create_ucbin( ucBin, opt.tHtmlTrans, str );
	 if ( error!=0 ) {
	     fprintf(fRepErr,"Creation of THB failed, error-code %d: %s\n",error,str);
	 }
	 if ( isVerbose ) {
	     if ( ucBin.HasHtml()==false ) fprintf(fRepErr,"Warning: no HTML!\n");
	     if ( error==0 ) {
		 fprintf(fRepErr,"Created THB, author: %s\n",sAuthor);
	     }
	 }
     }
     return error;
 }


 if ( strcmp(cmdStr,"dump")==0 ) {
     if ( n<2 ) {
	 fprintf(fRepErr,"'Dump' must have at least one input file.\n");
	 return 1;
     }
     str = opt.sCodePage.Str();
     DBGPRINT_MIN("DBG: opt.sCodePage='%s'\n",str);
     error = 0;
     if ( isVerbose ) {
	 fprintf(fRepErr,"Using the default code-page: %lu\n",possiblyRelevantCodePage);
	 error = report_cpbin( opt.isVeryVerbose, stderr, fRepErr, str );
     }
     gString sFile( arg.Str( 2 ) );
     if ( sFile.IsEmpty() ) return error;

     sTcbBin tcBin;
     if ( tcBin.ReadFromFile( str )!=0 ) {
	 fprintf(stderr,"Invalid tcb: %s\n",str);
	 return 2;
     }
     level = (int)isVerbose + (int)opt.isVeryVerbose;
     dump_file_use_cp( tcBin, sFile.Str(), fOut, level );
     return error;
 }

#ifdef DEBUG
 if ( strcmp(cmdStr,"strings")==0 ) {
     adapt_def_table( opt, NULL );
     error = read_opt_ucbin( opt, ucBin );
     if ( error!=0 ) return 2;
     error = intext_strings_many( arg, opt, ucBin );
     return error;
 }
#endif //DEBUG (not ready...)

 fprintf(stderr,"Invalid command: %s\n",cmdStr);
 return 1;
}
////////////////////////////////////////////////////////////
int main (int argc, char** argv, char** envp)
{
 int error, result=0;

 gINIT;

 error = go( argv, envp, result );

#ifdef DEBUG_CHECK
 sUnCodeBin ucChk;
 error = ucChk.ReadFromFile( "test.tbt" );
 print_ucbin( ucChk, false, stdout );
#endif //DEBUG_...

 gEND;

 DBGPRINT_MIN("go: %d, result=%d\n",error,result);

 return result;
}

