// gUnweb.cpp -- Version 1.0


#define thisProgramVersion "Version 1.0"
#define thisProgramCopyright "Prized Season & Sons"
#define thisProgramYear 2009
#define thisProgramCopy "This is free software (GPL)\n\
There is no warranty, not even for MERCHANTABILITY or\n\
FITNESS FOR A PARTICULAR PURPOSE."

#include <stdio.h>
#include <string.h>
#include "garg.h"
#include "gweb.h"
#include "gstringext.h"
#include "gHtmlParser.h"
#include "gHtmlCtrl.h"

#ifdef gDOS_SPEC
#include <windows.h>
#endif //gDOS_SPEC
////////////////////////////////////////////////////////////
struct sOptUnweb {
    sOptUnweb ()
	: isVerbose( false ),
	  isVeryVerbose( false ),
	  doViewAll( false ),
	  doSkip( false ),
	  isOutDOS( false ),
	  isDumpLinear( true ),
	  isTagUppercase( false ),
	  doShowTree( false ) {
	gHtmlCtrl::Self().Init();
    }
    ~sOptUnweb () {
	gHtmlCtrl::Self().ReleaseHash();
    }
    bool isVerbose, isVeryVerbose;
    bool doViewAll;
    bool doSkip;
    bool isOutDOS;
    bool isDumpLinear;
    gString sOutOptFile;
    gString sDumpFile;
    // HTML-opt-ins
    bool isTagUppercase;
    bool doShowTree;  // Mainly for debug
    gHtmlOpt htmlOpt;
    // ...other
    gString sCurFile;
    gString sRange;
    gFileOut fOutAll;
    // Get methods
    bool HasDumpFile () {
	return sDumpFile.IsEmpty()==false;
    }
    char* NewLine () {
	return fOutAll.NewLine();
    }
    // Set methods
    bool SetOutDOS (bool isDOS) {
	fOutAll.SetModeDOS( isDOS );
	return isDOS;
    }
};
////////////////////////////////////////////////////////////
int print_help (char* progStr)
{
 char
     *msgHelp = "%s - %s\n\n\
Usage:\n\
        %s command [options] [file ...]\n\
\n\
Commands are:\n\
   check        Check correctness.\n\
   dump         Dump text linearly.\n\
   show         Show text.\n\
\n\
Options are:\n\
   -h              This help (or --help / --version)\n\
   -v              Verbose (use twice, more verbose)\n\
   -o X            Output options (or --out-options)\n\
   -d X            Checking or dump to file X (or --dump)\n\
   --base-href X   Set BASE HREF as X\n\
   --skip          Skip syntax check\n\
   --view          View all details\n\
   --dos-mode      DOS-mode (CR&LF)\n\
   --range X       Dump in range X; @N suppress line n\n\
   --tag-uppercase Convert tags to uppercase.\n\
   --dump-linear   Linear dump (non-tags in the same line)\n\
\n";

 // ===> Debugging: --show-tree
 printf(msgHelp,
	progStr,
	thisProgramVersion,
	progStr);
 return 0;
}
////////////////////////////////////////////////////////////
int print_version (char* progStr)
{
 char
     *msgVersion = "%s - %s\n\
Written by Henrique Moreira.\n\
\n\
Copyright (C) %u %s.\n\
%s\n";

 printf(msgVersion,
	progStr,
	thisProgramVersion,
	thisProgramYear,thisProgramCopyright,
	thisProgramCopy);
 return 0;
}
////////////////////////////////////////////////////////////
// glib required functions
////////////////////////////////////////////////////////////
int gglobal_EventTerminate (int v)
{
 return 0;
}
////////////////////////////////////////////////////////////
int dump_list_normal (gUnweb& unweb, FILE* fOut, unsigned startLine, unsigned endLine, sOptUnweb& opt, gHtmlOpt& htmlOpt)
{
 unsigned i;
 char* str;

 for (i=startLine; i<=endLine; i++) {
     str = unweb.Str( i );
     DBGPRINT_MIN("DBG: %u/%u: %s|%s\n",i,endLine,htmlOpt.outOpt.HasToSkip( str )?"SKIP-OUT":".",htmlOpt.inOpt.HasToSkip( i )?"SKIP-IN":".");
     if ( htmlOpt.outOpt.HasToSkip( str ) || htmlOpt.inOpt.HasToSkip( i ) ) continue;
     fprintf(fOut,"%s%s",str,opt.NewLine());
 }
 return 0;
}

int flush_linear_list (FILE* fOut, gList& lNonTag, char* newLineStr)
{
 unsigned k, n;
 for (k=1, n=lNonTag.N(); k<=n; k++) {
     fprintf(fOut,"%s%s",lNonTag.Str(k),k<n?" ":"\0");
 }
 if ( n>0 ) fprintf(fOut,"%s",newLineStr);
 return 0;
}

int dump_list_linear (gUnweb& unweb, FILE* fOut, unsigned startLine, unsigned endLine, sOptUnweb& opt, gHtmlOpt& htmlOpt)
{
 unsigned i;
 char* str;
 bool isTag;
 bool inSkip;
 gList lNonTag;

 for (i=startLine; i<=endLine; i++) {
     str = unweb.Str( i );
     inSkip = htmlOpt.inOpt.HasToSkip( i );
     ///if ( inSkip ) printf("SKIP: %d: %s\n",i,str)
     if ( htmlOpt.outOpt.HasToSkip( str ) || inSkip ) continue;
     isTag = str[0]=='<';
     if ( isTag ) {
	 flush_linear_list( fOut, lNonTag, opt.NewLine() );
	 lNonTag.Delete();
	 fprintf(fOut,"%s%s",str,opt.NewLine());
     }
     else {
	 lNonTag.Add( str );
     }
 }
 flush_linear_list( fOut, lNonTag, opt.NewLine() );
 return 0;
}
////////////////////////////////////////////////////////////
// Specific functions
////////////////////////////////////////////////////////////
int check_web_list (gUnweb& unweb, FILE* fOut, FILE* fRepErr, sOptUnweb& opt)
{
 int error;
 unsigned i, n;
 unsigned nPost;
 char* thisFilename;
 bool isOk;
 bool doSuppressPickyWarns = opt.isVerbose==false;
 FILE* fOutErr;

 ASSERTION(fOut!=nil,"fOut!=nil");

 if ( opt.doViewAll || fOut!=stdout )
     fOutErr = fOut;
 else
     fOutErr = fRepErr;

 opt.fOutAll.f = fOutErr;
 // Refer to bookmark ptr(gcc): when the pointer-to is referred, also clean this one.

 error = unweb.Dump( fOutErr );

 thisFilename = opt.sCurFile.Str();
 n = unweb.aL.N();
 nPost = unweb.coordSerial.N();

 if ( fOut!=nil && fOut!=stdout && opt.doSkip==true ) {
     for (i=1; i<=nPost; i++) fprintf(fOut,"%s%s",unweb.Str( i ),opt.NewLine());
 }

 if ( opt.isVerbose ) {
     if ( unweb.coordParseErr.N()>0 ) {
	 if ( opt.isVeryVerbose ) fprintf(fOutErr,"Parse errors follow below for: %s\n",thisFilename);
	 unweb.coordParseErr.Report(fOutErr,true);
     }
 }
 if ( opt.isVeryVerbose ) {
     fprintf(fOutErr,"::: %s :::\n",thisFilename);
     fprintf(fOutErr,"Lines-in:    %u\n",n);
     fprintf(fOutErr,"Lines-out:   %u\n",nPost);
     if ( error!=0 ) {
	 fprintf(fOutErr,"HTML errors: (n=%d, line0=%d, lineL=%d)\n",
		 unweb.nHtmlErrors,
		 unweb.firstHtmlErrLNr,
		 unweb.lastHtmlErrLNr);
     }
 }

 if ( opt.doSkip ) return error;
 if ( error!=0 ) return error;

 DBGPRINT("DBG:SYN:START%d\n",error);

 gHtmlParser hParse( &unweb );
 hParse.dbgLevel = opt.doShowTree ? 9 : (opt.isVerbose ? 3 : 0);

 //
 // Interesting: HOW TO ADD an optional-ended tag by configuration:
 opt.htmlOpt.AddIdTagOptEnd( 25 /*DIV*/ );

 opt.htmlOpt.oErrStateSuppress[e_HS_Html].SetOn( doSuppressPickyWarns ); // Avoid "Warning: Line N: Not in LTD sequence: <HEAD> (int-seq:104)
 opt.htmlOpt.oErrStateSuppress[e_HS_Body].SetOn( doSuppressPickyWarns ); // Avoid "Warning: Line N: Not in LTD sequence (int-seq:106)
 //

 // Avoid warning when <HTML> tag is missing:
 opt.htmlOpt.oErrStateSuppress[e_HS_HeadAfter].SetOn( true );

 isOk = hParse.SetOptions( opt.htmlOpt );
 if ( isOk==false ) {
     fprintf(stderr,"Invalid HTML-options.\n");
     return -1;
 }

 error = hParse.Parse( fOutErr );

 DBGPRINT("DBG:SYN:END error=%d\n",error);

 if ( opt.isVerbose==true && opt.doViewAll==false ) {
     for (i=1, n=hParse.lOut.N(); i<=n; i++) printf("%s%s",hParse.lOut.Str(i),opt.NewLine());
 }

 if ( error!=0 ) return error;

 if ( opt.doShowTree ) {
     hParse.ShowTree( fRepErr );
 }

 return error;
}

int dump_web_list (gUnweb& unweb, FILE* fOut, FILE* fRepErr, sOptUnweb& opt)
{
 int error;
 unsigned nPost;
 char* thisFilename;
 //char* str = nil;
 FILE* fOutErr = fRepErr;

 thisFilename = opt.sCurFile.Str();

 ASSERTION(fOutErr!=nil,"fOutErr!=nil");

 error = unweb.Dump( fRepErr );
 nPost = unweb.coordSerial.N();

#ifdef DEBUG_EASY
 if ( opt.isVeryVerbose ) fprintf(fOut,"::DBG:Html:START:%u lines\n",nPost);

 if ( opt.isVerbose==true ) {
     for (i=1; i<=nPost; i++) {
	 unsigned idx;
	 int x, y;
	 //str = unweb.pL.Str( i );
	 str = unweb.coordSerial.Str( i );
	 short iAux = unweb.coordSerial.GetCoord( i ).iAux;
	 x = unweb.coordSerial.GetCoord(i).GetX();
	 y = unweb.coordSerial.GetCoord(i).GetY();

	 // Comment-printing (draft)
	 gList foundL, foundTxtL;
	 idx = unweb.coordComments.FindCoordX( x, foundL, foundTxtL );
	 if ( idx!=0 ) {
	     unsigned ilIdx;
	     printf("\nidx:%u {",idx);
	     for (ilIdx=1; ilIdx<=foundL.N(); ilIdx++)
		 printf("%s%s",foundL.Str(ilIdx),ilIdx==foundL.N()?"\0":",");
	     printf("} ");
	     printf("|%s| (1of%u)\n",foundTxtL.Str(1),foundTxtL.N());
	 }

	 // ...and the rest...
	 gString sDbg( str );
	 sDbg.TrimRight();
	 // Just checking the trimming of output-html
	 printf("%s:%u,%u",str,x,y);
	 if ( iAux!=0 ) printf(" (iAux=%d)",iAux);
	 printf("\n");
	 ASSERTION(sDbg.Length()==(unsigned)strlen(str),"non-trimmed");
     }
     printf("[[EOF]]\n\n");
 }
#endif //DEBUG_EASY

 if ( opt.isDumpLinear ) {
     dump_list_linear( unweb, fOut, 1, nPost, opt, opt.htmlOpt );
 }
 else {
     dump_list_normal( unweb, fOut, 1, nPost, opt, opt.htmlOpt );
 }

#ifdef DEBUG_EASY
 if ( opt.isVeryVerbose ) fprintf(fOut,"::DBG:Html:END:%u lines\n",nPost);
 if ( opt.isVerbose ) {
     fprintf(fOut,"::DBG:coordComments:START\n");
     unweb.coordComments.Report( stdout, opt.isVeryVerbose );
     fprintf(fOut,"::DBG:coordComments:END\n");
 }
#endif

 if ( opt.isVerbose ) {
     if ( unweb.coordParseErr.N()>0 ) {
	 if ( opt.isVeryVerbose ) fprintf(fOutErr,"Parse errors follow below for: %s\n",thisFilename);
	 unweb.coordParseErr.Report(fOutErr,true);
     }
 }

 return error;
}

int show_file (gList& aL, FILE* fOut, sOptUnweb& opt, gHtmlOpt& htmlOpt)
{
 unsigned i, n;

 for (i=1, n=aL.N(); i<=n; i++) {
     if ( htmlOpt.inOpt.HasToSkip( i )==false )
	 fprintf(fOut,"%s%s",aL.Str(i),opt.NewLine());
 }
 return 0;
}
////////////////////////////////////////////////////////////
int go_file_cmd (char* inFilename, FILE* fRepErr, sOptUnweb& opt, int aCmd)
{
 int error=0, thisError;
 gUnweb unweb( inFilename, inFilename!=nil && opt.isVeryVerbose );
 FILE* fOut = stdout;

 // Note: inFilename may be nil: in this case (stdin) no progress can be made

 switch ( aCmd ) {
 case 0:
 case 1:
 case 2:
     break;
 default:
     ASSERTION_FALSE("aCmd?");
     break;
 }

 thisError = unweb.IsOpened()==false;
 if ( thisError!=0 ) return 2;

 if ( unweb.IsBufferOk()==false ) {
     ASSERTION(unweb.IsDevice(),"unweb.IsBufferOk()");
     // gFileFetch is unable to recover long line buffer ->if using stdin !
     fprintf(stderr,"Internal-Error: line buffer overun (exceeded %u)\n",unweb.BufferSize());
     return 3;
 }

 if ( opt.isTagUppercase ) unweb.SetTagCaseChange( gUnweb::e_CaseConvUpper );
 unweb.showKind = gUnweb::e_ShowNonscript;

 if ( opt.HasDumpFile() ) {
     fOut = fopen( opt.sDumpFile.Str(), "wb" );
     if ( fOut==nil ) {
	 fprintf(stderr,"Unable to write file: %s\n",opt.sDumpFile.Str());
	 return 8;
     }
 }

 switch ( aCmd ) {
 case 0:
     if ( opt.sRange.IsEmpty()==false ) {
	 fprintf(fRepErr,"Invalid --range option: only valid with 'dump' command\n");
	 error = 1;
     }
     else {
	 thisError = check_web_list( unweb,  fOut, fRepErr, opt );
	 DBGPRINT("DBG: check_web_list returned %d\n",thisError);
     }
     break;
 case 1:
     if ( opt.htmlOpt.sBaseHREF.IsEmpty()==false ) {
	 fprintf(fRepErr,"Invalid --base-href option: only valid with 'check' command\n");
	 error = 1;
     }
     else {
	 thisError = dump_web_list( unweb, fOut, fRepErr, opt );
     }
     break;
 case 2:
     thisError = show_file( unweb.aL, fOut, opt, opt.htmlOpt );
 default:
     break;
 }

 DBGPRINT("DBG: output%s, error=%d\n",(fOut==stdout || fOut==stderr)?"(stdout)":"\0",error);

 if ( error!=0 ) return error;
 error = thisError;

 DBGPRINT("DBG: 'go_file_cmd'(%d) is returning error=%d\n",aCmd,error);

 // unweb.lastOpError...
 return error;
}
////////////////////////////////////////////////////////////
int go_cmd (gArg& arg, unsigned argIdx, FILE* fRepErr, sOptUnweb& opt, int aCmd)
{
 int error = 0, thisError, countIter = 0;
 char* str;
 unsigned i, n = arg.N(), nArgs;

 argIdx++;

 for (i=argIdx; i<=n; i++) {
     str = arg.Str( i );
     nArgs = n - i;
     if ( str[0]==0 ) continue;
     opt.sCurFile = str;
     if ( opt.isVeryVerbose || (opt.isVerbose==true && nArgs>1) ) fprintf(stderr,":::%s\n",str);
     thisError = go_file_cmd( str, fRepErr, opt, aCmd );
     countIter++;
     if ( thisError!=0 ) {
	 fprintf(stderr,"Error (%d) handling: %s\n",thisError,str);
	 if ( error==0 ) error = thisError;
     }
     DBGPRINT("DBG: 'go_file_cmd'(%d) count=%d\n",aCmd,countIter);
 }

 if ( countIter==0 ) {
     error = go_file_cmd( nil, fRepErr, opt, aCmd );
 }

 return error;
}
////////////////////////////////////////////////////////////
int go_check (gArg& arg, unsigned argIdx, FILE* fRepErr, sOptUnweb& opt)
{
 return go_cmd( arg, argIdx, fRepErr, opt, 0 );
}

int go_dump (gArg& arg, unsigned argIdx, FILE* fRepErr, sOptUnweb& opt)
{
 return go_cmd( arg, argIdx, fRepErr, opt, 1 );
}

int go_show (gArg& arg, unsigned argIdx, FILE* fRepErr, sOptUnweb& opt)
{
 return go_cmd( arg, argIdx, fRepErr, opt, 2 );
}
////////////////////////////////////////////////////////////
int go (char** argv, char** envp, int& result)
{
 int error, thisError = -1;
 gArg arg( argv, envp );
 sOptUnweb opt;

 unsigned n;
 FILE* fRepErr = stderr;

 result = 0;

 arg.AddParams( 1, "-vv|--verbose -h|--help --version\
 -o:%s|--out-options\
 -d:%s|--dump\
 -~:%s|--base-href\
 --view\
 --skip\
 --dos-mode\
 -?:%s|--range\
 --tag-uppercase\
 --dump-linear\
 --show-tree" );
 error = arg.FlushParams();
 n = arg.N();
 arg.FindOptionOccurr( "verbose", opt.isVerbose, opt.isVeryVerbose );

 // Option --view views all errors in stdout and also parsed lines, so that parsing errors can be
 // viewed while parsing...

 if ( arg.FindOption('h') || error!=0 ) {
     print_help( arg.Program() );
     return 0;
 }
 if ( arg.FindOption("version") ) {
     print_version( arg.Program() );
     return 0;
 }

 arg.FindOption('o',opt.sOutOptFile);
 arg.FindOption('d',opt.sDumpFile);
 arg.FindOption('~',opt.htmlOpt.sBaseHREF);
 arg.FindOption('?',opt.sRange);

 // if ( arg.FindOption("html",opt.sHtmlFilename) ) ...

 opt.doSkip = arg.FindOption("skip");
 opt.doViewAll = arg.FindOption("view");
 opt.SetOutDOS( arg.FindOption("dos-mode") );
 opt.isTagUppercase = arg.FindOption("tag-uppercase");
 opt.isDumpLinear = arg.FindOption("dump-linear");
 opt.doShowTree = arg.FindOption("show-tree");

 if ( n==0 ) {
     fprintf(stderr,"No command.\n");
     return 0;
 }
 // parsing nearly ends...

 // Compute out-options
 if ( opt.sOutOptFile.IsEmpty()==false ) {
     int iLine;
     char* str = opt.sOutOptFile.Str();
     thisError = opt.htmlOpt.outOpt.ConfigFromFile( str );
     iLine = opt.htmlOpt.outOpt.Config().iLineError;
     if ( thisError!=0 ) {
	 if ( iLine!=0 )
	     fprintf(fRepErr,"Error (%d) on line %d: out-options file: %s\n",
		     thisError,
		     iLine,
		     str);
	 else
	     fprintf(fRepErr,"Error (%d) on out-options file: %s\n",
		     thisError,
		     str);
	 return 1;
     }
 }

 if ( opt.htmlOpt.inOpt.SetFromString( opt.sRange.Str() )==false ) {
     fprintf(fRepErr,"Invalid range specified: %s\nUse e.g.:\n\
  @3-5 suppresses line 3 to 5\n\
  12-13 shows only line 12 to 13\n\
  45- shows line 45 up to the end.\n",
	     opt.sRange.Str());
     return 1;
 }

 // Go 'command'
 char* cmdStr = arg.Str( 1 );

 if ( strcmp(cmdStr,"check")==0 ) {
     thisError = go_check( arg, 1, fRepErr, opt );
 }
 if ( strcmp(cmdStr,"dump")==0 ) {
     thisError = go_dump( arg, 1, fRepErr, opt );
 }
 if ( strcmp(cmdStr,"show")==0 ) {
     thisError = go_show( arg, 1, fRepErr, opt );
 }

 switch ( thisError ) {
 case 0:
     return 0;
 case -1:
     fprintf(stderr,"Invalid command: %s\n",cmdStr);
 case 3:
     // Buffer overflow
     return result = 1;
 case 2:
 case 4:
     return result = thisError;
 default:
     break;
 }

 // Error-code mapping:
 //   8: pre-processing (...-199), post-consolidate
 result = (thisError<=199) ? 8 : thisError;

 return thisError;
}
////////////////////////////////////////////////////////////
int main (int argc, char** argv, char** envp)
{
 int error, result=0;

 gINIT;

 error = go( argv, envp, result );

 gEND;
 DBGPRINT_MIN("go: %d, result=%d\n",error,result);

 return result;
}

